/*
DelFEM (Finite Element Analysis)
Copyright (C) 2009  Nobuyuki Umetani    n.umetani@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*! @file
@brief ەIuWFNg̃C^[tF[X
@author Nobuyuki Umetani
*/

#if !defined(EQN_SYS_H)
#define EQN_SYS_H

#include <cassert>

#include <vector>
#include <map>

#if defined(__VISUALC__)
#pragma warning( disable : 4786 )
#endif

namespace LsSol{
	class CPreconditioner;	// Os
}
namespace Fem{
namespace Ls{
	class CLinearSystem_Field;					// Aꎟ
	class CLinearSystem_Save;					// Aꎟ(sۑ)
	class CLinearSystem_SaveDiaM_NewmarkBeta;	// Aꎟ(NewmarkBeta@ōsۑ)
}
namespace Field{
	class CField;
	class CFieldWorld;
}
namespace Eqn{

/*! 
@brief ۘANX
@ingroup FemEqnSystem
*/
class CEqnSystem
{
public:		
	//! ftHgERXgN^
	CEqnSystem() : m_gamma_newmark(0.6), m_beta_newmark(0.3025), m_dt(0.1), pLS(0), pPrec(0){}
	//! fXgN^
	virtual ~CEqnSystem(){ this->Clear(); }
	virtual void Clear();

	//! 
	virtual bool Solve(Fem::Field::CFieldWorld& world) = 0;

	const std::vector< std::pair<unsigned int, double> >& GetAry_ItrNormRes() const{ return m_aItrNormRes; }

	////////////////////////////////
	// Œ苫Eǉ

	//! @{
	//! (id_field)̗vfzSČŒ苫Eɂ
	virtual bool         AddFixField(                    unsigned int id_field, Fem::Field::CFieldWorld& world, int idof = -1) = 0;
	//! (m_id_val)̗vfz(id_ea)Œ苫Eɂ
	virtual unsigned int AddFixElemAry(                  unsigned int id_ea,    Fem::Field::CFieldWorld& world, int idof = -1) = 0;
	//! (m_id_val)̗vfz̔z(aIdEA)Œ苫Eɂ
	virtual unsigned int AddFixElemAry( const std::vector<unsigned int>& aIdEA, Fem::Field::CFieldWorld& world, int idof = -1) = 0;
	//! Œ苫E폜
	virtual bool ClearFixElemAry( unsigned int id_ea, Fem::Field::CFieldWorld& world ) = 0;
	//! SĂ̌Œ苫E폜
	virtual void ClearFixElemAry() = 0;
	//! @}

	/*! 
	@brief ԐϕɂẴp[^ݒ肷
	@param [in] dt ԍ
	@param [in] gamma Newmark@gamma(ȗ0.6ɃZbg)
	@param [in] beta Newmark@beta(gammaΏȗ)
	*/
	void SetTimeIntegrationParameter(double dt, double gamma=0.6, double beta=-1.0){
		m_dt = dt;
		m_gamma_newmark = gamma;
		if( beta < 0.0 ){ m_beta_newmark = 0.25*(m_gamma_newmark+0.5)*(m_gamma_newmark+0.5); }
		else{             m_beta_newmark = beta; }
	}

	////////////////////////////////
	// AꎟNXONX̍ĕ]Cč\zw֐

	//! @{
	virtual void ClearValueLinearSystemPreconditioner(){ 
		this->ClearValueLinearSystem();
		this->ClearPreconditioner();
	}
	virtual void ClearValueLinearSystem(){   m_is_cleared_value_ls   = true; }	// tO𗧂ĂSolve̎ɒlĕ]
	virtual void ClearValuePreconditioner(){ m_is_cleared_value_prec = true; }	// tO𗧂ĂSolve̎ɒlĕ]

	virtual void ClearLinearSystemPreconditioner(){
		this->ClearLinearSystem(); 
		this->ClearPreconditioner();
	}
	virtual void ClearPreconditioner();
	virtual void ClearLinearSystem();
	//! @}
protected:
	std::vector< std::pair<unsigned int, double> > m_aItrNormRes;
	////////////////
	double m_gamma_newmark, m_beta_newmark, m_dt;
	Fem::Ls::CLinearSystem_Field* pLS;	// AꎟNX
	LsSol::CPreconditioner* pPrec;	// ONX
	bool m_is_cleared_value_ls;
	bool m_is_cleared_value_prec;
};

}	// end namespace Eqn
}	// end namespace Fem

#endif
