/*
DelFEM (Finite Element Analysis)
Copyright (C) 2009  Nobuyuki Umetani    n.umetani@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*! @file
@brief ڗgU̗vf쐬֐̃C^[tF[X
@author Nobuyuki Umetani
*/


#if !defined(EQN_ADVECTION_H)
#define EQN_ADVECTION_H

#if defined(__VISUALC__)
#pragma warning( disable : 4786 )
#endif

namespace Fem{
namespace Ls
{
	class CLinearSystem_Field;
	class CLinearSystem_Save;
	class CLinearSystem_SaveDiaM_Newmark;
}
namespace Field
{
	class CField;
	class CFieldWorld;
}
namespace Eqn
{

/*! @defgroup eqn_advection_diffusion ڗgU̘Aꎟւ̃}[W֐Q
@ingroup FemEqnMargeFunction

@@
ÓIȈڗgU
@f$ \rho\frac{\partial\phi}{\partial t} = -\mu \nabla^2 \phi + f @f$

IȈڗgU
@f$\rho\frac{\partial\phi}{\partial t}+\rho v \cdot(\nabla\phi) = -\mu \nabla^2 \phi + f @f$

ڗgÚClɂĈڗƂƂɁCgU悤Ȍۂ\łD
Ⴆ΁u̗ɃCNLĂvCu˂o鉌ɂėvȂǂ̗lq\܂D

@pre ̐̕ƂāC̏㗬ɌŒ苫Eݒ肳ĂȂƘAꎟƂł܂D

*/
/*!@{*/


/*!
@brief ÓIȂQڗgU̘Aꎟւ̃}[W
@param [in,out] ls }[WAꎟ
@param [in] myu gUW
@param [in] source ʐςɔႷ\[X
@param [in] world ǗNX
@param [in] id_field_val lID
@param [in] id_field_velo xID
@pre l̓XJ[łȂ΂Ȃ܂Dx͂QxNgłȂ΂Ȃ܂D
*/
bool AddLinSys_AdvectionDiffusion_Static(
    Fem::Ls::CLinearSystem_Field& ls,
	double myu, double source,
	const Fem::Field::CFieldWorld& world,
	unsigned int id_field_val, unsigned int id_field_velo, 
	unsigned int id_ea = 0 );

/*!
@brief ÓIȈڗgU̘Aꎟւ̃}[W(sۑ)
@param [in,out] }[WAꎟ
@param [in] myu gUW
@param [in] source ʐςɔႷ\[X
@param [in] world ǗNX
@param [in] id_field_val lID
@param [in] id_field_velo xID
@pre l̓XJ[łȂ΂Ȃ܂Dx͂QxNgłȂ΂Ȃ܂D
*/
bool AddLinSys_AdvectionDiffusion_Static(
	Fem::Ls::CLinearSystem_Save& ls,
	double myu, double source,
	const Fem::Field::CFieldWorld& world,
	unsigned int id_field_val, unsigned int id_field_velo, 
	unsigned int id_ea = 0 );

/*!
@brief IȈڗgU̘Aꎟւ̃}[W
@param [in,out] }[WAꎟ
@param [in] rho  @f$\rho@f$
@param [in] myu gUW @f$\mu@f$
@param [in] source ʐςɔႷ\[X
@param [in] id_field_val lID
@param [in] id_field_velo xID
@pre l̓XJ[łȂ΂Ȃ܂Dx͂QxNgłȂ΂Ȃ܂D
*/	
bool AddLinSys_AdvectionDiffusion_NonStatic_Newmark(
	double dt, double gamma, 
	Fem::Ls::CLinearSystem_Field& ls,
	double rho, double myu, double source,
	const Fem::Field::CFieldWorld& world,
	unsigned int id_field_val, unsigned int id_field_velo, 
	unsigned int id_ea = 0 );

/*!
@brief IȈڗgU̘Aꎟւ̃}[W(sۑ)
@param [in,out] }[WAꎟ(sۑ^)
@param [in] rho 
@param [in] myu gUW
@param [in] source ʐςɔႷ\[X
@param [in] id_field_val lID
@param [in] id_field_velo xID
@pre l̓XJ[łȂ΂Ȃ܂Dx͂QxNgłȂ΂Ȃ܂D
*/	
bool AddLinSys_AdvectionDiffusion_NonStatic_Newmark(
	Fem::Ls::CLinearSystem_SaveDiaM_Newmark& ls,
	double rho, double myu, double source,
	const Fem::Field::CFieldWorld& world,
	unsigned int id_field_val, unsigned int id_field_velo, 
	unsigned int id_ea = 0 );

/*!@}*/
}
}

#endif
