/*
DelFEM (Finite Element Analysis)
Copyright (C) 2009  Nobuyuki Umetani    n.umetani@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*! @file
@brief gU̗vf쐬̃C^[tF[X
@author Nobuyuki Umetani
*/

#if !defined(EQN_DIFFUSION_H)
#define EQN_DIFFUSION_H

#if defined(__VISUALC__)
#pragma warning( disable : 4786 )
#endif

#include <vector>

namespace Fem
{
namespace Ls{
	class CLinearSystem_Field;
	class CLinearSystem_SaveDiaM_Newmark;
	class CPreconditioner;
}
namespace Field{
	class CField;
	class CFieldWorld;
}

namespace Eqn
{

/*! @defgroup eqn_diffusion gUAꎟɃ}[W֐Q
@ingroup FemEqnMargeFunction
@@
@f$ \rho \frac{\partial\phi}{\partial t} = -\mu \nabla^2\phi + f@f$

gUƂ́CZxzɔႵė(Flux)肳悤ȏ̕LqΔłD
ႦΔMCڗ̂Ȃ̔ZxωȂǂɏ]܂D
*/
//! @{

/*!
@brief gŨ}[W
@param [in,out] ls Aꎟ
@param [in] rho Me @f$\rho@f$
@param [in] alpha MgUW @f$\mu@f$
@param [in] source \[X @f$ f @f$
@param [in] world ǗNX
@param [in] id_field_val lID
*/	
bool AddLinSys_Diffusion(
	double dt, double gamma,
	Fem::Ls::CLinearSystem_Field& ls,
	double rho, double alpha, double source,
	const Fem::Field::CFieldWorld& world, unsigned int id_field_val,
	unsigned int id_ea = 0 );

/*!
@brief Ώ́CgŨ}[W
@param [in,out] ls Aꎟ
@param [in] rho Me @f$\rho@f$
@param [in] alpha MgUW @f$\mu@f$
@param [in] source \[X @f$ f @f$
@param [in] world ǗNX
@param [in] id_field_val lID
@remark ͎͂ɈvĂƂ邪Ĉ͂ɂD
*/	
bool AddLinSys_Diffusion_AxSym(
	double dt, double gamma,
	Fem::Ls::CLinearSystem_Field& ls,
	double rho, double alpha, double source,
	const Fem::Field::CFieldWorld& world, unsigned int id_field_val,
	unsigned int id_ea = 0 );


/*!
@brief gU(sۑ)̃}[W
@param [in,out] ls Aꎟ(sۑ^)
@param [in] rho Me @f$\rho@f$
@param [in] alpha MgUW @f$\mu@f$
@param [in] source \[X @f$ f @f$
@param [in] world ǗNX
@param [in] id_field_val lID
*/
bool AddLinSys_Diffusion(
	Fem::Ls::CLinearSystem_SaveDiaM_Newmark& ls,
	double rho, double alpha, double source,
	const Fem::Field::CFieldWorld& world,
	unsigned int id_field_val, unsigned int id_ea = 0 );

//! @}

}	// end namespace Eqn
}	// end namespace Fem

#endif
