/*
Copyright (C) 2009  Nobuyuki Umetani    n.umetani@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*! @file
@brief Interface of class (MatVec::CMat_BlkCrs)
@author Nobuyuki Umetani
*/

#if !defined(MAT_BLK_CRS_H)
#define MAT_BLK_CRS_H

#include <assert.h>
#include <vector>

namespace Com{
    class CIndexedArray;
}
	
namespace MatVec{

class CVector_Blk;
class COrdering_Blk;
class CBCFlag;

/*! 
@brief ubNCRS\̍sNX
@ingroup MatVec
*/
class CMat_BlkCrs  
{
public:
	/*!
	@brief TCYƂRXgN^
	@param[in] nblk_col ̃ubN
	@param[in] len_col  ubN̗̒
	@param[in] nblk_row s̃ubN
	@param[in] len_row  ubN̍s̒
	*/
	CMat_BlkCrs(unsigned int nblk_col, unsigned int len_col, 
		        unsigned int nblk_row, unsigned int len_row );
    
  CMat_BlkCrs(unsigned int nblk_col, const std::vector<unsigned int>& alen_col, 
              unsigned int nblk_row, const std::vector<unsigned int>& alen_row );

	/*!
	@brief sRs[邽߂̃RXgN^
	@param[in] rhs s
	@param[in] is_value l܂ŃRs[邩ǂ
	@param[in] isnt_trans ]uƂăRs[Ȃ
	*/
	CMat_BlkCrs(const CMat_BlkCrs& rhs, bool is_value, bool isnt_trans);

	CMat_BlkCrs(); 
	virtual ~CMat_BlkCrs();

	virtual bool Initialize(unsigned int nblk_col, unsigned int len_col, 
		                    unsigned int nblk_row, unsigned int len_row );
  // DiaŉB邽߂virtualɂĂ
  virtual bool Initialize(unsigned int nblk_col, const std::vector<unsigned int>& alen_col, 
                          unsigned int nblk_row, const std::vector<unsigned int>& alen_row );

	inline unsigned int NBlkMatCol() const { return m_nblk_MatCol; }	//!< sɊubN邩
	inline unsigned int NBlkMatRow() const { return m_nblk_MatRow; }	//!< ɊubN邩
	
	//! ubN̍s̃TCY(c̒)擾.FixĖꍇ-1Ԃ
	inline int LenBlkCol() const { return m_len_BlkCol; }
	//! ubN̍s̃TCY(c̒)擾.
	inline unsigned int LenBlkCol(unsigned int iblk) const { 
    assert( iblk < m_nblk_MatCol );
    if( m_len_BlkCol == -1 ){
      assert( m_DofPtrCol != 0 );
      return m_DofPtrCol[iblk+1]-m_DofPtrCol[iblk];
    }
    return m_len_BlkCol; 
  }
	//! ubÑ̗TCY(̒)擾.FixĖꍇ-1Ԃ
	inline int LenBlkRow() const { return m_len_BlkRow; }
	//! ubN̍s̃TCY(̒)擾.
	inline unsigned int LenBlkRow(unsigned int iblk) const { 
    assert( iblk < m_nblk_MatRow );
    if( m_len_BlkRow == -1 ){
      assert( m_DofPtrRow != 0 );
      return m_DofPtrRow[iblk+1]-m_DofPtrRow[iblk];
    }
    return m_len_BlkRow; 
  }

	//! lɂOݒDf[^̈悪mۂĂȂ΁Cmۂ
	virtual bool SetZero();	
	//! p^[Sď@RowPtr,Val̓
	virtual bool DeletePattern();	

	virtual void FillPattern();
  virtual bool AddPattern(const Com::CIndexedArray& crs);	//!< [p^[ǉ
	virtual bool AddPattern(const CMat_BlkCrs& rhs, const bool isnt_trans);	//!< [p^[ǉ
	virtual bool AddPattern(const CMat_BlkCrs& rhs, 
                          const COrdering_Blk& order_col, const COrdering_Blk& order_row);	//!< [p^[ǉ

	// ͔hNXɔC
	virtual bool SetPatternBoundary(const CMat_BlkCrs& rhs, const CBCFlag& bc_flag_col, const CBCFlag& bc_flag_row);
	virtual bool SetPatternDia(const CMat_BlkCrs& rhs);
	
	virtual bool SetValue(const CMat_BlkCrs& rhs, const bool isnt_trans);
	virtual bool SetValue(const CMat_BlkCrs& rhs, 
                        const COrdering_Blk& order_col, const COrdering_Blk& order_row);
	
	//! vfs}[W
	virtual bool Mearge
  (unsigned int nblkel_col, const unsigned int* blkel_col,
   unsigned int nblkel_row, const unsigned int* blkel_row,
   unsigned int blksize, const double* emat);
  //! Ɨpz̗̈
  void DeleteMargeTmpBuffer(){
    if( m_marge_tmp_buffer == 0 ) return;
    delete[]  m_marge_tmp_buffer;
    m_marge_tmp_buffer = 0;
  }

	//! sxNg
	virtual bool MatVec(double alpha, const CVector_Blk& x, double beta, CVector_Blk& b, const bool isnt_trans) const;

	//! bc_flagP̍s̗vfOɂ
	bool SetBoundaryCondition_Row(const CBCFlag& bc_flag);
	//! bc_flagP̗̗vfOɂ
	bool SetBoundaryCondition_Colum(const CBCFlag& bc_flag);
	//! bc_flagO̗vfOɂiS̍sۑɂ鍂@j
	bool SetBoundaryConditionInverse_Colum(const CBCFlag& bc_flag);
	
	////////////////////////////////////////////////
	// Crs Original Function ( Non-virtual )

	//! CRS̃TCY擾
	const unsigned int NCrs() const { return m_ncrs_Blk; }
	const unsigned int* GetPtrIndPSuP(const unsigned int ipoin, unsigned int& npsup) const
	{
		if( m_rowPtr_Blk == 0 ){ npsup = 0; return 0; }
		assert( ipoin < m_nblk_MatCol );
		npsup = m_colInd_Blk[ipoin+1]-m_colInd_Blk[ipoin];
		const unsigned int icrs = m_colInd_Blk[ipoin];
		return &m_rowPtr_Blk[icrs];
	}
	const double* GetPtrValPSuP(const unsigned int ipoin, unsigned int& npsup) const
	{
		if( m_rowPtr_Blk == 0 ){ npsup = 0; return 0; }
		assert( ipoin < m_nblk_MatCol );
		npsup = m_colInd_Blk[ipoin+1]-m_colInd_Blk[ipoin];
		const unsigned int icrs = m_colInd_Blk[ipoin];
		return &m_valCrs_Blk[icrs*m_len_BlkCol*m_len_BlkRow];
	}
	double* GetPtrValPSuP(const unsigned int ipoin, unsigned int& npsup)
	{
		if( m_rowPtr_Blk == 0 ){ npsup = 0; return 0; }
		assert( ipoin < m_nblk_MatCol );
		npsup = m_colInd_Blk[ipoin+1]-m_colInd_Blk[ipoin];
		const unsigned int icrs = m_colInd_Blk[ipoin];
		return &m_valCrs_Blk[icrs*m_len_BlkCol*m_len_BlkRow];
	}

private:
	unsigned int m_nblk_MatCol;	//!< sɊubN邩
	unsigned int m_nblk_MatRow;	//!< ɊubN邩

	int m_len_BlkCol;	//!< ubN̗TCY
	int m_len_BlkRow;	//!< ubN̍sTCY

protected:
	unsigned int  m_ncrs_Blk;	//!< CRS̃TCY
	unsigned int* m_colInd_Blk;	//!< CRSColum Index
	unsigned int* m_rowPtr_Blk;	//!< CRSRow Pointer

  int* m_marge_tmp_buffer;	//!< }[W̎ɕKvȍƃobt@

  // Flex̎ɒ`l
  unsigned int* m_DofPtrCol;
  unsigned int* m_DofPtrRow;
  unsigned int* m_ValPtr;

	double* m_valCrs_Blk;	//!< s̒l
};

}	// end namespace 'Ls'

#endif 
