/*
DelFEM (Finite Element Analysis)
Copyright (C) 2009  Nobuyuki Umetani    n.umetani@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*! @file
@brief QbVUtility֐Q
@author Nobuyuki Umetani
*/


#if !defined(MESH_KERNEL_2D_H)
#define MESH_KERNEL_2D_H

#if defined(__VISUALC__)
	#pragma warning( disable : 4786 )
#endif

#include <vector>

#include "delfem/vector2d.h"

namespace Msh
{

/*! 
@addtogroup Msh2D
*/
//! @{

////////////////////////////////////////////////
// {͋L

const double MIN_TRI_AREA = 1.0e-10;

const unsigned int nNoEd  = 2;	//!< ɂ_邩

const unsigned int nNoTri = 3;	//!< Rp`ɂ璸_邩
const unsigned int nEdTri = 3;	//!< Rp`ɂӂ邩
//! Rp`̊eӂ̒_ԍ
const unsigned int noelTriEdge[nEdTri][nNoEd] = {	
	{ 1, 2 },
	{ 2, 0 },
	{ 0, 1 },
};
//! Rp`̗אڊ֌W
const unsigned int relTriTri[3][3] = {
	{ 0, 2, 1 }, //  0
	{ 2, 1, 0 }, //  1 
	{ 1, 0, 2 }, //  2
};

const unsigned int nNoQuad = 4;	//!< Sp`ɂ璸_邩
const unsigned int nEdQuad = 4;	//!< Sp`ɂӂ邩
//! Sp`̊eӂ̒_ԍ
const unsigned int noelQuadEdge[nEdQuad][nNoEd] = {	
	{ 0, 1 },
	{ 1, 2 },
	{ 2, 3 },
	{ 3, 0 }
};
//! Sp`̗אڊ֌W
static const unsigned int relQuadQuad[nNoQuad][nNoQuad] = {
	{ 0, 3, 2, 1 }, //  0
	{ 1, 0, 3, 2 }, //  1
	{ 2, 1, 0, 3 }, //  2
	{ 3, 2, 1, 0 }, //  3
};

//! ӗvf\
struct SBar{
	unsigned int v[2];	//!< _ԍ
	////////////////
    // OE1
	unsigned int s2[2];
	unsigned int r2[2];
};


////////////////////////////////////////////////////////////////

//! Sp`vf\
struct SQuad2D{
	unsigned int v[4];	//!< _Index
	int g2[4];			//!< אڂvfzID(-1:אڗvfȂA-2:̗vfzɗא)
	unsigned int s2[4];	//!< אڗvfIndex
	unsigned int r2[4];			//!< אڊ֌W
};


//! QRp`vf\
struct STri2D{
	unsigned int v[3];	//!< _Index
    int g2[3];			//!< אڂvfzID(-1:אڗvfȂA-2:̗vfzɗא)
	unsigned int s2[3];	//!< אڗvfIndex
    unsigned int r2[3];	//!< אڊ֌W
};

/*!
@brief Q_NX
(CPoint2D.e!=-1)Ȃ(aTri[e].no[d])̓_̑S̐ߓ_ԍł͂
*/
class CPoint2D{
public:
	CPoint2D(){}
	CPoint2D( const CPoint2D& rhs )
		: p(rhs.p), e(rhs.e), d(rhs.d){}
	CPoint2D(double x, double y, int ielem, unsigned int idir)
		: p(x,y), e(ielem), d(idir){}
public:
    Com::CVector2D p;   //!< _̍W
    int e;              //!< _͂ޗvf̂̔ԍ(ǗĂ_Ȃ-1)
    unsigned int d;     //!< _͂ޗvfe̗vfߓ_ԍ
};

////////////////////////////////////////////////////////////////

//! _͂ގOp`̃Xg
bool MakePointSurTri(const std::vector<Msh::STri2D>& aTri, const unsigned int npoin, 
		unsigned int* const elsup_ind, unsigned int& nelsup, unsigned int*& elsup );

//! _͂ނSp`̃Xg
bool MakePointSurQuad(const std::vector<Msh::SQuad2D>& aQuad, const unsigned int npoin, 
		unsigned int* const elsup_ind, unsigned int& nelsup, unsigned int*& elsup );
//! _͂ސvf̃Xg
bool MakePointSurBar( const std::vector<Msh::SBar>& aBar, const unsigned int npoin, 
		unsigned int* const elsup_ind, unsigned int& nelsup, unsigned int*& elsup );

//! Rp`̓(אڊ֌W)
bool MakeInnerRelationTri( std::vector<Msh::STri2D>& aTri, const unsigned int npoin, 
	const unsigned int* elsup_ind, const unsigned int nelsup, const unsigned int* elsup);
//! Sp`̓(אڊ֌W)
bool MakeInnerRelationQuad( std::vector<Msh::SQuad2D>& aQuad, const unsigned int npoin, 
	const unsigned int* elsup_ind, const unsigned int nelsup, const unsigned int* elsup);
//! ̓(אڊ֌W).TA͊܂܂ȂƂ
//bool MakeInnerRelationBar( std::vector<Msh::SBar>& aBar );

//! Op`̊O͂ޕӗvfz
bool MakeOuterBoundTri( const std::vector<Msh::STri2D>& aTri, std::vector<Msh::SBar>& aBar );
//! Sp`̊O͂ޕӗvfz
bool MakeOuterBoundQuad( const std::vector<Msh::SQuad2D>& aQuad, std::vector<Msh::SBar>& aBar );

//! Op`̐`FbN
bool CheckTri( const std::vector<Msh::STri2D>& aTri );
//! Op`̐`FbN
bool CheckTri( const std::vector<Msh::CPoint2D>& po, const std::vector<Msh::STri2D>& tri );

/*!
@brief _vfɒǉ
@param[in] ipo_ins  vfɒǉ钸_̔ԍ
@param[in] itri_ins ǉvf̔ԍ
@retval true    
@retval false   s
*/
bool InsertPoint_Elem( const unsigned int ipo_ins, const unsigned int itri_ins,
					  std::vector<Msh::CPoint2D>& po, std::vector<Msh::STri2D>& tri );
//! _vf̕ӂɒǉ
bool InsertPoint_ElemEdge( const unsigned int ipo_ins,
    const unsigned int itri_ins, const unsigned int ied_ins,
	std::vector<Msh::CPoint2D>& po, std::vector<Msh::STri2D>& tri );

//! _̃f[j
bool DelaunayAroundPoint( unsigned int ipo0,
						 std::vector<Msh::CPoint2D>& po, std::vector<Msh::STri2D>& tri );

/*!
@brief _̃f[j
num_flip͕ӂFlip񐔂ɑ킹DOŏ͂ȂD
*/
bool DelaunayAroundPoint( unsigned int itri, unsigned int inotri, 
						 std::vector<Com::CVector2D>& aVec, std::vector<Msh::STri2D>& tri,
						 unsigned int& num_flip);
//! ӂtbv
bool FlipEdge( unsigned int itri0, unsigned int ied0,  std::vector<Msh::STri2D>& aTri );
//! ӂtbv
bool FlipEdge( unsigned int itri0, unsigned int ied0,
			  std::vector<Msh::CPoint2D>& aPo, std::vector<Msh::STri2D>& aTri );
//! ӂtbv
bool FindEdge( const unsigned int& ipo0, const unsigned int& ipo1,
	unsigned int& itri0, unsigned int& inotri0, unsigned int& inotri1,
	const std::vector<Msh::CPoint2D>& po, const std::vector<Msh::STri2D>& tri );
//! ܂ӂ
bool FindEdgePoint_AcrossEdge( const unsigned int& ipo0, const unsigned int& ipo1,
	unsigned int& itri0, unsigned int& inotri0, unsigned int& inotri1, double& ratio,
	std::vector<Msh::CPoint2D>& po, std::vector<Msh::STri2D>& tri );

//! MtgEbsO@ɂOp`
bool Tesselate_GiftWrapping(
	unsigned int ipo0, unsigned int ipo1,
	std::vector< std::pair<unsigned int,unsigned int> >& aTriEd,
	const std::vector<Com::CVector2D>& aPo, std::vector<Msh::STri2D>& aTri);

//! Op`P_폜
bool DeletePointFromMesh(unsigned int ipoin, std::vector<Msh::CPoint2D>& aPo, std::vector<Msh::STri2D>& aTri);
//! Op`Pvf폜
bool DeleteTri(unsigned int itri0, std::vector<Msh::CPoint2D>& aPo, std::vector<Msh::STri2D>& aTri);

//! vVAX[WO
void LaplacianSmoothing( std::vector<Msh::CPoint2D>& aPo, const std::vector<Msh::STri2D>& aTri,
	const std::vector<unsigned int>& aflg_isnt_move );
//! vVAX[WO
void LaplaceDelaunaySmoothing( std::vector<Msh::CPoint2D>& aPo, std::vector<Msh::STri2D>& aTri,
	const std::vector<unsigned int>& aflg_isnt_move );
//! oubVpX[WO
void PliantBossenHeckbertSmoothing( double elen, std::vector<Msh::CPoint2D>& aPo, std::vector<Msh::STri2D>& aTri );


//! BarAryF
void ColorCodeBarAry( const std::vector<SBar>& aBar, const std::vector<Com::CVector2D>& aVec, 
					 std::vector< std::vector<unsigned int> >& aIndBarAry );

//! @}
} // end name space mesh;



#endif // MSH_H
