/*
 *  edit_furniture.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 11/2/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include "delfem/vector2d.h"

#include "rigid_contact.h"
#include "gl_drawer_furniture.h"
#include "edit_mode.h"
#include "edit_furniture.h"
#include "rigid_pointset.h"
#include "joint_connectivity.h"


double DragAPlank
(unsigned int ind_picked_obj,
 int rigid_edit_mode,           
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,
 std::vector<CJoint_Fix>& aFix,       
 std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                
 std::vector<CJoint_Fix_Force>& aJFixForce,                                
 std::vector<CContactPoints_RigidBody3D>& aFP,               
 double mov_begin_x, double mov_begin_y, 
 double mov_end_x, double mov_end_y,
 const Com::CMatrix3& cam_rot,
 const Com::CVector3D& cam_trans,
 const Com::CVector3D& cam_objc)
{
  Com::CMatrix3 cr = cam_rot;
  unsigned int irb0 = ind_picked_obj;
  CRigid3D_Ini& ri = aRI[irb0];
  const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
  double delx = (mov_end_x - mov_begin_x);//*hvh*asp;
  double dely = (mov_end_y - mov_begin_y);//*hvh; 
  const bool flg[4] = {
    aFP[irb0].aFP[0].is_force_cont,         
    aFP[irb0].aFP[1].is_force_cont,        
    aFP[irb0].aFP[2].is_force_cont,         
    aFP[irb0].aFP[3].is_force_cont };                  
  if( rigid_edit_mode >= 0 && rigid_edit_mode < 3 ){          
    Com::CVector3D sensEdit;
    if( rigid_edit_mode == 0 ){ sensEdit = Com::CVector3D(1,0,0); }
    if( rigid_edit_mode == 1 ){ sensEdit = Com::CVector3D(0,1,0); }
    if( rigid_edit_mode == 2 ){ sensEdit = Com::CVector3D(0,0,1); }      
    sensEdit = Q.MatVec(sensEdit);
    Com::CVector3D v1 = cr.MatVec(sensEdit);
    Com::CVector2D snsdir(v1.x,v1.y);
    snsdir *= 1.0/SquareLength(snsdir);
    const double del_mode = snsdir.x*delx + snsdir.y*dely;   
    return del_mode;
  }
  if( rigid_edit_mode >= 3 && rigid_edit_mode < 6 ){  // rotation
    unsigned int iaxis = rigid_edit_mode-3;
    Com::CVector3D axis(0,0,0); axis[iaxis] = 1;
    axis = Q.MatVec(axis);
    Com::CVector3D rot_cent = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
    {
      if(      flg[0] && flg[1] && !flg[2] && !flg[3] ){
        if( iaxis != 0 ){}
        else {}
      }
      else if( !flg[0] && !flg[1] && flg[2] && flg[3] ){
        if( iaxis != 0 ){}
        else {}          
      }        
      else if( flg[0] && !flg[1] && !flg[2] && flg[3] ){ 
        if( iaxis != 1 ){}
        else {
          rot_cent -= Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) );                
        } 
      }
      else if( !flg[0] && flg[1] && flg[2] && !flg[3] ){
        if( iaxis != 1 ){}
        else {
          rot_cent += Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) );                                
        }             
      }         
    }
    Com::CVector3D dir = cam_rot.MatVecTrans( Com::CVector3D(0,0,1) );
    Com::CVector3D a0;
    {
      Com::CVector3D src = cam_rot.MatVecTrans( Com::CVector3D(mov_begin_x,mov_begin_y,0) - cam_trans ) + cam_objc;
      double t = Dot(rot_cent-src,axis)/Dot(dir,axis);
      a0 = src+t*dir-rot_cent;
    }
    Com::CVector3D a1;
    {
      Com::CVector3D src = cam_rot.MatVecTrans( Com::CVector3D(mov_end_x,mov_end_y,0) - cam_trans ) + cam_objc;
      double t = Dot(rot_cent-src,axis)/Dot(dir,axis);
      a1 = src+t*dir-rot_cent;
    }
    Com::CVector3D a2 = Cross(a0,a1);
    double del_mode = Dot(a2,axis);
    del_mode /= (a0.Length()*a1.Length());
    return del_mode;
  }
  if( rigid_edit_mode >= 6 ){ // change length
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
    Com::CVector3D sensEdit;
    unsigned int imode_l = rigid_edit_mode-6;
    if( imode_l == 0 ){ sensEdit = Q.MatVec( Com::CVector3D(-0.5, 0,    0) ); }
    if( imode_l == 1 ){ sensEdit = Q.MatVec( Com::CVector3D(+0.5, 0,    0) ); }
    if( imode_l == 2 ){ sensEdit = Q.MatVec( Com::CVector3D( 0,   -0.5, 0) ); }
    if( imode_l == 3 ){ sensEdit = Q.MatVec( Com::CVector3D( 0,   +0.5, 0) ); }                                 
    Com::CVector3D v1 = cr.MatVec(sensEdit);
    Com::CVector2D snsdir(v1.x,v1.y);
    snsdir *= 1.0/SquareLength(snsdir);          
    const double del_mode = snsdir.x*delx + snsdir.y*dely;
    return del_mode;
  }
  return 0;
}

void PickAPlank_Ini
(int& rigid_edit_mode,
 int& ind_picked_obj,                
 unsigned int& iface_picked_obj,
 Com::CVector3D& pic_pos_3d,                
 CDesignSensitivity& desSens,
 const Com::CVector3D& src, 
 const Com::CVector3D& dir,                
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CJoint_Fix>& aFix,       
 std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                
 std::vector<CContactPoints_RigidBody3D>& aFP,
 double hvh,
 const CContactProp& contProp)
{  
  unsigned int nRB = aRI.size();
  const int irb_pic = PickBox_Ini(src,-dir,pic_pos_3d,iface_picked_obj,aRI);
  desSens.Initialize( nRB, aFix.size() );
  if( rigid_edit_mode == -1 || ind_picked_obj < 0 || ind_picked_obj >= nRB ){
    ind_picked_obj = irb_pic;
    rigid_edit_mode = -2;
    return;
  } 
  unsigned int irb = ind_picked_obj;
  int icont_flg = 0;
  {
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x
    if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x    
    if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y    
    if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y    
  }
  int ijnt_flg = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;      
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
    if( irb0 != irb && irb1 != irb ) continue;
    unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
    if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
    if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
    if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
    if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y        
  }      
  if( rigid_edit_mode == -2 || (rigid_edit_mode >= 0 && rigid_edit_mode < 3 ) ){ // translation
    assert( ind_picked_obj >=0 && ind_picked_obj < nRB ); 
    int imode = -2;
    if( rigid_edit_mode != -2 ){
      CRigid3D_Ini& ri = aRI[irb];
      const Com::CVector3D& cg = ri.GetIniPosCG();
      const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();      
      Com::CVector3D e(0,0,0); e[rigid_edit_mode] = 1;
      e = Q.MatVec(e);
      double dist = Com::GetMinDist_LineLine(cg,e,src,dir);
     if( dist < 0.03*hvh && dist > 0 ){
        imode = rigid_edit_mode;
      }
    }
    if( imode == -2 ){
      imode = PickRigidBodyEditMode(src,dir,aRI,aFP[irb],irb,aFix,aFixDir,BOX_POS,hvh);
    }
    if( imode == -2 ){
      if( irb_pic == ind_picked_obj ){ // move on to another edit mode in the same plank
        rigid_edit_mode = -3;            
      }
      else{ // move on to another plank
        ind_picked_obj = irb_pic;
        rigid_edit_mode = -2;
      }
    }
    else{ // translate with this mode
      GetDesignSensitivity(desSens, aRI,aFix,aFixDir,irb, imode, icont_flg,contProp);
      rigid_edit_mode = imode;
    }
    return;
  }
  if( rigid_edit_mode == -3 || (rigid_edit_mode >= 3 && rigid_edit_mode < 6)){ // rotation
    assert( ind_picked_obj >=0 && ind_picked_obj < nRB );
    unsigned int imode = PickRigidBodyEditMode(src,dir,aRI,aFP[irb],irb,aFix,aFixDir,BOX_ROT,hvh);
    if( imode == -3 ){
      if( irb_pic == ind_picked_obj ){ // move on to another edit mode in the same plank
        rigid_edit_mode = -4;    
      }
      else{  // move on to another plank
        ind_picked_obj = irb_pic;
        rigid_edit_mode = -2;
      }
    }
    else{ // rotate with this mode
      GetDesignSensitivity(desSens, aRI,aFix,aFixDir,irb, imode, icont_flg,contProp);
      rigid_edit_mode = imode;          
    }
    return;
  }      
  if( rigid_edit_mode == -4 || rigid_edit_mode >= 6 ){
    assert( ind_picked_obj >=0 && ind_picked_obj < nRB );        
    unsigned int imode = PickRigidBodyEditMode(src,dir,aRI,aFP[irb],irb,aFix,aFixDir,BOX_LEN,hvh);
    if( imode == -4 ){
      if( irb_pic == ind_picked_obj ){ // move on to another edit mode in the same plank
        rigid_edit_mode = -2;            
      }
      else{  // move on to another plank
        ind_picked_obj = irb_pic;
      }
    }
    else{ // change size with this mode
      unsigned int imode_l = (imode-6)/2+6;
      GetDesignSensitivity(desSens, aRI,aFix,aFixDir,irb, imode_l, icont_flg,contProp);      
      rigid_edit_mode = imode;
    }
    return;
  }  
}

bool CheckJointInside
(unsigned int ifix,
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CJoint_Fix>& aFix,       
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir)
{  
  const Com::CVector3D aFaceNormal[6][3] = {
    { Com::CVector3D(+1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0,+1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-0,-1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0, 0,+1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) },
    { Com::CVector3D(-0, 0,-1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) } };    
  assert( ifix < aFix.size() );
  unsigned int idir0 = aFixDir[ifix].first;
  unsigned int idir1 = aFixDir[ifix].second;  
  const CRigid3D_Ini& ri0 = aRI[ aFix[ifix].irb0 ];
  const CRigid3D_Ini& ri1 = aRI[ aFix[ifix].irb1 ];   
  assert( idir0 == 4 || idir0 == 5 );
  if( idir1!=4 || idir1!=5 ){
    Com::CVector3D a,b;    
    bool res = GetIntersectionOfTwoPlanks(a,b,
                                          ri0,ri1,
                                          idir0,idir1);
//    std::cout << res << "   " << a << "   " << b << std::endl;
    return res;
  }
}


void FlipJoint
(unsigned int ifix,
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CJoint_Fix>& aFix,       
 std::vector< std::pair<unsigned int,unsigned int> >& aFixDir)                                       
{
  const Com::CVector3D aFaceNormal[6][3] = {
    { Com::CVector3D(+1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0,+1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-0,-1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0, 0,+1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) },
    { Com::CVector3D(-0, 0,-1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) } };    
  
  if( ifix >= aFix.size() ) return;
  unsigned int idir0 = aFixDir[ifix].first;
  unsigned int irb0 = aFix[ifix].irb0;
  const CRigid3D_Ini& ri0 = aRI[irb0];
  Com::CVector3D l0(ri0.xlen,ri0.ylen,ri0.zlen);
  Com::CMatrix3 B0(aFaceNormal[idir0][0].x*0.5, aFaceNormal[idir0][0].y*0.5, aFaceNormal[idir0][0].z*0.5);
  Com::CMatrix3 L0(l0.x,l0.y,l0.z);  
  Com::CMatrix3 R0 = ri0.GetRotMatrix_Ini();
  Com::CMatrix3 RB0 = R0.MatMat(B0);
  Com::CMatrix3 RL0 = R0.MatMat(L0);
  const Com::CVector3D& n0 = R0.MatVec( aFaceNormal[idir0][0] );      
  const Com::CVector3D& o0 = RB0.MatVec(l0) + ri0.GetIniPosCG();  
  ////    
  unsigned int idir1 = aFixDir[ifix].second;    
  unsigned int irb1 = aFix[ifix].irb1;  
  const CRigid3D_Ini& ri1 = aRI[irb1];     
  Com::CVector3D l1(ri1.xlen,ri1.ylen,ri1.zlen);
  Com::CMatrix3 B1(aFaceNormal[idir1][0].x*0.5, aFaceNormal[idir1][0].y*0.5, aFaceNormal[idir1][0].z*0.5);  
  Com::CMatrix3 L1(l1.x,l1.y,l1.z);
  Com::CMatrix3 R1 = ri1.GetRotMatrix_Ini();
  Com::CMatrix3 RB1 = R1.MatMat(B1);
  Com::CMatrix3 RL1 = R1.MatMat(L1);  
  const Com::CVector3D& n1 = R1.MatVec( aFaceNormal[idir1][0] );      
  const Com::CVector3D& o1 = RB1.MatVec(l1) + ri1.GetIniPosCG();      
  ////
  assert( idir0 == 4 || idir0 == 5 );
  if( idir1!=4||idir1!=5 ){
    const Com::CVector3D& h0 = RL0.MatVec(aFaceNormal[idir0][1]*0.5);
    const Com::CVector3D& v0 = RL0.MatVec(aFaceNormal[idir0][2]*0.5);
    double rh, rv;
    bool res = IsRayCollideQuad(o1,n0, o0,n0, h0,v0, rv,rh);    
    Com::CVector3D vec_e;
    unsigned int idir0a, idir1a;
    if(      rh > rv && rh > -rv ){ // +y
      idir1a = 2;      
      vec_e = v0;
    }
    else if( rh < rv && rh < -rv ){ // -y
      idir1a = 3;      
      vec_e = -v0;      
    }
    else if( rv > rh && rv > -rh ){ // +x
      idir1a = 0;      
      vec_e = h0;
    }   
    else if( rv < rh && rv < -rh ){ // -x
      idir1a = 1;      
      vec_e = -h0;      
    }
    //    Com::CMatrix3 R1 = ri1.GetRotMatrix_Ini();
    const Com::CVector3D& n1 = R1.MatVec( Com::CVector3D(0,0,1) );    
    if( Com::Dot(n1,vec_e) > 0 ){
      idir0a = 5;      
    }
    else{
      idir0a = 4;            
    }
    aFix[ifix].irb0 = irb1;
    aFix[ifix].irb1 = irb0;
    aFixDir[ifix].first = idir0a;
    aFixDir[ifix].second = idir1a;    
  }  
}


static void CalcInvMat(double* a, const unsigned int n, int& info )
{
	double tmp1;
  
	info = 0;
	unsigned int i,j,k;
	for(i=0;i<n;i++){
		if( fabs(a[i*n+i]) < 1.0e-30 ){
			info = 1;
			return;
		}
		if( a[i*n+i] < 0.0 ){
			info--;
		}
		tmp1 = 1.0 / a[i*n+i];
		a[i*n+i] = 1.0;
		for(k=0;k<n;k++){
			a[i*n+k] *= tmp1;
		}
		for(j=0;j<n;j++){
			if( j!=i ){
				tmp1 = a[j*n+i];
				a[j*n+i] = 0.0;
				for(k=0;k<n;k++){
					a[j*n+k] -= tmp1*a[i*n+k];
				}
			}
		}
	}
}



