/*
 *  edit_mode.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 11/4/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */


#if defined(__APPLE__) && defined(__MACH__)
#  include <GLUT/glut.h>
#else
#  include <GL/glut.h>
#endif

#include "edit_mode.h"

// Set neibouring planks's sensitivty that maintain joint connectivity by changing their length
void NeighboringPlankSensitivity
(unsigned int irb, // manipulated rigid body
 CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CJoint_Fix>& aFix,       
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir)
{
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };
  
  assert( irb < desSens.aPlank.size() );
  const CDesignSensPlank& desSensPlank = desSens.aPlank[irb];
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];
    
    const unsigned int irb0 = cjs.irb0;
    const CRigid3D_Ini& ri0 = aRI[irb0];    
    unsigned int idir0 = aFixDir[ifix].first;    
    Com::CMatrix3 Q0 = ri0.GetRotMatrix_Ini();
    Com::CVector3D n0 = Q0.MatVec(aFaceNormal[idir0]);
    Com::CVector3D _n0 = Cross(desSens.aPlank[irb0].r,n0);
    Com::CVector3D l0(ri0.xlen, ri0.ylen, ri0.zlen);    
    Com::CMatrix3 B0(0.5*aFaceNormal[idir0].x,0.5*aFaceNormal[idir0].y,0.5*aFaceNormal[idir0].z);    
    Com::CMatrix3 RB0 = Q0.MatMat(B0);
    Com::CVector3D b0 = RB0.MatVec(l0)+ri0.GetIniPosCG();
    Com::CVector3D _b0 = desSens.aPlank[irb0].p + Cross(desSens.aPlank[irb0].r,RB0.MatVec(l0)) + RB0.MatVec(desSens.aPlank[irb0].l);
    
    const unsigned int irb1 = cjs.irb1;
    const CRigid3D_Ini& ri1 = aRI[irb1];  
    unsigned int idir1 = aFixDir[ifix].second;    
    Com::CMatrix3 Q1 = ri1.GetRotMatrix_Ini();
    Com::CVector3D n1 = Q1.MatVec(aFaceNormal[idir1]);
    Com::CVector3D _n1 = Cross(desSens.aPlank[irb1].r,n1);
    Com::CVector3D l1(ri1.xlen, ri1.ylen, ri1.zlen);
    Com::CMatrix3 B1(0.5*aFaceNormal[idir1].x,0.5*aFaceNormal[idir1].y,0.5*aFaceNormal[idir1].z);
    Com::CMatrix3 RB1 = Q1.MatMat(B1);
    Com::CVector3D b1 = RB1.MatVec(l1)+ri1.GetIniPosCG();
    Com::CVector3D _b1 = desSens.aPlank[irb1].p + Cross(desSens.aPlank[irb1].r,RB1.MatVec(l1)) + RB1.MatVec(desSens.aPlank[irb1].l);

    assert( (idir0==4 || idir0==5) && ( idir1!=4 || idir1!=5 ) );
    if( irb0 == irb ){  // irb1 is neighbor
      const double n0tn1 = Com::Dot(n0,n1);
//      double t = Com::Dot(n0,b0-b1)/Com::Dot(n0,n1);
      const double _t = (Com::Dot(n0,_b0) + Com::Dot(_n0,b0-b1))/n0tn1 - (Com::Dot(n0,b0-b1)*Dot(n1,_n0))/(n0tn1*n0tn1);
      double dlen1 = _t;
      if( idir1 == 0 || idir1 == 1 ){
        desSens.aPlank[irb1].l[0] = dlen1;
        desSens.aPlank[irb1].p = dlen1*0.5*n1;
      }
      else{
        desSens.aPlank[irb1].l[1] = dlen1;
        desSens.aPlank[irb1].p = dlen1*0.5*n1;
      }
      desSens.aJoint[ifix] = n1*dlen1;
    }
    if( irb1 == irb ){  // irb0 is neighbor
      Com::CMatrix3 sns_p,sns_r,sns_l;
      GetSensitivityInitialPointOnRigidBody(sns_p,sns_r,sns_l, ri1,B1.MatVec(l1),B1);
      Com::CVector3D v1 = sns_p.MatVec(desSensPlank.p) + sns_r.MatVec(desSensPlank.r) + sns_l.MatVec(desSensPlank.l);
      const double dlen1 = Com::Dot(v1,n0);
      if( idir1 == 0 || idir1 == 1 ){
        desSens.aPlank[irb1].l[0] += dlen1;
        desSens.aPlank[irb1].p += dlen1*0.5*n1;
      }
      else{
        desSens.aPlank[irb1].l[1] += dlen1;
        desSens.aPlank[irb1].p += dlen1*0.5*n1;
      }      
      GetSensitivityInitialPointOnRigidBody(sns_p,sns_r,sns_l, ri1,B1.MatVec(l1),B1);
      Com::CVector3D v2 = sns_p.MatVec(desSensPlank.p) + sns_r.MatVec(desSensPlank.r) + sns_l.MatVec(desSensPlank.l);
      desSens.aJoint[ifix] += v2;      
    }
  }
}

void ContactPlankSensitivity
(unsigned int irb0,
 int icont_flg,
 CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,      
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const CContactProp& contProp)
{
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };

  unsigned int idir0;
  if(      icont_flg == 1 ){ idir0 = 0; } // +x
  else if( icont_flg == 2 ){ idir0 = 1; } // -x  
  else if( icont_flg == 4 ){ idir0 = 2; }// +y
  else if( icont_flg == 8 ){ idir0 = 3; }// -y
  else{ return; }    
  
  const CRigid3D_Ini& ri0 = aRI[irb0];    
  Com::CMatrix3 R0 = ri0.GetRotMatrix_Ini();
  Com::CVector3D n0 = R0.MatVec(aFaceNormal[idir0]);
  Com::CVector3D l0(ri0.xlen, ri0.ylen, ri0.zlen);    
  Com::CMatrix3 B0(0.5*aFaceNormal[idir0].x,0.5*aFaceNormal[idir0].y,0.5*aFaceNormal[idir0].z);    
  Com::CMatrix3 RB0 = R0.MatMat(B0);
  Com::CVector3D t0 = RB0.MatVec(l0)+ri0.GetIniPosCG();  
  
  Com::CVector3D b0(aFaceNormal[idir0].x,aFaceNormal[idir0].y,aFaceNormal[idir0].z); 
  Com::CVector3D Rb0 = R0.MatVec(b0);
  
  Com::CMatrix3 sns_p,sns_r,sns_l;
  GetSensitivityInitialPointOnRigidBody(sns_p,sns_r,sns_l, ri0,B0.MatVec(l0),B0);
  Com::CVector3D v0 = sns_p.MatVec(desSens.aPlank[irb0].p) + sns_r.MatVec(desSens.aPlank[irb0].r) + sns_l.MatVec(desSens.aPlank[irb0].l);
  const double dlen0 = Com::Dot(v0,contProp.cont_norm);
  
  if( idir0 == 0 || idir0 == 1 ){
    desSens.aPlank[irb0].l[0] += dlen0;
    desSens.aPlank[irb0].p += dlen0*0.5*Rb0;      
  }
  else {
    desSens.aPlank[irb0].l[1] += dlen0;
    desSens.aPlank[irb0].p += dlen0*0.5*Rb0;      
  }  
}

void GetDesignSensitivity
(CDesignSensitivity& desSens,                    
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CJoint_Fix>& aFix,       
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                
 unsigned int irb,
 unsigned int imode,
 unsigned int icont_config, // 0-8
 const CContactProp& contProp)
{ 
  int ijnt_flg = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;      
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
    if( irb0 != irb && irb1 != irb ) continue;
    unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
    if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
    if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
    if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
    if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y        
  }        
  desSens.Initialize(aRI.size(), aFix.size());  
  if( !IsAliveMode(aRI,aFix,aFixDir,irb,imode,icont_config,ijnt_flg) ){ return; }
  unsigned int iflg = icont_config | ijnt_flg;
  if(      imode>=0 && imode<3 ){
    {
      const CRigid3D_Ini& ri = aRI[irb];
      const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
      Com::CVector3D e(0,0,0); e[imode] = 1;
      desSens.aPlank[irb].p = Q.MatVec(e);
    }
    ContactPlankSensitivity(irb,icont_config, desSens,aRI,aFix,aFixDir,contProp);
    NeighboringPlankSensitivity(irb, desSens, aRI,aFix,aFixDir);
//    SetJointSensitivity(desSens,aRI,aFix,aFixDir);
  }
  else if( imode>=3 && imode<6 ){
    const CRigid3D_Ini& ri = aRI[irb];
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();          
    unsigned int iaxis = imode-3;
    {
      Com::CVector3D mov(0,0,0); mov[iaxis] = 1;                    
      desSens.aPlank[irb].r = Q.MatVec( mov);
    }
    if( iaxis == 0 ){
    }        
    if( iaxis == 1 ){
      if( iflg == 2 ){ // -x
        Com::CVector3D Qa = Q.MatVec( Com::CVector3D(0,0,ri.xlen*0.5) );
        desSens.aPlank[irb].p = -Qa;
      }          
      if( iflg == 1 ){ // +x
        Com::CVector3D Qa = Q.MatVec( Com::CVector3D(0,0,ri.xlen*0.5) );
        desSens.aPlank[irb].p = +Qa;              
      }          
    }            
    ContactPlankSensitivity(irb,icont_config, desSens,aRI,aFix,aFixDir,contProp);    
    NeighboringPlankSensitivity(irb, desSens, aRI,aFix,aFixDir);
//    SetJointSensitivity(desSens,aRI,aFix,aFixDir);
  }
  else if( imode>=6 && imode<8 ){    
    const CRigid3D_Ini& ri = aRI[irb];
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();          
    unsigned int imode_l = imode-6;
    desSens.aPlank[irb].l[imode_l] = 1;                    
    if(      icont_config == 4 ){ // +y
    }
    else if( icont_config == 8 ){ // -y
    }        
    else if( icont_config == 1 ){ // +x
      if( imode_l == 0 ){ 
        Com::CVector3D Qa = Q.MatVec( Com::CVector3D(0.5,0,0) );
        desSens.aPlank[irb].p = -Qa*0.5;
        desSens.aPlank[irb].l[imode_l] = 0.5;              
      }
    }
    else if( icont_config == 2 ){ // -x
      if( imode_l == 0 ){     
        Com::CVector3D Qa = Q.MatVec( Com::CVector3D(0.5,0,0) );
        desSens.aPlank[irb].p = +Qa*0.5;              
        desSens.aPlank[irb].l[imode_l] = 0.5;
      }          
    }            
    ContactPlankSensitivity(irb,icont_config, desSens,aRI,aFix,aFixDir,contProp);
    NeighboringPlankSensitivity(irb, desSens, aRI,aFix,aFixDir);    
//    SetJointSensitivity(desSens,aRI,aFix,aFixDir);    
  }  
}

bool IsAliveMode
(const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CJoint_Fix>& aFix,        
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                
 unsigned int irb0,
 unsigned int imode, // trans:0,1,2 rot:3,4,5, len:6,7
 unsigned int icont_flg,
 unsigned int ijnt_flg) // 0-8
{
  unsigned int  iflg = icont_flg | ijnt_flg;
  if(      imode>=0 && imode<3 ){
    if( iflg &  3 && imode==0 ) return false;
    if( iflg & 12 && imode==1 ) return false;                
    return true;
  }
  else if( imode>=3 && imode<6 ){ // rotation
    unsigned int iaxis = imode-3;    
    if( iaxis == 0 ){
      if(   iflg & 4    // +y
         || iflg & 8 ){ // -y
        return true;
      }
      else { return false; }
    }
    if( iaxis == 1 ){
      if(   iflg & 1    // +x
         || iflg & 2 ){ // -x
        return true;
      }
      else { return false; }      
    }
    return false;
  }
  else if( imode>=6 && imode<8 ){ // length
    unsigned int ilen = imode-6;
    const int iflg = ijnt_flg | icont_flg;
    if( (iflg&3 )==3  && ilen == 0 ) return false; // both -x,+x jointed
    if( (iflg&12)==12 && ilen == 1 ) return false; // both -y,+y jointed
    return true;
  }      
  return false;
}


static void myGlVertex3d(const Com::CVector3D& pos){
  ::glVertex3d(pos.x, pos.y, pos.z);
}


void DrawRing(unsigned int ndiv, 
              double radius, double thick,
              const Com::CVector3D& d, const Com::CMatrix3& Q,               
              const Com::CVector3D& ax, const Com::CVector3D& ay, const Com::CVector3D& az )
{
  const double rdiv = 3.1415*2 / ndiv;
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D v0 = sin(rdiv* idiv   )*ax + cos(rdiv* idiv   )*ay;
    Com::CVector3D v1 = sin(rdiv*(idiv+1))*ax + cos(rdiv*(idiv+1))*ay;        
    Com::CVector3D x0 = d + Q.MatVec( radius*v0 + thick*az );
    Com::CVector3D x1 = d + Q.MatVec( radius*v0 - thick*az );
    Com::CVector3D x2 = d + Q.MatVec( radius*v1 - thick*az ); 
    Com::CVector3D x3 = d + Q.MatVec( radius*v1 + thick*az );
    myGlVertex3d(x0);
    myGlVertex3d(x1);
    myGlVertex3d(x2);
    myGlVertex3d(x3);    
  }  
}

void DrawDotRing
(double dr, 
 double radius,
 const Com::CVector3D& d, const Com::CMatrix3& Q,               
 const Com::CVector3D& ax, const Com::CVector3D& ay, const Com::CVector3D& az )
{
  const double ndiv = 3.1415*2 / dr;
  double rdiv = dr;
  ::glBegin(GL_LINES);
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D v0 = sin(rdiv* idiv   )*ax + cos(rdiv* idiv   )*ay;
//    Com::CVector3D v1 = sin(rdiv*(idiv+1))*ax + cos(rdiv*(idiv+1))*ay;        
    Com::CVector3D x0 = d + Q.MatVec( radius*v0 );
//    Com::CVector3D x1 = d + Q.MatVec( radius*v0 - thick*az );
//    Com::CVector3D x2 = d + Q.MatVec( radius*v1 - thick*az ); 
//    Com::CVector3D x3 = d + Q.MatVec( radius*v1 + thick*az );
    myGlVertex3d(x0);
//    myGlVertex3d(x1);
//    myGlVertex3d(x2);
//    myGlVertex3d(x3);    
  }  
  ::glEnd();
}


void DrawRingArrow(unsigned int ndiv, 
                   double radius, double thick,
                   const Com::CVector3D& d, const Com::CMatrix3& Q,               
                   const Com::CVector3D& ax, const Com::CVector3D& ay, const Com::CVector3D& az,
                   double t0, double t1)
{
  double dt = atan2(thick,radius);
  if( t1-t0 < 0 ){ dt*=-1; }
  if( fabs(t1-t0) < 2*fabs(dt) ){ 
    thick = sin(fabs(t0-t1))*radius*0.5;
    dt = atan2(thick,radius);
    if( t1-t0 < 0 ){ dt*=-1; }
  }
  const double tdiv = (t1-dt-t0)/ndiv;
  ::glBegin(GL_QUADS);
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D v0 = sin(t0+tdiv* idiv   )*ax + cos(t0+tdiv* idiv   )*ay;
    Com::CVector3D v1 = sin(t0+tdiv*(idiv+1))*ax + cos(t0+tdiv*(idiv+1))*ay;  
    Com::CVector3D x0 = d + Q.MatVec( radius*v0 + thick*az );
    Com::CVector3D x1 = d + Q.MatVec( radius*v0 - thick*az );
    Com::CVector3D x2 = d + Q.MatVec( radius*v1 - thick*az ); 
    Com::CVector3D x3 = d + Q.MatVec( radius*v1 + thick*az );
    myGlVertex3d(x0);
    myGlVertex3d(x1);
    myGlVertex3d(x2);
    myGlVertex3d(x3);
  }  
  ::glEnd();
  ::glBegin(GL_TRIANGLES);
  {
    double t2 = t1-dt;
    Com::CVector3D v0 = sin(t2)*ax + cos(t2)*ay;
    Com::CVector3D x0 = d + Q.MatVec( radius*v0 + 2*thick*az );
    Com::CVector3D x1 = d + Q.MatVec( radius*v0 - 2*thick*az );        
    double t3 = t1+2*dt;    
    Com::CVector3D v2 = sin(t3)*ax + cos(t3)*ay;
    Com::CVector3D x2 = d + Q.MatVec( radius*v2 );            
    myGlVertex3d(x0);
    myGlVertex3d(x1);
    myGlVertex3d(x2);    
  }
  ::glEnd();
}

void DrawRingArrow2
(unsigned int ndiv, 
 double radius, double thick,
 const Com::CVector3D& d, const Com::CMatrix3& Q,               
 const Com::CVector3D& ax, const Com::CVector3D& ay, const Com::CVector3D& az,
 double t0, double t1)
{
  double dt = atan2(thick,radius);
  if( t1-t0 < 0 ){ dt*=-1; }
  if( fabs(t1-t0) < 2*fabs(dt) ){ 
    thick = sin(fabs(t0-t1))*radius*0.5;
    dt = atan2(thick,radius);
    if( t1-t0 < 0 ){ dt*=-1; }
  }
  const double tdiv = (t1-dt-t0)/ndiv;
  unsigned int nrdiv = 16;
  const double rdiv = 3.1415*2/nrdiv;
  ::glBegin(GL_QUADS);
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D v0 = sin(t0+tdiv* idiv   )*ax + cos(t0+tdiv* idiv   )*ay;
    Com::CVector3D v1 = sin(t0+tdiv*(idiv+1))*ax + cos(t0+tdiv*(idiv+1))*ay;
//    Com::CVector3D an = sin(t0+tdiv* idiv   )*ax + cos(t0+tdiv* idiv   )*ay;
    for(unsigned int jdiv=0;jdiv<nrdiv;jdiv++){
      Com::CVector3D w0 = radius*v0 + thick*sin(rdiv* jdiv   )*v0 + thick*cos(rdiv* jdiv   )*az;
      Com::CVector3D w1 = radius*v0 + thick*sin(rdiv*(jdiv+1))*v0 + thick*cos(rdiv*(jdiv+1))*az;
      Com::CVector3D w2 = radius*v1 + thick*sin(rdiv* jdiv   )*v1 + thick*cos(rdiv* jdiv   )*az;
      Com::CVector3D w3 = radius*v1 + thick*sin(rdiv*(jdiv+1))*v1 + thick*cos(rdiv*(jdiv+1))*az;      
      Com::CVector3D x0 = d + Q.MatVec( w0 );
      Com::CVector3D x1 = d + Q.MatVec( w1 );
      Com::CVector3D x2 = d + Q.MatVec( w2 ); 
      Com::CVector3D x3 = d + Q.MatVec( w3 );
      myGlVertex3d(x0);
      myGlVertex3d(x1);
      myGlVertex3d(x3);
      myGlVertex3d(x2);      
    }
  }  
  ::glEnd();
  ::glBegin(GL_TRIANGLES);
  {
    double t2 = t1-dt;
    Com::CVector3D v0 = sin(t2)*ax + cos(t2)*ay;
//    Com::CVector3D x0 = d + Q.MatVec( radius*v0 + 2*thick*az );
//    Com::CVector3D x1 = d + Q.MatVec( radius*v0 - 2*thick*az );        
    double t3 = t1+2*dt;    
    Com::CVector3D v1 = sin(t3)*ax + cos(t3)*ay;
//    Com::CVector3D x2 = d + Q.MatVec( radius*v2 );            
    for(unsigned int jdiv=0;jdiv<nrdiv;jdiv++){
      Com::CVector3D w0 = radius*v0 + 2*thick*sin(rdiv* jdiv   )*v0 + 2*thick*cos(rdiv* jdiv   )*az;
      Com::CVector3D w1 = radius*v0 + 2*thick*sin(rdiv*(jdiv+1))*v0 + 2*thick*cos(rdiv*(jdiv+1))*az;
      Com::CVector3D w2 = radius*v1;
      Com::CVector3D x0 = d + Q.MatVec( w0 );
      Com::CVector3D x1 = d + Q.MatVec( w1 );      
      Com::CVector3D x2 = d + Q.MatVec( w2 );            
      myGlVertex3d(x0);
      myGlVertex3d(x1);
      myGlVertex3d(x2);    
    }
  }
  {
    Com::CVector3D v0 = sin(t0)*ax + cos(t0)*ay;
    for(unsigned int jdiv=0;jdiv<nrdiv;jdiv++){
      Com::CVector3D w0 = radius*v0 + thick*sin(rdiv* jdiv   )*v0 + thick*cos(rdiv* jdiv   )*az;
      Com::CVector3D w1 = radius*v0 + thick*sin(rdiv*(jdiv+1))*v0 + thick*cos(rdiv*(jdiv+1))*az;
      Com::CVector3D x0 = d + Q.MatVec( radius*v0 );
      Com::CVector3D x1 = d + Q.MatVec( w0 );      
      Com::CVector3D x2 = d + Q.MatVec( w1 );                  
      myGlVertex3d(x0);
      myGlVertex3d(x1);
      myGlVertex3d(x2);          
    }
  }
  {
    Com::CVector3D v0 = sin(t1-dt)*ax + cos(t1-dt)*ay;
    for(unsigned int jdiv=0;jdiv<nrdiv;jdiv++){
      Com::CVector3D w0 = radius*v0 + 2*thick*sin(rdiv* jdiv   )*v0 + 2*thick*cos(rdiv* jdiv   )*az;
      Com::CVector3D w1 = radius*v0 + 2*thick*sin(rdiv*(jdiv+1))*v0 + 2*thick*cos(rdiv*(jdiv+1))*az;
      Com::CVector3D x0 = d + Q.MatVec( radius*v0 );
      Com::CVector3D x1 = d + Q.MatVec( w0 );      
      Com::CVector3D x2 = d + Q.MatVec( w1 );                  
      myGlVertex3d(x0);
      myGlVertex3d(x1);
      myGlVertex3d(x2);          
    }
  }  
  ::glEnd();
}



bool PickRing
(double& depth, 
 unsigned int ndiv, 
 double radius, double dist_threthold,
 const Com::CVector3D& d, const Com::CMatrix3& Q,              
 const Com::CVector3D& ax, const Com::CVector3D& ay,
 const Com::CVector3D& src, const Com::CVector3D& dir)
{
  const double rdiv = 3.1415*2 / ndiv;
  bool flg = false;
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D v0 = sin(rdiv* idiv   )*ax + cos(rdiv* idiv   )*ay;
    const Com::CVector3D x0 = d + Q.MatVec(radius*v0);
    double depth0 = -Dot(src-x0,dir)/Dot(dir,dir);
    double dist0  = Com::Length(src+depth0*dir-x0);
    std::cout << "dist threth depth" << dist0 << " " << dist_threthold << " " << depth << std::endl;
    if( dist0 < dist_threthold ){
      if( !flg ){
        flg = true;
        depth = depth0;
      }
      else{
        if( depth0 < depth ){
          depth = depth0;
        }
      }
    }
  }
  return flg;
  /*
   for(unsigned int idiv=0;idiv<ndiv;idiv++){
   Com::CVector3D v0 = sin(rdiv* idiv   )*ax + cos(rdiv* idiv   )*ay;
   Com::CVector3D v1 = sin(rdiv*(idiv+1))*ax + cos(rdiv*(idiv+1))*ay;        
   Com::CVector3D x0 = d + Q.MatVec( radius*v0 + thick*az );
   Com::CVector3D x1 = d + Q.MatVec( radius*v0 - thick*az );
   Com::CVector3D x2 = d + Q.MatVec( radius*v1 - thick*az ); 
   Com::CVector3D x3 = d + Q.MatVec( radius*v1 + thick*az );
   glVertex3d(x0.x,x0.y,x0.z);
   glVertex3d(x1.x,x1.y,x1.z);
   glVertex3d(x2.x,x2.y,x2.z);
   glVertex3d(x3.x,x3.y,x3.z);
   }  
   */
}



void DrawArrow(const Com::CVector3D& pos0, 
               const Com::CVector3D& pos1,
               double rad)
{
  bool is_lighting = glIsEnabled(GL_LIGHTING);
  ::glDisable(GL_LIGHTING);
  Com::CVector3D n = pos1-pos0; 
  double len_tot = n.Length();
  if( len_tot < 1.0e-20 ) return;
  n.SetNormalizedVector();
  Com::CVector3D h0,h1;
  GetVertical2Vector(n, h0, h1);
  unsigned int ndiv = 32;
  double rdiv = 3.1415*2/ndiv;
  ::glBegin(GL_QUADS);
  Com::CVector3D pos2 = pos1 + rad*n;
  double len_cylin = len_tot-rad;  
  if( len_cylin < rad ){
    rad = (len_tot+rad)/3;
    len_cylin = rad;
  }
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D a0 = pos0+rad*sin(rdiv*(idiv+0))*h0 + rad*cos(rdiv*(idiv+0))*h1;
    Com::CVector3D a1 = pos0+rad*sin(rdiv*(idiv+1))*h0 + rad*cos(rdiv*(idiv+1))*h1;
    Com::CVector3D a2 = a1 + len_cylin*n;
    Com::CVector3D a3 = a0 + len_cylin*n;    
    myGlVertex3d(a0);
    myGlVertex3d(a1);
    myGlVertex3d(a2);
    myGlVertex3d(a3);    
  }
  ::glEnd();
  ::glBegin(GL_TRIANGLES);
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    double rad2 = rad*1.5;
    Com::CVector3D a0 = pos0+rad2*sin(rdiv*(idiv+0))*h0 + rad2*cos(rdiv*(idiv+0))*h1;
    Com::CVector3D a1 = pos0+rad2*sin(rdiv*(idiv+1))*h0 + rad2*cos(rdiv*(idiv+1))*h1;
    Com::CVector3D a2 = a1 + len_cylin*n;
    Com::CVector3D a3 = a0 + len_cylin*n;    
    myGlVertex3d(a2);
    myGlVertex3d(a3);  
    myGlVertex3d(pos2);
  }
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D a0 = pos0+rad*sin(rdiv*(idiv+0))*h0 + rad*cos(rdiv*(idiv+0))*h1;
    Com::CVector3D a1 = pos0+rad*sin(rdiv*(idiv+1))*h0 + rad*cos(rdiv*(idiv+1))*h1;
    myGlVertex3d(a0);
    myGlVertex3d(a1);
    myGlVertex3d(pos0);    
    double rad2 = rad*1.5;    
    Com::CVector3D b0 = pos0+rad2*sin(rdiv*(idiv+0))*h0 + rad2*cos(rdiv*(idiv+0))*h1;
    Com::CVector3D b1 = pos0+rad2*sin(rdiv*(idiv+1))*h0 + rad2*cos(rdiv*(idiv+1))*h1;    
    myGlVertex3d(b0+len_cylin*n);
    myGlVertex3d(b1+len_cylin*n);    
    myGlVertex3d(pos0+len_cylin*n);    
  }
  ::glEnd();
  if( is_lighting ){ ::glEnable(GL_LIGHTING); }
}



void DrawRigidBody_Ini_ModeDist
(unsigned int irb,
 const std::vector<CRigid3D_Ini>& aRI,                         
 const std::vector<CJoint_Fix>& aFix,       
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 const CContactPoints_RigidBody3D& aCP,
 int imode_edit,
 double tr,
 double tmode, 
 double hvh)  // -1 none,  
{  
  const bool is_light = glIsEnabled(GL_LIGHTING);
  ::glDisable(GL_LIGHTING);
  int icont_flg = 0;
  if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x  
  if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x
  if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y  
  if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y
  int ijnt_flg = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;      
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
    if( irb0 != irb && irb1 != irb ) continue;
    unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
    if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
    if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
    if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
    if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y 
  }
  unsigned int iflg = ijnt_flg | icont_flg;
  const CRigid3D_Ini& ri = aRI[irb];
  if( imode_edit >= 0 && imode_edit < 3 ){ // 0,1,2
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
    const Com::CVector3D& cg = ri.GetIniPosCG();
    Com::CVector3D e(0,0,0); e[imode_edit] = tr;
    e = Q.MatVec(e);
    ::glColor3d(0.3,0.3,1);
    DrawArrow(cg-tmode*e,cg-e*ri.zlen*0.8,ri.zlen);
  }
  if( imode_edit >= 3 && imode_edit < 6 ){ // 3,4,5
    unsigned int iaxis = imode_edit-3;
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();          
    unsigned int ndiv = 12;
    double thick = ri.zlen;
    double ave = (ri.xlen + ri.ylen + ri.zlen)/8;
    double radius = ave*(1.0 + iaxis*0.05);        
    Com::CVector3D aAxis[3][3] = {
      { Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), Com::CVector3D(1,0,0) },
      { Com::CVector3D(0,0,1), Com::CVector3D(1,0,0), Com::CVector3D(0,1,0) },
      { Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1) } };
    ::glColor3d(0.3,0.3,1);
    if( iflg == 2 && iaxis == 1 ){      
      DrawRingArrow2(ndiv*2,ri.xlen*1.02,ri.zlen,
                    d-Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                    aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],
                    tmode*tr,0);
    }
    else if( iflg == 1 && iaxis == 1 ){
      DrawRingArrow2(ndiv*2,ri.xlen*1.02,ri.zlen,  
                    d+Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                    -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],
                    tmode*tr,0);
    }        
    else{
      double dt = atan2(ri.zlen,radius);
      if( tmode < 0 ){ dt*=-1; }
      DrawRingArrow2(ndiv,radius,ri.zlen,
                     d,Q,
                     aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],
                     tmode*tr+dt,0+dt);      
      DrawRingArrow2(ndiv,radius,ri.zlen,
                     d,Q,
                     -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],
                     tmode*tr+dt,0+dt);    
      ::glLineWidth(2);
      DrawDotRing(0.1,radius,
                  d,Q,
                  aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2]);
    }
  }
  if( imode_edit >= 6  ){ // 6,7
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();    
    const Com::CMatrix3 L(ri.xlen*0.5, ri.ylen*0.5, ri.zlen*0.5);    
    if( imode_edit == 6 ){ // x
      Com::CVector3D e(0,0,0); e[0] = tr; e = Q.MatVec(e);          
      ::glColor3d(0.3,0.3,1);      
      if(      (iflg&2) && !(iflg&1) ){ // touch -x
        const Com::CVector3D v0 = d + Q.MatVec( L.MatVec( Com::CVector3D(+1.1,0,0) ) );
        DrawArrow(v0-tmode*e*0.5,v0,ri.zlen);
      }
      else if( (iflg&1) && !(iflg&2) ){ // touch +x
        const Com::CVector3D v1 = d + Q.MatVec( L.MatVec( Com::CVector3D(-1.1,0,0) ) );
        DrawArrow(v1+tmode*e*0.5,v1,ri.zlen);        
      }
      else {
        const Com::CVector3D v0 = d + Q.MatVec( L.MatVec( Com::CVector3D(+1.1,0,0) ) );
        DrawArrow(v0-tmode*e*0.5,v0,ri.zlen);      
        const Com::CVector3D v1 = d + Q.MatVec( L.MatVec( Com::CVector3D(-1.1,0,0) ) );
        DrawArrow(v1-tmode*e*0.5,v1,ri.zlen);                
      }
    }
    else if( imode_edit == 7 ){
      Com::CVector3D e(0,0,0); e[1] = tr; e = Q.MatVec(e);          
      ::glColor3d(0.3,0.3,1);      
      const Com::CVector3D v0 = d + Q.MatVec( L.MatVec( Com::CVector3D(0,+1.1,0) ) );
      DrawArrow(v0-tmode*e*0.5,v0,ri.zlen);      
      const Com::CVector3D v1 = d + Q.MatVec( L.MatVec( Com::CVector3D(0,-1.1,0) ) );
      DrawArrow(v1+tmode*e*0.5,v1,ri.zlen);                
    }
    /*
    for(unsigned int i=0;i<4;i++){
      if( (icont_flg & 2) && i == 0 ) continue; // touch -x
      if( (icont_flg & 1) && i == 1 ) continue; // touch +x      
      if( (icont_flg & 8) && i == 2 ) continue; // touch -y
      if( (icont_flg & 4) && i == 3 ) continue; // touch +y
      if( (ijnt_flg & 2) &&  i == 0 ) continue;
      if( (ijnt_flg & 1) &&  i == 1 ) continue;        
      if( (ijnt_flg & 8) &&  i == 2 ) continue;
      if( (ijnt_flg & 4) &&  i == 3 ) continue;

      const Com::CVector3D v = d + Q.MatVec( B.MatVec( Com::CVector3D(tabl[i][0]*0.9,tabl[i][1]*0.9,0) ) );
      ::glPushMatrix();      
      ::glTranslated(v.x,v.y,v.z);    
      if( imode_edit - 6 == i ){ ::glColor3d(0.5, 0.5, 1.0); }
      else{                      ::glColor3d(0.3, 0.3, 1.0); }        
      ::glutSolidSphere(hvh*0.05, 24,24); 
      ::glPopMatrix();          
    }    
     */
  }
  
  if( !is_light ){ ::glDisable(GL_LIGHTING); }
  else{ ::glEnable(GL_LIGHTING); }
}


void DrawRigidBody_Ini_BoundMode
(unsigned int irb,
 const std::vector<CRigid3D_Ini>& aRI,                         
 const std::vector<CJoint_Fix>& aFix,       
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 const CContactPoints_RigidBody3D& aCP,
 int imode_edit,
 double tr,
 double tmode, double tmin, double tmax,
 double hvh,
 bool is_range_ok)  // -1 none,  
{  
  const bool is_light = glIsEnabled(GL_LIGHTING);
  ::glDisable(GL_LIGHTING);
  int icont_flg = 0;
  if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x  
  if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x
  if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y  
  if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y
  int ijnt_flg = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;      
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
    if( irb0 != irb && irb1 != irb ) continue;
    unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
    if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
    if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
    if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
    if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y 
  }
  unsigned int iflg = ijnt_flg | icont_flg;
  const CRigid3D_Ini& ri = aRI[irb];
  if( imode_edit >= 0 && imode_edit < 3 ){
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
    const Com::CVector3D& cg = ri.GetIniPosCG();
    Com::CVector3D e(0,0,0); e[imode_edit] = tr;
    e = Q.MatVec(e);
    if( is_range_ok ){ ::glColor3d(0,0,0); }
    else{ ::glColor3d(1,0,0); }
    double tmode0 = (tmin+tmax)*0.5;
//    if( tmode < tmax ){ DrawArrow(cg,cg+(tmax-tmode0)*e,ri.zlen); }
//    if( tmode > tmin ){ DrawArrow(cg,cg+(tmin-tmode0)*e,ri.zlen); }
    DrawArrow(cg+(tmode0-tmode)*e,cg+(tmax-tmode)*e,ri.zlen);
    DrawArrow(cg+(tmode0-tmode)*e,cg+(tmin-tmode)*e,ri.zlen);  
    {
      double span = hvh*0.03;
      unsigned int ndiv = 10/span;
      ::glLineWidth(2);
      ::glBegin(GL_LINES);
      for(unsigned int idiv=0;idiv<ndiv;idiv++){
        Com::CVector3D v0 = cg+(idiv-ndiv*0.5)*e*span;
        myGlVertex3d(v0);
      }
      ::glEnd();
    }  
  }
  if( imode_edit >= 3 && imode_edit < 6 ){
    unsigned int iaxis = imode_edit-3;
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();          
    unsigned int ndiv = 12;
    double thick = ri.zlen;
    double ave = (ri.xlen + ri.ylen + ri.zlen)/8;
    double radius = ave*(1.0 + iaxis*0.05);        
    Com::CVector3D aAxis[3][3] = {
      { Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), Com::CVector3D(1,0,0) },
      { Com::CVector3D(0,0,1), Com::CVector3D(1,0,0), Com::CVector3D(0,1,0) },
      { Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1) } };
    if( is_range_ok ){ ::glColor3d(0,0,0); }
    else{ ::glColor3d(1,0,0); }
    double tmode0 = (tmin+tmax)*0.5;    
    if( iflg == 2 && iaxis == 1 ){      
      /*
      if( tmode < tmax ){
        DrawRingArrow2(ndiv*2,ri.xlen*1.03,ri.zlen,
                      d-Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                      aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],
                      0,-(tmax-tmode0)*tr);
      }
      if( tmode > tmin ){
        DrawRingArrow2(ndiv*2,ri.xlen*1.03,ri.zlen,
                      d-Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                      aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],
                      0,-(tmin-tmode0)*tr);        
      }*/
      DrawRingArrow2(ndiv*2,ri.xlen*1.03,ri.zlen,
                     d-Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                     aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],
                     -(tmode0-tmode)*tr,-(tmax-tmode)*tr);
      DrawRingArrow2(ndiv*2,ri.xlen*1.03,ri.zlen,
                     d-Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                     aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],
                     -(tmode0-tmode)*tr,-(tmin-tmode)*tr);        
    }
    else if( iflg == 1 && iaxis == 1 ){
      /*
      if( tmode < tmax ){       
        DrawRingArrow2(ndiv*2,ri.xlen*1.03,ri.zlen,  
                      d+Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                      -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],
                      0,-(tmax-tmode)*tr);            
      }
      if( tmode > tmin ){      
        DrawRingArrow2(ndiv*2,ri.xlen*1.03,ri.zlen,  
                      d+Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                      -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],
                      0,-(tmin-tmode)*tr);                    
      }*/
      DrawRingArrow2(ndiv*2,ri.xlen*1.03,ri.zlen,  
                     d+Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                     -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],
                     -(tmode0-tmode)*tr,-(tmax-tmode)*tr);            
      DrawRingArrow2(ndiv*2,ri.xlen*1.03,ri.zlen,  
                     d+Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                     -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],
                     -(tmode0-tmode)*tr,-(tmin-tmode)*tr);                    
    }    
    else{
      double dt = atan2(ri.zlen,radius);
      double dt0 = dt*0.5;
      /*
      if( tmode < tmax ){       
        if( (tmax-tmode) < 0 ){ dt0*=-1; }
        DrawRingArrow2(ndiv,radius,ri.zlen,
                       d,Q,
                       aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],
                       0-dt0,-(tmax-tmode)*tr-dt0);            
        DrawRingArrow2(ndiv,radius,ri.zlen,
                       d,Q,
                       -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],                     
                       0-dt0,-(tmax-tmode)*tr-dt0);           
      }
      if( tmode > tmin ){   
        if( (tmin-tmode) < 0 ){ dt0*=-1; }
        DrawRingArrow2(ndiv,radius,ri.zlen,
                       d,Q,
                       aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],                     
                       0-dt0,-(tmin-tmode)*tr-dt0);
        DrawRingArrow2(ndiv,radius,ri.zlen,
                       d,Q,
                       -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],                     
                       0-dt0,-(tmin-tmode)*tr-dt0);                                    
      }
       */
      DrawRingArrow2(ndiv,radius,ri.zlen,
                     d,Q,
                     aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],
                     -(tmode0-tmode)*tr-dt0,-(tmax-tmode)*tr-dt0);            
      DrawRingArrow2(ndiv,radius,ri.zlen,
                     d,Q,
                     -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],                     
                     -(tmode0-tmode)*tr-dt0,-(tmax-tmode)*tr-dt0);                 
      DrawRingArrow2(ndiv,radius,ri.zlen,
                     d,Q,
                     aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2],                     
                     -(tmode0-tmode)*tr-dt0,-(tmin-tmode)*tr-dt0);
      DrawRingArrow2(ndiv,radius,ri.zlen,
                     d,Q,
                     -aAxis[iaxis][0],-aAxis[iaxis][1],-aAxis[iaxis][2],                     
                     -(tmode0-tmode)*tr-dt0,-(tmin-tmode)*tr-dt0);                                          
      ::glLineWidth(2);
      DrawDotRing(0.1,radius,
                  d,Q,
                  aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2]);      
    }
  }
  if( imode_edit >= 6  ){ // 6,7,8,9
    if( is_range_ok ){ ::glColor3d(0,0,0); }
    else{ ::glColor3d(1,0,0); }
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();    
    const Com::CMatrix3 L(ri.xlen*0.5, ri.ylen*0.5, ri.zlen*0.5);    
    double tmode0 = (tmin+tmax)*0.5;        
    if(      imode_edit == 6 ){
      Com::CVector3D e(0,0,0); e[0] = tr; e = Q.MatVec(e);          
      const Com::CVector3D v0 = d + Q.MatVec( L.MatVec( Com::CVector3D(-1.0,0,0) ) );      
      /*
      if( tmode < tmax ){ DrawArrow(v0,v0-(tmax-tmode)*e*0.5,ri.zlen); }
      if( tmode > tmin ){ DrawArrow(v0,v0-(tmin-tmode)*e*0.5,ri.zlen); }
       */
      DrawArrow(v0-(tmode0-tmode)*e*0.5,v0-(tmax-tmode)*e*0.5,ri.zlen);
      DrawArrow(v0-(tmode0-tmode)*e*0.5,v0-(tmin-tmode)*e*0.5,ri.zlen);
    }
    else if( imode_edit == 7){ // x
      Com::CVector3D e(0,0,0); e[0] = tr; e = Q.MatVec(e);          
      const Com::CVector3D v0 = d + Q.MatVec( L.MatVec( Com::CVector3D(+1.0,0,0) ) );      
      /*
      if( tmode < tmax ){ DrawArrow(v0,v0+(tmax-tmode)*e*0.5,ri.zlen); }
      if( tmode > tmin ){ DrawArrow(v0,v0+(tmin-tmode)*e*0.5,ri.zlen); }
       */
      DrawArrow(v0+(tmode0-tmode)*e*0.5,v0+(tmax-tmode)*e*0.5,ri.zlen);
      DrawArrow(v0+(tmode0-tmode)*e*0.5,v0+(tmin-tmode)*e*0.5,ri.zlen);
    }
    else if( imode_edit == 8 ){
      Com::CVector3D e(0,0,0); e[1] = tr; e = Q.MatVec(e);          
      const Com::CVector3D v1 = d + Q.MatVec( L.MatVec( Com::CVector3D(0,-1.0,0) ) );      
      /*
      if( tmode < tmax ){ DrawArrow(v1,v1-(tmax-tmode)*0.5*e,ri.zlen); }
      if( tmode > tmin ){ DrawArrow(v1,v1-(tmin-tmode)*0.5*e,ri.zlen); }
       */
      DrawArrow(v1-(tmode0-tmode)*e*0.5,v1-(tmax-tmode)*0.5*e,ri.zlen);
      DrawArrow(v1-(tmode0-tmode)*e*0.5,v1-(tmin-tmode)*0.5*e,ri.zlen);
    }
    else if( imode_edit == 9 ){
      Com::CVector3D e(0,0,0); e[1] = tr; e = Q.MatVec(e);          
      const Com::CVector3D v0 = d + Q.MatVec( L.MatVec( Com::CVector3D(0,+1.0,0) ) );      
//      if( tmode < tmax ){ DrawArrow(v0,v0+(tmax-tmode)*0.5*e,ri.zlen); }
//      if( tmode > tmin ){ DrawArrow(v0,v0+(tmin-tmode)*0.5*e,ri.zlen); }
      DrawArrow(v0+(tmode0-tmode)*e,v0+(tmax-tmode)*0.5*e,ri.zlen);
      DrawArrow(v0+(tmode0-tmode)*e,v0+(tmin-tmode)*0.5*e,ri.zlen);      
    }    
  }
  
  if( !is_light ){ ::glDisable(GL_LIGHTING); }
  else{ ::glEnable(GL_LIGHTING); }
}

void DrawRigidBody_Ini_Mode
(unsigned int irb,
 const std::vector<CRigid3D_Ini>& aRI,                         
 const std::vector<CJoint_Fix>& aFix,       
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 const CContactPoints_RigidBody3D& aCP,
 int imode_edit,
 double hvh)  // -1 none,  
{  
  const bool is_light = glIsEnabled(GL_LIGHTING);
  ::glDisable(GL_LIGHTING);
  int icont_flg = 0;
  if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x  
  if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x
  if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y  
  if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y
  int ijnt_flg = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;      
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
    if( irb0 != irb && irb1 != irb ) continue;
    unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
    if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
    if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
    if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
    if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y 
  }
  unsigned int iflg = ijnt_flg | icont_flg;
  const CRigid3D_Ini& ri = aRI[irb];
  if( imode_edit == -2 || ( imode_edit >= 0 && imode_edit < 3) ){ // translation
    double tabp[6][3] = { 
      {-1,  0,  0}, 
      {+1,  0,  0}, 
      { 0, -1,  0}, 
      { 0, +1,  0},
      { 0,  0, -1},
      { 0,  0, +1} };        
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
    for(unsigned int i=0;i<6;i++){
      if( iflg &  3 && (i==0||i==1) ) continue;
      if( iflg & 12 && (i==2||i==3) ) continue;      
      // color
      if(   ( imode_edit == 0  && ( i==0 || i==1 ) ) 
         || ( imode_edit == 1  && ( i==2 || i==3 ) )  
         || ( imode_edit == 2  && ( i==4 || i==5 ) ) ){
        ::glColor3d(0.5, 0.5, 1.0); 
      }
      else{ ::glColor3d(0.3, 0.3, 1.0); }        
      // transation
      double ave = (ri.xlen + ri.ylen + ri.zlen)/10;
      Com::CVector3D v = d + Q.MatVec( Com::CVector3D( tabp[i][0], tabp[i][1], tabp[i][2] ) )*ave;
      ::glPushMatrix();        
      ::glTranslated(v.x,v.y,v.z);               
      ::glutSolidSphere(hvh*0.05, 24,24);         
      ::glPopMatrix();                
      ::glColor3d(0,0,0);
      ::glLineWidth(3);
      ::glBegin(GL_LINES);
      ::glVertex3d(d.x,d.y,d.z);
      ::glVertex3d(v.x,v.y,v.z);        
      ::glEnd();
    }
  }  
  if( imode_edit == -3 || ( imode_edit >= 3 && imode_edit < 6 ) ){
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();          
    unsigned int ndiv = 12;
    double thick = ri.zlen;
    double ave = (ri.xlen + ri.ylen + ri.zlen)/8;
    Com::CVector3D aAxis[3][3] = {
      { Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), Com::CVector3D(1,0,0) },
      { Com::CVector3D(0,0,1), Com::CVector3D(1,0,0), Com::CVector3D(0,1,0) },
      { Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1) } };
    ::glBegin(GL_QUADS);
    for(unsigned int iaxis=0;iaxis<3;iaxis++){
      if( !IsAliveMode(aRI,aFix,aFixDir,irb,3+iaxis,icont_flg,ijnt_flg) ){ continue; }
      double radius = ave*(1.0 + iaxis*0.05);
      if( iaxis + 3 == imode_edit ){ ::glColor3d(0.5,0.5,1.0); }
      else {                         ::glColor3d(0.3,0.3,1.0); }
      if( iflg == 8 ){ // -y
        if( iaxis != 0 ) continue;          
        else {          
          continue;
        }
      }
      if( iflg == 4 ){ // +y
        if( iaxis != 0 ) continue;          
        else {     
          continue;
        }          
      }        
      if( iflg == 2  ){ // -x
        if( iaxis != 1 ) continue;
        else {     
          DrawRing(ndiv*2,ri.xlen,ri.ylen*0.3,
                   d-Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                   aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2]);                        
          continue;
        }          
      }
      if( iflg == 1 ){ // +x
        if( iaxis != 1 ) continue;          
        else {     
          DrawRing(ndiv*2,ri.xlen,ri.ylen*0.3,  
                   d+Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                   aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2]);            
          continue;
        }          
      }        
      DrawRing(ndiv,radius,thick,
               d,Q,
               aAxis[iaxis][0],aAxis[iaxis][1],aAxis[iaxis][2]);
    }
    ::glEnd();             
  }
  if( imode_edit == -4 || imode_edit >= 6  ){ // 0-10
    double tabl[4][2] = { {-1, 0}, {+1, 0}, { 0,-1}, { 0,+1} };
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();    
    const Com::CMatrix3 B(ri.xlen*0.5, ri.ylen*0.5, ri.zlen*0.5);
    for(unsigned int i=0;i<4;i++){
      if( (icont_flg & 2) && i == 0 ) continue; // touch -x
      if( (icont_flg & 1) && i == 1 ) continue; // touch +x      
      if( (icont_flg & 8) && i == 2 ) continue; // touch -y
      if( (icont_flg & 4) && i == 3 ) continue; // touch +y
      if( (ijnt_flg & 2) &&  i == 0 ) continue;
      if( (ijnt_flg & 1) &&  i == 1 ) continue;        
      if( (ijnt_flg & 8) &&  i == 2 ) continue;
      if( (ijnt_flg & 4) &&  i == 3 ) continue;
      if( !IsAliveMode(aRI,aFix,aFixDir,irb,6+i/2,icont_flg,ijnt_flg) ){ continue; }      
      const Com::CVector3D v = d + Q.MatVec( B.MatVec( Com::CVector3D(tabl[i][0]*0.9,tabl[i][1]*0.9,0) ) );
      ::glPushMatrix();      
      ::glTranslated(v.x,v.y,v.z);    
      if( imode_edit - 6 == i ){ ::glColor3d(0.5, 0.5, 1.0); }
      else{                      ::glColor3d(0.3, 0.3, 1.0); }        
      ::glutSolidSphere(hvh*0.05, 24,24); 
      ::glPopMatrix();          
    }
  }
  
  if( !is_light ){ ::glDisable(GL_LIGHTING); }
  else{ ::glEnable(GL_LIGHTING); }
}


int PickRigidBodyEditMode
(const Com::CVector3D& src, const Com::CVector3D& dir,
 const std::vector<CRigid3D_Ini>& aRI,
 const CContactPoints_RigidBody3D& aCP,
 unsigned int irb,
 const std::vector<CJoint_Fix>& aFix,       
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,  
 PARAM_TYPE itype,
 double hvh)
{
  //  const CRigid3D_Ini& ri = rb.ri;      
  int icont_flg = 0;
  if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x  
  if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x
  if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y  
  if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y
  int ijnt_flg = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;      
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
    if( irb0 != irb && irb1 != irb ) continue;
    unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
    if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
    if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
    if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
    if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y        
  }        
  unsigned int iflg = ijnt_flg | icont_flg;  
  const CRigid3D_Ini& ri = aRI[irb];
  if( itype == BOX_POS ){
    double tabp[6][3] = { 
      {-1,  0,  0}, 
      {+1,  0,  0}, 
      { 0, -1,  0}, 
      { 0, +1,  0},
      { 0,  0, -1},
      { 0,  0, +1} };        
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
    double min_depth = -1;
    int imin = -1;
    for(unsigned int i=0;i<6;i++){
      if( !IsAliveMode(aRI,aFix,aFixDir,irb,i/2,icont_flg,ijnt_flg) ) continue;
      if( iflg &  3 && (i==0||i==1) ) continue;
      if( iflg & 12 && (i==2||i==3) ) continue;            
      double ave = (ri.xlen + ri.ylen + ri.zlen)/10;
      Com::CVector3D v = d + Q.MatVec( Com::CVector3D( tabp[i][0], tabp[i][1], tabp[i][2] ) )*ave;
      double t = -Dot(src-v,dir)/Dot(dir,dir);
      double dist = Com::Length(src+t*dir-v);
      if( dist < hvh*0.05 ){
        if( imin == -1 ){ 
          imin = i;           
          min_depth = t; 
        }
        else if( t < min_depth ){
          imin = i;
          min_depth = t;
        } 
      }
    }  
    if( imin != -1 ) return imin/2;
    return -2;
  }
  if( itype == BOX_ROT ){
    double ave = (ri.xlen + ri.ylen + ri.zlen)/8;    
    unsigned int ndiv = 2*3.1415*ave/ri.zlen;
    double rdiv = 3.14115*2/ndiv;
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();        
    double mindepth = 0;
    int nearestaxis = -3;
    Com::CVector3D aAxis[3][3] = {
      { Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), Com::CVector3D(1,0,0) },
      { Com::CVector3D(0,0,1), Com::CVector3D(1,0,0), Com::CVector3D(0,1,0) },
      { Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1) } };    
    for(unsigned int iaxis=0;iaxis<3;iaxis++){   
      if( !IsAliveMode(aRI,aFix,aFixDir,irb,iaxis+3,icont_flg,ijnt_flg) ) continue;
      double radius = ave*(1+iaxis*0.05);
      double depth;
      bool res;
      if( iflg == 4 && iaxis == 0 ){
      }
      else if( iflg == 8 && iaxis == 0 ){
      }        
      else if( iflg == 2 && iaxis == 1 ){
        res = PickRing(depth,
                       ndiv*2,ri.xlen,ri.ylen*0.15,
                       d-Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                       aAxis[iaxis][0],aAxis[iaxis][1],
                       src,dir);
      }
      else if( iflg == 1 && iaxis == 1 ){
        res = PickRing(depth,
                       ndiv*2,ri.xlen,ri.ylen*0.15,  
                       d+Q.MatVec( Com::CVector3D(ri.xlen*0.5,0,0) ),Q,
                       aAxis[iaxis][0],aAxis[iaxis][1],
                       src,dir);            
      }              
      else{
        res = PickRing(depth,
                       ndiv, radius, ri.zlen,
                       d, Q,
                       aAxis[iaxis][0], aAxis[iaxis][1],
                       src, dir);                
      }
      if( res ){
//        std::cout << "picked something" << std::endl;
        if( nearestaxis == -3 ){
          nearestaxis = iaxis+3;
          mindepth = depth;
        }
        else if( depth > mindepth ){
          nearestaxis = iaxis+3;
          mindepth = depth;
        }
      }
    }    
    return nearestaxis;
  }
  if( itype == BOX_LEN ){
    double tabl[4][2] = { {-1, 0}, {+1, 0}, { 0,-1}, { 0,+1} };
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();    
    const Com::CMatrix3 B(ri.xlen*0.5, ri.ylen*0.5, ri.zlen*0.5);
    for(unsigned int i=0;i<4;i++){
      if( !IsAliveMode(aRI,aFix,aFixDir,irb,6+i/2,icont_flg,ijnt_flg) ) continue;
      const Com::CVector3D v = d + Q.MatVec( B.MatVec( Com::CVector3D(tabl[i][0]*0.9,tabl[i][1]*0.9,0) ) );
      double t = -Dot(src-v,dir)/Dot(dir,dir);
      double dist = Com::Length(src+t*dir-v);
      //      std::cout << i << " " << dist << std::endl;
      if( dist < hvh*0.05 ) return 6+i;
    }    
    return -4;    
  }
  return -1;
  
  /*  
   const Com::CVector3D& d = rb.ri.GetIniPosCG();
   const Com::CMatrix3 A(ri.xlen,ri.ylen,ri.zlen);
   const Com::CMatrix3& Q = rb.ri.GetRotMatrix_Ini();
   const Com::CMatrix3& QA = Q.MatMat(A);
   */
  /*
   double flg[4][2] = {
   {1,1},
   {-1,1},
   {1,-1},
   {-1,-1} };
   for(unsigned int i=0;i<4;i++){
   double rh,rv;  
   Com::CVector3D org   = d + QA.MatVec( Com::CVector3D(0.35, 0, 0.0) );
   Com::CVector3D vec_h = d + QA.MatVec( Com::CVector3D(0.00, 0, 1.5) );    
   Com::CVector3D vec_v = d + QA.MatVec( Com::CVector3D(0.15, 0, 0.0) ); 
   bool res = IsRayCollideQuad(src,dir, org,n,vec_h,vec_v, rh,rv);    
   }
   */
}
