/*
 *  jointed_rigidbody.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/12/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if defined(__APPLE__) && defined(__MACH__)
#  include <GLUT/glut.h>
#else
#  include <GL/glut.h>
#endif

#include <stdio.h>

#include "delfem/quaternion.h"
#include "delfem/indexed_array.h"
#include "delfem/matvec/matdia_blkcrs.h"
#include "delfem/matvec/matdiafrac_blkcrs.h"
#include "delfem/matvec/vector_blk.h"
#include "delfem/matvec/bcflag_blk.h"


#include "furniture.h"
#include "rigid_pointset.h"
#include "furniture_utility.h"
#include "joint_connectivity.h"
#include "rigid_contact_ls.h"
#include "nail_bound_furniture.h"
#include "gl_drawer_furniture.h"
#include "joint_connectivity.h"


const Com::CVector3D aFaceNormal[6] = {
  Com::CVector3D(+1, 0, 0),
  Com::CVector3D(-1, 0, 0),
  Com::CVector3D(+0,+1, 0),
  Com::CVector3D(-0,-1, 0),
  Com::CVector3D(+0, 0,+1),
  Com::CVector3D(-0, 0,-1) };





unsigned int CJointedRigidBody::AddRigidBody()
{
  assert( aFP.size() == aRI.size() );  
  unsigned int irb0 = aRI.size();
  aRI.resize(irb0+1);
  aFP.resize(irb0+1);
  aRD.resize(irb0+1);
  return irb0;
}



Com::CBoundingBox3D CJointedRigidBody::GetBoundingBox(bool is_current) const
{
  Com::CBoundingBox3D bb;
  if( aRI.empty() ){
    bb = Com::CBoundingBox3D(-2,+2, -2,+2, -2,+2);
  }
  for(unsigned int irb=0;irb<aRI.size();irb++){
    //    const CRigidBody3D& rb = aRB[irb];
    const CRigid3D_Ini& ri = aRI[irb];    
    const CRigid3D_Def& rd = aRD[irb];     
    const Com::CMatrix3& R = rd.GetRotMatrix_Inc().MatMat( ri.GetRotMatrix_Ini() );
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);                    
      Com::CVector3D v1 = ri.GetRotMatrix_Ini().MatVec(v0) + ri.GetIniPosCG();
      Com::CVector3D v2 = R.MatVec(v0) + ri.GetIniPosCG() + rd.GetDispCG();
      if( is_current ){ bb += v2; }
      else{ bb += v1; }
    }
  }
  return bb;
}



/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////




void CJointedRigidBody::Draw(bool is_current,
                             std::vector<unsigned int> aIndRBPicked,
                             const float* colorPicked,
//                             double r, double g, double b, double a,
                             const Com::CVector3D& zscr) const
{  
//  std::cout << "Enter Draw" << std::endl;
  const bool is_light = glIsEnabled(GL_LIGHTING);          
  
  ::glDisable(GL_LIGHTING);

  if( is_current ) 
  { // draw force on joint
    ::glLineWidth(2);
    ::glBegin(GL_LINES);
    assert( aRI.size() == aFP.size() );        
    for(unsigned int irb=0;irb<aRI.size();irb++){        
//      DrawContactForce(aRI[irb],aRD[irb],aFP[irb],contProp);
    }
    ::glEnd();
  }
  else{
    for(unsigned int irb=0;irb<aRI.size();irb++){
      const CContactPoints_RigidBody3D& aCP = aFP[irb],contProp;
      for(unsigned int i=0;i<aCP.Size();i++){
//        std::cout << irb << " " << i << " " << aCP.aFP[i].is_force_cont << std::endl;
        if( !aCP.aFP[i].is_force_cont ) continue;
        const CRigid3D_Ini& ri = aRI[irb];
        Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);         
        Com::CVector3D v1 = ri.GetInitialPosition(v0);
        ::glColor3d(1,0,0);
        ::glPointSize(5);
        ::glBegin(GL_POINTS);
        ::glVertex3d(v1.x,v1.y,v1.z);
        ::glEnd();
      }
    }
  }
  
//  std::cout << "Enter Draw 0" << std::endl;  

  
//  std::cout << "Enter Draw 1" << std::endl;  
  
  if( !is_light ){ ::glDisable(GL_LIGHTING); }
  else{ ::glEnable(GL_LIGHTING); }  
  
  {
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      double shear, pull;
      DisplayJointCapacity(ifix, shear, pull, 
                           aRI, aRD,
                           aFix, 
                           aJFixForce, 
                           aFixDir, 
                           aNailJoint);
      bool isnt_brake = (shear<1) && (pull<1);
      assert( ifix < aFix.size() );
      assert( ifix < aFixDir.size() );      
      assert( ifix < aNailJoint.size() );
      DrawNails(aFix[ifix], aFixDir[ifix].first, aFixDir[ifix].second, 
                aRI,aRD, is_current, aNailJoint[ifix],
                isnt_brake);
      /*
      const CNailJoint& nj = aNailJoint[ifix];
      ::glDisable(GL_LIGHTING);
      ::glPointSize(5);
      ::glBegin(GL_POINTS);
      assert( nj.aPosHead.size() == nj.nnail );
      for(unsigned int inail=0;inail<nj.nnail;inail++){
        Com::CVector3D p = nj.aPosHead[inail];
        ::glVertex3d(p.x,p.y,p.z);
      }
      ::glEnd();
      ::glEnable(GL_LIGHTING);
       */
    }
  }

//  std::cout << "Enter Draw 2" << std::endl;  
  
  {
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
//      bool is_hilight = ( ind_picked_obj == ifix && is_rigid_picked_obj == false );
      bool is_hilight = false;
      const CJoint_Fix& cjs = aFix[ifix];
      const CJoint_Fix_Force& jff = aJFixForce[ifix];
//      DrawFixJoint(!is_current,is_hilight,cjs,jff,aRI,aRD);
    }    
    for(unsigned int irb=0;irb<aRI.size();irb++){   
      bool is_pick = false;
      for(unsigned int ip=0;ip<aIndRBPicked.size();ip++){
        if( irb == aIndRBPicked[ip] ){
          is_pick = true;
          break;
        }
      }
//      unsigned int imode = ( ind_picked_obj == irb && is_rigid_picked_obj == true ) ? 1 : 0;
      const float colorDef[4] = { 1.0f,1.0f,1.0f,0.7f };
      const float* color = (is_pick) ? colorPicked : colorDef;
      if( is_current ){
        DrawRigidBody_Def(aRI[irb],aRD[irb],color);        
      }
      else{
        DrawRigidBody_Ini(aRI[irb],color);        
      }
    }
  }   
//  std::cout << "Exit Draw" << std::endl;
  // draw external force
  for(unsigned int ief=0;ief<aExForce.size();ief++){      
    const CExternalForceOnRigidBody& ef = aExForce[ief];
    unsigned int irb = ef.irb;
    DrawExternalForce(aRI[irb],aRD[irb],ef,is_current, zscr);
  }  
}

void CJointedRigidBody::MoveInitialConfig_UnchangeCurrent
(const Com::CVector3D& ave_cg_ini,
 const Com::CVector3D& ave_cg_inc,
 const Com::CMatrix3& avP)
{
  for(unsigned int irb=0;irb<aRI.size();irb++){        
    CRigid3D_Ini& ri = aRI[irb];
    CRigid3D_Def& rd = aRD[irb];    
    ri.AddRotation_IniLeft(avP);
    rd.AddRotation_IncRight(avP.Trans());
    Com::CVector3D ini_cg = ri.GetIniPosCG();
    ri.SetIniPosCG( avP.MatVec(ini_cg-ave_cg_ini) + ave_cg_ini + ave_cg_inc );
    rd.AddDispCG( -( avP.MatVec(ini_cg-ave_cg_ini) + ave_cg_ini + ave_cg_inc) + ini_cg );
  }    
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    CJoint_Fix& cjs = aFix[ifix];
    Com::CVector3D ini_jp = cjs.ini_pos_joint;
    cjs.ini_pos_joint = avP.MatVec(ini_jp-ave_cg_ini) + ave_cg_ini + ave_cg_inc;
  }        
}




void CJointedRigidBody::GetSensitivityMove(CSimulationSensitivity& simSens,
                                           const CDesignSensitivity& desSens) const
{
  GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
  (simSens,
   ////
   desSens,
   ////
   aRI, 
   aRD,
   aFix,
   aJFixForce,
   aExForce,
   ////
   gravity,
//   cont_base, cont_norm, stiff_n, stiff_f,
   contProp,
   aFP);  
}

void CJointedRigidBody::DeleteRigidBody(unsigned int irb_del)
{
  std::cout << "Delete " << irb_del << std::endl;
  CJointedRigidBody jrb1;
  {
    jrb1.gravity = gravity;
    jrb1.contProp = contProp;
    jrb1.myu_s = myu_s;
    jrb1.myu_k = myu_k;
    jrb1.rho = rho;
    jrb1.lenz = lenz;
    jrb1.invstiff_pos = invstiff_pos;
    jrb1.invstiff_rot = invstiff_rot;
  }
  std::vector<int> aMap;
  {
    aMap.reserve(aRI.size());  
    unsigned int icnt = 0;    
    for(unsigned int irb=0;irb<aRI.size();irb++){
      if( irb == irb_del ){
        aMap.push_back(-1);
        continue;
      }      
      jrb1.aRI.push_back( aRI[irb] );
      jrb1.aRD.push_back( aRD[irb] );
      jrb1.aFP.push_back( aFP[irb] );
      aMap.push_back(icnt);
      icnt++;
    }    
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    if( aFix[ifix].irb0 == irb_del || aFix[ifix].irb1 == irb_del ){
      continue;
    }
    unsigned int ifix1 = jrb1.aFix.size();
    jrb1.aFix.push_back( aFix[ifix] );
    CJoint_Fix& fix1 = jrb1.aFix[ifix1];
    fix1.irb0 = aMap[fix1.irb0];
    fix1.irb1 = aMap[fix1.irb1];      
    ////
    jrb1.aJFixForce.push_back( aJFixForce[ifix] );
    ////
    jrb1.aNailJoint.push_back( aNailJoint[ifix] );
    ////
    jrb1.aFixDir.push_back( aFixDir[ifix] );        
  }
  for(unsigned int iforce=0;iforce<aExForce.size();iforce++){
    if( aExForce[iforce].irb == irb_del ) continue;
    jrb1.aExForce.push_back( aExForce[iforce] );
  }
  *this = jrb1;
}

int CJointedRigidBody::AddRigidBody_Box
(double lenx, double leny, double lenz, double rho, 
 const Com::CVector3D& ini_pos_cg,
 const Com::CMatrix3& rot)
{
  assert( aFP.size() == aRI.size() ); 
  assert( aFP.size() == aRD.size() );     
  if( lenx < lenz*0.9 ) return -1;
  if( leny < lenz*0.9 ) return -1;  
  aFP.resize( aFP.size()+1 );    
  unsigned int irb = aRI.size();
  aRI.resize( irb+1 );
  aRI[irb].SetModeBox(lenx,leny,lenz,rho);
  aRI[irb].SetIniPosCG(ini_pos_cg);
  aRI[irb].SetIniRotation(rot);
  aRD.resize( irb+1 );
  { // set is_force_flag
    CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      const CRigid3D_Ini& ri = aRI[irb];
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);         
      Com::CVector3D v1 = ri.GetInitialPosition(v0);
      double pd = Com::Dot(contProp.cont_base-v1,contProp.cont_norm);
      if( pd > -lenz*0.1 ){
        aCP.aFP[i].is_force_cont = true;
      }
    }    
  }  
  return irb;
}  

int CJointedRigidBody::AddRigidBody_BoxLeg
(double lenx, double leny, double lenz, double rho, 
 const Com::CVector3D& ini_pos_cg,
 const Com::CMatrix3& rot, 
 unsigned int irb0, unsigned int iface0)
{
  int irb_add = this->AddRigidBody_Box(lenx,leny,lenz,rho, ini_pos_cg, rot);
  if( irb_add == -1 ) return -1;
  if( iface0 == 4 || iface0 == 5 ){
    const CRigid3D_Ini& ri0 = aRI[irb0];
    const CRigid3D_Ini& ri1 = aRI[irb_add];
    const Com::CVector3D& n0 = ri0.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );      
    const Com::CVector3D& x1 = ri1.GetRotMatrix_Ini().MatVec( Com::CVector3D(1,0,0) );
    Com::CVector3D pos0 = ini_pos_cg - x1*lenx*0.5;
    double dot = Com::Dot(n0,x1);
    unsigned int idir = ( dot > 0 ) ? 4 : 5;
    this->AddFixJoint(irb0, irb_add, 
                      idir, 1,
                      pos0 , Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), 
                      invstiff_pos, invstiff_rot);    
  }    
  else{
    const CRigid3D_Ini& ri0 = aRI[irb0];
    const CRigid3D_Ini& ri1 = aRI[irb_add];
    const Com::CVector3D& n1 = ri1.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );    
    const Com::CVector3D& x1 = ri1.GetRotMatrix_Ini().MatVec( Com::CVector3D(1,0,0) );          
    const Com::CVector3D& n0 = ri0.GetRotMatrix_Ini().MatVec( aFaceNormal[iface0] );
    Com::CVector3D pos0 = ini_pos_cg - x1*lenx*0.5;
    double dot = Com::Dot(n0,n1);
    unsigned int idir = ( dot < 0 ) ? 4 : 5;
    this->AddFixJoint(irb_add, irb0, 
                      idir, iface0,
                      pos0 , Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), 
                      invstiff_pos, invstiff_rot);            
  }

  EnforceContact_ChangeBoxLen(aRI,aFP,contProp.cont_norm,contProp.cont_base);
  EnforceJointConnectivity_ChangeBoxLen(aRI, aFix, aFixDir);  
  PlaceNails();
  return irb_add;
}

int CJointedRigidBody::AddRigidBody_BoxConnect
(double lenx, double leny, double lenz, double rho, 
 const Com::CVector3D& ini_pos_cg,
 const Com::CMatrix3& rot, 
 unsigned int irb0, unsigned int iface0,
 unsigned int irb1, unsigned int iface1)
{
  int irb_add = this->AddRigidBody_Box(lenx,leny,lenz,rho, ini_pos_cg, rot);
  if( irb_add == -1 ) return -1;
  if( iface0 == 4 || iface0 == 5 ){
    const CRigid3D_Ini& ri0 = aRI[irb0];
    const CRigid3D_Ini& ria = aRI[irb_add];
    const Com::CVector3D& n0 = ri0.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );      
    const Com::CVector3D& xa = ria.GetRotMatrix_Ini().MatVec( Com::CVector3D(1,0,0) );
    Com::CVector3D pos0 = ini_pos_cg - xa*lenx*0.5;
    double dot = Com::Dot(n0,xa);
    unsigned int idir = ( dot > 0 ) ? 4 : 5;
    this->AddFixJoint(irb0, irb_add, 
                      idir, 1, 
                      pos0 , Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), 
                      invstiff_pos, invstiff_rot);    
  }
  else {
    const CRigid3D_Ini& ri0 = aRI[irb0];
    const CRigid3D_Ini& ria = aRI[irb_add];
    const Com::CVector3D& na = ria.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );    
    const Com::CVector3D& xa = ria.GetRotMatrix_Ini().MatVec( Com::CVector3D(1,0,0) );          
    const Com::CVector3D& n0 = ri0.GetRotMatrix_Ini().MatVec( aFaceNormal[iface0] );
    Com::CVector3D pos0 = ini_pos_cg - xa*lenx*0.5;
    double dot = Com::Dot(n0,na);
    unsigned int idir = ( dot < 0 ) ? 4 : 5;
    this->AddFixJoint(irb_add, irb0, 
                      idir, iface0,
                      pos0 , Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), 
                      invstiff_pos, invstiff_rot);                
  }
  if( iface1 == 4 || iface1 == 5 ){
    const CRigid3D_Ini& ri1 = aRI[irb1];
    const CRigid3D_Ini& ria = aRI[irb_add];
    const Com::CVector3D& n1 = ri1.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );      
    const Com::CVector3D& xa = ria.GetRotMatrix_Ini().MatVec( Com::CVector3D(1,0,0) );
    Com::CVector3D pos1 = ini_pos_cg + xa*lenx*0.5;
    double dot = Com::Dot(n1,xa);
    unsigned int idir = ( dot < 0 ) ? 4 : 5;
    this->AddFixJoint(irb1, irb_add, 
                      idir, 0, 
                      pos1 , Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), 
                      invstiff_pos, invstiff_rot);    
  }      
  else{
    const CRigid3D_Ini& ri1 = aRI[irb1];
    const CRigid3D_Ini& ria = aRI[irb_add];
    const Com::CVector3D& na = ria.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );    
    const Com::CVector3D& xa = ria.GetRotMatrix_Ini().MatVec( Com::CVector3D(1,0,0) );          
    const Com::CVector3D& n1 = ri1.GetRotMatrix_Ini().MatVec( aFaceNormal[iface1] );
    Com::CVector3D pos1 = ini_pos_cg - xa*lenx*0.5;
    double dot = Com::Dot(na,n1);
    unsigned int idir = ( dot < 0 ) ? 4 : 5;
    this->AddFixJoint(irb_add, irb1, 
                      idir, iface1,
                      pos1 , Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), 
                      invstiff_pos, invstiff_rot);                    
  }
  EnforceContact_ChangeBoxLen(aRI,aFP,contProp.cont_norm,contProp.cont_base);  
  EnforceJointConnectivity_ChangeBoxLen(aRI, aFix, aFixDir);  
  PlaceNails();  
  return irb_add;    
}

unsigned int CJointedRigidBody::AddLeg
(unsigned int irb0, 
 const Com::CVector3D& pos0, const Com::CVector3D& pos1)
{
  assert( aRI.size() > 0 );
  assert( aFix.size() > 0 );
  /////  
  const CRigid3D_Ini& ri0 = aRI[irb0];  
  double lenx = (pos0-pos1).Length();
  double leny = lenz;
  Com::CVector3D cg = (pos0+pos1)*0.5;
  const Com::CVector3D& n0 = ri0.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );  
  Com::CVector3D x0 = pos0-pos1;  
  x0.SetNormalizedVector();
  Com::CVector3D y0 = Cross(x0,n0);
  y0.SetNormalizedVector();
  Com::CVector3D z0 = Cross(x0,y0);
  Com::CMatrix3 rot(1.0);  
  for(unsigned int i=0;i<3;i++){ 
    rot.mat[3*i+0] = x0[i];
    rot.mat[3*i+1] = y0[i];
    rot.mat[3*i+2] = z0[i];
  }  
  unsigned int irb_add = this->AddRigidBody_Box(lenx,leny,lenz,rho, cg, rot);    
  {
    double dot = Com::Dot(n0,-x0);
    unsigned int idir = ( dot > 0 ) ? 4 : 5;
    this->AddFixJoint(irb_add, irb0, 
                      0, idir, 
                      pos0 , Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), 
                      invstiff_pos, invstiff_rot);    
  }  
  return irb_add;
}

unsigned int CJointedRigidBody::ConnectPlanks
(unsigned int irb0, unsigned int irb1,
 const Com::CVector3D& pos0, const Com::CVector3D& pos1)
{
  assert( aRI.size() > 0 );
  assert( aFix.size() > 0 );
  /////
  const CRigid3D_Ini& ri0 = aRI[irb0];
  const CRigid3D_Ini& ri1 = aRI[irb1];    
  const Com::CVector3D& n0 = ri0.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );
  const Com::CVector3D& n1 = ri1.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );    
  Com::CVector3D y0 = Com::Cross(n0,n1);
  if( SquareLength(y0) < 1.0e-5 ){
    y0 = Cross(n1,Com::CVector3D(0,0,1));          
  }  
  y0.SetNormalizedVector();
  Com::CVector3D x0 = (pos0-pos1);
  x0 -= Dot(y0,x0)*y0;
  double lenx = x0.Length();  
  x0.SetNormalizedVector();
  Com::CVector3D z0 = Com::Cross(x0,y0);
  Com::CMatrix3 rot(1.0);  
  for(unsigned int i=0;i<3;i++){ 
    rot.mat[3*i+0] = x0[i];
    rot.mat[3*i+1] = y0[i];
    rot.mat[3*i+2] = z0[i];
  }
  Com::CVector3D cg = (pos0+pos1)*0.5;
  double leny = fabs(Dot(pos0-pos1,y0));
  if( leny < lenz ){ leny = lenz; }
  std::cout << lenx << " " << leny << " " << lenz << std::endl;
  assert( lenx > 0 && leny > 0 && lenz > 0 );
  unsigned int irb_add = this->AddRigidBody_Box(lenx,leny,lenz,rho, cg, rot);  
  {
    double dot = Com::Dot(n0,-x0);
    unsigned int idir = ( dot > 0 ) ? 4 : 5;
    this->AddFixJoint(irb0, irb_add, 
                      idir, 0, 
                      pos0 , Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), 
                      invstiff_pos, invstiff_rot);    
  }
  {
    double dot = Com::Dot(n1,+x0);
    unsigned int idir = ( dot > 0 ) ? 4 : 5;    
    this->AddFixJoint(irb1, irb_add, 
                      idir, 1, 
                      pos1 , Com::CVector3D(1,0,0), Com::CVector3D(0,1,0), Com::CVector3D(0,0,1), 
                      invstiff_pos, invstiff_rot);      
  }
  {
    const CRigid3D_Ini& ria = aRI[irb_add]; 
//    std::cout << ria.xlen << " " << ria.ylen << " " << ria.zlen << std::endl;   
  }  
  EnforceContact_ChangeBoxLen(aRI,aFP,contProp.cont_norm,contProp.cont_base);  
  EnforceJointConnectivity_ChangeBoxLen(aRI, aFix, aFixDir);
  {
    const CRigid3D_Ini& ria = aRI[irb_add]; 
//    std::cout << ria.xlen << " " << ria.ylen << " " << ria.zlen << std::endl;   
  }
  return irb_add;
}

/*
 std::vector<CRigid3D_Ini> aRI;  
 std::vector<CRigid3D_Def> aRD;    
 std::vector<CContactPoints_RigidBody3D> aFP;  
 
 std::vector<CJoint_Fix> aFix;    
 std::vector< std::pair<unsigned int, unsigned int> > aFixDir;     
 std::vector<CJoint_Fix_Force> aJFixForce;   
 std::vector<CNailJoint> aNailJoint;   
 
 std::vector<CExternalForceOnRigidBody> aExForce;  

 
 // constant variable 
 double rho;
 double lenz;
 double invstiff_pos, invstiff_rot;
 
 /// enviroment
 Com::CVector3D gravity;
 CContactProp contProp;
 double myu_s;
 double myu_k;   
 */


void myFPrintf(FILE* fp,const Com::CVector3D& vec){
  fprintf(fp,"%lf %lf %lf\n",vec.x, vec.y, vec.z);    
}

void CJointedRigidBody::writeFile(const std::string& fname) const
{  
  std::cout << "CJointedRigidBody::writeFile" << fname << std::endl;
  FILE *fp;
  if ((fp = fopen(fname.c_str(), "w")) == NULL) {
    std::cout << "cannot open file " << fname << std::endl;
    return;
	}
  fprintf(fp,"%d\n",aRI.size());
  for(unsigned int irb=0;irb<aRI.size();irb++){        
    {
      /*
       double crv_ini[3]; // initial_rotation    
       Com::CVector3D ini_pos_cg;  // center of gravity
       double mass;
       double mineatia[3];      
       ////
       double xlen, ylen, zlen; // separate this line
       */      
      const CRigid3D_Ini& ri = aRI[irb];
      fprintf(fp,"%lf %lf %lf\n",ri.crv_ini[0],ri.crv_ini[1],ri.crv_ini[2]);
      myFPrintf(fp, ri.ini_pos_cg);
      fprintf(fp,"%lf %lf %lf %lf\n",ri.mass, ri.mineatia[0],ri.mineatia[1],ri.mineatia[2]);
      fprintf(fp,"%lf %lf %lf\n",ri.xlen,ri.ylen,ri.zlen);
    }
    {
      /*
       Com::CVector3D disp_cg;    
       double crv_inc[3]; // incremental_rotation  
       Com::CVector3D velo_cg;  
       Com::CVector3D Omega; 
       */      
      const CRigid3D_Def& rd = aRD[irb];
      myFPrintf(fp, rd.disp_cg);
      fprintf(fp,"%lf %lf %lf\n", rd.crv_inc[0], rd.crv_inc[1], rd.crv_inc[2]);
      myFPrintf(fp, rd.velo_cg);
      myFPrintf(fp, rd.Omega);
    }    
    {
      /*
       double pd;  // penetration depth
       Com::CVector3D aloc; // anchor location
       int itype_contact;	// 0:not_contact 1:static contact 2:dynamic contact  
       bool is_force_cont; 
       */      
      const CContactPoints_RigidBody3D& aCP = aFP[irb];
      fprintf(fp,"%d\n",aCP.aFP.size());
      for(unsigned int ifp=0;ifp<aCP.aFP.size();ifp++){
        const CFrictionPoint& frp = aCP.aFP[ifp];
        fprintf(fp,"%lf\n",frp.pd);
        myFPrintf(fp, frp.aloc);
        fprintf(fp,"%d %d\n",frp.itype_contact,frp.is_force_cont);
      }      
    }
  }
  fprintf(fp,"%d\n",aFix.size());
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){        
    {
      /*
       double invstiff_pos, invstiff_rot;
       Com::CVector3D ini_pos_joint;
       Com::CVector3D loc_coord[3];
       unsigned int irb0;
       unsigned int irb1;             
       */
      const CJoint_Fix& fix = aFix[ifix];
      fprintf(fp,"%lf %lf\n",fix.invstiff_pos,fix.invstiff_rot);
      myFPrintf(fp, fix.ini_pos_joint);
      myFPrintf(fp, fix.loc_coord[0]);
      myFPrintf(fp, fix.loc_coord[1]);
      myFPrintf(fp, fix.loc_coord[2]);
      fprintf(fp,"%d %d\n",fix.irb0,fix.irb1);
    }
    {
      fprintf(fp,"%d %d\n",aFixDir[ifix].first, aFixDir[ifix].second);
    }
    {
      /*
       double lambda[6];
       double dlambda[6];           
       */
      const CJoint_Fix_Force& jff = aJFixForce[ifix];
      fprintf(fp,"%lf %lf %lf %lf %lf %lf\n",jff.lambda[0],jff.lambda[1],jff.lambda[2],jff.lambda[3],jff.lambda[4],jff.lambda[5]);
      fprintf(fp,"%lf %lf %lf %lf %lf %lf\n",jff.dlambda[0],jff.dlambda[1],jff.dlambda[2],jff.dlambda[3],jff.dlambda[4],jff.dlambda[5]);      
    }
    {
      /*
       double shear_max;
       double pull_max_perlen;
       double intervnail;
       double lennail;
       std::vector<Com::CVector3D> aPosHead;       
       */
      const CNailJoint& nj = aNailJoint[ifix];
      fprintf(fp,"%lf\n",nj.shear_max);
      fprintf(fp,"%lf\n",nj.pull_max_perlen);
      fprintf(fp,"%lf\n",nj.intervnail);
      fprintf(fp,"%lf\n",nj.lennail);  
      fprintf(fp,"%d\n",nj.aPosHead.size());
      for(unsigned int iph=0;iph<nj.aPosHead.size();iph++){
        myFPrintf(fp, nj.aPosHead[iph]);
      }
    }
  }
  fprintf(fp,"%d\n",aExForce.size());
  for(unsigned int ief=0;ief<aExForce.size();ief++){
    /*
     unsigned int irb;
     Com::CVector3D pos;  
     Com::CVector3D force;     
     */
    const CExternalForceOnRigidBody& ef = aExForce[ief];    
    fprintf(fp,"%d\n",ef.irb);
    myFPrintf(fp,ef.pos);
    myFPrintf(fp,ef.force);
  }
  /*
   
   // constant variable 
   double rho;
   double lenz;
   double invstiff_pos, invstiff_rot;
   
   /// enviroment
   Com::CVector3D gravity;
   CContactProp contProp;
   double myu_s;
   double myu_k;    
   */
  /*
   Com::CVector3D cont_base;
   Com::CVector3D cont_norm; 
   double stiff_n;
   double stiff_f;  
   */
  fprintf(fp,"%lf %lf %lf %lf\n",rho,lenz,invstiff_pos,invstiff_rot);
  myFPrintf(fp, gravity);
  myFPrintf(fp, contProp.cont_base);
  myFPrintf(fp, contProp.cont_norm);
  fprintf(fp,"%lf %lf\n",contProp.stiff_n,contProp.stiff_f);
  fprintf(fp,"%lf %lf\n",myu_s,myu_k);
  
  fclose(fp);
}

void myFScanf(FILE* fp,Com::CVector3D& vec){
  fscanf(fp,"%lf %lf %lf\n",&vec.x, &vec.y, &vec.z);    
}

bool CJointedRigidBody::loadFile(const std::string& fname)
{
  FILE *fp;  
  if ((fp = fopen(fname.c_str(), "r")) == NULL) {
    std::cout << "cannot open file " << fname << std::endl;
    return false;
	}  
  
  aRI.clear();
  aRD.clear();
  aFP.clear();
  int nrb;
  fscanf(fp,"%d\n",&nrb);
  for(unsigned int irb=0;irb<nrb;irb++){
    {
      /*
       double crv_ini[3]; // initial_rotation    
       Com::CVector3D ini_pos_cg;  // center of gravity
       double mass;
       double mineatia[3];      
       ////
       double xlen, ylen, zlen; // separate this line
       */      
      CRigid3D_Ini ri;
      fscanf(fp,"%lf %lf %lf\n",ri.crv_ini+0,ri.crv_ini+1,ri.crv_ini+2);
      myFScanf(fp, ri.ini_pos_cg);
      fscanf(fp,"%lf %lf %lf %lf\n",&ri.mass, ri.mineatia+0,ri.mineatia+1,ri.mineatia+2);
      fscanf(fp,"%lf %lf %lf\n",&ri.xlen,&ri.ylen,&ri.zlen);
      aRI.push_back(ri);
    }
    {
      /*
       Com::CVector3D disp_cg;    
       double crv_inc[3]; // incremental_rotation  
       Com::CVector3D velo_cg;  
       Com::CVector3D Omega; 
       */      
      CRigid3D_Def rd;
      myFScanf(fp, rd.disp_cg);
      fscanf(fp,"%lf %lf %lf\n", rd.crv_inc+0, rd.crv_inc+1, rd.crv_inc+2);
      myFScanf(fp, rd.velo_cg);
      myFScanf(fp, rd.Omega);
      aRD.push_back(rd);
    }    
    {
      /*
       double pd;  // penetration depth
       Com::CVector3D aloc; // anchor location
       int itype_contact;	// 0:not_contact 1:static contact 2:dynamic contact  
       bool is_force_cont; 
       */      
      CContactPoints_RigidBody3D aCP;
      int nFP;
      fscanf(fp,"%d",&nFP);
      for(unsigned int ifp=0;ifp<nFP;ifp++){
        CFrictionPoint frp;
        fscanf(fp,"%lf",&frp.pd);
        myFScanf(fp, frp.aloc);
        fscanf(fp,"%d %d",&frp.itype_contact,&frp.is_force_cont);
        std::cout << "read cp" << frp.itype_contact << " " << frp.is_force_cont << std::endl;
        aCP.aFP[ifp] = frp;
      }      
      aFP.push_back(aCP);
    }    
  }
  
  aFix.clear();
  aFixDir.clear();
  aJFixForce.clear();
  aNailJoint.clear();
  int nFix;
  fscanf(fp,"%d",&nFix);
  std::cout << "nfix " << nFix << std::endl;  
  for(unsigned int ifix=0;ifix<nFix;ifix++){        
    {
      /*
       double invstiff_pos, invstiff_rot;
       Com::CVector3D ini_pos_joint;
       Com::CVector3D loc_coord[3];
       unsigned int irb0;
       unsigned int irb1;             
       */
      CJoint_Fix fix(0,0);
      fscanf(fp,"%lf %lf",&fix.invstiff_pos,&fix.invstiff_rot);
      myFScanf(fp, fix.ini_pos_joint);
      myFScanf(fp, fix.loc_coord[0]);
      myFScanf(fp, fix.loc_coord[1]);
      myFScanf(fp, fix.loc_coord[2]);
      fscanf(fp,"%d %d",&fix.irb0,&fix.irb1);
      aFix.push_back(fix);
    }
    {
      int idir0,idir1;
      fscanf(fp,"%d %d\n",&idir0,&idir1);
      aFixDir.push_back( std::make_pair(idir0,idir1) );
    }
    {
      /*
       double lambda[6];
       double dlambda[6];           
       */
      CJoint_Fix_Force jff;
      fscanf(fp,"%lf %lf %lf %lf %lf %lf\n",jff.lambda+ 0,jff.lambda +1,jff.lambda +2,jff.lambda +3,jff.lambda +4,jff.lambda +5);
      fscanf(fp,"%lf %lf %lf %lf %lf %lf\n",jff.dlambda+0,jff.dlambda+1,jff.dlambda+2,jff.dlambda+3,jff.dlambda+4,jff.dlambda+5);      
      aJFixForce.push_back(jff);
    }
    {
      /*
       double shear_max;
       double pull_max_perlen;
       double intervnail;
       double lennail;
       std::vector<Com::CVector3D> aPosHead;       
       */
      CNailJoint nj;
      fscanf(fp,"%lf",&nj.shear_max);
      fscanf(fp,"%lf",&nj.pull_max_perlen);
      fscanf(fp,"%lf",&nj.intervnail);
      fscanf(fp,"%lf",&nj.lennail);  
      int nph;
      fscanf(fp,"%d",&nph);
      for(unsigned int iph=0;iph<nph;iph++){
        Com::CVector3D ph;
        myFScanf(fp, ph);
        nj.aPosHead.push_back(ph);
      }
      aNailJoint.push_back(nj);
    }
  }
  int nef;
  fscanf(fp,"%d",&nef);
  aExForce.clear();
  for(unsigned int ief=0;ief<nef;ief++){
    /*
     unsigned int irb;
     Com::CVector3D pos;  
     Com::CVector3D force;     
     */
    CExternalForceOnRigidBody ef(0,Com::CVector3D(0,0,0), Com::CVector3D(0,0,0));
    fscanf(fp,"%d",&ef.irb);
    myFScanf(fp,ef.pos);
    myFScanf(fp,ef.force);
    aExForce.push_back(ef);
  }  
  
  fscanf(fp,"%lf %lf %lf %lf",&rho,&lenz,&invstiff_pos,&invstiff_rot);
  myFScanf(fp, gravity);
  myFScanf(fp, contProp.cont_base);
  myFScanf(fp, contProp.cont_norm);
  fscanf(fp,"%lf %lf",&contProp.stiff_n,&contProp.stiff_f);
  fscanf(fp,"%lf %lf",&myu_s,&myu_k);  
  
  fclose(fp);  
  return true;
}



void myFPrintfObj(FILE* fp,const Com::CVector3D& vec){
  fprintf(fp,"v %lf %lf %lf\n",vec.x, vec.y, vec.z);    
}

void CJointedRigidBody::writeFileObj(const std::string& fname) const
{
  std::cout << "CJointedRigidBody::writeFile : " << fname << std::endl;
  FILE *fp;
  if ((fp = fopen(fname.c_str(), "w")) == NULL) {
    std::cout << "cannot open file " << fname << std::endl;
    return;
	}  
  fprintf(fp,"g furniture\n");  
  for(unsigned int irb=0;irb<aRI.size();irb++){        
    const CRigid3D_Ini& ri = aRI[irb];
    Com::CMatrix3 R; R.SetRotMatrix_CRV(ri.crv_ini);
    Com::CVector3D cg = ri.ini_pos_cg;
    const Com::CVector3D& v0 =  cg + R.MatVec( Com::CVector3D(-ri.xlen*0.5,-ri.ylen*0.5,-ri.zlen*0.5) );
    const Com::CVector3D& v1 =  cg + R.MatVec( Com::CVector3D(+ri.xlen*0.5,-ri.ylen*0.5,-ri.zlen*0.5) );
    const Com::CVector3D& v2 =  cg + R.MatVec( Com::CVector3D(-ri.xlen*0.5,+ri.ylen*0.5,-ri.zlen*0.5) );
    const Com::CVector3D& v3 =  cg + R.MatVec( Com::CVector3D(+ri.xlen*0.5,+ri.ylen*0.5,-ri.zlen*0.5) );    
    const Com::CVector3D& v4 =  cg + R.MatVec( Com::CVector3D(-ri.xlen*0.5,-ri.ylen*0.5,+ri.zlen*0.5) );
    const Com::CVector3D& v5 =  cg + R.MatVec( Com::CVector3D(+ri.xlen*0.5,-ri.ylen*0.5,+ri.zlen*0.5) );
    const Com::CVector3D& v6 =  cg + R.MatVec( Com::CVector3D(-ri.xlen*0.5,+ri.ylen*0.5,+ri.zlen*0.5) );
    const Com::CVector3D& v7 =  cg + R.MatVec( Com::CVector3D(+ri.xlen*0.5,+ri.ylen*0.5,+ri.zlen*0.5) );        
    myFPrintfObj(fp, v0);
    myFPrintfObj(fp, v1);
    myFPrintfObj(fp, v2);
    myFPrintfObj(fp, v3);
    myFPrintfObj(fp, v4);
    myFPrintfObj(fp, v5);    
    myFPrintfObj(fp, v6);
    myFPrintfObj(fp, v7);
  }    
  for(unsigned int irb=0;irb<aRI.size();irb++){        
    const CRigid3D_Ini& ri = aRI[irb];
    Com::CMatrix3 R; R.SetRotMatrix_CRV(ri.crv_ini);
    Com::CVector3D n0 = R.MatVec( Com::CVector3D(-1, 0, 0) );
    Com::CVector3D n1 = R.MatVec( Com::CVector3D(+1, 0, 0) );
    Com::CVector3D n2 = R.MatVec( Com::CVector3D( 0,-1, 0) );
    Com::CVector3D n3 = R.MatVec( Com::CVector3D( 0,+1, 0) );    
    Com::CVector3D n4 = R.MatVec( Com::CVector3D( 0, 0,-1) );
    Com::CVector3D n5 = R.MatVec( Com::CVector3D( 0, 0,+1) );        
    fprintf(fp,"vn %lf %lf %lf\n",n0.x, n0.y, n0.z);
    fprintf(fp,"vn %lf %lf %lf\n",n1.x, n1.y, n1.z);
    fprintf(fp,"vn %lf %lf %lf\n",n2.x, n2.y, n2.z);
    fprintf(fp,"vn %lf %lf %lf\n",n3.x, n3.y, n3.z);
    fprintf(fp,"vn %lf %lf %lf\n",n4.x, n4.y, n4.z);
    fprintf(fp,"vn %lf %lf %lf\n",n5.x, n5.y, n5.z);    
  }    
  for(unsigned int irb=0;irb<aRI.size();irb++){     
    const unsigned int iv0 = 1+irb*8;
    const unsigned int in0 = 1+irb*6; 
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+0,in0+4,iv0+3,in0+4, iv0+1,in0+4);
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+0,in0+4,iv0+2,in0+4, iv0+3,in0+4);
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+4,in0+5,iv0+5,in0+5, iv0+7,in0+5);
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+4,in0+5,iv0+7,in0+5, iv0+6,in0+5);    
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+0,in0+2,iv0+1,in0+2, iv0+5,in0+2);
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+0,in0+2,iv0+5,in0+2, iv0+4,in0+2);    
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+3,in0+3,iv0+2,in0+3, iv0+7,in0+3);
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+2,in0+3,iv0+6,in0+3, iv0+7,in0+3);      
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+0,in0+0,iv0+6,in0+0, iv0+2,in0+0);
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+0,in0+0,iv0+4,in0+0, iv0+6,in0+0);          
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+1,in0+1,iv0+3,in0+1, iv0+7,in0+1);
    fprintf(fp,"f %d//%d %d//%d %d//%d\n",iv0+1,in0+1,iv0+7,in0+1, iv0+5,in0+1);           
  }
  fclose(fp);
}




