/*
 *  jointed_rigidbody.h
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/12/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if !defined(FURNITURE_H)
#define FURNITURE_H

#include "delfem/vector3d.h"

#include "rigidbody.h"
#include "rigid_contact.h"
#include "rigid_contact_ls.h"
#include "joint_fix.h"
#include "joint_connectivity.h"
#include "jointed_rigidbody_contact.h"
#include "nail_bound_furniture.h"
#include "jointed_rigidbody_contact_test.h"
#include "nail_bound_furniture.h"

class CJointedRigidBody
{
public:
  void Clear(){
    aFix.clear();
    aFixDir.clear();    
    aNailJoint.clear();
    ////
    aRI.clear();
    aRD.clear();    
    aJFixForce.clear();    
    aFP.clear();
    aExForce.clear();
//    aOptDof.clear();    
  }
  void ClearValue(){    
    for(unsigned int irb=0;irb<aRD.size();irb++){
      aRD[irb].Clear();
    }
    for(unsigned int icst=0;icst<aFix.size();icst++){
      aJFixForce[icst].Clear();
    }    
  }
  
  /////
  unsigned int AddRigidBody();
  void DeleteRigidBody(unsigned int irb);
  unsigned int AddFixJoint(unsigned int irb0, unsigned int irb1,
                           unsigned int idir0, unsigned int idir1, 
                           const Com::CVector3D& pos_j, 
                           const Com::CVector3D& ex,
                           const Com::CVector3D& ey,
                           const Com::CVector3D& ez, 
                           double invstiff_pos, double invstiff_rot)
  {
    unsigned int ifix = aFix.size();
    assert( idir0 == 4 || idir0 == 5 );
    assert( aFixDir.size() == ifix );
    assert( aNailJoint.size() == ifix );
    aFixDir.push_back( std::make_pair(idir0,idir1) );      
    CJoint_Fix fix(irb0,irb1); 
    fix.SetIniPosJoint(pos_j);
    fix.SetIniJointDirection(ex, ey, ez);
    fix.SetInvStiffJoint(invstiff_pos, invstiff_rot);            
    aFix.push_back( fix );          
    aJFixForce.resize( aJFixForce.size()+1 );
    aNailJoint.resize( aNailJoint.size() + 1 );
    PlaceNails();
    return ifix;
  }
  unsigned int ConnectPlanks(unsigned int irb0, unsigned int irb1,
                             const Com::CVector3D& pos0, const Com::CVector3D& pos1);
  
  int AddRigidBody_BoxLeg(double lenx, double leny, double lenz, double rho, 
                          const Com::CVector3D& ini_pos_cg,
                          const Com::CMatrix3& rot, 
                          unsigned int irb0, unsigned int iface0);
  int AddRigidBody_BoxConnect(double lenx, double leny, double lenz, double rho, 
                              const Com::CVector3D& ini_pos_cg,
                              const Com::CMatrix3& rot, 
                              unsigned int irb0, unsigned int iface0,
                              unsigned int irb1, unsigned int iface1);
  
  int AddRigidBody_Box(double lenx, double leny, double lenz, double rho, 
                       const Com::CVector3D& ini_pos_cg,
                       const Com::CMatrix3& rot);  
  
  unsigned int AddLeg(unsigned int irb0, 
                      const Com::CVector3D& pos0, const Com::CVector3D& pos1);
  
  unsigned int AddExternalForce(unsigned int irb,
                        const Com::CVector3D& pos,                               
                        const Com::CVector3D& force ){      
    unsigned int iexf = aExForce.size();
    aExForce.push_back( CExternalForceOnRigidBody(irb,pos,force) );    
    return iexf;
  }   
  double GetTotalEnergy() const{
    double eng = 0;
    for(unsigned int irb=0;irb<aRI.size();irb++){
      double e = GetKineticEnergy(aRI[irb],aRD[irb]);
      e -= Com::Dot(gravity,aRD[irb].GetDispCG())*aRI[irb].GetMass();
      eng += e;
    }
    return eng;
  }
  void Draw(bool is_current,
            std::vector<unsigned int> aIndRBPicked,
//            double r, double g, double b, double a, 
            const float* colorPicked,
            const Com::CVector3D& zscr) const;
  
  void DrawPickedPlank(const CDesignSensitivity& desSens,
                       const CSimulationSensitivity& simSens,
                       double eps,                       
                       bool is_rigid_picked_obj,
                       int ind_picked_obj);
                       
  Com::CBoundingBox3D GetBoundingBox(bool is_current) const;  
  
  void GetSensitivityMove(CSimulationSensitivity& ds,
                          const CDesignSensitivity& desSens) const;
  
  void MoveInitialConfig_UnchangeCurrent
  (const Com::CVector3D& ave_cg_ini,
   const Com::CVector3D& ave_cg_inc,
   const Com::CMatrix3& avP);
  
  void PlaceNails(){
    PlaceNailsToJoint(aNailJoint,aRI,aFix,aFixDir);
  }
  
  void writeFile(const std::string& fname) const;
  bool loadFile(const std::string& fname);
  void writeFileObj(const std::string& fname) const;
public:
  std::vector<CRigid3D_Ini> aRI;  
  std::vector<CJoint_Fix> aFix;    
  std::vector< std::pair<unsigned int, unsigned int> > aFixDir;    
  std::vector<CNailJoint> aNailJoint;  
  
  std::vector<CRigid3D_Def> aRD;    
  std::vector<CJoint_Fix_Force> aJFixForce;  
  std::vector<CExternalForceOnRigidBody> aExForce;  
  std::vector<CContactPoints_RigidBody3D> aFP;  
  
  // constant variable 
  double rho;
  double lenz;
  double invstiff_pos, invstiff_rot;
  
  /// enviroment
  Com::CVector3D gravity;
  CContactProp contProp;
  double myu_s;
  double myu_k;  
};

static void StepTime_BackwardEular_UntactAnchor(CJointedRigidBody& jrb,
                                                double dt, double& res_norm, double& upd_norm)
{
  StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(dt,
                                                                    jrb.aRI,jrb.aRD,jrb.aFix,jrb.aJFixForce,jrb.aExForce,
                                                                    jrb.gravity,
                                                                    jrb.contProp,                                                                    
                                                                    jrb.aFP,
                                                                    res_norm, upd_norm);
}

static void StepTime_BackwardEular(CJointedRigidBody& jrb,
                                   double dt, double& res_norm, double& upd_norm)
{
  StepTime_BackwardEular_JointedRigidBodyContactSystem(dt,
                                                       jrb.aRI,jrb.aRD,jrb.aFix,jrb.aJFixForce,jrb.aExForce,
                                                       jrb.gravity,
                                                       jrb.contProp,
                                                       jrb.myu_s,jrb.myu_k,
                                                       jrb.aFP,
                                                       res_norm, upd_norm);    
}

static void TestSensitiveResidual(CJointedRigidBody& jrb,
                                  PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s,
                                  ANCHOR_RELOCATION_TYPE is_release_anchor)
{    
  TestSensitiveResidual_JointedRigidBodyContactSystem(itype_s,iobj_s,idim_s,
                                                      is_release_anchor,
                                                      jrb.aRI,jrb.aRD,jrb.aFix,jrb.aJFixForce,jrb.aExForce,
                                                      jrb.gravity,
                                                      jrb.contProp,
                                                      jrb.aFP);
}
static void TestSensitiveResidual(CJointedRigidBody& jrb,
                           const CDesignSensitivity& desSens){
  TestSensitiveResidual_JointedRigidBodyContactSystem(desSens,
                                                      jrb.aRI,jrb.aRD,jrb.aFix,jrb.aJFixForce,jrb.aExForce,
                                                      jrb.gravity,
                                                      jrb.contProp,
                                                      jrb.aFP);
}  
static void TestSensitive_PRLJ(CJointedRigidBody& jrb,
                               PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s, 
                               ANCHOR_RELOCATION_TYPE is_release_anchor)
{
  TestSensitivePRLM_InitialConfigBox_JointedRigidBodyContactSystem(itype_s,iobj_s,idim_s,
                                                                   is_release_anchor,
                                                                   jrb.aRI,jrb.aRD,jrb.aFix,jrb.aJFixForce,jrb.aExForce,
                                                                   jrb.gravity,
                                                                   jrb.contProp,
                                                                   jrb.aFP);                                                                     
}

static void TestSensitive_PRLJ(CJointedRigidBody& jrb, 
                               CDesignSensitivity& desSens)
{
  TestSensitivePRLM_InitialConfigBox_JointedRigidBodyContactSystem(desSens,
                                                                   jrb.aRI,jrb.aRD,jrb.aFix,jrb.aFixDir,jrb.aJFixForce,jrb.aExForce,
                                                                   jrb.gravity,
                                                                   jrb.contProp,
                                                                   jrb.aFP);                                                                     
}  


#endif