/*
 *  furniture_optimizer.h
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/25/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include "furniture.h"
//#include "jointed_rigidbody_contact.h"


class CEvalGradSensPRLM
{
public:
  double eval;
  double deval;
  CSimulationSensitivity simSens;
};

////　
// Exclude functions from here.
////
static void JointConnectivityEnforcement_ChangeBoxLen(CJointedRigidBody& jrb)
{
  EnforceJointConnectivity_ChangeBoxLen(jrb.aRI,jrb.aFix,jrb.aFixDir);
}

static void ContactEnforcement_ChangeBoxLen(CJointedRigidBody& jrb)
{
  EnforceContact_ChangeBoxLen(jrb.aRI,jrb.aFP,jrb.contProp.cont_norm,jrb.contProp.cont_base);
}

static void JointConnectivityEnforcement_ChangeBoxPosRotLen(CJointedRigidBody& jrb)
{
  EnforceJointConnectivity_ChangeBoxPosRotLen(jrb.aRI,jrb.aFix,jrb.aFixDir);
}
////
////  
void MoveJointTwardBendless(CJointedRigidBody& jrb, unsigned int ifix, double ratio);
/////
void FindInitialConfiguration_SameFrictionAnchor(CJointedRigidBody& jrb);  
////
void OptimizeEvalFunc_Grad(CJointedRigidBody& jrb, PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s, 
                           double alpha, ANCHOR_RELOCATION_TYPE is_release_anchor);
void OptimizeEvalFunc_GradPrj(CJointedRigidBody& jrb, double alpha, 
                              const std::vector<COptimizationDof>& aOptDof);
void OptimizeEvalFunc_Sym_Grad(CJointedRigidBody& jrb, PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s, 
                               unsigned int iobj_m, unsigned int iaxis_m,
                               double alpha, ANCHOR_RELOCATION_TYPE is_release_anchor);
void ItrOptimizeEvalFunc_Grad(CJointedRigidBody& jrb, double eps, 
                              const std::vector<COptimizationDof>& aOptDof,                              
                              double penalty_coeff_joint, double penalty_coeff_cont);
void ItrOptimizeEvalFunc_GradMultDir(CJointedRigidBody& jrb, double eps, 
                                     const std::vector<COptimizationDof>& aOptDof,                                 
                                     double penalty_coeff_joint, double penalty_coeff_cont);  
void OptimizeEvalFunc_GradMultDirPrj(CJointedRigidBody& jrb, double alpha, 
                                     const std::vector<COptimizationDof>& aOptDof,                                     
                                     double penalty_coeff_joint, double penalty_coeff_cont);

void GetGradEvaluationFunctionAndSens
(double& eval, double& deval,
 CSimulationSensitivity& simSens,
 ////
 double penalty_coeff_joint, double penalty_coeff_cont,
 ////
 PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s, 
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 ////
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,                                        
 std::vector<CJoint_Fix>& aFix, 
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir,
 const std::vector<CExternalForceOnRigidBody>& aMyExForce,     
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP);  

double GetDerivativeEvaluationFunction
(PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s,
 const CSimulationSensitivity& simSens,
 double coeff_joint, double coeff_cont,
 /////
 std::vector<CRigid3D_Ini>& aRI,
 std::vector<CRigid3D_Def>& aRD,                        
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir, 
 ////
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 const std::vector<CContactPoints_RigidBody3D>& aFP);

void JointDegeneration
(const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                       
 const std::vector<COptimizationDof>& aOptDof,
 const std::vector<Com::CVector3D>& aGradJnt,
 std::vector<double>& upd );

void Projection_JointConnectivity
(const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CRigid3D_Def>& aRD,  
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const std::vector<COptimizationDof>& aOptDof,
 const std::vector<double>& upd_in, // same size as aOptDof
 std::vector<double>& upd_out,      // same size as aEvalDevalSensPRLM
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_cont);

void GetDesignSensitivity
(CDesignSensitivity& desSens, 
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CRigid3D_Def>& aRD,                          
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const std::vector<COptimizationDof>& aOptDof,                          
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm,
 const std::vector<CContactPoints_RigidBody3D>& aFP,                          
 unsigned int irb_pic, const Com::CVector3D& v);

void EditInitialConfig_JointedRigidBodyContactSystem_SlaveJoint
(std::vector<CRigid3D_Ini>& aRI,
 std::vector<CRigid3D_Def>& aRD, 
 std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 ////
 const std::vector<COptimizationDof>& aOptDof,
 const std::vector<CEvalGradSensPRLM>& aEvalDevalSensPRLM,
 const std::vector<double>& upd,  // dim of nOptDof
 const Com::CVector3D& cont_norm);

void PrintJointBendForceSensitivty(CJointedRigidBody& jrb,
                                   const std::vector<COptimizationDof>& aOptDof);

//void MakeSuggestionSupportMaterial(CJointedRigidBody& jrb);

void MakeSuggestionSupportMaterial2
(const CJointedRigidBody& jrb, 
 std::vector<CJointedRigidBody>& aSgg);
