/*
 *  ls_utility.h
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/24/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if !defined(FURNITURE_UTILITY_H)
#define FURNITURE_UTILITY_H

#include "delfem/matvec/matdia_blkcrs.h"
#include "delfem/matvec/vector_blk.h"


enum PARAM_TYPE { BOX_POS, BOX_ROT, BOX_LEN, JNT_POS };

class COptimizationDof
{
public:
  COptimizationDof(PARAM_TYPE itype, unsigned int iobj, unsigned int idim){
    this->itype = itype;
    this->iobj = iobj;
    this->idim = idim;
    //    this->iobj_m = -1;
  }
  /*  
   COptimizationDof(PARAM_TYPE itype, unsigned int iobj, unsigned int idim,
   unsigned int iobj_m, unsigned int iaxis_m)
   {
   this->itype = itype;
   this->iobj = iobj;
   this->idim = idim;
   
   //    this->iobj_m = iobj_m;
   //    this->iaxis_m = iaxis_m;
   }
   */ 
public:
  PARAM_TYPE itype;
  unsigned int iobj;
  unsigned int idim;
  
  //  int iobj_m;
  //  unsigned int iaxis_m;
};



static void AddMatrix
(MatVec::CMatDia_BlkCrs& mat, 
 unsigned int iblk0, unsigned int offsetr, unsigned int sizer,
 unsigned int iblk1, unsigned int offsetl, unsigned int sizel,
 const double* emat, double d)
{
  //  const unsigned int iblk0 = (is_rb_r) ? indr : indr + nRB;
  //  const unsigned int iblk1 = (is_rb_l) ? indl : indl + nRB;
  const unsigned int lencol = mat.LenBlkCol(iblk0);
  const unsigned int lenrow = mat.LenBlkRow(iblk1);
  assert( lencol*lenrow <= 81 );
  double tmp[81];
  for(unsigned int i=0;i<lencol*lenrow;i++){ tmp[i] = 0; }
  for(unsigned int i=0;i<sizer;i++){
    for(unsigned int j=0;j<sizel;j++){
      tmp[(i+offsetr)*lenrow + (j+offsetl)] += emat[i*sizel+j]*d;
    }
  }
  mat.Mearge(1,&iblk0, 1,&iblk1, lencol*lenrow, tmp );
}

static void Add_Vector(MatVec::CMatDia_BlkCrs& mat, 
                       unsigned int iblk0, unsigned int offsetr,
                       unsigned int iblk1, unsigned int offsetl,
                       const Com::CVector3D& vec, double d, bool is_column)
{
  const unsigned int lencol = mat.LenBlkCol(iblk0);
  const unsigned int lenrow = mat.LenBlkRow(iblk1);
  assert( lencol*lenrow <= 36 );
  double tmp[36];
  for(unsigned int i=0;i<lencol*lenrow;i++){ tmp[i] = 0; }
  if( is_column ){
    tmp[ (0+offsetr)*lenrow + (offsetl)] += vec.x*d;
    tmp[ (1+offsetr)*lenrow + (offsetl)] += vec.y*d;
    tmp[ (2+offsetr)*lenrow + (offsetl)] += vec.z*d;
  }
  else{
    tmp[ (offsetr)*lenrow + (0+offsetl)] += vec.x*d;
    tmp[ (offsetr)*lenrow + (1+offsetl)] += vec.y*d;
    tmp[ (offsetr)*lenrow + (2+offsetl)] += vec.z*d;
  }
  mat.Mearge(1,&iblk0, 1, &iblk1, lencol*lenrow, tmp );
}


static void AddMatrix
(MatVec::CMatDia_BlkCrs& mat, 
 unsigned int iblk0, unsigned int offsetr,
 unsigned int iblk1, unsigned int offsetl,
 const Com::CMatrix3& m, double d, bool isnt_trans )
{
  const unsigned int lencol = mat.LenBlkCol(iblk0);
  const unsigned int lenrow = mat.LenBlkRow(iblk1);
  assert( lencol*lenrow <= 81 );
  double tmp[81];
  for(unsigned int i=0;i<lencol*lenrow;i++){ tmp[i] = 0; }
  if( isnt_trans ){
    for(unsigned int i=0;i<3;i++){
      for(unsigned int j=0;j<3;j++){
        tmp[(i+offsetr)*lenrow + (j+offsetl)] += m.mat[i*3+j]*d;
      }
    }
  }
  else{
    for(unsigned int i=0;i<3;i++){
      for(unsigned int j=0;j<3;j++){
        tmp[(i+offsetr)*lenrow + (j+offsetl)] += m.mat[j*3+i]*d;
      }
    }
  }
  mat.Mearge(1,&iblk0, 1, &iblk1, lencol*lenrow, tmp );
}

static void AddVector(MatVec::CVector_Blk& v,
                 unsigned int iblk, unsigned int offset, unsigned int size,
                 const double* eres, double d )
{
  assert( offset+3 <= v.Len(iblk) );  
  for(unsigned int i=0;i<size;i++){
    v.AddValue(iblk,offset+i,eres[i]*d);
  }
}

static void AddVector(MatVec::CVector_Blk& v,
                 unsigned int iblk, unsigned int offset, 
                 const Com::CVector3D& vres, double d )
{
  assert( offset+3 <= v.Len(iblk) );
  v.AddValue(iblk,offset+0,vres.x*d);
  v.AddValue(iblk,offset+1,vres.y*d);
  v.AddValue(iblk,offset+2,vres.z*d);
}

static bool IsRayCollideQuad
(const Com::CVector3D& src, const Com::CVector3D& dir,
 ////
 const Com::CVector3D& org, const Com::CVector3D& n, 
 const Com::CVector3D& vec_h, const Com::CVector3D& vec_v, 
 double& rh, double& rv)
{  
  if( fabs( Com::Dot(dir,n) ) < 0.01 ) return false; 
  double t = Dot(org-src,n) / Dot(dir,n);
  Com::CVector3D p = src + t*dir - org;
  //  std::cout << Com::Dot(dir,n) << " " << t << " " << fabs( Dot(p,vec_h) ) / SquareLength(vec_h) << "  " << fabs( Dot(p,vec_v) ) / SquareLength(vec_v) << std::endl;
  rh = Dot(p,vec_h) / SquareLength(vec_h);
  rv = Dot(p,vec_v) / SquareLength(vec_v);
  if( fabs(rh) > 1 ) return false;
  if( fabs(rv) > 1 ) return false;
  return true;
}


static bool SensRayCollideQuad
(const Com::CVector3D& s, const Com::CVector3D& d,
 const Com::CVector3D& o, const Com::CVector3D& n, 
 const Com::CVector3D& h, const Com::CVector3D& v,
 double& rh, double& rv,
 ////
 const Com::CVector3D& _s, const Com::CVector3D& _d,
 const Com::CVector3D& _o, const Com::CVector3D& _n, 
 const Com::CVector3D& _h, const Com::CVector3D& _v, 
 double& _rh, double& _rv)
{  
  if( fabs( Com::Dot(d,n) ) < 0.01 ) return false; 
  double t = Dot(o-s,n) / Dot(d,n);
  double _t = (Dot(o-s,n)*( Dot(_d,n) + Dot(d,_n) ) - ( Dot(_o-_s,n) + Dot(o-s,_n) )*Dot(d,n) ) / ( Dot(d,n)*Dot(d,n) );
  Com::CVector3D p = s + t*d - o;
  Com::CVector3D _p = _s + _t*d + t*_d - _o;
  //  std::cout << Com::Dot(dir,n) << " " << t << " " << fabs( Dot(p,vec_h) ) / SquareLength(vec_h) << "  " << fabs( Dot(p,vec_v) ) / SquareLength(vec_v) << std::endl;
  rh = Dot(p,h) / SquareLength(h);
  _rh = ((Dot(_p,h) + Dot(p,_h))*Dot(h,h) - Dot(p,h)*2*Dot(_h,h) ) / (Dot(h,h)*Dot(h,h));
  rv = Dot(p,v) / SquareLength(v);  
  _rv = ((Dot(_p,v) + Dot(p,_v))*Dot(v,v) - Dot(p,v)*2*Dot(_v,v) ) / (Dot(v,v)*Dot(v,v));
  if( fabs(rh) > 1 ) return false;
  if( fabs(rv) > 1 ) return false;
  return true;
}

#endif
