/*
 *  gl_drawer_furniture.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/31/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if defined(__APPLE__) && defined(__MACH__)
#  include <GLUT/glut.h>
#else
#  include <GL/glut.h>
#endif


#include "delfem/vector3d.h"
#include "delfem/matrix3d.h"

#include "furniture_utility.h"
#include "rigidbody.h"
#include "joint_fix.h"
#include "jointed_rigidbody_contact.h"
#include "nail_bound_furniture.h"
#include "gl_drawer_furniture.h"


void DrawRigidBody_Ini
(const CRigid3D_Ini& ri,              
 const float* color )
// double r, double g, double b, double a)
// unsigned int imode)  // mode 0:normal, mode 1:hilight, mode 2:gohst
{  
  const bool is_light = glIsEnabled(GL_LIGHTING);
  const bool is_blend = glIsEnabled(GL_BLEND);
  
  ::glPushMatrix();
  const Com::CVector3D& v = ri.GetIniPosCG();
  ::glTranslated(v.x,v.y,v.z);
  Com::CMatrix3 m = ri.GetRotMatrix_Ini();
  double rot0[16];
  m.GetAffineTransMatElements(rot0);
  ::glMultMatrixd(rot0);          
  ::glScaled(ri.xlen,ri.ylen,ri.zlen);
  ::glEnable(GL_BLEND);
  
  {
    ::glDisable(GL_LIGHTING);
    ::glColor3d(0,0,0);
    ::glLineWidth(1);          
    ::glutWireCube(1.0);     
    if( is_light ){
      ::glEnable(GL_LIGHTING);          
      glMaterialfv(GL_FRONT, GL_DIFFUSE, color);                 
      GLfloat shininess[] = {0.0};    
      glMaterialfv(GL_FRONT, GL_SHININESS, shininess);    
      glEnable(GL_NORMALIZE);
    }
    else{
      glColor4fv(color);
    }
    
    ::glEnable(GL_CULL_FACE);
    ::glCullFace(GL_BACK);
    ::glutSolidCube(1.0);                     
    ::glDisable(GL_CULL_FACE);
    ::glPopMatrix();        
  }
  if( !is_light ){ ::glDisable(GL_LIGHTING); }
  else{ ::glEnable(GL_LIGHTING); }  
  if( !is_blend ){ ::glDisable(GL_BLEND); }
  else{ ::glEnable(GL_BLEND); }  
}


void DrawRigidBody_Def(const CRigid3D_Ini& ri, 
                       const CRigid3D_Def& rd,                         
                       const float* color)
//                       double r, double g, double b, double a)  // mode 0:normal, mode 1:hilight, mode 2:gohst
{  
  const bool is_light = glIsEnabled(GL_LIGHTING);

  ::glPushMatrix();
  const Com::CVector3D& v = ri.GetIniPosCG() + rd.GetDispCG();
  ::glTranslated(v.x,v.y,v.z);
  Com::CMatrix3 m = rd.GetRotMatrix_Inc().MatMat( ri.GetRotMatrix_Ini() );
  double rot0[16];
  m.GetAffineTransMatElements(rot0);
  ::glMultMatrixd(rot0);  
  ::glScaled(ri.xlen,ri.ylen,ri.zlen);

  
  {
    ::glDisable(GL_LIGHTING);
    ::glColor3d(0,0,0);
    ::glLineWidth(1);          
    ::glutWireCube(1.0);     
    if( is_light ){
      ::glEnable(GL_LIGHTING);         
      glMaterialfv(GL_FRONT, GL_DIFFUSE, color);          
      GLfloat shininess[] = {0.0};    
      glMaterialfv(GL_FRONT, GL_SHININESS, shininess);            
      glEnable(GL_NORMALIZE);      
    }
    else{
      ::glColor4fv(color);
//      ::glColor4d(0.9,0.9,0.9,0.5);
    }
    
    ::glEnable(GL_CULL_FACE);
    ::glCullFace(GL_BACK);
    ::glutSolidCube(1.0);                     
    ::glDisable(GL_CULL_FACE);
  }
    
  ::glPopMatrix();      
  
    if( !is_light ){ ::glDisable(GL_LIGHTING); }
  else{ ::glEnable(GL_LIGHTING); }
}


void DrawFixJoint(bool is_initial, bool is_selscted, 
                  const CJoint_Fix& cjs, 
                  const CJoint_Fix_Force& jff,
                  const std::vector<CRigid3D_Ini>& aRI,
                  const std::vector<CRigid3D_Def>& aRD)
{	
  //		std::cout << "CJoint_Spherical" << std::endl;			
  const bool is_light = glIsEnabled(GL_LIGHTING);  
  
  const CRigid3D_Ini& ri0 = aRI[cjs.irb0];
  const CRigid3D_Ini& ri1 = aRI[cjs.irb1];
  const CRigid3D_Def& rd0 = aRD[cjs.irb0];
  const CRigid3D_Def& rd1 = aRD[cjs.irb1];  
  
  Com::CVector3D vec_cg0, vec_cg1;
  if( !is_initial ){
    vec_cg0 = ri0.GetIniPosCG() + rd0.GetDispCG();
    vec_cg1 = ri1.GetIniPosCG() + rd1.GetDispCG();      
  }
  else{
    vec_cg0 = ri0.GetIniPosCG();
    vec_cg1 = ri1.GetIniPosCG();      
  }
  
  Com::CVector3D vec_j;
  if( !is_initial ){
    vec_j = GetCurrentPositionFromInitial(cjs.ini_pos_joint,ri0,rd0); // joint pos
  }
  else{
    vec_j = cjs.ini_pos_joint;
  }
  
  double scale = 10;
  const Com::CVector3D lx = rd0.GetRotMatrix_Inc().MatVec(cjs.loc_coord[0])*scale*jff.lambda[3];
  const Com::CVector3D ly = rd0.GetRotMatrix_Inc().MatVec(cjs.loc_coord[1])*scale*jff.lambda[4];
  const Com::CVector3D lz = rd0.GetRotMatrix_Inc().MatVec(cjs.loc_coord[2])*scale*jff.lambda[5];
  const Com::CVector3D l = (lx+ly+lz)*0.01;
  
  Com::CVector3D lpos(jff.lambda[0],jff.lambda[1],jff.lambda[2]);
  lpos *= 0.001;
  
  {
    ::glPushMatrix();
    ::glTranslated( vec_j.x, vec_j.y, vec_j.z );  // joint position  
    
    GLfloat pink[4] = {1,0,1,1};
    GLfloat coba[4] = {0,1,1,1};
    
    if( is_light ){
      if( is_selscted ){ glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, pink); }
      else{ glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, coba); }
    }
    else{
      if( is_selscted ){ ::glColor3fv(pink); }
      else{ ::glColor3fv(coba); }
    }
    ::glutSolidSphere(0.03,10,5);
    ::glPopMatrix();
  }
  ::glDisable(GL_LIGHTING);
  { // draw connectivity
    ::glColor3d(1,1,1);
    ::glLineWidth(2);
    ::glBegin(GL_LINES);
    ::glVertex3d(vec_j.x,  vec_j.y,  vec_j.z  );
    ::glVertex3d(vec_cg0.x,vec_cg0.y,vec_cg0.z);
    ::glVertex3d(vec_j.x,  vec_j.y,  vec_j.z);
    ::glVertex3d(vec_cg1.x,vec_cg1.y,vec_cg1.z);
    ::glEnd();
  }
  
  { // draw force
    ::glColor3d(1,0,0);    
    ::glBegin(GL_LINES);
    ::glVertex3d(vec_j.x,  vec_j.y,  vec_j.z  );
    ::glVertex3d(vec_j.x+l.x,  vec_j.y+l.y,  vec_j.z+l.z  );      
    ::glEnd();
    ::glColor3d(0,0,1); 
    ::glBegin(GL_LINES);
    ::glVertex3d(vec_j.x,  vec_j.y,  vec_j.z  );
    ::glVertex3d(vec_j.x+lpos.x,  vec_j.y+lpos.y,  vec_j.z+lpos.z  );      
    ::glEnd();      
  }
  
  if( !is_light ){ ::glDisable(GL_LIGHTING); }
  else{ ::glEnable(GL_LIGHTING); }
  return;  
}

void myGLVertex3d(const Com::CVector3D& v){
  ::glVertex3d(v.x,v.y,v.z);
}

void DrawNails
(unsigned int idir0, unsigned int idir1,
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0, 
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1, 
 bool is_current,
 const CNailJoint& nj,
 bool isnt_brake)  // nail is go through rb1 to rb0 
{
  bool is_light = glIsEnabled(GL_LIGHTING);
//  std::cout << "Enter DrawNails" << std::endl;
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };
  
  assert( idir0==4 || idir0==5 );
  assert( idir1!=4 && idir1!=5 );
  const unsigned int nnail = nj.aPosHead.size();
  const double lennail = nj.lennail;
  
  const Com::CVector3D& b1 = aFaceNormal[idir1]*0.5;
//  const Com::CVector3D& h1 = Cross( Com::CVector3D(0,0,1), b1 )*2;    
//  const Com::CVector3D& Ba1 = Com::CVector3D(ri1.xlen*b1.x, ri1.ylen*b1.y, ri1.zlen*b1.z);
  const Com::CMatrix3& P1 = rd1.GetRotMatrix_Inc().MatMat(ri1.GetRotMatrix_Ini());
  const Com::CMatrix3& Q1 = ri1.GetRotMatrix_Ini();
//  const Com::CVector3D& Rh1 = R1.MatVec(h1);
  
  ::glDisable(GL_LIGHTING);
  ::glLineWidth(3);  
  if( isnt_brake ){ ::glColor3d(0,0,0); } 
  else{ ::glColor3d(1,0,0); }
  
  bool is_cull = glIsEnabled(GL_CULL_FACE);
  glDisable(GL_CULL_FACE);
  assert( nnail == nj.aPosHead.size() );
  for(unsigned int i=0;i<nnail;i++){    
    Com::CVector3D ph2 = nj.aPosHead[i];
    Com::CVector3D Rb1 = Q1.MatVec(b1)*2;
    if( is_current ){
      ph2 = rd1.GetRotMatrix_Inc().MatVec(ph2-ri1.GetIniPosCG())+ri1.GetIniPosCG()+rd1.GetDispCG();
      Rb1 = rd1.GetRotMatrix_Inc().MatVec(Rb1);
    }
    const Com::CVector3D& v3 = ph2 - lennail*Rb1; 
    const Com::CVector3D& v4 = ph2 + (ri1.zlen*0.05)*Rb1;
    ::glBegin(GL_LINES);    
    ::myGLVertex3d(v4);
    ::myGLVertex3d(v3);    
    ::glEnd(); 
    const Com::CVector3D& n1 = Q1.MatVec(b1)*2;
    Com::CVector3D h0,h1;
    GetVertical2Vector(n1, h0, h1);
    unsigned int ndiv = 20;
    double h = 3.1415*2/ndiv;
    double rad = nj.dia_head*0.5*2;
    ::glBegin(GL_TRIANGLES);
    for(unsigned int i=0;i<ndiv;i++){      
      Com::CVector3D v5 = v4 + rad*(sin(h*i)*h0 + cos(h*i)*h1);
      Com::CVector3D v6 = v4 + rad*(sin(h*i+h)*h0 + cos(h*i+h)*h1);
      myGLVertex3d(v4);
      myGLVertex3d(v5);
      myGLVertex3d(v6);     
    }
    ::glEnd();
  }
  if( is_cull ){ glEnable(GL_CULL_FACE); }
  if( is_light ){ ::glEnable(GL_LIGHTING); }
}


void DrawNails(const CJoint_Fix& cjs, 
               unsigned int idir0, unsigned int idir1,
               const std::vector<CRigid3D_Ini>& aRI, 
               const std::vector<CRigid3D_Def>& aRD,                
               bool is_current,
               const CNailJoint& nj, 
               bool isnt_brake)
{
  unsigned int irb0 = cjs.irb0;
  unsigned int irb1 = cjs.irb1;  
  DrawNails(idir0,idir1,
            aRI[irb0],aRD[irb0],
            aRI[irb1],aRD[irb1],
            is_current,
            nj,
            isnt_brake);    
}




int PickBox_Ini
(const Com::CVector3D& src, const Com::CVector3D& dir, 
 Com::CVector3D& hitpos, unsigned int& iface,
 const std::vector<CRigid3D_Ini>& aRI)
{  
  unsigned int irb_min; double mindepth; bool is_some_hit = false;  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    Com::CVector3D hitp;
    unsigned int iface0;
    bool is_hit = IsRayCollide_Ini(src,dir,hitp,iface0,aRI[irb]);
    if( !is_hit ) continue;
    const double depth = Dot(hitp-src,dir);    
    if( !is_some_hit ){
      is_some_hit = true;
      mindepth = depth;
      irb_min  = irb;      
      hitpos = hitp;
      iface = iface0;
    }
    else{
      if( depth < mindepth ){ 
        mindepth = depth;
        irb_min  = irb;
        hitpos = hitp;        
        iface = iface0;
      }
    }
  }
  if( is_some_hit ){ return irb_min; }
  return -1;  
}


void DrawPickedPlank(const CDesignSensitivity& desSens,
                     const CSimulationSensitivity& simSens,
                     double eps,
                     unsigned int irb,
                     const CRigid3D_Ini& ri0, 
                     const CRigid3D_Def& rd0,
                     const CContactPoints_RigidBody3D& aCP)
{
//  if( !is_rigid_picked_obj ) return;
//  if( ind_picked_obj >= aRI.size() ) return;
//  unsigned int irb = ind_picked_obj;
  const CDesignSensPlank& dsp = desSens.aPlank[irb];
  const Com::CVector3D& d = dsp.p;
  const Com::CVector3D& q = dsp.r;  
  const Com::CVector3D& l = dsp.l;    
  //  const Com::CVector3D& u = simSens.aPlankPR[irb].first;
  //  const Com::CVector3D& p = simSens.aPlankPR[irb].second;  
  CRigid3D_Ini ri = ri0;
  CRigid3D_Def rd = rd0;  
  ri.SetIniPosCG( ri.GetIniPosCG() + d*eps );
//  DrawRigidBody_Ini(ri,2);
}


void DrawExternalForce
(const CRigid3D_Ini& ri,
 const CRigid3D_Def& rd,
 const CExternalForceOnRigidBody& ef,
 bool is_current,
 const Com::CVector3D& zscr)
{
  bool is_lighting = glIsEnabled(GL_LIGHTING);
  ::glDisable(GL_LIGHTING);
  
  ::glEnable(GL_CULL_FACE);
  ::glCullFace(GL_BACK);
  ::glColor4d(0,0,0,0.8);
  Com::CVector3D p;
  if( is_current ){ p = GetCurrentPosition(ef.pos,ri,rd); }
  else{ p = ri.GetInitialPosition(ef.pos); }
  double size = sqrt(Com::Length(ef.force));
  size *= 0.005;
  unsigned int ndiv = 24;
  double rad = size*0.5;  
  Com::CVector3D ev = -ef.force; ev.SetNormalizedVector();
  Com::CVector3D ex,ey; GetVertical2Vector(ev,ex,ey);
  double rdiv = 3.1415*2/ndiv;
  ::glBegin(GL_QUADS);  
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D v0 = p + rad*ex*sin(rdiv*(idiv+0)) + rad*ey*cos(rdiv*(idiv+0));
    Com::CVector3D v1 = p + rad*ex*sin(rdiv*(idiv+1)) + rad*ey*cos(rdiv*(idiv+1));
    Com::CVector3D v2 = v0 + size*ev;
    Com::CVector3D v3 = v1 + size*ev;
    myGLVertex3d(v2);    
    myGLVertex3d(v3);
    myGLVertex3d(v1);
    myGLVertex3d(v0);
  }
  std::vector< std::pair<double,double> > prof;
  {
    prof.push_back( std::make_pair(1.0,1.0) );
    prof.push_back( std::make_pair(0.5,1.0) );    
    prof.push_back( std::make_pair(0.5,1.1) );
    unsigned int ni = 10;
    double h = 0.2;
    double r = h/ni;
    double a = (0.5*h*0.5*h);
    for(unsigned int i=0;i<=ni;i++){
      prof.push_back( std::make_pair(0.5+0.2*(a-(r*i-0.5*h)*(r*i-0.5*h))/a,1.1+r*i) );    
    }
  }
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D r0 = ex*sin(rdiv*(idiv+0)) + ey*cos(rdiv*(idiv+0));
    Com::CVector3D r1 = ex*sin(rdiv*(idiv+1)) + ey*cos(rdiv*(idiv+1)); 
    for(unsigned int jdiv=0;jdiv<prof.size()-1;jdiv++){
      double rad2 = rad*prof[jdiv+0].first;
      double h2   = size*prof[jdiv+0].second; 
      double rad3 = rad*prof[jdiv+1].first;
      double h3   = size*prof[jdiv+1].second;          
//      std::cout << jdiv << " " << prof[jdiv].first << " " << prof[jdiv].second << std::endl;
      Com::CVector3D v0 = p + rad2*r0 + ev*h2;
      Com::CVector3D v2 = p + rad3*r0 + ev*h3;       
      Com::CVector3D v1 = p + rad2*r1 + ev*h2;
      Com::CVector3D v3 = p + rad3*r1 + ev*h3;
      myGLVertex3d(v1);    
      myGLVertex3d(v0);
      myGLVertex3d(v2);
      myGLVertex3d(v3);      
    }
  }  
  ::glEnd();
  ::glBegin(GL_TRIANGLES);  
  for(unsigned int idiv=0;idiv<ndiv;idiv++){
    Com::CVector3D r0 = ex*sin(rdiv*(idiv+0)) + ey*cos(rdiv*(idiv+0));
    Com::CVector3D r1 = ex*sin(rdiv*(idiv+1)) + ey*cos(rdiv*(idiv+1)); 
    /*
    Com::CVector3D v2 = p + rad*r0 + size*ev;
    Com::CVector3D v3 = p + rad*r1 + size*ev;
    Com::CVector3D v4 = p + size*ev;
    ::myGLVertex3d(v3);
    ::myGLVertex3d(v2);    
    ::myGLVertex3d(v4);
     */
    Com::CVector3D v5 = p + 0.5*rad*r0 + 1.3*size*ev;
    Com::CVector3D v6 = p + 0.5*rad*r1 + 1.3*size*ev;
    Com::CVector3D v7 = p +              1.3*size*ev;    
    ::myGLVertex3d(v6);
    ::myGLVertex3d(v5);    
    ::myGLVertex3d(v7);    
  }
  ::glEnd();
  ::glDisable(GL_CULL_FACE);    
  if( is_lighting ){
    ::glEnable(GL_LIGHTING);
  }
  /*
  Com::CVector3D p1 = p+a;
  const double lenf = Com::Length(ef.force);
  Com::CVector3D e1 = ef.force / lenf;
  Com::CVector3D e2 = Com::Cross(e1,zscr);  
  ::glColor3d(1,0,1);              
  ::glLineWidth(2);
  ::glBegin(GL_LINES);
  ::myGLVertex3d(p);
  ::myGLVertex3d(p1);
  double ratio = 0.2*lenf*0.001;
  ::myGLVertex3d(p1);
  ::myGLVertex3d(p1-(e1+e2)*ratio);
  ::myGLVertex3d(p1);
  ::myGLVertex3d(p1-(e1-e2)*ratio);  
  ::glEnd();  
   */
  
}


void DrawContactForce
(const CRigid3D_Ini& ri,
 const CRigid3D_Def& rd,
 const CContactPoints_RigidBody3D& aCP,
 const CContactProp& contProp)
{
  for(unsigned int i=0;i<aCP.Size();i++){
    if( aCP.aFP[i].is_force_cont != 1 ) continue;    
    Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);   
    Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);        
    Com::CVector3D v2 = aCP.aFP[i].aloc;
    Com::CVector3D dir_f = v2-v1;
    const Com::CVector3D& cont_norm = contProp.cont_norm;
    const Com::CVector3D& cont_base = contProp.cont_base;        
    double stiff_f = contProp.stiff_f;
    double stiff_n = contProp.stiff_n;        
    dir_f -= cont_norm*Com::Dot(cont_norm,dir_f);
    Com::CVector3D v3 = v1 + (stiff_f*dir_f)*0.001;
    ::glColor3d(1,1,0);
    ::myGLVertex3d(v1);
    ::myGLVertex3d(v3);    
    Com::CVector3D dir_n = cont_norm*Com::Dot(cont_norm,cont_base-v1);
    Com::CVector3D v4 = v1 + (stiff_n*dir_n)*0.001;
    ::glColor3d(1,0,1);        
    ::myGLVertex3d(v1);
    ::myGLVertex3d(v4);
  }        
}

void DrawFloor(const CContactProp& contProp,
               float r, float g, float b)
{ // draw floor
  double len = 10.0;
  bool is_light = glIsEnabled(GL_LIGHTING);
  if( is_light ){
    GLfloat material[] = { r, g, b, 1.0f };      
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material);      
    GLfloat spec[] = { 0.0f, 0.0f, 0.0f, 1.0f };      
    glMaterialfv(GL_FRONT, GL_SPECULAR, spec);            
    GLfloat shininess[] = {125.0f};    
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);                  
  }
  else{
    ::glColor3d(r,g,b);
  }
  const Com::CVector3D& cont_norm = contProp.cont_norm;
  const Com::CVector3D& cont_base = contProp.cont_base;    
  Com::CVector3D evec_x, evec_y;
  GetVertical2Vector(cont_norm, evec_x,evec_y);
  Com::CVector3D vec_x = len*evec_x;
  Com::CVector3D vec_y = len*evec_y;    
  ::glBegin(GL_QUADS);
  ::glNormal3d(cont_norm.x,cont_norm.y,cont_norm.z);    
  myGLVertex3d(cont_base+vec_x+vec_y);
  myGLVertex3d(cont_base-vec_x+vec_y);
  myGLVertex3d(cont_base-vec_x-vec_y);    
  myGLVertex3d(cont_base+vec_x-vec_y);    
  ::glEnd();
  
  ::glDisable(GL_LIGHTING);
  ::glLineWidth(1);
  double elen = 1;
  int ndiv = (double)len/elen;    
  ::glColor3d(0.5,0.5,0.5);
  ::glBegin(GL_LINES);
  for(int i=-ndiv;i<ndiv;i++){
    myGLVertex3d(cont_base+vec_x+(i*elen)*evec_y);
    myGLVertex3d(cont_base-vec_x+(i*elen)*evec_y);
    myGLVertex3d(cont_base+vec_y+(i*elen)*evec_x);
    myGLVertex3d(cont_base-vec_y+(i*elen)*evec_x);      
  }
  ::glEnd();
  if( is_light ){ glEnable(GL_LIGHTING); }
}  
