/*
 *  joint_fix.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/24/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if defined(__VISUALC__)
#pragma warning ( disable : 4786 ) 
#pragma warning ( disable : 4996 )
#endif
#define for if(0);else for

#if defined(_WIN32)
#  include <windows.h>
#if defined(__VISUALC__)
#  pragma comment (lib, "winmm.lib")      /* link with Windows MultiMedia lib */
#  pragma comment (lib, "opengl32.lib")  /* link with Microsoft OpenGL lib */
#  pragma comment (lib, "glu32.lib")     /* link with Microsoft OpenGL Utility lib */
#endif  /* _WIN32 */
#endif

#if defined(__APPLE__) && defined(__MACH__)
#  include <OpenGL/gl.h>
#  include <OpenGL/glu.h>
#else
#  include <GL/gl.h>
#  include <GL/glu.h>
#endif

#include <iostream>

#include "delfem/matrix3d.h"

#include "rigidbody.h"
#include "joint_fix.h"
#include "furniture_utility.h"
/*
void CJoint_Fix::UpdateLambda_NewmarkBetaAPrime
(const double* upd, double dt, double newmark_gamma, double newmark_beta ){
  double tmp = dt*dt*newmark_beta;
  lambda[0] += tmp*upd[0];
  lambda[1] += tmp*upd[1];
  lambda[2] += tmp*upd[2];
  lambda[3] += tmp*upd[3];
  lambda[4] += tmp*upd[4];
  lambda[5] += tmp*upd[5];    
}

void CJoint_Fix::AddLinearSystem_NewmarkBetaAPrime
(Ls::CLinearSystem_RigidBody& ls, unsigned int icst,
 const double dt, const double newmark_gamma, const double newmark_beta,
 const std::vector<CRigidBody3D>& aRB,
 bool is_initial ) const
{
  const unsigned int irb0 = aIndRB[0];
  assert( irb0 < ls.GetSizeRigidBody() );
  const CRigidBody3D& rb0 = aRB[irb0];
  
  const unsigned int irb1 = aIndRB[1];
  assert( irb1 < ls.GetSizeRigidBody() );
  const CRigidBody3D& rb1 = aRB[irb1];
  
  const Com::CMatrix3& mrot0 = rb0.GetRotMatrix_Inc();
  const Com::CMatrix3& mrot1 = rb1.GetRotMatrix_Inc();
  const Com::CVector3D vlambda(lambda[0],lambda[1],lambda[2]);
  const Com::CVector3D rlambda(lambda[3],lambda[4],lambda[5]);  
  const Com::CVector3D Xdistfix0 = rb0.ri.GetIniPosCG() - ini_pos_joint;
  const Com::CVector3D Xdistfix1 = rb1.ri.GetIniPosCG() - ini_pos_joint;  
  
  ////////////////////////////////////////////////
  // 拘束力残差（物体０）
	// 並進の残差
  ls.AddResidual(irb0,true,0,  vlambda,+1);
	{   // 変位拘束力から来る回転の残差（物体０）
    const Com::CVector3D& tmp_vec = mrot0.MatVecTrans(vlambda);
    const Com::CMatrix3 wX(Xdistfix0);
    ls.AddResidual(irb0,true,3,  wX.MatVec(tmp_vec),-1);
	}
	{   // 変位拘束条件の変分（物体０）
    Com::CVector3D rot_pos_cg = mrot0.MatVec(Xdistfix0);
    ls.AddResidual(icst,false,0,   rb0.ri.GetIniPosCG()+rb0.rd.GetDispCG()-rot_pos_cg,1 );
	}
  ////////////////////////////////////////////////
  // 拘束力残差（物体１）
	// 並進の残差
  ls.AddResidual(irb1,true,0,  vlambda,-1);
	{ // relative translation constraint
    const Com::CVector3D& tmp_vec = mrot1.MatVecTrans(vlambda);
    const Com::CMatrix3 wX(Xdistfix1);
    ls.AddResidual(irb1,true,3,  wX.MatVec(tmp_vec),1 );
	}
	{ // relative translation constraint
    const Com::CVector3D& rot_pos_cg = mrot1.MatVec(Xdistfix1);
    ls.AddResidual(icst,false,0,  rb1.ri.GetIniPosCG()+rb1.rd.GetDispCG()-rot_pos_cg,-1);
	}
  
  /////
  { // relaxation term
    ls.AddResidual(icst,false,0,invstiff_pos*vlambda,+1);
  }
  
  const Com::CVector3D& R0x = mrot0.MatVec(loc_coord[0]);
  const Com::CVector3D& R0y = mrot0.MatVec(loc_coord[1]);
  const Com::CVector3D& R0z = mrot0.MatVec(loc_coord[2]);
  ////
  const Com::CVector3D& R1x = mrot1.MatVec(loc_coord[0]);
  const Com::CVector3D& R1y = mrot1.MatVec(loc_coord[1]);
  const Com::CVector3D& R1z = mrot1.MatVec(loc_coord[2]);      
  
  Com::CMatrix3 Rtwma[2];  
  { // rotational constraint
    const Com::CVector3D& wR0yR1z = Com::CMatrix3(R0y).MatVec(R1z);
    const Com::CVector3D& wR0zR1x = Com::CMatrix3(R0z).MatVec(R1x);
    const Com::CVector3D& wR0xR1y = Com::CMatrix3(R0x).MatVec(R1y);
    Com::CMatrix3 mtmp(wR0yR1z,wR0zR1x,wR0xR1y);
    Rtwma[0] = mrot0.MatMatTrans(mtmp);
    Rtwma[1] = mrot1.MatMatTrans(mtmp);    
  }
  
  ls.AddResidual(irb0,true,3, Rtwma[0].MatVec(rlambda), -1 );
  ls.AddResidual(irb1,true,3, Rtwma[1].MatVec(rlambda), +1 );
  {
    double res[6] = {0,0,0,0,0,0};
    res[3] = Com::Dot(R0y,R1z) - invstiff_rot*lambda[3];
    res[4] = Com::Dot(R0z,R1x) - invstiff_rot*lambda[4];
    res[5] = Com::Dot(R0x,R1y) - invstiff_rot*lambda[5];    
    ls.SubResidual(icst,false,res);    
  }
  
  ////////////
  Com::CMatrix3 RotwX0 = mrot0.MatMat( Com::CMatrix3(Xdistfix0) );
  Com::CMatrix3 wXwRtL0;
	{
    const Com::CVector3D& RtL = mrot0.MatVecTrans(vlambda);
    wXwRtL0 = Com::CMatrix3(Xdistfix0).MatMat( Com::CMatrix3(RtL) );
	}         
  Com::CMatrix3 RotwX1 = mrot1.MatMat( Com::CMatrix3(Xdistfix1) );
  Com::CMatrix3 wXwRtL1;
	{
    const Com::CVector3D& RtL = mrot1.MatVecTrans(vlambda);
    wXwRtL1 = Com::CMatrix3(Xdistfix1).MatMat( Com::CMatrix3(RtL) );
	}         
  
	{   // 剛性行列を作る
		const double dtmp1 = dt*dt*newmark_beta;
    Com::CMatrix3 mtmp;
    mtmp.SetIdentity();
    ls.AddMatrix(irb0,true, 0, icst,false,0, mtmp,   -dtmp1, true);
    ls.AddMatrix(icst,false,0, irb0,true, 0, mtmp,   -dtmp1, true);
    ls.AddMatrix(irb1,true, 0, icst,false,0, mtmp,    dtmp1, true);
    ls.AddMatrix(icst,false,0, irb1,true, 0, mtmp,    dtmp1, true);
    
    ls.AddMatrix(irb0,true,3, irb0,true,3, wXwRtL0,  dtmp1, true);
    ls.AddMatrix(irb1,true,3, irb1,true,3, wXwRtL1, -dtmp1, true);
    
    ls.AddMatrix(irb0,true, 3,  icst,false,0,  RotwX0, -dtmp1, false);
    ls.AddMatrix(icst,false,0,  irb0,true, 3,  RotwX0, -dtmp1, true );
    ls.AddMatrix(irb1,true, 3,  icst,false,0,  RotwX1, +dtmp1, false);
    ls.AddMatrix(icst,false,0,  irb1,true, 3,  RotwX1, +dtmp1, true );    
    
    ls.AddMatrix(       irb0,true, 3,  icst,false,3, +Rtwma[0],    dtmp1, true );
    ls.AddMatrix(       irb1,true, 3,  icst,false,3, -Rtwma[1],    dtmp1, true );    
    ls.AddMatrix(       icst,false,3,  irb0,true, 3, +Rtwma[0],    dtmp1, false);    
    ls.AddMatrix(       icst,false,3,  irb1,true, 3, -Rtwma[1],    dtmp1, false);            
    
    {
      Com::CMatrix3 esp_pos;
      esp_pos.SetIdentity();
      esp_pos *= -invstiff_pos;
      ls.AddMatrix(icst,false,0, icst,false,0, esp_pos, dtmp1, true);      
    }
    {
      Com::CMatrix3 esp_rot;
      esp_rot.SetIdentity();
      esp_rot *= -invstiff_rot;      
      ls.AddMatrix(icst,false,3, icst,false,3, esp_rot, dtmp1, true);            
    }
	}
}
*/
/*
void CJoint_Fix::UpdateLambda_BackwardEular
(const double* upd, double dt)
{
  for(unsigned int i=0;i<6;i++){ dlambda[i] += upd[i]; }
  for(unsigned int i=0;i<6;i++){ lambda[i] += dt*dlambda[i]; };    
}
 */

void AddResidual_JointFix
(std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResPR,
 std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResLM,
 unsigned int nRB, 
 unsigned int icst, const CJoint_Fix& jF, const CJoint_Fix_Force& jFF, 
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1 )
{
//  const unsigned int irb0 = aIndRB[0];
//  const CRigidBody3D& rb0 = aRB[irb0];
  
//  const unsigned int irb1 = aIndRB[1];
//  const CRigidBody3D& rb1 = aRB[irb1];
  
  const Com::CMatrix3& mrot0 = rd0.GetRotMatrix_Inc();
  const Com::CMatrix3& mrot1 = rd1.GetRotMatrix_Inc();
  const Com::CVector3D vlambda(jFF.lambda[0],jFF.lambda[1],jFF.lambda[2]);  // position constraint
  const Com::CVector3D rlambda(jFF.lambda[3],jFF.lambda[4],jFF.lambda[5]);  // rotation constraint
  const Com::CVector3D Xdistfix0 = ri0.GetIniPosCG() - jF.ini_pos_joint;
  const Com::CVector3D Xdistfix1 = ri1.GetIniPosCG() - jF.ini_pos_joint;  
  
  const Com::CVector3D& R0x = mrot0.MatVec(jF.loc_coord[0]);
  const Com::CVector3D& R0y = mrot0.MatVec(jF.loc_coord[1]);
  const Com::CVector3D& R0z = mrot0.MatVec(jF.loc_coord[2]);
  ////
  const Com::CVector3D& R1x = mrot1.MatVec(jF.loc_coord[0]);
  const Com::CVector3D& R1y = mrot1.MatVec(jF.loc_coord[1]);
  const Com::CVector3D& R1z = mrot1.MatVec(jF.loc_coord[2]);    
  
  Com::CMatrix3 Rtwma[2];  
  { // rotational constraint
    const Com::CVector3D& wR0yR1z = Com::CMatrix3(R0y).MatVec(R1z);
    const Com::CVector3D& wR0zR1x = Com::CMatrix3(R0z).MatVec(R1x);
    const Com::CVector3D& wR0xR1y = Com::CMatrix3(R0x).MatVec(R1y);
    Com::CMatrix3 mtmp(wR0yR1z,wR0zR1x,wR0xR1y);
    Rtwma[0] = mrot0.MatMatTrans(mtmp);
    Rtwma[1] = mrot1.MatMatTrans(mtmp);    
  }
  
  unsigned int irb0 = jF.irb0;
  unsigned int irb1 = jF.irb1;  
  
  { // rv0
    aResPR[irb0].first += vlambda;
  }
	{ // rr0
    Com::CVector3D rr0 = -Com::CMatrix3(Xdistfix0).MatVec( mrot0.MatVecTrans(vlambda) );
    rr0 -= Rtwma[0].MatVec(rlambda); 
    aResPR[irb0].second += rr0;
	}
  { // rv1
    aResPR[irb1].first -= vlambda;
  }
	{ // rr1
    Com::CVector3D rr1 = Com::CMatrix3(Xdistfix1).MatVec( mrot1.MatVecTrans(vlambda) );
    rr1 += Rtwma[1].MatVec(rlambda);
    aResPR[irb1].second += rr1;
	}
  { // lv
    Com::CVector3D rcv(0,0,0);
    rcv += ri0.GetIniPosCG()+rd0.GetDispCG()-mrot0.MatVec(Xdistfix0);
    rcv -= ri1.GetIniPosCG()+rd1.GetDispCG()-mrot1.MatVec(Xdistfix1);
    rcv += jF.invstiff_pos*vlambda;
    aResLM[icst].first += rcv;
	}  
  { // lr
    Com::CVector3D rcr(0,0,0);
    rcr.x = - Com::Dot(R0y,R1z) + jF.invstiff_rot*jFF.lambda[3];
    rcr.y = - Com::Dot(R0z,R1x) + jF.invstiff_rot*jFF.lambda[4];
    rcr.z = - Com::Dot(R0x,R1y) + jF.invstiff_rot*jFF.lambda[5];    
    aResLM[icst].second += rcr;
  }
}

void AddSensitivityResidual_StaticJoint
(std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResPR,
 std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResLM,
 ////
 unsigned int nRB,
 unsigned int icst, const CJoint_Fix& jF,  const CJoint_Fix_Force& jFF, 
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1,
 bool is_rb_pos, int iobj_s, unsigned int idim_s, double alpha)
{
  const unsigned int irb0 = jF.irb0;  
  const unsigned int irb1 = jF.irb1;
  
  const Com::CMatrix3& mrot0 = rd0.GetRotMatrix_Inc();
  const Com::CMatrix3& mrot1 = rd1.GetRotMatrix_Inc();
  const Com::CVector3D vlambda(jFF.lambda[0],jFF.lambda[1],jFF.lambda[2]);
  const Com::CVector3D Xdistfix0 = ri0.GetIniPosCG() - jF.ini_pos_joint;
  const Com::CVector3D Xdistfix1 = ri1.GetIniPosCG() - jF.ini_pos_joint;  
  
  ////////////////////////////////////////////////

  Com::CVector3D e_s(0,0,0); e_s[idim_s] = alpha;
  
  if( is_rb_pos ){
    if( iobj_s == irb0 ){
      const Com::CVector3D& Ptl = mrot0.MatVecTrans(vlambda);
      aResPR[irb0].second +=  Com::CMatrix3(Ptl).MatVec(e_s);
      ////    
      Com::CMatrix3 dg_dd = Com::CMatrix3(1.0)-mrot0;
      aResLM[icst].first += dg_dd.MatVec(e_s);
    }
    if( iobj_s == irb1 ){
      const Com::CVector3D& Ptl = mrot1.MatVecTrans(vlambda);
      aResPR[irb1].second -= Com::CMatrix3(Ptl).MatVec(e_s);
      ////    
      Com::CMatrix3 dg_dd = Com::CMatrix3(1.0)-mrot1;
      aResLM[icst].first -= dg_dd.MatVec(e_s);
    }      
  }
  else{    
    if( iobj_s == icst ){
      {
        const Com::CVector3D& Ptl = mrot0.MatVecTrans(vlambda);
        aResPR[irb0].second -= Com::CMatrix3(Ptl).MatVec(e_s);
        aResLM[icst].first  += mrot0.MatVec(e_s);
      }
      {
        const Com::CVector3D& Ptl = mrot1.MatVecTrans(vlambda);
        aResPR[irb1].second += Com::CMatrix3(Ptl).MatVec(e_s);
        aResLM[icst].first  -= mrot1.MatVec(e_s);
      }
    }
  }  
}

void AddSensitivityResidual_StaticJoint
(std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResPR,
 std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResLM,
 ////
 unsigned int nRB,
 unsigned int icst, const CJoint_Fix& jF,  const CJoint_Fix_Force& jFF, 
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1,
 bool is_rb_pos, int iobj_s, const Com::CVector3D& e_s)
{
  const unsigned int irb0 = jF.irb0;  
  const unsigned int irb1 = jF.irb1;
  
  const Com::CMatrix3& mrot0 = rd0.GetRotMatrix_Inc();
  const Com::CMatrix3& mrot1 = rd1.GetRotMatrix_Inc();
  const Com::CVector3D vlambda(jFF.lambda[0],jFF.lambda[1],jFF.lambda[2]);
  const Com::CVector3D Xdistfix0 = ri0.GetIniPosCG() - jF.ini_pos_joint;
  const Com::CVector3D Xdistfix1 = ri1.GetIniPosCG() - jF.ini_pos_joint;  
  
  ////////////////////////////////////////////////
  
  if( is_rb_pos ){
    if( iobj_s == irb0 ){
      const Com::CVector3D& Ptl = mrot0.MatVecTrans(vlambda);
      aResPR[irb0].second +=  Com::CMatrix3(Ptl).MatVec(e_s);
      ////    
      Com::CMatrix3 dg_dd = Com::CMatrix3(1.0)-mrot0;
      aResLM[icst].first += dg_dd.MatVec(e_s);
    }
    if( iobj_s == irb1 ){
      const Com::CVector3D& Ptl = mrot1.MatVecTrans(vlambda);
      aResPR[irb1].second -= Com::CMatrix3(Ptl).MatVec(e_s);
      ////    
      Com::CMatrix3 dg_dd = Com::CMatrix3(1.0)-mrot1;
      aResLM[icst].first -= dg_dd.MatVec(e_s);
    }      
  }
  else{    
    if( iobj_s == icst ){
      {
        const Com::CVector3D& Ptl = mrot0.MatVecTrans(vlambda);
        aResPR[irb0].second -= Com::CMatrix3(Ptl).MatVec(e_s);
        aResLM[icst].first  += mrot0.MatVec(e_s);
      }
      {
        const Com::CVector3D& Ptl = mrot1.MatVecTrans(vlambda);
        aResPR[irb1].second += Com::CMatrix3(Ptl).MatVec(e_s);
        aResLM[icst].first  -= mrot1.MatVec(e_s);
      }
    }
  }  
}



/*
void CJoint_Fix::AddLinearSystem_BackwardEular
(Ls::CLinearSystem_RigidBody& ls, unsigned int icst,
 const double dt, 
 const std::vector<CRigidBody3D>& aRB,
 bool is_initial ) const
{
  const unsigned int irb0 = aIndRB[0];
  assert( irb0 < ls.GetSizeRigidBody() );
  const CRigidBody3D& rb0 = aRB[irb0];
  
  const unsigned int irb1 = aIndRB[1];
  assert( irb1 < ls.GetSizeRigidBody() );
  const CRigidBody3D& rb1 = aRB[irb1];
  
  const Com::CMatrix3& mrot0 = rb0.GetRotMatrix_Inc();
  const Com::CMatrix3& mrot1 = rb1.GetRotMatrix_Inc();
  const Com::CVector3D vlambda(lambda[0],lambda[1],lambda[2]);  // position constraint
  const Com::CVector3D rlambda(lambda[3],lambda[4],lambda[5]);  // rotation constraint
  const Com::CVector3D Xdistfix0 = rb0.ri.GetIniPosCG() - ini_pos_joint;
  const Com::CVector3D Xdistfix1 = rb1.ri.GetIniPosCG() - ini_pos_joint;  
  
  const Com::CVector3D& R0x = mrot0.MatVec(loc_coord[0]);
  const Com::CVector3D& R0y = mrot0.MatVec(loc_coord[1]);
  const Com::CVector3D& R0z = mrot0.MatVec(loc_coord[2]);
  ////
  const Com::CVector3D& R1x = mrot1.MatVec(loc_coord[0]);
  const Com::CVector3D& R1y = mrot1.MatVec(loc_coord[1]);
  const Com::CVector3D& R1z = mrot1.MatVec(loc_coord[2]);    
  
  Com::CMatrix3 Rtwma[2];  
  { // rotational constraint
    const Com::CVector3D& wR0yR1z = Com::CMatrix3(R0y).MatVec(R1z);
    const Com::CVector3D& wR0zR1x = Com::CMatrix3(R0z).MatVec(R1x);
    const Com::CVector3D& wR0xR1y = Com::CMatrix3(R0x).MatVec(R1y);
    Com::CMatrix3 mtmp(wR0yR1z,wR0zR1x,wR0xR1y);
    Rtwma[0] = mrot0.MatMatTrans(mtmp);
    Rtwma[1] = mrot1.MatMatTrans(mtmp);    
  }
  
  ////////////
  Com::CMatrix3 RotwX0 = mrot0.MatMat( Com::CMatrix3(Xdistfix0) );
  Com::CMatrix3 wXwRtL0;
	{
    const Com::CVector3D& RtL = mrot0.MatVecTrans(vlambda);
    wXwRtL0 = Com::CMatrix3(Xdistfix0).MatMat( Com::CMatrix3(RtL) );
	}         
  Com::CMatrix3 RotwX1 = mrot1.MatMat( Com::CMatrix3(Xdistfix1) );
  Com::CMatrix3 wXwRtL1;
	{
    const Com::CVector3D& RtL = mrot1.MatVecTrans(vlambda);
    wXwRtL1 = Com::CMatrix3(Xdistfix1).MatMat( Com::CMatrix3(RtL) );
	}         
  
  const Com::CMatrix3 mInd(1.0);
	{
		const double dtmp1 = dt*dt;
    ls.AddMatrix(       irb0,true, 0,  icst,false,0, -mInd,        dtmp1, true);
    
    ls.AddMatrix(       irb0,true, 3,  irb0,true, 3,  wXwRtL0,     dtmp1, true);
    ls.AddMatrix(       irb0,true, 3,  icst,false,0, -RotwX0,      dtmp1, false);        
    ls.AddMatrix(       irb0,true, 3,  icst,false,3, Rtwma[0],    dtmp1, true);
    
    ls.AddMatrix(       irb1,true, 0,  icst,false,0,  mInd,        dtmp1, true);
    
    ls.AddMatrix(       irb1,true, 3,  irb1,true,3,  -wXwRtL1,     dtmp1, true);    
    ls.AddMatrix(       irb1,true, 3,  icst,false,0,  RotwX1,      dtmp1, false);    
    ls.AddMatrix(       irb1,true, 3,  icst,false,3, -Rtwma[1],    dtmp1, true);
    
    ls.AddMatrix(       icst,false,0,  irb0,true, 0, -mInd,        dtmp1, true);
    ls.AddMatrix(       icst,false,0,  irb1,true, 0,  mInd,        dtmp1, true);        
    ls.AddMatrix(       icst,false,0,  irb0,true, 3, -RotwX0,      dtmp1, true );
    ls.AddMatrix(       icst,false,0,  irb1,true, 3,  RotwX1,      dtmp1, true );    
    
    ls.AddMatrix(       icst,false, 3,  irb0,true,3, Rtwma[0],    dtmp1, false);    
    ls.AddMatrix(       icst,false, 3,  irb1,true,3,-Rtwma[1],    dtmp1, false);        
    
    ls.AddMatrix(icst,false,0, icst,false,0, Com::CMatrix3(-invstiff_pos), dtmp1, true);      
    ls.AddMatrix(icst,false,3, icst,false,3, Com::CMatrix3(-invstiff_rot), dtmp1, true);                
	}
  
  
  ////////////////////////////////////////////////
  const Com::CVector3D vdlam(dlambda[0],dlambda[1],dlambda[2]);  // position constraint
  const Com::CVector3D rdlam(dlambda[3],dlambda[4],dlambda[5]);  // rotation constraint  
  {
    Com::CVector3D rv0 = vlambda;
    rv0 -= dt*(-mInd).MatVec(vdlam);
    ls.AddResidual(irb0,true,0,  rv0,+dt);
  }
	{ // relative translation constraint
    Com::CVector3D rr0 = -Com::CMatrix3(Xdistfix0).MatVec( mrot0.MatVecTrans(vlambda) );
    rr0 -= Rtwma[0].MatVec(rlambda); 
    rr0 -= dt*wXwRtL0.MatVec(rb0.GetOmega());
    rr0 -= dt*(-RotwX0).MatVecTrans(vdlam);
    rr0 -= dt*(Rtwma[0]).MatVec(rdlam);
    ls.AddResidual(irb0,true,3,  rr0,+dt);    
	}
  {
    Com::CVector3D rv1 = -vlambda;    
    rv1 -= dt*(mInd).MatVec(vdlam);
    ls.AddResidual(irb1,true,0,  rv1,+dt);
  }
	{ // relative translation constraint
    Com::CVector3D rr1 = Com::CMatrix3(Xdistfix1).MatVec( mrot1.MatVecTrans(vlambda) );
    rr1 += Rtwma[1].MatVec(rlambda);
    rr1 -= dt*(-wXwRtL1).MatVec(rb1.GetOmega());
    rr1 -= dt*(RotwX1).MatVecTrans(vdlam);
    rr1 -= dt*(-Rtwma[1]).MatVec(rdlam);
    ls.AddResidual(irb1,true,3, rr1,+dt );
	}
  { // relative translation constraint
    Com::CVector3D rcv(0,0,0);
    rcv += rb0.ri.GetIniPosCG()+rb0.rd.GetDispCG()-mrot0.MatVec(Xdistfix0);
    rcv -= rb1.ri.GetIniPosCG()+rb1.rd.GetDispCG()-mrot1.MatVec(Xdistfix1);
    rcv += invstiff_pos*vlambda;
    rcv -= dt*(-mInd).MatVec(rb0.GetVeloCG());
    rcv -= dt*(-RotwX0).MatVec(rb0.GetOmega());    
    rcv -= dt*( mInd).MatVec(rb1.GetVeloCG());
    rcv -= dt*( RotwX1).MatVec(rb1.GetOmega());
    rcv -= dt*(-invstiff_pos)*vdlam;
    ls.AddResidual(icst,false,0, rcv,+dt );
	}  
  {
    Com::CVector3D rcr(0,0,0);
    rcr.x = - Com::Dot(R0y,R1z) + invstiff_rot*lambda[3];
    rcr.y = - Com::Dot(R0z,R1x) + invstiff_rot*lambda[4];
    rcr.z = - Com::Dot(R0x,R1y) + invstiff_rot*lambda[5];    
    rcr -= dt*( Rtwma[0]).MatVecTrans(rb0.GetOmega());
    rcr -= dt*(-Rtwma[1]).MatVecTrans(rb1.GetOmega());
    rcr -= dt*(-invstiff_rot)*rdlam;
    ls.AddResidual(icst,false,3,rcr,dt);
  }  
}
*/
void UpdateLambda_BackwardEular
(CJoint_Fix_Force& jff,
 const double* upd, double dt)
{
  for(unsigned int i=0;i<6;i++){ jff.dlambda[i] += upd[i]; }
  for(unsigned int i=0;i<6;i++){ jff.lambda[i] += dt*jff.dlambda[i]; };    
}

void AddLinearSystem_JointFix_Static
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 ////
 unsigned int nRB, 
 unsigned int icst, const CJoint_Fix& cjf, const CJoint_Fix_Force& jff,
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1)
{
  
  const unsigned int irb0 = cjf.irb0;
  assert( irb0 < mat.NBlkMatCol() );
  //  const CRigidBody3D& rb0 = aRB[irb0];
  
  const unsigned int irb1 = cjf.irb1;
  assert( irb1 < mat.NBlkMatCol() );
  //  const CRigidBody3D& rb1 = aRB[irb1];
  
  const Com::CMatrix3& mrot0 = rd0.GetRotMatrix_Inc();
  const Com::CMatrix3& mrot1 = rd1.GetRotMatrix_Inc();
  const Com::CVector3D vlambda(jff.lambda[0],jff.lambda[1],jff.lambda[2]);  // position constraint
  const Com::CVector3D rlambda(jff.lambda[3],jff.lambda[4],jff.lambda[5]);  // rotation constraint
  const Com::CVector3D Xdistfix0 = ri0.GetIniPosCG() - cjf.ini_pos_joint;
  const Com::CVector3D Xdistfix1 = ri1.GetIniPosCG() - cjf.ini_pos_joint;  
  
  const Com::CVector3D& R0x = mrot0.MatVec(cjf.loc_coord[0]);
  const Com::CVector3D& R0y = mrot0.MatVec(cjf.loc_coord[1]);
  const Com::CVector3D& R0z = mrot0.MatVec(cjf.loc_coord[2]);
  ////
  const Com::CVector3D& R1x = mrot1.MatVec(cjf.loc_coord[0]);
  const Com::CVector3D& R1y = mrot1.MatVec(cjf.loc_coord[1]);
  const Com::CVector3D& R1z = mrot1.MatVec(cjf.loc_coord[2]);    
  
  Com::CMatrix3 Rtwma[2];  
  { // rotational constraint
    const Com::CVector3D& wR0yR1z = Com::CMatrix3(R0y).MatVec(R1z);
    const Com::CVector3D& wR0zR1x = Com::CMatrix3(R0z).MatVec(R1x);
    const Com::CVector3D& wR0xR1y = Com::CMatrix3(R0x).MatVec(R1y);
    Com::CMatrix3 mtmp(wR0yR1z,wR0zR1x,wR0xR1y);
    Rtwma[0] = mrot0.MatMatTrans(mtmp);
    Rtwma[1] = mrot1.MatMatTrans(mtmp);    
  }
  
  ////////////
  Com::CMatrix3 RotwX0 = mrot0.MatMat( Com::CMatrix3(Xdistfix0) );
  Com::CMatrix3 wXwRtL0;
	{
    const Com::CVector3D& RtL = mrot0.MatVecTrans(vlambda);
    wXwRtL0 = Com::CMatrix3(Xdistfix0).MatMat( Com::CMatrix3(RtL) );
	}         
  Com::CMatrix3 RotwX1 = mrot1.MatMat( Com::CMatrix3(Xdistfix1) );
  Com::CMatrix3 wXwRtL1;
	{
    const Com::CVector3D& RtL = mrot1.MatVecTrans(vlambda);
    wXwRtL1 = Com::CMatrix3(Xdistfix1).MatMat( Com::CMatrix3(RtL) );
	}         
  
  //  const unsigned int nRB = aRB.size();
  const Com::CMatrix3 mInd(1.0);
	{    
    AddMatrix(mat,       irb0,     0,  icst+nRB, 0, -mInd,      1, true);
    
    AddMatrix(mat,       irb0,     3,  irb0,     3,  wXwRtL0,   1, true);
    AddMatrix(mat,       irb0,     3,  icst+nRB, 0, -RotwX0,    1, false);        
    AddMatrix(mat,       irb0,     3,  icst+nRB, 3, Rtwma[0],   1, true);
    
    AddMatrix(mat,       irb1,     0,  icst+nRB, 0,  mInd,      1, true);
    
    AddMatrix(mat,       irb1,     3,  irb1,     3,  -wXwRtL1,  1, true);    
    AddMatrix(mat,       irb1,     3,  icst+nRB, 0,  RotwX1,    1, false);    
    AddMatrix(mat,       irb1,     3,  icst+nRB, 3, -Rtwma[1],  1, true);
    
    AddMatrix(mat,       icst+nRB, 0,  irb0,     0, -mInd,      1, true);
    AddMatrix(mat,       icst+nRB, 0,  irb1,     0,  mInd,      1, true);        
    AddMatrix(mat,       icst+nRB, 0,  irb0,     3, -RotwX0,    1, true );
    AddMatrix(mat,       icst+nRB, 0,  irb1,     3,  RotwX1,    1, true );    
    
    AddMatrix(mat,       icst+nRB, 3,  irb0,     3, Rtwma[0],   1, false);    
    AddMatrix(mat,       icst+nRB, 3,  irb1,     3,-Rtwma[1],   1, false);        
    
    AddMatrix(mat,       icst+nRB, 0, icst+nRB,  0, Com::CMatrix3(-cjf.invstiff_pos), 1, true);      
    AddMatrix(mat,       icst+nRB, 3, icst+nRB,  3, Com::CMatrix3(-cjf.invstiff_rot), 1, true);            
    /*      
     ls.AddMatrix(       irb0,true, 0,  icst,false,0, -mInd,        dtmp1, true);
     
     ls.AddMatrix(       irb0,true, 3,  irb0,true, 3,  wXwRtL0,     dtmp1, true);
     ls.AddMatrix(       irb0,true, 3,  icst,false,0, -RotwX0,      dtmp1, false);        
     ls.AddMatrix(       irb0,true, 3,  icst,false,3, Rtwma[0],    dtmp1, true);
     
     ls.AddMatrix(       irb1,true, 0,  icst,false,0,  mInd,        dtmp1, true);
     
     ls.AddMatrix(       irb1,true, 3,  irb1,true,3,  -wXwRtL1,     dtmp1, true);    
     ls.AddMatrix(       irb1,true, 3,  icst,false,0,  RotwX1,      dtmp1, false);    
     ls.AddMatrix(       irb1,true, 3,  icst,false,3, -Rtwma[1],    dtmp1, true);
     
     ls.AddMatrix(       icst,false,0,  irb0,true, 0, -mInd,        dtmp1, true);
     ls.AddMatrix(       icst,false,0,  irb1,true, 0,  mInd,        dtmp1, true);        
     ls.AddMatrix(       icst,false,0,  irb0,true, 3, -RotwX0,      dtmp1, true );
     ls.AddMatrix(       icst,false,0,  irb1,true, 3,  RotwX1,      dtmp1, true );    
     
     ls.AddMatrix(       icst,false, 3,  irb0,true,3, Rtwma[0],    dtmp1, false);    
     ls.AddMatrix(       icst,false, 3,  irb1,true,3,-Rtwma[1],    dtmp1, false);        
     
     ls.AddMatrix(icst,false,0, icst,false,0, Com::CMatrix3(-invstiff_pos), dtmp1, true);      
     ls.AddMatrix(icst,false,3, icst,false,3, Com::CMatrix3(-invstiff_rot), dtmp1, true);                
     */
	}
  
  
  ////////////////////////////////////////////////
  const Com::CVector3D vdlam(jff.dlambda[0],jff.dlambda[1],jff.dlambda[2]);  // position constraint
  const Com::CVector3D rdlam(jff.dlambda[3],jff.dlambda[4],jff.dlambda[5]);  // rotation constraint  
  {
    Com::CVector3D rv0 = vlambda;
//    rv0 -= dt*(-mInd).MatVec(vdlam);
    AddVector(r, irb0,0,  rv0,+1);
  }
	{ // relative translation constraint
    Com::CVector3D rr0 = -Com::CMatrix3(Xdistfix0).MatVec( mrot0.MatVecTrans(vlambda) );
    rr0 -= Rtwma[0].MatVec(rlambda); 
//    rr0 -= dt*wXwRtL0.MatVec(rd0.GetOmega());
//    rr0 -= dt*(-RotwX0).MatVecTrans(vdlam);
//    rr0 -= dt*(Rtwma[0]).MatVec(rdlam);
    AddVector(r, irb0,3,  rr0,+1);    
	}
  {
    Com::CVector3D rv1 = -vlambda;    
//    rv1 -= dt*(mInd).MatVec(vdlam);
    AddVector(r, irb1,0,  rv1,+1);
  }
	{ // relative translation constraint
    Com::CVector3D rr1 = Com::CMatrix3(Xdistfix1).MatVec( mrot1.MatVecTrans(vlambda) );
    rr1 += Rtwma[1].MatVec(rlambda);
//    rr1 -= dt*(-wXwRtL1).MatVec(rd1.GetOmega());
//    rr1 -= dt*(RotwX1).MatVecTrans(vdlam);
//    rr1 -= dt*(-Rtwma[1]).MatVec(rdlam);
    AddVector(r, irb1,3, rr1,+1 );    
	}
  { // relative translation constraint
    Com::CVector3D rcv(0,0,0);
    rcv += ri0.GetIniPosCG()+rd0.GetDispCG()-mrot0.MatVec(Xdistfix0);
    rcv -= ri1.GetIniPosCG()+rd1.GetDispCG()-mrot1.MatVec(Xdistfix1);
    rcv += cjf.invstiff_pos*vlambda;
//    rcv -= dt*(-mInd).MatVec(rd0.GetVeloCG());
//    rcv -= dt*(-RotwX0).MatVec(rd0.GetOmega());    
//    rcv -= dt*( mInd).MatVec(rd1.GetVeloCG());
//    rcv -= dt*( RotwX1).MatVec(rd1.GetOmega());
//    rcv -= dt*(-cjf.invstiff_pos)*vdlam;
    AddVector(r, icst+nRB,0, rcv,+1 );
	}  
  {
    Com::CVector3D rcr(0,0,0);
    rcr.x = - Com::Dot(R0y,R1z) + cjf.invstiff_rot*jff.lambda[3];
    rcr.y = - Com::Dot(R0z,R1x) + cjf.invstiff_rot*jff.lambda[4];
    rcr.z = - Com::Dot(R0x,R1y) + cjf.invstiff_rot*jff.lambda[5];    
//    rcr -= dt*( Rtwma[0]).MatVecTrans(rd0.GetOmega());
//    rcr -= dt*(-Rtwma[1]).MatVecTrans(rd1.GetOmega());
//    rcr -= dt*(-cjf.invstiff_rot)*rdlam;
    AddVector(r,icst+nRB,3,rcr,1);    
  }          
}

void AddLinearSystem_JointFix_BackwardEular
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 ////
 unsigned int nRB,
 unsigned int icst, const CJoint_Fix& cjf, const CJoint_Fix_Force& jff, 
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1,
 const double dt)
{
  const unsigned int irb0 = cjf.irb0;
  assert( irb0 < mat.NBlkMatCol() );
//  const CRigidBody3D& rb0 = aRB[irb0];
  
  const unsigned int irb1 = cjf.irb1;
  assert( irb1 < mat.NBlkMatCol() );
//  const CRigidBody3D& rb1 = aRB[irb1];
  
  const Com::CMatrix3& mrot0 = rd0.GetRotMatrix_Inc();
  const Com::CMatrix3& mrot1 = rd1.GetRotMatrix_Inc();
  const Com::CVector3D vlambda(jff.lambda[0],jff.lambda[1],jff.lambda[2]);  // position constraint
  const Com::CVector3D rlambda(jff.lambda[3],jff.lambda[4],jff.lambda[5]);  // rotation constraint
  const Com::CVector3D Xdistfix0 = ri0.GetIniPosCG() - cjf.ini_pos_joint;
  const Com::CVector3D Xdistfix1 = ri1.GetIniPosCG() - cjf.ini_pos_joint;  
  
  const Com::CVector3D& R0x = mrot0.MatVec(cjf.loc_coord[0]);
  const Com::CVector3D& R0y = mrot0.MatVec(cjf.loc_coord[1]);
  const Com::CVector3D& R0z = mrot0.MatVec(cjf.loc_coord[2]);
  ////
  const Com::CVector3D& R1x = mrot1.MatVec(cjf.loc_coord[0]);
  const Com::CVector3D& R1y = mrot1.MatVec(cjf.loc_coord[1]);
  const Com::CVector3D& R1z = mrot1.MatVec(cjf.loc_coord[2]);    
  
  Com::CMatrix3 Rtwma[2];  
  { // rotational constraint
    const Com::CVector3D& wR0yR1z = Com::CMatrix3(R0y).MatVec(R1z);
    const Com::CVector3D& wR0zR1x = Com::CMatrix3(R0z).MatVec(R1x);
    const Com::CVector3D& wR0xR1y = Com::CMatrix3(R0x).MatVec(R1y);
    Com::CMatrix3 mtmp(wR0yR1z,wR0zR1x,wR0xR1y);
    Rtwma[0] = mrot0.MatMatTrans(mtmp);
    Rtwma[1] = mrot1.MatMatTrans(mtmp);    
  }
  
  ////////////
  Com::CMatrix3 RotwX0 = mrot0.MatMat( Com::CMatrix3(Xdistfix0) );
  Com::CMatrix3 wXwRtL0;
	{
    const Com::CVector3D& RtL = mrot0.MatVecTrans(vlambda);
    wXwRtL0 = Com::CMatrix3(Xdistfix0).MatMat( Com::CMatrix3(RtL) );
	}         
  Com::CMatrix3 RotwX1 = mrot1.MatMat( Com::CMatrix3(Xdistfix1) );
  Com::CMatrix3 wXwRtL1;
	{
    const Com::CVector3D& RtL = mrot1.MatVecTrans(vlambda);
    wXwRtL1 = Com::CMatrix3(Xdistfix1).MatMat( Com::CMatrix3(RtL) );
	}         
  
//  const unsigned int nRB = aRB.size();
  const Com::CMatrix3 mInd(1.0);
	{    
		const double dtmp1 = dt*dt;
    AddMatrix(mat,       irb0,     0,  icst+nRB, 0, -mInd,      dtmp1, true);
    
    AddMatrix(mat,       irb0,     3,  irb0,     3,  wXwRtL0,   dtmp1, true);
    AddMatrix(mat,       irb0,     3,  icst+nRB, 0, -RotwX0,    dtmp1, false);        
    AddMatrix(mat,       irb0,     3,  icst+nRB, 3, Rtwma[0],   dtmp1, true);
    
    AddMatrix(mat,       irb1,     0,  icst+nRB, 0,  mInd,      dtmp1, true);
    
    AddMatrix(mat,       irb1,     3,  irb1,     3,  -wXwRtL1,  dtmp1, true);    
    AddMatrix(mat,       irb1,     3,  icst+nRB, 0,  RotwX1,    dtmp1, false);    
    AddMatrix(mat,       irb1,     3,  icst+nRB, 3, -Rtwma[1],  dtmp1, true);
    
    AddMatrix(mat,       icst+nRB, 0,  irb0,     0, -mInd,      dtmp1, true);
    AddMatrix(mat,       icst+nRB, 0,  irb1,     0,  mInd,      dtmp1, true);        
    AddMatrix(mat,       icst+nRB, 0,  irb0,     3, -RotwX0,    dtmp1, true );
    AddMatrix(mat,       icst+nRB, 0,  irb1,     3,  RotwX1,    dtmp1, true );    
    
    AddMatrix(mat,       icst+nRB, 3,  irb0,     3, Rtwma[0],   dtmp1, false);    
    AddMatrix(mat,       icst+nRB, 3,  irb1,     3,-Rtwma[1],   dtmp1, false);        
    
    AddMatrix(mat,       icst+nRB, 0, icst+nRB,  0, Com::CMatrix3(-cjf.invstiff_pos), dtmp1, true);      
    AddMatrix(mat,       icst+nRB, 3, icst+nRB,  3, Com::CMatrix3(-cjf.invstiff_rot), dtmp1, true);            
/*      
    ls.AddMatrix(       irb0,true, 0,  icst,false,0, -mInd,        dtmp1, true);
    
    ls.AddMatrix(       irb0,true, 3,  irb0,true, 3,  wXwRtL0,     dtmp1, true);
    ls.AddMatrix(       irb0,true, 3,  icst,false,0, -RotwX0,      dtmp1, false);        
    ls.AddMatrix(       irb0,true, 3,  icst,false,3, Rtwma[0],    dtmp1, true);
    
    ls.AddMatrix(       irb1,true, 0,  icst,false,0,  mInd,        dtmp1, true);
    
    ls.AddMatrix(       irb1,true, 3,  irb1,true,3,  -wXwRtL1,     dtmp1, true);    
    ls.AddMatrix(       irb1,true, 3,  icst,false,0,  RotwX1,      dtmp1, false);    
    ls.AddMatrix(       irb1,true, 3,  icst,false,3, -Rtwma[1],    dtmp1, true);
    
    ls.AddMatrix(       icst,false,0,  irb0,true, 0, -mInd,        dtmp1, true);
    ls.AddMatrix(       icst,false,0,  irb1,true, 0,  mInd,        dtmp1, true);        
    ls.AddMatrix(       icst,false,0,  irb0,true, 3, -RotwX0,      dtmp1, true );
    ls.AddMatrix(       icst,false,0,  irb1,true, 3,  RotwX1,      dtmp1, true );    
    
    ls.AddMatrix(       icst,false, 3,  irb0,true,3, Rtwma[0],    dtmp1, false);    
    ls.AddMatrix(       icst,false, 3,  irb1,true,3,-Rtwma[1],    dtmp1, false);        
    
    ls.AddMatrix(icst,false,0, icst,false,0, Com::CMatrix3(-invstiff_pos), dtmp1, true);      
    ls.AddMatrix(icst,false,3, icst,false,3, Com::CMatrix3(-invstiff_rot), dtmp1, true);                
 */
	}
  
  
  ////////////////////////////////////////////////
  const Com::CVector3D vdlam(jff.dlambda[0],jff.dlambda[1],jff.dlambda[2]);  // position constraint
  const Com::CVector3D rdlam(jff.dlambda[3],jff.dlambda[4],jff.dlambda[5]);  // rotation constraint  
  {
    Com::CVector3D rv0 = vlambda;
    rv0 -= dt*(-mInd).MatVec(vdlam);
    AddVector(r, irb0,0,  rv0,+dt);
  }
	{ // relative translation constraint
    Com::CVector3D rr0 = -Com::CMatrix3(Xdistfix0).MatVec( mrot0.MatVecTrans(vlambda) );
    rr0 -= Rtwma[0].MatVec(rlambda); 
    rr0 -= dt*wXwRtL0.MatVec(rd0.GetOmega());
    rr0 -= dt*(-RotwX0).MatVecTrans(vdlam);
    rr0 -= dt*(Rtwma[0]).MatVec(rdlam);
    AddVector(r, irb0,3,  rr0,+dt);    
	}
  {
    Com::CVector3D rv1 = -vlambda;    
    rv1 -= dt*(mInd).MatVec(vdlam);
    AddVector(r, irb1,0,  rv1,+dt);
  }
	{ // relative translation constraint
    Com::CVector3D rr1 = Com::CMatrix3(Xdistfix1).MatVec( mrot1.MatVecTrans(vlambda) );
    rr1 += Rtwma[1].MatVec(rlambda);
    rr1 -= dt*(-wXwRtL1).MatVec(rd1.GetOmega());
    rr1 -= dt*(RotwX1).MatVecTrans(vdlam);
    rr1 -= dt*(-Rtwma[1]).MatVec(rdlam);
    AddVector(r, irb1,3, rr1,+dt );    
	}
  { // relative translation constraint
    Com::CVector3D rcv(0,0,0);
    rcv += ri0.GetIniPosCG()+rd0.GetDispCG()-mrot0.MatVec(Xdistfix0);
    rcv -= ri1.GetIniPosCG()+rd1.GetDispCG()-mrot1.MatVec(Xdistfix1);
    rcv += cjf.invstiff_pos*vlambda;
    rcv -= dt*(-mInd).MatVec(rd0.GetVeloCG());
    rcv -= dt*(-RotwX0).MatVec(rd0.GetOmega());    
    rcv -= dt*( mInd).MatVec(rd1.GetVeloCG());
    rcv -= dt*( RotwX1).MatVec(rd1.GetOmega());
    rcv -= dt*(-cjf.invstiff_pos)*vdlam;
    AddVector(r, icst+nRB,0, rcv,+dt );
	}  
  {
    Com::CVector3D rcr(0,0,0);
    rcr.x = - Com::Dot(R0y,R1z) + cjf.invstiff_rot*jff.lambda[3];
    rcr.y = - Com::Dot(R0z,R1x) + cjf.invstiff_rot*jff.lambda[4];
    rcr.z = - Com::Dot(R0x,R1y) + cjf.invstiff_rot*jff.lambda[5];    
    rcr -= dt*( Rtwma[0]).MatVecTrans(rd0.GetOmega());
    rcr -= dt*(-Rtwma[1]).MatVecTrans(rd1.GetOmega());
    rcr -= dt*(-cjf.invstiff_rot)*rdlam;
    AddVector(r,icst+nRB,3,rcr,dt);    
  }      
}




void AddMatrix_StaticJoint
(MatVec::CMatDia_BlkCrs& mat, 
 /////
 unsigned int nRB,
 unsigned int icst, const CJoint_Fix& jF, const CJoint_Fix_Force& jFF, 
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1)
{
  const unsigned int irb0 = jF.irb0;
  assert( irb0 < mat.NBlkMatCol() );
//  const CRigidBody3D& rb0 = aRB[irb0];
  
  const unsigned int irb1 = jF.irb1;
  assert( irb1 < mat.NBlkMatCol() );
//  const CRigidBody3D& rb1 = aRB[irb1];
  
  const Com::CMatrix3& P0 = rd0.GetRotMatrix_Inc();
  const Com::CMatrix3& P1 = rd1.GetRotMatrix_Inc();
  const Com::CVector3D vlambda(jFF.lambda[0],jFF.lambda[1],jFF.lambda[2]);
  const Com::CVector3D Xdistfix0 = ri0.GetIniPosCG() - jF.ini_pos_joint;
  const Com::CVector3D Xdistfix1 = ri1.GetIniPosCG() - jF.ini_pos_joint;  
  
  const Com::CVector3D& P0x = P0.MatVec(jF.loc_coord[0]);
  const Com::CVector3D& P0y = P0.MatVec(jF.loc_coord[1]);
  const Com::CVector3D& P0z = P0.MatVec(jF.loc_coord[2]);
  const Com::CMatrix3& P0wx = P0.MatMat( Com::CMatrix3(jF.loc_coord[0]) );
  const Com::CMatrix3& P0wy = P0.MatMat( Com::CMatrix3(jF.loc_coord[1]) );
  const Com::CMatrix3& P0wz = P0.MatMat( Com::CMatrix3(jF.loc_coord[2]) );
  ////
  const Com::CVector3D& P1x = P1.MatVec(jF.loc_coord[0]);
  const Com::CVector3D& P1y = P1.MatVec(jF.loc_coord[1]);
  const Com::CVector3D& P1z = P1.MatVec(jF.loc_coord[2]);    
  const Com::CMatrix3& P1wx = P1.MatMat( Com::CMatrix3(jF.loc_coord[0]) );
  const Com::CMatrix3& P1wy = P1.MatMat( Com::CMatrix3(jF.loc_coord[1]) );
  const Com::CMatrix3& P1wz = P1.MatMat( Com::CMatrix3(jF.loc_coord[2]) );  
  
  const Com::CMatrix3& P1tP0 = P1.MatMatTrans(P0);
  const Com::CMatrix3& P0tP1 = P0.MatMatTrans(P1);
  
  Com::CMatrix3 Rtwma[2] = {
    -Com::CMatrix3( P0wy.MatVecTrans(P1z), P0wz.MatVecTrans(P1x), P0wx.MatVecTrans(P1y) ),
    -Com::CMatrix3( P1wy.MatVecTrans(P0z), P1wz.MatVecTrans(P0x), P1wx.MatVecTrans(P0y) ) };
  
  const Com::CVector3D rlambda(jFF.lambda[3],jFF.lambda[4],jFF.lambda[5]);
  Com::CMatrix3 P0weP1we[3] = { P0wy.MatMatTrans(P1wz), P0wz.MatMatTrans(P1wx), P0wx.MatMatTrans(P1wy) };
  Com::CMatrix3 P01 = P0weP1we[0]*rlambda.x + P0weP1we[1]*rlambda.y + P0weP1we[2]*rlambda.z;
  Com::CMatrix3 P00 = P0weP1we[0].MatMat(P1tP0)*(-rlambda.x) 
                    + P0weP1we[1].MatMat(P1tP0)*(-rlambda.y) 
                    + P0weP1we[2].MatMat(P1tP0)*(-rlambda.z);  
  Com::CMatrix3 P11 = P0weP1we[0].MatMatTrans(P0tP1)*(-rlambda.x) 
                    + P0weP1we[1].MatMatTrans(P0tP1)*(-rlambda.y) 
                    + P0weP1we[2].MatMatTrans(P0tP1)*(-rlambda.z);  
  
  
  ////////////
  Com::CMatrix3 RotwX0 = P0.MatMat( Com::CMatrix3(Xdistfix0) );
  Com::CMatrix3 wXwRtL0;
	{
    const Com::CVector3D& RtL = P0.MatVecTrans(vlambda);
    wXwRtL0 = Com::CMatrix3(Xdistfix0).MatMat( Com::CMatrix3(RtL) );
	}         
  Com::CMatrix3 RotwX1 = P1.MatMat( Com::CMatrix3(Xdistfix1) );
  Com::CMatrix3 wXwRtL1;
	{
    const Com::CVector3D& RtL = P1.MatVecTrans(vlambda);
    wXwRtL1 = Com::CMatrix3(Xdistfix1).MatMat( Com::CMatrix3(RtL) );
	}         
  
  const Com::CMatrix3 mInd(1.0);
  const double dtmp1 = 1;
  AddMatrix(mat,       irb0,     0,  icst+nRB, 0, -mInd,      dtmp1, true);
  
  AddMatrix(mat,       irb0,     3,  irb0,     3,  wXwRtL0,   dtmp1, true);
//  AddMatrix(mat,       irb0,     3,  irb0,     3,  P00,       dtmp1, true);  
//  AddMatrix(mat,       irb0,     3,  irb1,     3,  P01,       dtmp1, true);
  AddMatrix(mat,       irb0,     3,  icst+nRB, 0, -RotwX0,    dtmp1, false);        
  AddMatrix(mat,       irb0,     3,  icst+nRB, 3, Rtwma[0],   dtmp1, true);
  
  AddMatrix(mat,       irb1,     0,  icst+nRB, 0,  mInd,      dtmp1, true);
  
  AddMatrix(mat,       irb1,     3,  irb1,     3,  -wXwRtL1,  dtmp1, true);    
//  AddMatrix(mat,       irb1,     3,  irb0,     3,  P01,       dtmp1, false);      
//  AddMatrix(mat,       irb1,     3,  irb1,     3,  P11 ,      dtmp1, true);    
  AddMatrix(mat,       irb1,     3,  icst+nRB, 0,  RotwX1,    dtmp1, false);    
  AddMatrix(mat,       irb1,     3,  icst+nRB, 3, -Rtwma[1],  dtmp1, true);
  
  AddMatrix(mat,       icst+nRB, 0,  irb0,     0, -mInd,      dtmp1, true);
  AddMatrix(mat,       icst+nRB, 0,  irb1,     0,  mInd,      dtmp1, true);        
  AddMatrix(mat,       icst+nRB, 0,  irb0,     3, -RotwX0,    dtmp1, true );
  AddMatrix(mat,       icst+nRB, 0,  irb1,     3,  RotwX1,    dtmp1, true );    
  
  AddMatrix(mat,       icst+nRB, 3,  irb0,     3, Rtwma[0],   dtmp1, false);    
  AddMatrix(mat,       icst+nRB, 3,  irb1,     3,-Rtwma[1],   dtmp1, false);        
  
  AddMatrix(mat,       icst+nRB, 0, icst+nRB,  0, Com::CMatrix3(-jF.invstiff_pos), 1, true);      
  AddMatrix(mat,       icst+nRB, 3, icst+nRB,  3, Com::CMatrix3(-jF.invstiff_rot), 1, true);            
}



