/*
 *  joint_fix.h
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/24/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if !defined(JOINT_FIX_H)
#define JOINT_FIX_H

#include "delfem/MatVec/matdia_blkcrs.h"

#include "rigidbody.h"

class CJoint_Fix
{
public:  
  CJoint_Fix(){
    this->irb0 = 0;
    this->irb1 = 0;
    invstiff_pos = 0;
    invstiff_rot = 0;    
  }
  CJoint_Fix(unsigned int irb0, unsigned int irb1){
    this->irb0 = irb0;
    this->irb1 = irb1;
    invstiff_pos = 0;
    invstiff_rot = 0;
  }
  
  ////////////////////////////
  // virtual functions
  virtual unsigned int GetDOF() const { return 6; }
  /*
  void UpdateLambda_NewmarkBetaAPrime(const double* upd, const double dt, const double newmark_gamma, const double newmark_beta);        
  void AddLinearSystem_NewmarkBetaAPrime(Ls::CLinearSystem_RigidBody& ls, unsigned int icst,
                                         const double dt, const double newmark_gamma, const double newmark_beta,
                                         const std::vector<CRigidBody3D>& aRB,
                                         bool is_initial = false ) const;
  
  virtual void UpdateLambda_BackwardEular(const double* upd, double dt);    
  virtual void AddLinearSystem_BackwardEular(Ls::CLinearSystem_RigidBody& ls, unsigned int icst,
                                             const double dt, 
                                             const std::vector<CRigidBody3D>& aRB, 
                                             bool is_initial ) const;        
   */
  //
  
  void SetInvStiffJoint(double invstiff_pos, double invstiff_rot){
    this->invstiff_pos = invstiff_pos;
    this->invstiff_rot = invstiff_rot;
  }
  void SetIniPosJoint(double px, double py, double pz){
    ini_pos_joint.SetVector(px, py, pz);
  }
  void SetIniPosJoint(const Com::CVector3D& pos_j){
    ini_pos_joint = pos_j;
  }    
  void SetIniJointDirection(double sx, double sy, double sz,
                            double tx, double ty, double tz,
                            double ux, double uy, double uz){
    loc_coord[0].SetVector(sx,sy,sz);
    loc_coord[1].SetVector(tx,ty,tz);
    loc_coord[2].SetVector(ux,uy,uz);
  }
  void SetIniJointDirection(const Com::CVector3D& s, const Com::CVector3D& t, const Com::CVector3D& u){
    loc_coord[0] = s;
    loc_coord[1] = t;
    loc_coord[2] = u;
  }      
public:
  double invstiff_pos, invstiff_rot;
  Com::CVector3D ini_pos_joint;
  Com::CVector3D loc_coord[3];
  unsigned int irb0;
  unsigned int irb1;
}; 

class CJoint_Fix_Force
{
public:
  CJoint_Fix_Force(){
    lambda[0] = 0;   lambda[1] = 0;  lambda[2] = 0;  lambda[3] = 0;  lambda[4] = 0;  lambda[5] = 0;
    dlambda[0] = 0; dlambda[1] = 0; dlambda[2] = 0; dlambda[3] = 0; dlambda[4] = 0; dlambda[5] = 0;      
  }
  virtual void Clear(){
    lambda[0] = 0;   lambda[1] = 0;  lambda[2] = 0;  lambda[3] = 0;  lambda[4] = 0;  lambda[5] = 0;
    dlambda[0] = 0; dlambda[1] = 0; dlambda[2] = 0; dlambda[3] = 0; dlambda[4] = 0; dlambda[5] = 0;            
  }    
public:
  double lambda[6];
  double dlambda[6];    
};

void UpdateLambda_BackwardEular
(CJoint_Fix_Force& jff,
 const double* upd, double dt);

void AddResidual_JointFix
(std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResPR,
 std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResLM,       
 ////
 unsigned int nRB,
 unsigned int icst, const CJoint_Fix& jF, const CJoint_Fix_Force& jFF,
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1);

void AddMatrix_StaticJoint
(MatVec::CMatDia_BlkCrs& m_mat, 
 ////
 unsigned int nRB,
 unsigned int icst, const CJoint_Fix& jF, const CJoint_Fix_Force& jFF,
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1);

void AddLinearSystem_JointFix_BackwardEular
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 ////
 unsigned int nRB, 
 unsigned int icst, const CJoint_Fix& jF, const CJoint_Fix_Force& jFF,
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1,
 const double dt);

void AddLinearSystem_JointFix_Static
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 ////
 unsigned int nRB, 
 unsigned int icst, const CJoint_Fix& jF, const CJoint_Fix_Force& jFF,
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1);

void AddSensitivityResidual_StaticJoint
(std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResPR,
 std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResLM,
 ////
 unsigned int nRB,
 unsigned int icst, const CJoint_Fix& jF, const CJoint_Fix_Force& jFF,
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1,  
 bool is_rb_pos, int irb_s, unsigned int idim_s, double alpha);  

void AddSensitivityResidual_StaticJoint
(std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResPR,
 std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResLM,
 ////
 unsigned int nRB,
 unsigned int icst, const CJoint_Fix& jF, const CJoint_Fix_Force& jFF,
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0,
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1,  
 bool is_rb_pos, int irb_s, const Com::CVector3D& e_s);  

#endif