/*
 *  jointed_rigidbody_contact.h
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/2/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if !defined(JOINTED_RIGID_BODY_CONTACT_H)
#define JOINTED_RIGID_BODY_CONTACT_H

#include "delfem/vector3d.h"
#include "delfem/indexed_array.h"
#include "delfem/matvec/matdia_blkcrs.h"
#include "delfem/matvec/vector_blk.h"
#include "delfem/matvec/matdiafrac_blkcrs.h"


#include "rigidbody.h"
#include "joint_fix.h"
#include "rigid_contact.h"
#include "rigid_contact_ls.h"

// put functinos including everything except for evaluation fuction

class CExternalForceOnRigidBody
{
public:
  CExternalForceOnRigidBody(unsigned int irb, const Com::CVector3D& pos, const Com::CVector3D& force){
    this->irb = irb;
    this->pos = pos;
    this->force = force;
  }
public:
  unsigned int irb;
  Com::CVector3D pos;  
  Com::CVector3D force;
};

class CDesignSensPlank{
public:
  Com::CVector3D p;
  Com::CVector3D r;
  Com::CVector3D l;
};

class CDesignSensitivity
{
public:
  void SetZero(){
    for(unsigned int irb=0;irb<aPlank.size();irb++){
      aPlank[irb].p.SetZero();
      aPlank[irb].r.SetZero();
      aPlank[irb].l.SetZero();      
    }
    for(unsigned int ijt=0;ijt<aJoint.size();ijt++){
      aJoint[ijt].SetZero();
    }
  }
  void AddScaled(const CDesignSensitivity& desSens, double alpha){
    for(unsigned int irb=0;irb<aPlank.size();irb++){
      aPlank[irb].p += alpha*desSens.aPlank[irb].p;
      aPlank[irb].r += alpha*desSens.aPlank[irb].r;
      aPlank[irb].l += alpha*desSens.aPlank[irb].l;      
    }
    for(unsigned int ifix=0;ifix<aJoint.size();ifix++){
      aJoint[ifix] += alpha*desSens.aJoint[ifix];
    }
  }
  void Initialize(unsigned int nRB, unsigned int nJoint){
    aPlank.resize(nRB);
    aJoint.resize(nJoint);
    SetZero();
  }  
public:
  std::vector< CDesignSensPlank > aPlank;
  std::vector< Com::CVector3D > aJoint;
};

class CSimulationSensitivity{
public:
  CSimulationSensitivity(){}
  void Initialize(unsigned int np, unsigned int nj){ 
    aPlankPR.clear();
    aJointLM.clear();
    for(unsigned int irb=0;irb<np;irb++){    
      aPlankPR.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) ); 
    }
    for(unsigned int ifix=0;ifix<nj;ifix++){ 
      aJointLM.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) ); 
    }    
  }
  void AddScaled(const CSimulationSensitivity& simSens, double alpha){  
    assert( aPlankPR.size() == simSens.aPlankPR.size() );
    assert( aJointLM.size() == simSens.aJointLM.size() );    
    for(unsigned int irb=0;irb<aPlankPR.size();irb++){    
      aPlankPR[irb].first  += alpha*simSens.aPlankPR[irb].first;
      aPlankPR[irb].second += alpha*simSens.aPlankPR[irb].second;      
    }
    for(unsigned int ifix=0;ifix<aJointLM.size();ifix++){ 
      aJointLM[ifix].first  += alpha*simSens.aJointLM[ifix].first;
      aJointLM[ifix].second += alpha*simSens.aJointLM[ifix].second;            
    }    
  }
public:
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aPlankPR;
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aJointLM;
};

class CContactProp
{
public:
  Com::CVector3D cont_base;
  Com::CVector3D cont_norm; 
  double stiff_n;
  double stiff_f;  
};


void GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
(CSimulationSensitivity& desSens,
 ////
 PARAM_TYPE itype_s, unsigned int irb_s, unsigned int idim_s,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 ////
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,    
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aFixForce,  
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP);

void GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
(CSimulationSensitivity& simSens,
 ////
 const CDesignSensitivity& desSens,
 ////
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,    
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 const std::vector<CContactPoints_RigidBody3D>& aFP);

void SetPatternDynamicSystem
(MatVec::CMatDia_BlkCrs& m_mat,     
 MatVec::CVector_Blk& m_residual,
 const std::vector<CRigidBody3D>& aRB,
 const std::vector<CJoint_Fix>& aFix );

void MoveInitialConfig_JointedRigidBodyContactSystem
(PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s, double eps,
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CJoint_Fix>& aFix );
/*
void MoveCurrentConfig_JointedRigidBodyContactSystem
(const CSimulationSensitivity& simSens,
 double eps,
 std::vector<CRigid3D_Def>& aRD,   
 std::vector<CJoint_Fix_Force>& aJFixForce );
*/
void RelocateAnchors_JointedRigidBodyContactSystem
(std::vector<CContactPoints_RigidBody3D>& aFP,
 PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc, 
 const std::pair< Com::CVector3D, Com::CVector3D>& sns_prlm, 
 double eps,
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,    
 const Com::CVector3D& cont_norm);

bool UpdateValueOfRigidSystem_BackwardEular_JointedRigidBodySystem
(std::vector<CRigidBody3D>& aRB, std::vector<CJoint_Fix>& aFix, 
 MatVec::CVector_Blk& upd,
 double dt, bool is_first);

void StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem
(double dt,
 const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,   
 const std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,  
 const std::vector<CContactPoints_RigidBody3D>& aFP,
 double& res_norm, double& upd_norm);


void StepTime_BackwardEular_JointedRigidBodyContactSystem
(double dt,
 const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,   
 const std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,  
 double myu_s, double myu_k,
 std::vector<CContactPoints_RigidBody3D>& aFP,
 double& res_norm, double& upd_norm); 

void AddSensitivityResidual_JointedRigidBodyContactSystem
(std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResPR,
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResLM,
 PARAM_TYPE itype_s, int iobj_s, unsigned int idim_s, double alpha,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 /////
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,    
 const std::vector<CJoint_Fix>& aFix, 
 const std::vector<CJoint_Fix_Force>& aJFixForce,   
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 /////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP);

void AddSensitivityResidual_JointedRigidBodyContactSystem
(std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResPR,
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResLM,
 const CDesignSensitivity& desSens,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 /////
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,   
 const std::vector<CJoint_Fix>& aFix, 
 const std::vector<CJoint_Fix_Force>& aJFixForce,  
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 /////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,  
 const std::vector<CContactPoints_RigidBody3D>& aFP);

void GetResidual_JointedRigidBodyContactSystem
(CSimulationSensitivity& resSens,
 ////
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,  
 const std::vector<CJoint_Fix>& aFix, 
 const std::vector<CJoint_Fix_Force>& aJFixForce,  
 const std::vector<CExternalForceOnRigidBody>& aExForce, 
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,  
 const std::vector<CContactPoints_RigidBody3D>& aFP);

void MakeLinearSystemFrac_JointedRigidBodyContactSystem
(MatVec::CMatDia_BlkCrs& m_mat,
 MatVec::CVector_Blk& m_r,
 MatVec::CMatDiaFrac_BlkCrs& m_frac,
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,   
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 const std::vector<CContactPoints_RigidBody3D>& aFP);

void MoveInitialDesignAnchor
(const CDesignSensitivity& desSens, double eps, 
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CJoint_Fix>& aFix,
 std::vector<CContactPoints_RigidBody3D>& aFP);

void MoveCurrentDeform
(const CSimulationSensitivity& simSens, double eps,
 std::vector<CRigid3D_Def>& aRD,
 std::vector<CJoint_Fix_Force>& aFixForce);

void SetJointSensitivity
(CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,                                                                       
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir);

void EnforceContact_ChangeBoxLen
(std::vector<CRigid3D_Ini>& aRI,  
 std::vector<CContactPoints_RigidBody3D>& aFP, 
 const Com::CVector3D& n, const Com::CVector3D& b);

void SolveStaticEquilibrium
(const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,   
 const std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 const std::vector<CContactPoints_RigidBody3D>& aFP);

#endif
