/*
 *  jointed_rigidbody_contact_test.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/24/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include "joint_connectivity.h"
#include "jointed_rigidbody_contact_test.h"

void TestSensitiveResidual_JointedRigidBodyContactSystem
(PARAM_TYPE itype_s, unsigned int irb_s, unsigned int idim_s, 
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 ////
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,   
 std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 std::vector<CContactPoints_RigidBody3D>& aCP)
{
  unsigned int nRB = aRI.size();
  assert( nRB == aRD.size() );
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensResPR, aSensResLM;
  for(unsigned int irb=0;irb<nRB;irb++){     aSensResPR.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) ); }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){ aSensResLM.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) ); }      
  AddSensitivityResidual_JointedRigidBodyContactSystem(aSensResPR,aSensResLM, 
                                                       itype_s,irb_s,idim_s, 1, itype_anchor_reloc,
                                                       aRI,aRD,aFix,aJFixForce,aExForce,
                                                       gravity,
                                                       contProp,
                                                       aCP);
  
  std::pair<Com::CVector3D, Com::CVector3D> ZeroPR( std::make_pair( Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) );
  
  const double eps = 1.0e-5;
  MoveInitialConfig_JointedRigidBodyContactSystem(itype_s, irb_s, idim_s, -eps, aRI, aFix );
  RelocateAnchors_JointedRigidBodyContactSystem(aCP,
                                                itype_s, irb_s, idim_s, 
                                                itype_anchor_reloc,
                                                ZeroPR, -eps, 
                                                aRI,aRD,contProp.cont_norm);  
  
  CSimulationSensitivity resSim0;
  GetResidual_JointedRigidBodyContactSystem(resSim0,
                                            aRI,aRD,aFix,aJFixForce,aExForce,
                                            gravity,
                                            contProp,
                                            aCP);
  
  MoveInitialConfig_JointedRigidBodyContactSystem(itype_s, irb_s, idim_s, 2*eps, aRI,aFix);
  RelocateAnchors_JointedRigidBodyContactSystem(aCP,
                                                itype_s, irb_s, idim_s, 
                                                itype_anchor_reloc,
                                                ZeroPR, 2*eps, 
                                                aRI,aRD,contProp.cont_norm);  
  
  CSimulationSensitivity resSim1;
  GetResidual_JointedRigidBodyContactSystem(resSim1,
                                            aRI,aRD,aFix,aJFixForce,aExForce,
                                            gravity,
                                            contProp,
                                            aCP);                                              
  
  std::cout << "##########################" << std::endl;  
  std::cout << "Test Sensitivity Residual" << std::endl;
//  std::cout << aSensResPR.size() << " " << aResPR1.size() << " " << aResPR0.size() << std::endl;
  std::cout << "moving " << itype_s << " " << irb_s << " " << idim_s << "  " << itype_anchor_reloc << std::endl;
  for(unsigned int irb=0;irb<nRB;irb++){
    std::cout << irb << std::endl;
    std::cout << " rv" << std::endl;
    std::cout << "  s " << aSensResPR[irb].first << std::endl;
    std::cout << "  n " << (resSim1.aPlankPR[irb].first - resSim0.aPlankPR[irb].first  )/(2*eps) << std::endl;
    std::cout << " rr" << std::endl;
    std::cout << "  s " << aSensResPR[irb].second << std::endl;
    std::cout << "  n " << (resSim1.aPlankPR[irb].second - resSim0.aPlankPR[irb].second)/(2*eps) << std::endl;
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    std::cout << ifix << std::endl;    
    std::cout << " lv" << std::endl;
    std::cout << "  s " << aSensResLM[ifix].first << std::endl;
    std::cout << "  n " << (resSim1.aJointLM[ifix].first - resSim0.aJointLM[ifix].first  )/(2*eps) << std::endl;
    std::cout << " lr" << std::endl;
    std::cout << "  s " << aSensResLM[ifix].second << std::endl;
    std::cout << "  n " << (resSim1.aJointLM[ifix].second - resSim0.aJointLM[ifix].second)/(2*eps) << std::endl;    
  } 
}

void TestSensitiveResidual_JointedRigidBodyContactSystem
(const CDesignSensitivity& desSens,
 ////
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,  
 std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 std::vector<CContactPoints_RigidBody3D>& aCP)
{
  unsigned int nRB = aRI.size();
  assert( aRD.size() == nRB );
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensResPR, aSensResLM;
  for(unsigned int irb=0;irb<nRB;irb++){     aSensResPR.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) ); }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){ aSensResLM.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) ); }      
  AddSensitivityResidual_JointedRigidBodyContactSystem(aSensResPR,aSensResLM, 
                                                       desSens, MOVE_INITIAL,
                                                       aRI,aRD,aFix,aJFixForce,aExForce,
                                                       gravity,
                                                       contProp,
                                                       aCP);
  
  std::pair<Com::CVector3D, Com::CVector3D> ZeroPR( std::make_pair( Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) );
  
  const double eps = 1.0e-5;
  MoveInitialDesignAnchor(desSens, -eps, aRI, aFix, aCP );
  
  CSimulationSensitivity resSim0;  
  GetResidual_JointedRigidBodyContactSystem(resSim0,
                                            aRI,aRD,aFix,aJFixForce,aExForce,
                                            gravity,
                                            contProp,
                                            aCP);
  
  MoveInitialDesignAnchor(desSens, 2*eps, aRI,aFix,aCP);
  
  CSimulationSensitivity resSim1;  
  GetResidual_JointedRigidBodyContactSystem(resSim1,
                                            aRI,aRD,aFix,aJFixForce,aExForce,
                                            gravity,
                                            contProp,
                                            aCP);                                              
  
  std::cout << "##########################" << std::endl;  
  std::cout << "Test Sensitivity Residual" << std::endl;
//  std::cout << aSensResPR.size() << " " << aResPR1.size() << " " << aResPR0.size() << std::endl;
  for(unsigned int irb=0;irb<nRB;irb++){
    std::cout << irb << std::endl;
    std::cout << " rv" << std::endl;
    std::cout << "  s " << aSensResPR[irb].first << std::endl;
    std::cout << "  n " << (resSim1.aPlankPR[irb].first - resSim0.aPlankPR[irb].first  )/(2*eps) << std::endl;
    std::cout << " rr" << std::endl;
    std::cout << "  s " << aSensResPR[irb].second << std::endl;
    std::cout << "  n " << (resSim1.aPlankPR[irb].second - resSim0.aPlankPR[irb].second)/(2*eps) << std::endl;
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    std::cout << ifix << std::endl;    
    std::cout << " lv" << std::endl;
    std::cout << "  s " << aSensResLM[ifix].first << std::endl;
    std::cout << "  n " << (resSim1.aJointLM[ifix].first - resSim0.aJointLM[ifix].first  )/(2*eps) << std::endl;
    std::cout << " lr" << std::endl;
    std::cout << "  s " << aSensResLM[ifix].second << std::endl;
    std::cout << "  n " << (resSim1.aJointLM[ifix].second - resSim0.aJointLM[ifix].second)/(2*eps) << std::endl;    
  } 
}



// position rotation box-length joit-position
void TestSensitivePRLM_InitialConfigBox_JointedRigidBodyContactSystem
(const CDesignSensitivity& desSens,
 ////
 std::vector<CRigid3D_Ini>& aRI,
 std::vector<CRigid3D_Def>& aRD,  
 std::vector<CJoint_Fix>& aFix,
 std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
// const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n, double stiff_f,
 const CContactProp& contProp, 
 std::vector<CContactPoints_RigidBody3D>& aFP)
{    
  std::cout << "##################" << std::endl;
  std::cout << "Test Sensitive PRLJ" << std::endl;
  
//  MoveInitialDesignAnchor(desSens,eps, aRI,aFix,aFP);    
  EnforceJointConnectivity_ChangeBoxLen(aRI,aFix,aFixDir,false);          
  EnforceContact_ChangeBoxLen(aRI,aFP,contProp.cont_norm,contProp.cont_base);
  
  //  MoveCurrentDeform(simSens,eps, jrb.aRD,jrb.aJFixForce);    
  //  jrb.PlaceNails();  
  
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);
/*  
  for(unsigned int i=0;i<1000;i++){
    double resn, updn;    
    StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                      aRI,aRD,aFix,aJFixForce,aExForce,
                                                                      gravity,
//                                                                      cont_base,cont_norm,stiff_n,stiff_f,
                                                                      contProp,
                                                                      aFP,
                                                                      resn,updn);	// solve rigid motion    
    if( updn < 1.0e-10 ) break;
  }  
*/  
  CSimulationSensitivity simSens;
  GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem(simSens,
                                                                    desSens,
                                                                    aRI,aRD,aFix,aJFixForce,aExForce,
                                                                    gravity,
//                                                                    cont_base,cont_norm,stiff_n,stiff_f,
                                                                    contProp,
                                                                    aFP);
  std::vector< Com::CVector3D > aDispCG0;
  for(unsigned int irb=0;irb<aRD.size();irb++){
    aDispCG0.push_back( aRD[irb].GetDispCG() );
  }  
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aLM0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){    
    const CJoint_Fix& cjs = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];    
    const double* lm = jff.lambda;
    aLM0.push_back( std::make_pair( Com::CVector3D(lm[0],lm[1],lm[2]),
                                   Com::CVector3D(lm[3],lm[4],lm[5]) ) );
  }
  
  double eps = 1.0e-5;
  /*
  assert( desSens.aPlank.size() == aRI.size() );
  for(unsigned int irb=0;irb<aRI.size();irb++){
    Com::CVector3D q = desSens.aPlank[irb].p*eps;
    Com::CVector3D s = desSens.aPlank[irb].r*eps;
    Com::CVector3D l = desSens.aPlank[irb].l*eps;
//    CRigidBody3D& rb = aRB[irb];
    CRigid3D_Ini& ri = aRI[irb];
    ri.SetIniPosCG( ri.GetIniPosCG() + q );
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(s);
    ri.AddRotation_IniLeft( rot );
    const double rho = ri.GetMass() / (ri.xlen*ri.ylen*ri.zlen);
    const double xlen = ri.xlen + l.x;
    const double ylen = ri.ylen + l.y;
    const double zlen = ri.zlen + l.z;
    ri.SetModeBox(xlen,ylen,zlen,rho);
    CContactPoints_RigidBody3D& aCP = aFP[irb]; 
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;    
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      Com::CVector3D v1 = ri.GetInitialPosition(v0);
      aCP.aFP[i].aloc = v1;
    }
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    Com::CVector3D p = desSens.aJoint[ifix]*eps;
    aFix[ifix].ini_pos_joint += p;
  }
   */
  MoveInitialDesignAnchor(desSens,eps, aRI,aFix,aFP);    
//  EnforceJointConnectivity_ChangeBoxLen(aRI,aFix,aFixDir,false);          
  EnforceContact_ChangeBoxLen(aRI,aFP,contProp.cont_norm,contProp.cont_base);

  //  MoveCurrentDeform(simSens,eps, jrb.aRD,jrb.aJFixForce);    
//  jrb.PlaceNails();
  
  /*
   MoveInitialConfig_JointedRigidBodyContactSystem(itype_s,irb_s,idim_s, eps, aRB,aFix);
   MoveCurrentConfig_JointedRigidBodyContactSystem(aSensPR,aSensLM,eps, aRB,aFix);
   RelocateAnchors_JointedRigidBodyContactSystem(aFP,
   itype_s,irb_s,idim_s, itype_anchor_reloc,
   aSensPR[irb_s], eps, 
   aRB,cont_norm);
   */
  
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);
  /*
  for(unsigned int i=0;i<1000;i++){
    double resn, updn;
    StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                      aRI,aRD,aFix,aJFixForce,aExForce,
                                                                      gravity,
//                                                                      cont_base,cont_norm,stiff_n,stiff_f,
                                                                      contProp,
                                                                      aFP,
                                                                      resn,updn);	// solve rigid motion    
    if( updn < 1.0e-10 ) break;
  }
   */
  
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aLM1;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){    
    const CJoint_Fix& cjs = aFix[ifix];    
    const CJoint_Fix_Force& jff = aJFixForce[ifix];        
    const double* lm = jff.lambda;
    aLM1.push_back( std::make_pair( Com::CVector3D(lm[0],lm[1],lm[2]),
                                   Com::CVector3D(lm[3],lm[4],lm[5]) ) );
  }  
  
  std::vector< Com::CVector3D > aDispCG1;
  for(unsigned int irb=0;irb<aRD.size();irb++){
    aDispCG1.push_back( aRD[irb].GetDispCG() );
  }  
  
  std::cout << "#################" << std::endl;
  std::cout << "Test Sensitive PRLJ" << std::endl;
  //  std::cout << "moving " << itype_s << " " << irb_s << " " << idim_s << std::endl;  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    std::cout << "rb " << irb << std::endl;
    Com::CVector3D dld = (aDispCG1[irb]  - aDispCG0[irb] )*(1.0/eps);
    Com::CVector3D drd = simSens.aPlankPR[irb].first;    
    std::cout << "cmp disp cg" << std::endl;
    std::cout << "n  " << dld.x << " " << dld.y << " " << dld.z << std::endl;
    std::cout << "s  " << drd.x << " " << drd.y << " " << drd.z << std::endl;        
  }
  ////////
  for(unsigned int iaFix=0;iaFix<aFix.size();iaFix++){
    std::cout << "fix " << iaFix << std::endl;
    Com::CVector3D dlv = (aLM1[iaFix].first  - aLM0[iaFix].first )*(1.0/eps);
    Com::CVector3D dlt = (aLM1[iaFix].second - aLM0[iaFix].second)*(1.0/eps);    
    Com::CVector3D drv = simSens.aJointLM[iaFix].first;
    Com::CVector3D drt = simSens.aJointLM[iaFix].second;
    std::cout << "cmp joint push/pull force change" << std::endl;
    std::cout << "n  " << dlv.x << " " << dlv.y << " " << dlv.z << std::endl;
    std::cout << "s  " << drv.x << " " << drv.y << " " << drv.z << std::endl;    
    ////
    std::cout << "cmp joint bend force change" << std::endl;
    std::cout << "n  " << dlt.x << " " << dlt.y << " " << dlt.z << std::endl;
    std::cout << "s  " << drt.x << " " << drt.y << " " << drt.z << std::endl;        
  }
}


// position rotation box-length joit-position
void TestSensitivePRLM_InitialConfigBox_JointedRigidBodyContactSystem
(PARAM_TYPE itype_s, unsigned int irb_s, unsigned int idim_s,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 ////
// std::vector<CRigidBody3D>& aRB, 
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,    
 std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
// const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n, double stiff_f,
 const CContactProp& contProp,
 std::vector<CContactPoints_RigidBody3D>& aFP)
{    
  unsigned int nRB = aRI.size();
  assert( aRD.size() == nRB );
  std::cout << "##################" << std::endl;
  std::cout << "Test Sensitive PRLJ" << std::endl;
  
  /*
  for(unsigned int i=0;i<1000;i++){
    double resn, updn;    
    StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                      aRI,aRD,aFix,aJFixForce,aExForce,
                                                                      gravity,
//                                                                      cont_base,cont_norm,stiff_n,stiff_f,
                                                                      contProp,
                                                                      aFP,
                                                                      updn,resn);	// solve rigid motion
    if( updn < 1.0e-10 ) break;
  } 
   */
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);  
  
  CSimulationSensitivity simSens;
  GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem(simSens,
                                                                    itype_s,irb_s,idim_s, itype_anchor_reloc,
                                                                    aRI,aRD,aFix,aJFixForce,aExForce,
                                                                    gravity,
//                                                                    cont_base,cont_norm,stiff_n,stiff_f,
                                                                    contProp,
                                                                    aFP);
  std::vector< Com::CVector3D > aDispCG0;
  for(unsigned int irb=0;irb<nRB;irb++){
    aDispCG0.push_back( aRD[irb].GetDispCG() );
  }  
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aLM0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){    
    const CJoint_Fix& cjs = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];    
    const double* lm = jff.lambda;
    aLM0.push_back( std::make_pair( Com::CVector3D(lm[0],lm[1],lm[2]),Com::CVector3D(lm[3],lm[4],lm[5]) ) );
  }
  
  double eps = 1.0e-5;
  MoveInitialConfig_JointedRigidBodyContactSystem(itype_s,irb_s,idim_s, eps, aRI,aFix);
  MoveCurrentDeform(simSens,eps, aRD,aJFixForce);
  RelocateAnchors_JointedRigidBodyContactSystem(aFP,
                                                itype_s,irb_s,idim_s, itype_anchor_reloc,
                                                simSens.aPlankPR[irb_s], eps, 
                                                aRI,aRD,contProp.cont_norm);
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);
  /*
  for(unsigned int i=0;i<1000;i++){
    double resn, updn;    
    StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                      aRI,aRD,aFix,aJFixForce,aExForce,
                                                                      gravity,
                                                                      contProp,
                                                                      aFP,
                                                                      resn,updn);	// solve rigid motion    
    if( updn < 1.0e-10 ) break;
  }
   */
  
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aLM1;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){    
    const CJoint_Fix& cjs = aFix[ifix];    
    const CJoint_Fix_Force& jff = aJFixForce[ifix];        
    const double* lm = jff.lambda;
    aLM1.push_back( std::make_pair( Com::CVector3D(lm[0],lm[1],lm[2]),Com::CVector3D(lm[3],lm[4],lm[5]) ) );
  }  
  
  std::vector< Com::CVector3D > aDispCG1;
  for(unsigned int irb=0;irb<nRB;irb++){
    aDispCG1.push_back( aRD[irb].GetDispCG() );
  }  
  
  std::cout << "#################" << std::endl;
  std::cout << "Test Sensitive PRLJ" << std::endl;
  std::cout << "moving " << itype_s << " " << irb_s << " " << idim_s << std::endl;  
  for(unsigned int irb=0;irb<nRB;irb++){
    std::cout << "rb " << irb << std::endl;
    Com::CVector3D dld = (aDispCG1[irb]  - aDispCG0[irb] )*(1.0/eps);
    Com::CVector3D drd = simSens.aPlankPR[irb].first;    
    std::cout << "cmp disp cg" << std::endl;
    std::cout << "n  " << dld.x << " " << dld.y << " " << dld.z << std::endl;
    std::cout << "s  " << drd.x << " " << drd.y << " " << drd.z << std::endl;        
  }
  ////////
  for(unsigned int iaFix=0;iaFix<aFix.size();iaFix++){
    std::cout << "fix " << iaFix << std::endl;
    Com::CVector3D dlv = (aLM1[iaFix].first  - aLM0[iaFix].first )*(1.0/eps);
    Com::CVector3D dlt = (aLM1[iaFix].second - aLM0[iaFix].second)*(1.0/eps);    
    Com::CVector3D drv = simSens.aJointLM[iaFix].first;
    Com::CVector3D drt = simSens.aJointLM[iaFix].second;
    std::cout << "cmp joint push/pull force change" << std::endl;
    std::cout << "n  " << dlv.x << " " << dlv.y << " " << dlv.z << std::endl;
    std::cout << "s  " << drv.x << " " << drv.y << " " << drv.z << std::endl;    
    ////
    std::cout << "cmp joint bend force change" << std::endl;
    std::cout << "n  " << dlt.x << " " << dlt.y << " " << dlt.z << std::endl;
    std::cout << "s  " << drt.x << " " << drt.y << " " << drt.z << std::endl;        
  }
}




