/*
 *  nail_furniture.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/31/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include "nail_bound_furniture.h"
#include "joint_connectivity.h"
#include "rigidbody.h"
#include "jointed_rigidbody_contact.h"

void PlaceNailsToJoint
(std::vector<CNailJoint>& aNailJoint,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir) 
{
  //  std::cout << "Enter DrawNails" << std::endl;
  
  const Com::CVector3D aFaceNormal[6][3] = {
    { Com::CVector3D(+1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0,+1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-0,-1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0, 0,+1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) },
    { Com::CVector3D(-0, 0,-1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) } };      
    
  assert( aNailJoint.size() == aFix.size() );
  assert( aNailJoint.size() == aFixDir.size() );
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){  
    const unsigned int iri0 = aFix[ifix].irb0;
    const unsigned int iri1 = aFix[ifix].irb1;    
    const CRigid3D_Ini& ri0 = aRI[iri0];
    const CRigid3D_Ini& ri1 = aRI[iri1];    
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;    
    assert( idir0==4 || idir0==5 );                
    Com::CVector3D a,b;
    bool res = GetIntersectionOfTwoPlanks(a,b,
                                          ri0,ri1,
                                          idir0, idir1);
    CNailJoint& nj = aNailJoint[ifix];        
    const double intervnail = nj.intervnail;
    nj.aPosHead.clear();
    if( res ){
      const double contlen = Distance(a,b);
      const unsigned int nnail = ( contlen-intervnail*2>0 ) ? (contlen-intervnail*2)/intervnail+1  : 1;
      double intervnail1 = (contlen-intervnail)/nnail;            
//      std::cout << "Nail Interval : " << intervnail << " " << intervnail1 << " " << nnail << " " << contlen << std::endl;
      Com::CVector3D nh = b-a; nh.SetNormalizedVector();
      Com::CVector3D ri1z = ri1.GetRotMatrix_Ini().MatVec(aFaceNormal[idir1][0] )*ri0.zlen;
      for(unsigned int inail=0;inail<nnail;inail++){            
        const Com::CVector3D& hp1 = a + nh*(intervnail1*inail+intervnail1);
        const Com::CVector3D& hp2 = hp1 + ri1z;
        nj.aPosHead.push_back(hp2);
      }      
    }
  }
}

void CheckNailJointTolerance
(double& smin, double& smax,
 const Com::CVector3D& sns_s, const Com::CVector3D& sns_b,
 const CJoint_Fix& cst, 
 const CJoint_Fix_Force& cff,    
 unsigned int idir0, unsigned int idir1,
 const CRigid3D_Ini& ri0, 
 const CRigid3D_Ini& ri1,
 const CNailJoint& nj)  // nail is go through rb1 to rb0
{  
  assert( idir0!=4 && idir0!=5 );
  assert( idir1==4 || idir1==5 );  
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };
  
  const Com::CVector3D vlambda(cff.lambda[0],cff.lambda[1],cff.lambda[2]); 
  const Com::CVector3D rlambda(cff.lambda[3],cff.lambda[4],cff.lambda[5]);      
  const unsigned int nnail = nj.aPosHead.size();
  const Com::CMatrix3& Q0 = ri0.GetRotMatrix_Ini();  
  const Com::CVector3D& Qn0 = Q0.MatVec( aFaceNormal[idir0] );
  const Com::CVector3D& Qh0 = Q0.MatVec( Cross( Com::CVector3D(0,0,1), aFaceNormal[idir0] ) );
  const Com::CVector3D& tang = (Com::CMatrix3(1)-Com::CMatrix3(Qn0,Qn0)).MatVec(vlambda);
  std::cout.setf(std::ios::fixed);
  //  sher = (tang.Length()/nnail)/(shear_max);
  
  //  std::cout << "shear " <<  sher*100 << "%     " <<  tang.Length()/nnail << " " << 19.0*10 << std::endl;
  double torq = Dot(Qh0,rlambda);
  double bend_pull = fabs(torq)/(ri0.zlen*0.5) - Dot(Qn0,vlambda);
  double pull_max = (nj.lennail-ri1.zlen)*nj.pull_max_perlen;
  //  double sns_bend_pull = Dot(Qh0,sns_b)/(ri0.zlen*0.5) - Dot(Qn0,sns_s);
  double k = (ri0.zlen*0.5);
  double a = pull_max*nnail;
  double c = Dot(Qn0,sns_s);
  double d = Dot(Qn0,vlambda);  
  double b = Dot(Qh0,sns_b);
  double r = Dot(Qh0,rlambda);
  const double s0 = +(r-(a+d)*k)/(k*c-b);  
  const double s1 = -(r+(a+d)*k)/(k*c+b);  
  if( s0 > s1 ){ smin = s1; smax = s0; }
  else {         smin = s0; smax = s1; }
  std::cout << fabs(r)/k-d << " " << pull_max*nnail << std::endl;
}

/*
void GetTorelanceSensitivity
(double& smin, double& smax,
 const CDesignSensitivity& desSens,
 const CSimulationSensitivity& simSens,
 const std::vector<CRigidBody3D>& aRB,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm,
 const std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector<CNailJoint>& aNailJoint)
{
  smin = -1.5;
  smax = +1.5;
  assert( aRB.size() == aFP.size() );        
  for(unsigned int irb=0;irb<aRB.size();irb++){        
    const CRigidBody3D& rb = aRB[irb];
    const CRigid3D_Ini& ri = rb.ri;        
    const CRigid3D_Def& rd = rb.rd;        
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);                      
      if( aCP.aFP[i].itype_contact != 1 ) continue;
      Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);        
      Com::CVector3D v2 = aCP.aFP[i].aloc;
      Com::CVector3D dir_f = v2-v1;
      dir_f -= cont_norm*Com::Dot(cont_norm,dir_f);
      const double depth = Com::Dot(cont_norm,cont_base-v1);
      Com::CMatrix3  m = aCP.GetSensMat_ContPointParam(i);
      Com::CVector3D sns = GetSensitivityCurrentPointOnRigidBody(desSens.aPlank[irb].p, 
                                                                 desSens.aPlank[irb].r, 
                                                                 desSens.aPlank[irb].l,
                                                                 simSens.aPlankPR[irb].first,
                                                                 simSens.aPlankPR[irb].second,
                                                                 rb.ri,
                                                                 rb.rd,
                                                                 v0,m);
      double sns_depth = -Com::Dot(cont_norm,sns);
      if( fabs(sns_depth) < 1.0e-5 ) continue;
      double t = -depth / sns_depth;
      if( t > 0 ){
        if( t < smax ){ smax = t; }
      }
      else{
        if( t > smin ){ smin = t; }        
      }
      //      std::cout << "depth " << depth << " " << t << std::endl;
    }      
  }
  {
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      const CJoint_Fix& cst = aFix[ifix];
      const CJoint_Fix_Force& cff = aJFixForce[ifix];
      unsigned int idir0 = aFixDir[ifix].first;
      unsigned int idir1 = aFixDir[ifix].second;    
      unsigned int irb0 = cst.irb0;
      unsigned int irb1 = cst.irb1;
      double tmin, tmax;
      if(      (idir0==4||idir0==5) && (idir1!=4&&idir1!=5) ){
        CheckNailJointTolerance(tmin,tmax,
                                -simSens.aJointLM[ifix].first,                       
                                -simSens.aJointLM[ifix].second,
                                cst, cff,
                                idir1,idir0,
                                aRB[irb1],aRB[irb0],
                                aNailJoint[ifix]);
      }
      else if( (idir0!=4&&idir0!=5) && (idir1==4||idir1==5) ){
        CheckNailJointTolerance(tmin,tmax,
                                simSens.aJointLM[ifix].first,                       
                                simSens.aJointLM[ifix].second,                      
                                cst, cff,
                                idir0,idir1,
                                aRB[irb0],aRB[irb1],
                                aNailJoint[ifix]);    
      }                
      if( smin < tmin ){ smin = tmin; }
      if( smax > tmax ){ smax = tmax; }
      std::cout << smin << " " << smax << std::endl;
      
    }
  }      
}
*/


void CheckCapacity
(double& sher_ratio, double& bend_ratio,
 const CJoint_Fix& cst,
 const Com::CVector3D& pos_force,
 const Com::CVector3D& bnd_force,
 unsigned int idir0, unsigned int idir1,
 const CRigid3D_Ini& ri0, const CRigid3D_Def& rd0, 
 const CRigid3D_Ini& ri1, const CRigid3D_Def& rd1, 
 const CNailJoint& nj)  // nail is go through rb1 to rb0
{  
  assert( idir0!=4 && idir0!=5 );
  assert( idir1==4 || idir1==5 );  
  
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };
  
//  const double shear_max = 19.0*10.0; // per newton
//  const double pull_max_perlen = 3500;
  const unsigned int nnail = nj.aPosHead.size();
  const double lennail = nj.lennail;
  const Com::CMatrix3& Q0 = ri0.GetRotMatrix_Ini();
  const Com::CVector3D& Qn0 = Q0.MatVec( aFaceNormal[idir0] );
  const Com::CVector3D& Qh0 = Q0.MatVec( Cross( Com::CVector3D(0,0,1), aFaceNormal[idir0] ) );
  const Com::CVector3D& tang = (Com::CMatrix3(1)-Com::CMatrix3(Qn0,Qn0)).MatVec(pos_force);
  std::cout.setf(std::ios::fixed);
  sher_ratio = (tang.Length()/nnail)/(nj.shear_max);
  ////
  double torq = fabs(Dot(Qh0,bnd_force));
  double bend_pull = torq/(ri0.zlen*0.5) + Dot(Qn0,pos_force);
//  std::cout << torq/(ri0.zlen*0.5) << " " << Dot(Qn0,pos_force) << std::endl;
  double pull_max = (lennail-ri1.zlen)*nj.pull_max_perlen;
  bend_ratio = (bend_pull/nnail)/(pull_max);  
  //  std::cout << "shear " <<  sher*100 << "%     " <<  tang.Length()/nnail << " " << 19.0*10 << std::endl;  
  //  std::cout << "pull  " << bend*100 << "%      " << bend_pull/nnail << "  " << pull_max << " " << std::endl;
  //  std::cout << Dot(Qn0,vlambda) << std::endl;  
}

void DisplayJointCapacity
(unsigned int ifix, double& shear, double& pull,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CRigid3D_Def>& aRD,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                          
 const std::vector<CNailJoint>& aNailJoint)
{
  const CJoint_Fix& cst = aFix[ifix];
  const CJoint_Fix_Force& jff = aJFixForce[ifix];
  unsigned int idir0 = aFixDir[ifix].first;
  unsigned int idir1 = aFixDir[ifix].second;   
  Com::CVector3D pos_force(jff.lambda[0],jff.lambda[1],jff.lambda[2]); 
  Com::CVector3D bnd_force(jff.lambda[3],jff.lambda[4],jff.lambda[5]); 
  assert( idir0 == 4 || idir0 == 5 );
  if(idir1!=4&&idir1!=5 ){
    CheckCapacity(shear, pull,
                  cst,-pos_force,-bnd_force,
                  idir1,idir0,
                  aRI[cst.irb1],aRD[cst.irb1],
                  aRI[cst.irb0],aRD[cst.irb0],
                  aNailJoint[ifix]);
  }
  else{
    assert(0);
  }          
}


void FindTopplingRangeSensitivity
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                       
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 /////
 CSimulationSensitivity& simSens,
 std::vector<CRigid3D_Def>& aRD,
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector<CExternalForceOnRigidBody>& aExForce)
{
  tmax = +10;
  tmin = -10;
  // conv
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);
  /*
  for(unsigned int i=0;i<1000;i++){
    double resn, updn;    
    StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                      aRI,aRD,aFix,aJFixForce,aExForce,
                                                                      gravity,
                                                                      contProp,
                                                                      aFP,
                                                                      resn,updn);	// solve rigid motion    
    if( updn < 1.0e-10 ) break;
  } 
   */
  // sens
  GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem(simSens,
                                                                    desSens,
                                                                    aRI,aRD,aFix,aJFixForce,aExForce,
                                                                    gravity,contProp,aFP);
  // check
  const Com::CVector3D& cont_norm = contProp.cont_norm;
  const Com::CVector3D& cont_base = contProp.cont_base;
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];        
    const CRigid3D_Def& rd = aRD[irb];        
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;      
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);  
      Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);
      const double depth = Com::Dot(cont_norm,cont_base-v1);
      std::cout << depth << "  " << v1 << "  " << v0 << "  " << rd.GetDispCG() << "  " << std::endl;
      Com::CMatrix3  B0 = aCP.GetSensMat_ContPointParam(i);      
      Com::CVector3D sns = GetSensitivityCurrentPointOnRigidBody(desSens.aPlank[irb].p,desSens.aPlank[irb].r,desSens.aPlank[irb].l,
                                                                 simSens.aPlankPR[irb].first, simSens.aPlankPR[irb].second,
                                                                 ri,rd,
                                                                 v0,
                                                                 B0);
      double sns_depth = -Com::Dot(cont_norm,sns);      
      if( fabs(sns_depth) < 1.0e-2*fabs(depth) ) continue;
      double t = -depth / sns_depth;
//      std::cout << "depth t " << irb << " " << i << "    " << depth << " " << t << std::endl;      
      if( depth > 0 ){
        if( t > 0 && t < tmax ){ tmax = t; }
        if( t < 0 && t > tmin ){ tmin = t; }        
      }
      if( depth < 0 ){
        if( t > 0 && t > tmin ){ tmin = t; }
        if( t < 0 && t < tmax ){ tmax = t; }        
      }
      //      std::cout << "depth " << depth << " " << t << std::endl;
    }          
  }
//  std::cout << " guess : " << tmin << " " << tmax << std::endl;
}
  

void FindTopplingRange
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 std::vector<CRigid3D_Ini>& aRI1,
 std::vector<CJoint_Fix>& aFix1,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 /////
 std::vector<CRigid3D_Def>& aRD0,
 std::vector<CJoint_Fix_Force>& aJFixForce0,
 std::vector<CContactPoints_RigidBody3D>& aFP0,
 const std::vector<CExternalForceOnRigidBody>& aExForce)
{
  std::cout << "input " << tmin << " " << tmax << std::endl;
  const double tmin_ini = tmin;
  const double tmax_ini = tmax;
  double tmin0, tmax0;
  // conv
  std::vector<CRigid3D_Def> aRD1 = aRD0;
  std::vector<CJoint_Fix_Force> aJFixForce1 = aJFixForce0;  
  CSimulationSensitivity simSens1;
  FindTopplingRangeSensitivity(tmin0, tmax0, desSens, 
                               aRI1, aFix1, aFixDir, gravity, contProp,
                               simSens1,
                               aRD1, aJFixForce1, aFP0, aExForce);
  tmin = (tmin0<tmin*0.5) ? tmin*0.5 : tmin0;
  tmax = (tmax0>tmax*0.5) ? tmax*0.5 : tmax0;  
  std::cout << "initial guess  " << tmin << " " << tmax << std::endl;
  {
    CSimulationSensitivity simSens2 = simSens1;
    std::vector<CRigid3D_Ini> aRI2 = aRI1;
    std::vector<CJoint_Fix> aFix2 = aFix1;    
    std::vector<CContactPoints_RigidBody3D> aFP2 = aFP0;
    std::vector<CRigid3D_Def> aRD2 = aRD1;
    std::vector<CJoint_Fix_Force> aJFixForce2 = aJFixForce1;
    double tmin2 = tmin;
    for(int itr=0;itr<1;itr++){ // find min
      MoveInitialDesignAnchor(desSens,tmin2, aRI2,aFix2,aFP2);
      MoveCurrentDeform(simSens2,tmin2,aRD2,aJFixForce2);
      double tmax2 = tmax;
      FindTopplingRangeSensitivity(tmin2, tmax2, desSens, 
                                   aRI2, aFix2, aFixDir, gravity, contProp,
                                   simSens2,
                                   aRD2, aJFixForce2, aFP2, aExForce);
      if( tmin+tmin2 < tmin_ini ){ tmin = tmin_ini - tmin2; }
      tmin += tmin2;
    }
  }
  std::cout << "minimum determined " << tmin << std::endl;
  {
    CSimulationSensitivity simSens2 = simSens1;
    std::vector<CRigid3D_Ini> aRI2 = aRI1;
    std::vector<CJoint_Fix> aFix2 = aFix1;    
    std::vector<CContactPoints_RigidBody3D> aFP2 = aFP0;
    std::vector<CRigid3D_Def> aRD2 = aRD1;
    std::vector<CJoint_Fix_Force> aJFixForce2 = aJFixForce1;
    double tmax2 = tmax;    
    for(int itr=0;itr<1;itr++){ // find min
      MoveInitialDesignAnchor(desSens,tmax2, aRI2,aFix2,aFP2);
      MoveCurrentDeform(simSens2,tmax0,aRD2,aJFixForce2);
      double tmin2 = tmin;
      FindTopplingRangeSensitivity(tmin2, tmax2, desSens, 
                                   aRI2, aFix2, aFixDir, gravity, contProp,
                                   simSens2,
                                   aRD2, aJFixForce2, aFP2, aExForce);      
      if( tmax+tmax2 > tmax_ini ){ tmax = tmax_ini - tmax2; }      
      tmax += tmax2;
    }    
  }  
  std::cout << "maximum determined " << tmin << std::endl;  
}



void GetJointStabilityRange
(double& smin, double& smax,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 /////
 CSimulationSensitivity& simSens, 
 std::vector<CRigid3D_Def>& aRD,
 std::vector<CJoint_Fix_Force>& aJFixForce,
 std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 const std::vector<CNailJoint>& aNailJoint)
{  
  assert( aRI.size() == aFP.size() );        
  assert( aRD.size() == aFP.size() );     
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);
  /*
  for(unsigned int i=0;i<1000;i++){
    double resn, updn;    
    StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                      aRI,aRD,aFix,aJFixForce,aExForce,
                                                                      gravity,
                                                                      contProp,
                                                                      aFP,
                                                                      resn,updn);	// solve rigid motion    
    if( updn < 1.0e-10 ) break;
  } 
   */
  // sens
  GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem(simSens,
                                                                    desSens,
                                                                    aRI,aRD,aFix,aJFixForce,aExForce,
                                                                    gravity,contProp,aFP);  
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cst = aFix[ifix];
    const CJoint_Fix_Force& cff = aJFixForce[ifix];
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;    
    unsigned int irb0 = cst.irb0;
    unsigned int irb1 = cst.irb1;
    double tmin, tmax;
    assert( idir0 == 4 || idir0 == 5 );    
    if( idir1!=4&&idir1!=5 ){
      CheckNailJointTolerance(tmin,tmax,
                              -simSens.aJointLM[ifix].first,                       
                              -simSens.aJointLM[ifix].second,
                              cst, cff,
                              idir1,idir0,
                              aRI[irb1],aRI[irb0],
                              aNailJoint[ifix]);
    }
    else{
      assert(0);
    }
    if( smin < tmin ){ smin = tmin; }
    if( smax > tmax ){ smax = tmax; }
    std::cout << smin << " " << smax << std::endl;        
  }      
}


void MoveInitialDesign(const CDesignSensitivity& desSens, double eps,
                       std::vector<CRigid3D_Ini>& aRI,
                       std::vector<CJoint_Fix>& aFix)
{
  assert( desSens.aPlank.size() == aRI.size() );
  for(unsigned int irb=0;irb<aRI.size();irb++){
    Com::CVector3D q = desSens.aPlank[irb].p*eps;
    Com::CVector3D s = desSens.aPlank[irb].r*eps;
    Com::CVector3D l = desSens.aPlank[irb].l*eps;
    CRigid3D_Ini& ri = aRI[irb];        
    ri.SetIniPosCG( ri.GetIniPosCG() + q );
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(s);
    ri.AddRotation_IniLeft( rot );
    const double rho = ri.GetMass() / (ri.xlen*ri.ylen*ri.zlen);
    const double xlen = ri.xlen + l.x;
    const double ylen = ri.ylen + l.y;
    const double zlen = ri.zlen + l.z;
    ri.SetModeBox(xlen,ylen,zlen,rho);
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    Com::CVector3D p = desSens.aJoint[ifix]*eps;
    aFix[ifix].ini_pos_joint += p;
  } 
}

bool IsJointInsidePlank
(double t,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI0,
 const std::vector<CJoint_Fix>& aFix0,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir)
{
  std::vector<CRigid3D_Ini> aRI = aRI0;
  std::vector<CJoint_Fix> aFix = aFix0;
  MoveInitialDesign(desSens, t, aRI, aFix);
  for(unsigned int irb=0;irb<aRI.size();irb++){
    CRigid3D_Ini& ri = aRI[irb]; 
    if( ri.xlen < ri.zlen ) return false;
    if( ri.ylen < ri.zlen ) return false;    
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;    
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;    
    Com::CVector3D a, b;
    bool res = GetIntersectionOfTwoPlanks(a,b, aRI[irb0], aRI[irb1], idir0,idir1);
    if( !res ){ return false; }
  }
  return true;
}

bool GetJointForceDepth
(std::vector<double>& aJForce,
 std::vector<double>& aDepth,
 double t,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI0,
 const std::vector<CJoint_Fix>& aFix0,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP0,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 const std::vector<CNailJoint>& aNailJoint0)
{
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };  
  ////
  std::vector<CRigid3D_Ini> aRI = aRI0;
  std::vector<CJoint_Fix> aFix = aFix0;
  MoveInitialDesign(desSens, t, aRI, aFix);    
  for(unsigned int irb=0;irb<aRI.size();irb++){
    CRigid3D_Ini& ri = aRI[irb]; 
    if( ri.xlen < ri.zlen ) return false;
    if( ri.ylen < ri.zlen ) return false;    
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;    
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;    
    Com::CVector3D a, b;
    bool res = GetIntersectionOfTwoPlanks(a,b, aRI[irb0], aRI[irb1], idir0,idir1);
    if( !res ){ return false; }
  }  
  EnforceJointConnectivity_ChangeBoxLen(aRI, aFix, aFixDir);  
  std::vector<CNailJoint> aNailJoint = aNailJoint0;
  PlaceNailsToJoint(aNailJoint,aRI,aFix,aFixDir);  
  std::vector<CContactPoints_RigidBody3D> aFP = aFP0;  
  EnforceContact_ChangeBoxLen(aRI,aFP,contProp.cont_norm,contProp.cont_base);      
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];
    CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;    
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      Com::CVector3D v1 = ri.GetInitialPosition(v0);
      aCP.aFP[i].aloc = v1;
    }    
  }

  std::vector<CRigid3D_Def> aRD; 
  std::vector<CJoint_Fix_Force> aJFixForce;  
  aRD.resize( aRI.size() );
  aJFixForce.resize( aFix.size() );
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,gravity,contProp,aFP);
  aDepth.clear();
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];        
    const CRigid3D_Def& rd = aRD[irb];        
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);  
      Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);
      const double depth1 = Com::Dot(contProp.cont_norm,contProp.cont_base-v1);
      aDepth.push_back(depth1);
      Com::CVector3D cg = aRD[irb].GetDispCG() + aRI[irb].GetIniPosCG();
      const double depth2 = Com::Dot(contProp.cont_norm,cg-contProp.cont_base);
      if( depth2 < 0 ){ return false; }      
      if( depth1 < 0 ){ return false; }
    }
  }
  aJForce.clear();
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cst = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];
//    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;   
    Com::CVector3D pos_force(jff.lambda[0],jff.lambda[1],jff.lambda[2]); 
    Com::CVector3D bnd_force(jff.lambda[3],jff.lambda[4],jff.lambda[5]); 
//    assert( idir0 == 4 || idir0 == 5 );
    const CNailJoint& nj = aNailJoint[ifix];
//    const unsigned int nnail = nj.aPosHead.size();
    const double lennail = nj.lennail;
    const CRigid3D_Ini& ri1 = aRI[cst.irb1];
    const Com::CMatrix3& Q1 = ri1.GetRotMatrix_Ini();
//    const Com::CVector3D& Qn1 = Q1.MatVec( aFaceNormal[idir1] );
    const Com::CVector3D& Qh1 = Q1.MatVec( Cross( Com::CVector3D(0,0,1), aFaceNormal[idir1] ) );
//    const Com::CVector3D& tang = (Com::CMatrix3(1)-Com::CMatrix3(Qn1,Qn1)).MatVec(pos_force);
//    std::cout.setf(std::ios::fixed);
//    sher_ratio = (tang.Length()/nnail)/(nj.shear_max);
    double jforce = Dot(Qh1,bnd_force);
    aJForce.push_back(jforce);
//    std::cout << Qh1 << "    " << bnd_force << std::endl;
//    std::cout << jforce << std::endl;
//    double torq = fabs(jforce);
//    double bend_pull = torq/(ri1.zlen*0.5) + Dot(Qn1,pos_force);
    //  std::cout << torq/(ri0.zlen*0.5) << " " << Dot(Qn0,pos_force) << std::endl;
//    double pull_max = (lennail-ri1.zlen)*nj.pull_max_perlen;
//    bend_ratio = (bend_pull/nnail)/(pull_max); 
  }
  return true;  
}

bool IsSoundStructure
(double t,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI0,
 const std::vector<CJoint_Fix>& aFix0,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP0,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 const std::vector<CNailJoint>& aNailJoint0,
 bool is_check_capa)
{
  std::vector<CRigid3D_Ini> aRI = aRI0;
  std::vector<CJoint_Fix> aFix = aFix0;
  MoveInitialDesign(desSens, t, aRI, aFix);    
  for(unsigned int irb=0;irb<aRI.size();irb++){
    CRigid3D_Ini& ri = aRI[irb]; 
    if( ri.xlen < ri.zlen ) return false;
    if( ri.ylen < ri.zlen ) return false;    
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;    
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;    
    Com::CVector3D a, b;
    bool res = GetIntersectionOfTwoPlanks(a,b, aRI[irb0], aRI[irb1], idir0,idir1);
    if( !res ){ return false; }
  }  
  EnforceJointConnectivity_ChangeBoxLen(aRI, aFix, aFixDir);  
  std::vector<CNailJoint> aNailJoint = aNailJoint0;
  PlaceNailsToJoint(aNailJoint,aRI,aFix,aFixDir);

  std::vector<CContactPoints_RigidBody3D> aFP = aFP0;  
  EnforceContact_ChangeBoxLen(aRI,aFP,contProp.cont_norm,contProp.cont_base);    
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];
    CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;    
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      Com::CVector3D v1 = ri.GetInitialPosition(v0);
      aCP.aFP[i].aloc = v1;
    }    
  }

  std::vector<CRigid3D_Def> aRD; 
  std::vector<CJoint_Fix_Force> aJFixForce;  
  aRD.resize( aRI.size() );
  aJFixForce.resize( aFix.size() );
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,gravity,contProp,aFP);
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];        
    const CRigid3D_Def& rd = aRD[irb];        
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);  
      Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);
      const double depth1 = Com::Dot(contProp.cont_norm,contProp.cont_base-v1);
      Com::CVector3D cg = aRD[irb].GetDispCG() + aRI[irb].GetIniPosCG();
      const double depth2 = Com::Dot(contProp.cont_norm,cg-contProp.cont_base);
      if( depth1 < 0 ){ return false; }
      if( depth2 < 0 ){ return false; }
    }
  }
  if( !is_check_capa ) return true;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cst = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;   
    Com::CVector3D pos_force(jff.lambda[0],jff.lambda[1],jff.lambda[2]); 
    Com::CVector3D bnd_force(jff.lambda[3],jff.lambda[4],jff.lambda[5]); 
    assert( idir0 == 4 || idir0 == 5 );
    double shear, pull;
    if(idir1!=4&&idir1!=5 ){
      CheckCapacity(shear, pull,
                    cst,-pos_force,-bnd_force,
                    idir1,idir0,
                    aRI[cst.irb1],aRD[cst.irb1],
                    aRI[cst.irb0],aRD[cst.irb0],
                    aNailJoint[ifix]);
    }
    if( shear > 1 ) return false;
    if( pull  > 1 ) return false;
  }
  return true;
}


void FindRangeSoundStructure_Bisec
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 const std::vector<CNailJoint>& aNailJoint)
{
  const double tmin0 = tmin;
  const double tmax0 = tmax;
  unsigned int n=30;  
  
  bool is_org_sound = IsSoundStructure(0, desSens, aRI,aFix,aFixDir, 
                                       gravity,contProp,aFP,aExForce,aNailJoint);
  if( is_org_sound ){
    tmin = -(tmax0-tmin0)*1.0e-5;
    tmax = +(tmax0-tmin0)*1.0e-5;
    for(unsigned int i=1;i<=n;i++){
      double t = tmax0/n*i;
      bool res = IsSoundStructure(t, desSens, aRI,aFix,aFixDir, 
                                  gravity,contProp,aFP,aExForce,aNailJoint);
      if( !res ) break;
      tmax = t;
    }
    for(unsigned int i=1;i<=n;i++){    
      double t = tmin0/n*i;
      bool res = IsSoundStructure(t, desSens, aRI,aFix,aFixDir,
                                  gravity,contProp,aFP,aExForce,aNailJoint);
      if( !res ) break;
      tmin = t;
    }      
  }
  else{
    unsigned int i,j;
    for(i=1;i<=n;i++){
      double t = tmax0/n*i;
      bool res = IsSoundStructure(t, desSens, aRI,aFix,aFixDir, 
                                  gravity,contProp,aFP,aExForce,aNailJoint);
      if( res ) break;
    }
    for(int j=1;j<=n;j++){    
      double t = tmin0/n*j;
      bool res = IsSoundStructure(t, desSens, aRI,aFix,aFixDir,
                                  gravity,contProp,aFP,aExForce,aNailJoint);
      if( res ) break;
    }
//    std::cout << "org unstable " << i << " "<< j << std::endl;
    if( i <= j && i != n){  // stable in i 
      tmin = tmax0/n*i-(tmax0-tmin0)*1.0e-5;
      for(int k=i+1;k<=n;k++){
        double t = tmax0/n*k;
        bool res = IsSoundStructure(t, desSens, aRI,aFix,aFixDir, 
                                    gravity,contProp,aFP,aExForce,aNailJoint);
        if( !res ) break;
        tmax = t;
      }      
    }
    else if( j < i ){ // stable in j
      tmax = tmin0/n*j+(tmax0-tmin0)*1.0e-5;
      for(int k=j+1;k<=n;k++){    
        double t = tmin0/n*k;
        bool res = IsSoundStructure(t, desSens, aRI,aFix,aFixDir,
                                    gravity,contProp,aFP,aExForce,aNailJoint);
        if( !res ) break;
        tmin = t;
      }      
    }
    else{
      tmin = tmax = 0;
    }
  }
}



bool IsntToppling
(double t,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI0,
 const std::vector<CJoint_Fix>& aFix0,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP0,
 const std::vector<CExternalForceOnRigidBody>& aExForce)
{
  std::vector<CRigid3D_Ini> aRI = aRI0;
  std::vector<CJoint_Fix> aFix = aFix0;
  MoveInitialDesign(desSens, t, aRI, aFix);    
  std::vector<CContactPoints_RigidBody3D> aFP = aFP0;  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];
    CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;    
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      Com::CVector3D v1 = ri.GetInitialPosition(v0);
      aCP.aFP[i].aloc = v1;
    }    
  }
  std::vector<CRigid3D_Def> aRD; 
  std::vector<CJoint_Fix_Force> aJFixForce;  
  aRD.resize( aRI.size() );
  aJFixForce.resize( aFix.size() );
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,gravity,contProp,aFP);
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];        
    const CRigid3D_Def& rd = aRD[irb];        
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);  
      Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);
      const double depth1 = Com::Dot(contProp.cont_norm,contProp.cont_base-v1);
      Com::CVector3D cg = aRD[irb].GetDispCG() + aRI[irb].GetIniPosCG();
      const double depth2 = Com::Dot(contProp.cont_norm,cg-contProp.cont_base);      
      if( depth1 < 0 ){ return false; }
      if( depth2 < 0 ){ return false; }
    }
  }
  return true;
}
                  

void FindTopplingRange_Bisec
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 std::vector<CRigid3D_Ini>& aRI,
 std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector<CExternalForceOnRigidBody>& aExForce)
{
  const double tmin0 = tmin;
  const double tmax0 = tmax;
  //  unsigned int n_min=2;
  unsigned int n=10;  
  for(unsigned int i=0;i<=n;i++){    
    double t = tmax0/n*i;
    bool res = IsntToppling(t, desSens, aRI,aFix,aFixDir, 
                            gravity,contProp,aFP,aExForce);
    if( !res ) break;
    std::cout << "max" << i << " " << t << std::endl;
    tmax = t;
  }
  for(unsigned int i=0;i<=n;i++){    
    double t = tmin0/n*i;
    bool res = IsntToppling(t, desSens, aRI,aFix,aFixDir,
                            gravity,contProp,aFP,aExForce);
    if( !res ) break;
    std::cout << "min" << i << " " << t << std::endl;    
    tmin = t;
  }      
}

void FindJointInsidePlankRange_Bisec
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir)
{
  const double tmin0 = tmin;
  const double tmax0 = tmax;
//  unsigned int n_min=2;
  unsigned int n=30;  
  for(unsigned int i=0;i<=n;i++){    
    double t = tmax0/n*i;
    bool res = IsJointInsidePlank(t, desSens, aRI,aFix,aFixDir);
    if( !res ) break;
    tmax = t;
  }
  for(unsigned int i=0;i<=n;i++){    
    double t = tmin0/n*i;
    bool res = IsJointInsidePlank(t, desSens, aRI,aFix,aFixDir);
    if( !res ) break;
    tmin = t;
  }  
}



void FindJointInsidePlankRange
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir)
{
  const Com::CVector3D aFaceNormal[6][3] = {
    { Com::CVector3D(+1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0,+1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-0,-1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0, 0,+1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) },
    { Com::CVector3D(-0, 0,-1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) } };      
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int irb0 = aFix[ifix].irb0;
    const CRigid3D_Ini& ri0 = aRI[irb0];
    const Com::CVector3D l0(ri0.xlen,ri0.ylen,ri0.zlen);
    Com::CMatrix3 B0(0.0); B0.SetDiag(aFaceNormal[idir0][0]*0.5);
    const Com::CMatrix3 L0(l0.x, l0.y, l0.z);
    const Com::CMatrix3 R0 = ri0.GetRotMatrix_Ini();
    const Com::CMatrix3& RB0 = R0.MatMat(B0);
    const Com::CVector3D& n0 = R0.MatVec( aFaceNormal[idir0][0] );      
    const Com::CVector3D& o0 = RB0.MatVec(l0) + ri0.GetIniPosCG();
    const Com::CVector3D& _n0 = Com::Cross(desSens.aPlank[irb0].r,n0);
    const Com::CVector3D& _o0 = desSens.aPlank[irb0].p + Com::Cross(desSens.aPlank[irb0].r,RB0.MatVec(l0)) + RB0.MatVec(desSens.aPlank[irb0].l);
    ////    
    unsigned int idir1 = aFixDir[ifix].second;    
    unsigned int irb1 = aFix[ifix].irb1;
    const CRigid3D_Ini& ri1 = aRI[ irb1 ];       
    const Com::CVector3D l1(ri1.xlen,ri1.ylen,ri1.zlen);
    Com::CMatrix3 B1(0.0); B1.SetDiag(aFaceNormal[idir1][0]*0.5);
    const Com::CMatrix3 L1(l1.x, l1.y, l1.z);
    const Com::CMatrix3 R1 = ri1.GetRotMatrix_Ini();
    const Com::CMatrix3& RB1 = R1.MatMat(B1);
    const Com::CVector3D& n1 = R1.MatVec( aFaceNormal[idir1][0] );              
    const Com::CVector3D& o1 = RB1.MatVec(l1) + ri1.GetIniPosCG(); 
    const Com::CVector3D& _n1 = Com::Cross(desSens.aPlank[irb1].r,n1);
    const Com::CVector3D& _o1 = desSens.aPlank[irb1].p + Com::Cross(desSens.aPlank[irb1].r,RB1.MatVec(l1)) + RB1.MatVec(desSens.aPlank[irb1].l);
    assert( idir0 == 4 || idir0 == 5 );    
    if( idir1!=4||idir1!=5 ){
      Com::CMatrix3 C0(0.0); C0.SetDiag(aFaceNormal[idir0][1]*0.5);
      Com::CMatrix3 D0(0.0); D0.SetDiag(aFaceNormal[idir0][2]*0.5);          
      const Com::CMatrix3& RC0 = R0.MatMat(C0);
      const Com::CMatrix3& RD0 = R0.MatMat(D0);      
      const Com::CVector3D& h0 = RC0.MatVec(l0);      
      const Com::CVector3D& v0 = RD0.MatVec(l0);
      const Com::CVector3D& _h0 = Com::Cross(desSens.aPlank[irb0].r,h0) + RC0.MatVec(desSens.aPlank[irb0].l);
      const Com::CVector3D& _v0 = Com::Cross(desSens.aPlank[irb0].r,v0) + RD0.MatVec(desSens.aPlank[irb0].l);
      double rh,rv, _rh,_rv;
      bool res = SensRayCollideQuad( o1, n1,  o0, n0,  h0, v0, rv, rh,
                                    _o1,_n1, _o0,_n0, _h0,_v0,_rv,_rh); 
      if( !res ) continue;
      if( fabs(_rh) > (ri0.xlen+ri0.ylen) * 0.01 ){
        const double tmp0 = (-1-rh)/_rh;
        const double tmp1 = (+1-rh)/_rh;
        double win_min_h = ( tmp0 < tmp1 ) ? tmp0 : tmp1;      
        double win_max_h = ( tmp0 > tmp1 ) ? tmp0 : tmp1;         
        tmin = ( tmin > win_min_h ) ? tmin : win_min_h;
        tmax = ( tmax < win_max_h ) ? tmax : win_max_h;        
      }
      if( fabs(_rv) > (ri0.xlen+ri0.ylen) * 0.01 ){
        const double tmp0 = (-1-rv)/_rv;
        const double tmp1 = (+1-rv)/_rv;
        double win_min_v = ( tmp0 < tmp1 ) ? tmp0 : tmp1;      
        double win_max_v = ( tmp0 > tmp1 ) ? tmp0 : tmp1;         
        tmin = ( tmin > win_min_v ) ? tmin : win_min_v;
        tmax = ( tmax < win_max_v ) ? tmax : win_max_v;        
      }      
    }
    std::cout << ifix << " " << tmin << " " << tmax << std::endl;
  } // for(int ifix;ifix<aFix.size();ifix++)
  std::cout << "min max joint  " << tmin << " " << tmax << std::endl;  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];
    const double _xlen = desSens.aPlank[irb].l[0];
    if( fabs(_xlen) > (ri.xlen+ri.ylen)*0.001 ){
      const double t = -ri.xlen/_xlen;
      if( t > 0 ){
        if( tmax > t ){ tmax = t; }
      }
      if( t < 0 ){
        if( tmin < t ){ tmin = t; }        
      }
    }
    const double _ylen = desSens.aPlank[irb].l[1];  
    if( fabs(_ylen) > (ri.xlen+ri.ylen)*0.001 ){
      const double t = -ri.ylen/_ylen;
      if( t > 0 ){
        if( tmax > t ){ tmax = t; }
      }
      if( t < 0 ){
        if( tmin < t ){ tmin = t; }        
      }
    }    
//    std::cout << ri.xlen/_xlen << " " << ri.ylen/_ylen << std::endl;    
  }
  std::cout << "min max final " << tmin << " " << tmax << std::endl;  
}
 




