/*
 *  nail_furniture.h
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/31/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if !defined(NAIL_FURNITURE_H)
#define NAIL_FURNITURE_H

#include "rigidbody.h"
#include "jointed_rigidbody_contact.h"
#include "joint_fix.h"


//  const double shear_max = 19.0*10.0; // per newton
//  const double pull_max_perlen = 3500;

class CNailJoint
{
public:
  CNailJoint(){
    lennail = 0.032;
    dia_head = 0.0045;
    intervnail = 0.02; // 2cm
    shear_max = 19.0*10.0; // per newton
    pull_max_perlen = 3.5*100*10.0;  // 4kg/cm = 4000N/m  
    /*
    lennail = 0.045;
    dia_head = 0.0058;
    intervnail = dia_head*5; // 2cm
    shear_max = 19.0*10.0; // per newton
    pull_max_perlen = 4.5*100*10.0;  // 4kg/cm = 4000N/m      
     */
  }
public:
  double shear_max;
  double pull_max_perlen;
  double intervnail;
  double lennail;
  double dia_head;
  std::vector<Com::CVector3D> aPosHead;
};


void DisplayJointCapacity(unsigned int ifix, double& shear, double& pull,
                          const std::vector<CRigid3D_Ini>& aRI,
                          const std::vector<CRigid3D_Def>& aRD,
                          const std::vector<CJoint_Fix>& aFix,
                          const std::vector<CJoint_Fix_Force>& aJFixForce,
                          const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                          
                          const std::vector<CNailJoint>& aNailJoint);
void FindTopplingRange
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 std::vector<CRigid3D_Ini>& aRI1,
 std::vector<CJoint_Fix>& aFix1,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 /////
 std::vector<CRigid3D_Def>& aRD0,
 std::vector<CJoint_Fix_Force>& aJFixForce0,
 std::vector<CContactPoints_RigidBody3D>& aFP0,
 const std::vector<CExternalForceOnRigidBody>& aExForce);

void FindJointInsidePlankRange
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir);

void FindJointInsidePlankRange_Bisec
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir);

void FindTopplingRange_Bisec
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 std::vector<CRigid3D_Ini>& aRI,
 std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector<CExternalForceOnRigidBody>& aExForce);

void GetJointStabilityRange
(double& smin, double& smax,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 /////
 CSimulationSensitivity& simSens, 
 std::vector<CRigid3D_Def>& aRD,
 std::vector<CJoint_Fix_Force>& aJFixForce,
 std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 const std::vector<CNailJoint>& aNailJoint);

void FindRangeSoundStructure_Bisec
(double& tmin, double& tmax,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 const std::vector<CNailJoint>& aNailJoint);


void PlaceNailsToJoint
(std::vector<CNailJoint>& aNailJoint,
 const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir);

bool IsSoundStructure
(double t,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI0,
 const std::vector<CJoint_Fix>& aFix0,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP0,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 const std::vector<CNailJoint>& aNailJoint0, 
 bool is_check_capa = true);

bool GetJointForceDepth
(std::vector<double>& aJForce,
 std::vector<double>& aDepth,
 double t,
 const CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI0,
 const std::vector<CJoint_Fix>& aFix0,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP0,
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 const std::vector<CNailJoint>& aNailJoint0);

void MoveInitialDesign(const CDesignSensitivity& desSens, double eps,
                       std::vector<CRigid3D_Ini>& aRI,
                       std::vector<CJoint_Fix>& aFix);

#endif
