/*
 *  optimize_furniture.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 11/6/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include <set>
#include <stdio.h>
//#include "Eigen/Dense"

#include "delfem/indexed_array.h"
#include "delfem/matvec/matdia_blkcrs.h"
#include "delfem/matvec/matdiafrac_blkcrs.h"
#include "delfem/matvec/vector_blk.h"

#include "rigid_contact.h"
#include "gl_drawer_furniture.h"
#include "edit_mode.h"
#include "edit_furniture.h"
#include "nail_bound_furniture.h"
#include "joint_connectivity.h"
#include "rigidbody.h"
#include "jointed_rigidbody_contact.h"


#include "optimize_furniture.h"

//using Eigen::MatrixXd;
//using Eigen::MatrixXd;

void Optimize_GradModes
(double alpha,
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,                                        
 std::vector<CJoint_Fix>& aFix, 
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir,
 const std::vector<CExternalForceOnRigidBody>& aExForce,     
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 std::vector<CContactPoints_RigidBody3D>& aFP)
{
  const unsigned int nRB = aRI.size();
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);
  std::vector< std::pair<unsigned int,unsigned int> > aRBMode;
  std::vector< CDesignSensitivity > aDesMode;    
  aRBMode.reserve( nRB*8 );
  aDesMode.reserve( nRB*8 );
  for(unsigned int irb=0;irb<aRI.size();irb++){
    int icont_flg = 0;
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x
    if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x
    if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y        
    if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y
    int ijnt_flg = 0;
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      unsigned int irb0 = aFix[ifix].irb0;
      unsigned int irb1 = aFix[ifix].irb1;      
      unsigned int idir0 = aFixDir[ifix].first;
      unsigned int idir1 = aFixDir[ifix].second;
      if( irb0 != irb && irb1 != irb ) continue;
      unsigned int idir2 = ( irb0 == irb) ? idir0 : idir1;
      if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
      if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
      if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
      if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y        
    }    
    for(unsigned int imode=0;imode<8;imode++){      
      bool is_alive = IsAliveMode(aRI,aFix,aFixDir,irb,imode,icont_flg,ijnt_flg);
      if( !is_alive ) continue;
      aRBMode.push_back( std::make_pair(irb,imode) );
      CDesignSensitivity desSens;
      GetDesignSensitivity(desSens,   aRI,aFix,aFixDir,   irb,imode,icont_flg,contProp);  
      aDesMode.push_back(desSens);
    }
  }
  std::vector<double> aGrad;
  std::vector<CSimulationSensitivity> aSimSens;
  {
    aGrad.resize(aRBMode.size());    
    aSimSens.resize(aRBMode.size());
    //  std::cout << "Get Sensitivity w.r.t initial configuration (improved version) " << std::endl;
    MatVec::CMatDia_BlkCrs m_mat;
    MatVec::CVector_Blk m_r;
    MatVec::CMatDiaFrac_BlkCrs m_frac;    
    MakeLinearSystemFrac_JointedRigidBodyContactSystem
    (m_mat,m_r,m_frac,
     aRI, aRD, aFix, aJFixForce, aExForce,
     gravity, contProp, aFP);    
    for(unsigned int irbm=0;irbm<aRBMode.size();irbm++){
      const CDesignSensitivity& desSens = aDesMode[irbm];
      CSimulationSensitivity& simSens = aSimSens[irbm];
      // residual 
      CSimulationSensitivity resSim; resSim.Initialize(aRI.size(), aFix.size());
      AddSensitivityResidual_JointedRigidBodyContactSystem(resSim.aPlankPR,resSim.aJointLM,
                                                           desSens,MOVE_INITIAL,
                                                           aRI,aRD,aFix,aJFixForce,aExForce,
                                                           gravity, contProp, aFP);
      m_r.SetVectorZero();
      for(unsigned int irb=0;irb<aRI.size();irb++){
        AddVector(m_r, irb,0, resSim.aPlankPR[irb].first,  1);
        AddVector(m_r, irb,3, resSim.aPlankPR[irb].second, 1);    
      }
      for(unsigned int ifix=0;ifix<aFix.size();ifix++){
        AddVector(m_r, ifix+nRB,0, resSim.aJointLM[ifix].first,  1);
        AddVector(m_r, ifix+nRB,3, resSim.aJointLM[ifix].second, 1);        
      }
      
      ////////////////////////////////////////////////////////////////
      // Solve Linear System
      m_frac.Solve( m_r );
      
      double tmp[6];
      simSens.aPlankPR.clear();
      for(unsigned int irb=0;irb<aRI.size();irb++){
        const unsigned int nlen = m_r.Len(irb); assert( nlen == 6 );
        for(unsigned int ilen=0;ilen<nlen;ilen++){ tmp[ilen] = m_r.GetValue(irb,ilen); }
        simSens.aPlankPR.push_back( std::make_pair(Com::CVector3D(tmp[0],tmp[1],tmp[2]), 
                                                   Com::CVector3D(tmp[3],tmp[4],tmp[5]) ) );    
      }    
      simSens.aJointLM.clear();
      for(unsigned int icst=0;icst<aFix.size();icst++){
        const unsigned int nlen = m_r.Len(icst+nRB); assert( nlen == 6 );
        for(unsigned int ilen=0;ilen<nlen;ilen++){ tmp[ilen] = m_r.GetValue(icst+nRB,ilen); }
        simSens.aJointLM.push_back( std::make_pair(Com::CVector3D(tmp[0],tmp[1],tmp[2]), 
                                                   Com::CVector3D(tmp[3],tmp[4],tmp[5]) ) );
      }      
      double grad = 0;
      for(unsigned int ifix=0;ifix<aFix.size();ifix++){
        const CJoint_Fix_Force& jff = aJFixForce[ifix];
        const Com::CVector3D& gfb = simSens.aJointLM[ifix].second; 
        const Com::CVector3D fb(jff.lambda[3],jff.lambda[4],jff.lambda[5]);      
        grad += Com::Dot(gfb,fb);
      }
      aGrad[irbm] = grad;
    }
  }      
  CDesignSensitivity desSens; desSens.Initialize(aRI.size(), aFix.size());
  CSimulationSensitivity simSens; simSens.Initialize(aRI.size(), aFix.size());
  for(unsigned int ids=0;ids<aDesMode.size();ids++){
    desSens.AddScaled(aDesMode[ids], -aGrad[ids]*alpha);
    simSens.AddScaled(aSimSens[ids], -aGrad[ids]*alpha);
  }
  MoveInitialDesignAnchor(desSens, 1, aRI, aFix, aFP);      
  MoveCurrentDeform(simSens, 1, aRD, aJFixForce);  
}


void GetBendingInvariantModes
(std::vector<CDesignSensitivity>& aDesMode,
 const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,                                        
 const std::vector<CJoint_Fix>& aFix, 
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir,
 const std::vector<CExternalForceOnRigidBody>& aExForce,     
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 std::vector<CContactPoints_RigidBody3D>& aFP)
{
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };  
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);  
  
  std::vector< std::pair<unsigned int,unsigned int> > aRBMode;
  std::vector< CDesignSensitivity > aDesSens;
  aRBMode.reserve(aRI.size()*8);  
  aDesSens.reserve(aRI.size()*8);  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    int icont_flg = 0;
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x    
    if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x
    if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y        
    if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y
    int ijnt_flg = 0;
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      unsigned int irb0 = aFix[ifix].irb0;
      unsigned int irb1 = aFix[ifix].irb1;      
      unsigned int idir0 = aFixDir[ifix].first;
      unsigned int idir1 = aFixDir[ifix].second;
      if( irb0 != irb && irb1 != irb ) continue;
      unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
      if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
      if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
      if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
      if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y        
    }    
    for(unsigned int imode=0;imode<8;imode++){      
      bool is_alive = IsAliveMode(aRI,aFix,aFixDir,irb,imode,icont_flg,ijnt_flg);
      if( !is_alive ) continue;
      aRBMode.push_back( std::make_pair(irb,imode) );
      CDesignSensitivity desSens;
      GetDesignSensitivity
      (desSens,
       aRI,aFix,aFixDir,                
       irb,imode,icont_flg,contProp);      
      aDesSens.push_back(desSens);
    }
  }
  const unsigned int nRBMode = aRBMode.size();
  std::vector<Com::CVector3D> aQh;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
//    const CRigid3D_Ini& ri0 = aRI[aFix[ifix].irb0];
    const CRigid3D_Ini& ri1 = aRI[aFix[ifix].irb1];    
    assert( idir0 == 4 || idir0 == 5 );
    if( idir1!=4 && idir1!=5 ){    
      const Com::CMatrix3& Q1 = ri1.GetRotMatrix_Ini();
      const Com::CVector3D& Qh1 = Q1.MatVec( Cross( Com::CVector3D(0,0,1), aFaceNormal[idir1] ) );
      aQh.push_back(Qh1);
      Com::CVector3D vtmp2(aJFixForce[ifix].lambda[3],aJFixForce[ifix].lambda[4],aJFixForce[ifix].lambda[5]);
//      std::cout << " fix : " << ifix << " " << Com::Dot(vtmp2,Qh1) << std::endl;
    }
  }  
//  std::cout << std::endl;
  std::vector<double> A;
  A.resize( nRBMode*aFix.size() );
  assert( aQh.size() == aFix.size() );  
  for(unsigned int irbm=0;irbm<nRBMode;irbm++){
    const CDesignSensitivity& desSens = aDesSens[irbm];
    CSimulationSensitivity simSens;
    GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
    (simSens,
     desSens,
     aRI, aRD, aFix, aJFixForce, aExForce,
     gravity, contProp, aFP);
//    std::cout << irbm << " ";
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      double dtmp1 = Com::Dot(simSens.aJointLM[ifix].second,aQh[ifix]);
      unsigned int irb1 = aFix[ifix].irb1;
      Com::CVector3D vtmp1 = Cross(desSens.aPlank[irb1].r,aQh[ifix]);
      Com::CVector3D vtmp2(aJFixForce[ifix].lambda[3],aJFixForce[ifix].lambda[4],aJFixForce[ifix].lambda[5]);
      double dtmp2 = Com::Dot(vtmp1,vtmp2);
      A[ifix*(nRBMode)+irbm] = +dtmp1 + dtmp2;      
//      std::cout << A[ifix*(nRBMode)+irbm] << " ";
    }
//    std::cout << std::endl;
  }
  /*
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    for(unsigned int iop=0;iop<nRBMode;iop++){
      std::cout << ifix << " " << iop << "  " << A[ifix*(nRBMode)+iop] << std::endl;
    }
  }
   */
/*
  FILE *fp = fopen("sens_bend.m","w");
  fprintf(fp,"A=[ ");
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    for(unsigned int iop=0;iop<nRBMode;iop++){
      fprintf(fp,"%lf ",A[ifix*(nRBMode)+iop]);
    }
    fprintf(fp,"; ");
  }
  fprintf(fp,"];");
  fclose(fp);  
*/  
  /*
  ////////////////////
  // perform SVD
  Eigen::MatrixXd m(aFix.size(),nRBMode);
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    for(unsigned int iop=0;iop<nRBMode;iop++){
      m(ifix,iop) = A[ifix*(nRBMode)+iop];
    }
  }  
  Eigen::JacobiSVD<Eigen::MatrixXd> svd(m, Eigen::ComputeFullU | Eigen::ComputeFullV);    
  svd.computeV();  
  const Eigen::MatrixXd& V1 = svd.matrixV();
   */
  ////////////////////
  std::vector< std::vector<double> > V2;
  /*
  V2.resize(nRBMode-aFix.size());
  for(unsigned int i=0;i<nRBMode-aFix.size();i++){
    V2[i].resize(nRBMode);
    for(unsigned int j=0;j<nRBMode;j++){
      V2[i][j] = V1(j,i+aFix.size());
//      V2[i][j] = V1(i+aFix.size(),j);      
    }
  } 
   */
  /*
  V2.resize(nRBMode);
  for(unsigned int i=0;i<nRBMode;i++){
    V2[i].resize(nRBMode,0);    
//    std::cout << i << "  ";
    for(unsigned int j=0;j<nRBMode;j++){
      V2[i][j] = V1(j,i);
//      V2[i][j] = V1(i,j);      
//      std::cout << V2[i][j] << " ";
 //      V2[i][j] = V1(i+aFix.size(),j);      
    }
//    std::cout << std::endl;
//    V2[i][i] = 1.0;
  } 
   */
  ////
  std::vector<unsigned int> aModeCount;
  aDesMode.clear();
  for(unsigned int i=0;i<V2.size();i++){
    /*
    double maxabs = 0;
    for(unsigned int j=0;j<V2[i].size();j++){
      if( maxabs < fabs(V2[i][j]) ){
        maxabs = fabs(V2[i][j]);
      }
    }
    unsigned int icnt = 0;
    for(unsigned int j=0;j<V2[i].size();j++){
      if( fabs(V2[i][j]) > maxabs*0.3 ){
        icnt++;
      }
    }      
    if( icnt == 1 ) continue;
     */
    ////
    CDesignSensitivity desSens;
    desSens.Initialize(aRI.size(), aFix.size());
//    std::cout << "mode : " << i;
    for(unsigned int j=0;j<nRBMode;j++){
      desSens.AddScaled(aDesSens[j],V2[i][j]);
//      std::cout << " " << V2[i][j];
    } 
//    std::cout << std::endl;
    aDesMode.push_back(desSens);
//    aModeCount.push_back(icnt);
  }
  for(unsigned int imo=0;imo<aDesMode.size();imo++){
    CSimulationSensitivity simSens;
    GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
    (simSens,
     aDesMode[imo],
     aRI, aRD, aFix, aJFixForce, aExForce,
     gravity, contProp, aFP);
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      double dtmp1 = Com::Dot(simSens.aJointLM[ifix].second,aQh[ifix]);
      unsigned int irb1 = aFix[ifix].irb1;
      Com::CVector3D vtmp1 = Cross(aDesMode[imo].aPlank[irb1].r,aQh[ifix]);
      Com::CVector3D vtmp2(aJFixForce[ifix].lambda[3],aJFixForce[ifix].lambda[4],aJFixForce[ifix].lambda[5]);
      double dtmp2 = Com::Dot(vtmp1,vtmp2);
//      A[ifix*(nRBMode)+irbm] = +dtmp1 + dtmp2;      
//      std::cout << imo << " "<< ifix << " " << dtmp1+dtmp2 << std::endl;
    }    
  }  
}
  /*
  assert( aModeCount.size() > 0 );
  unsigned int imo_max = 0;
  {
    unsigned int max_count = aModeCount[0];
    for(unsigned int imo=1;imo<aModeCount.size();imo++){
      if( aModeCount[imo] > max_count ){
        imo_max = imo;
      }
    }
  }
  std::cout << imo_max << std::endl;
  MoveInitialDesign
  (aDesMode[imo_max],alpha,
   aRI,
   aFix,
   aFixDir,                        
   aFP);        
   */
  /*
  CDesignSensitivity desSens;
  desSens.Initialize(aRI.size(), aFix.size());
  for(unsigned int j=0;j<nRBMode;j++){
    desSens.AddScaled(aDesSens[j],V1(j,0));
  }   
  MoveInitialDesign
  (desSens,alpha,
   aRI,
   aFix,
   aFixDir,                        
   aFP);        
  */
/*  

  for(unsigned int i=0;i<nRBMode-aFix.size();i++){

  }   
*/ 

class CRayCollide{
public:
  bool operator<(const CRayCollide& rc) const {
    return fabs(t) < fabs(rc.t);
  }  
public:    
  double t;
  double r0;
  double r1;
  double r2;
  int imo0;
  int imo1;
  int imo2;
};


static void CalcInvMat(double* a, const unsigned int n, int& info )
{
	double tmp1;
  
	info = 0;
	unsigned int i,j,k;
	for(i=0;i<n;i++){
		if( fabs(a[i*n+i]) < 1.0e-30 ){
			info = 1;
			return;
		}
		if( a[i*n+i] < 0.0 ){
			info--;
		}
		tmp1 = 1.0 / a[i*n+i];
		a[i*n+i] = 1.0;
		for(k=0;k<n;k++){
			a[i*n+k] *= tmp1;
		}
		for(j=0;j<n;j++){
			if( j!=i ){
				tmp1 = a[j*n+i];
				a[j*n+i] = 0.0;
				for(k=0;k<n;k++){
					a[j*n+k] -= tmp1*a[i*n+k];
				}
			}
		}
	}
}

void GetSuggestionModes_JointDurable
(std::vector<CDesModeBound>& aDesModeBound,
 const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,                                        
 const std::vector<CJoint_Fix>& aFix, 
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir,
 const std::vector<CNailJoint>& aNailJoint,
 const std::vector<CExternalForceOnRigidBody>& aExForce,  
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 std::vector<CContactPoints_RigidBody3D>& aFP)
{
  aDesModeBound.clear();
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };  
//  EnforceContact_ChangeBoxLen(
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);  
 
  clock_t clock_start,clock_middle,clock_end;
  clock_start = clock();            
  
  std::vector< std::pair<unsigned int,unsigned int> > aRBMode;
  std::vector< CDesignSensitivity > aDesSens;
  aRBMode.reserve(aRI.size()*8);  
  aDesSens.reserve(aRI.size()*8);  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    int icont_flg = 0;
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x    
    if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x
    if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y        
    if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y
    int ijnt_flg = 0;
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      unsigned int irb0 = aFix[ifix].irb0;
      unsigned int irb1 = aFix[ifix].irb1;      
      unsigned int idir0 = aFixDir[ifix].first;
      unsigned int idir1 = aFixDir[ifix].second;
      if( irb0 != irb && irb1 != irb ) continue;
      unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
      if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
      if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
      if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
      if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y        
    }    
    for(unsigned int imode=0;imode<8;imode++){      
      bool is_alive = IsAliveMode(aRI,aFix,aFixDir,irb,imode,icont_flg,ijnt_flg);
      if( !is_alive ) continue;
      aRBMode.push_back( std::make_pair(irb,imode) );
      CDesignSensitivity desSens;
      GetDesignSensitivity
      (desSens,
       aRI,aFix,aFixDir,                
       irb,imode,icont_flg,contProp);      
      aDesSens.push_back(desSens);
    }
  }
  const unsigned int nRBMode = aDesSens.size();
  std::vector<Com::CVector3D> aQh;
  std::vector<double> aBendCap;
  std::vector<double> aBend;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
    const CRigid3D_Ini& ri0 = aRI[aFix[ifix].irb0];
    const CRigid3D_Ini& ri1 = aRI[aFix[ifix].irb1];
    const CNailJoint& nj = aNailJoint[ifix];
    assert( idir0 == 4 || idir0 == 5 );
    if( idir1!=4 && idir1!=5 ){    
      const Com::CMatrix3& Q1 = ri1.GetRotMatrix_Ini();
      const Com::CVector3D& Qn1 = Q1.MatVec( aFaceNormal[idir1] );      
      const Com::CVector3D& Qh1 = Q1.MatVec( Cross( Com::CVector3D(0,0,1), aFaceNormal[idir1] ) );
      aQh.push_back(Qh1);
      Com::CVector3D lam_r(aJFixForce[ifix].lambda[3],aJFixForce[ifix].lambda[4],aJFixForce[ifix].lambda[5]);
      double pull_max = (nj.lennail-ri0.zlen)*nj.pull_max_perlen;
      Com::CVector3D lam_t(aJFixForce[ifix].lambda[0],aJFixForce[ifix].lambda[1],aJFixForce[ifix].lambda[2]);
      double bend_cap = (pull_max*nj.aPosHead.size()-Dot(Qn1,lam_t))*0.5*ri1.zlen; 
      double bend = Com::Dot(lam_r,Qh1);
      aBend.push_back(bend);
      aBendCap.push_back(bend_cap);
//      std::cout << " fix : " << ifix << " " << bend << " " << bend_cap << std::endl;
    }
  } 
  std::vector< std::pair<unsigned int,double> > aPrjIndGap;
  bool is_prj_need = false;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    if( fabs(aBend[ifix]) > fabs(aBendCap[ifix]) ){
      is_prj_need = true;
      double gap;
      if( aBend[ifix] < 0 ){
        gap = -aBendCap[ifix]-aBend[ifix];
      }
      else{
        gap = +aBendCap[ifix]-aBend[ifix];
      }
      aPrjIndGap.push_back( std::make_pair(ifix,gap) );
    }
  }
//  std::cout << std::endl;
  if( !is_prj_need ){ 
    std::cout << "no-need suggestion joint durable" << std::endl;
    return; 
  }
  std::cout << "generate suggestion joint durable" << std::endl;
  ////
  std::vector<double> A;
  const unsigned int nFix = aFix.size();
  A.resize( nRBMode*aFix.size() );
  assert( aQh.size() == aFix.size() );  
  for(unsigned int irbm=0;irbm<nRBMode;irbm++){
    const CDesignSensitivity& desSens = aDesSens[irbm];
    CSimulationSensitivity simSens;
    GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
    (simSens,
     desSens,
     aRI, aRD, aFix, aJFixForce, aExForce,
     gravity, contProp, aFP);
//    std::cout << irbm << " ";
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      double dtmp1 = Com::Dot(simSens.aJointLM[ifix].second,aQh[ifix]);
      unsigned int irb1 = aFix[ifix].irb1;
      Com::CVector3D vtmp1 = Cross(desSens.aPlank[irb1].r,aQh[ifix]);
      Com::CVector3D vtmp2(aJFixForce[ifix].lambda[3],aJFixForce[ifix].lambda[4],aJFixForce[ifix].lambda[5]);
      double dtmp2 = Com::Dot(vtmp1,vtmp2);
      A[ifix*(nRBMode)+irbm] = +dtmp1 + dtmp2;      
//      std::cout << A[ifix*(nRBMode)+irbm] << " ";
    }
//    std::cout << std::endl;
  }
  std::set<CRayCollide> setRayCollide;
  for(unsigned int  irbm=0;irbm<nRBMode;irbm++){
    double min, max;
    for(unsigned int iprj=0;iprj<aPrjIndGap.size();iprj++){
      unsigned int ind = aPrjIndGap[iprj].first;
      double gap = aPrjIndGap[iprj].second;
      double grad = A[ind*(nRBMode)+irbm];
      double t = gap/grad;
      if( iprj == 0 ){
        min = t; 
        max = t;
      }
      if( t < min ){ min = t; }
      if( t > max ){ max = t; }
    }
    if( min*max < 0 ){ 
//      std::cout << "out " << min << " " << max << std::endl;
      continue;
    }
    double t;
    if( fabs(min) < fabs(max) ){ t = max; }
    else{ t = min; }
    if( fabs(t) > 1 ){ continue; }
    bool is_out = false;
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      double bnd = aBend[ifix] + A[ifix*(nRBMode)+irbm]*t;
      if( fabs(bnd) > aBendCap[ifix] ){
//        std::cout << "   out " << ifix << " " << bnd << " " << aBendCap[ifix] << std::endl;
        is_out = true;
      }
    }
    if( is_out ){ continue; }
//    std::cout << "in  " << min << " " << max << " " << t << std::endl; 
    CRayCollide rc;
    rc.t = t;
    rc.r0 = 1;
    rc.imo0 = irbm;
    rc.imo1 = -1;
    rc.imo2 = -1;
    setRayCollide.insert(rc);
  }
  for(unsigned int  irbm=0;irbm<nRBMode;irbm++){      
    for(unsigned int  jrbm=irbm+1;jrbm<nRBMode;jrbm++){
      {
        bool is_explored = false;
        for(std::set<CRayCollide>::iterator itr=setRayCollide.begin();itr!=setRayCollide.end();itr++){
          if( (itr->imo0 == irbm || itr->imo0 == jrbm) && itr->imo1 == -1 ){
            is_explored = true;
            break;
          }
        }
        if( is_explored ) continue;
      }
      std::vector<double> B; B.resize(aFix.size()*2);
      for(unsigned int ifix=0;ifix<aFix.size();ifix++){
        B[ifix*2+0] = A[ifix*(nRBMode)+irbm];
        B[ifix*2+1] = A[ifix*(nRBMode)+jrbm];
      }
      double C[4] = {0,0,0,0};
      double b[2] = {0,0};
      for(unsigned int ifix=0;ifix<aFix.size();ifix++){
        C[0] += B[ifix*2+0]*B[ifix*2+0];                  
        C[1] += B[ifix*2+0]*B[ifix*2+1];                          
        C[2] += B[ifix*2+1]*B[ifix*2+0];                          
        C[3] += B[ifix*2+1]*B[ifix*2+1];                                  
        b[0] += -aBend[ifix]*B[ifix*2+0];
        b[1] += -aBend[ifix]*B[ifix*2+1];
      }
      double detC = C[0]*C[3]-C[1]*C[2];
      double invdetC = 1.0/detC;
      double invC[4] = {invdetC*C[3],-invdetC*C[1],-invdetC*C[2],invdetC*C[0]};
      double x[2] = { invC[0]*b[0]+invC[1]*b[1], invC[2]*b[0]+invC[3]*b[1] };
      std::vector<double> D; D.resize(nFix);
      double sqn_d = 0;
      for(unsigned int ifix=0;ifix<nFix;ifix++){
        D[ifix] = B[ifix*2+0]*x[0] + B[ifix*2+1]*x[1] + aBend[ifix];
        sqn_d = D[ifix]*D[ifix];
      }
      if( fabs(x[0]) < fabs(x[1])*0.1 || fabs(x[1]) < fabs(x[0])*0.1 ){
        continue;
      }
      double l = sqrt(x[0]*x[0]+x[1]*x[1]);
      if( l > 1 ){ continue; }
//      std::cout << "sqn " << irbm << " " << jrbm << "   " << sqrt(sqn_d) << " " << x[0] << " " << x[1] << std::endl;
      double r0 = x[0]/l;
      double r1 = x[1]/l;      
      double min, max;
      for(unsigned int iprj=0;iprj<aPrjIndGap.size();iprj++){
        unsigned int ind = aPrjIndGap[iprj].first;
        double gap = aPrjIndGap[iprj].second;
        double grad = A[ind*(nRBMode)+irbm]*r0 + A[ind*(nRBMode)+jrbm]*r1;
        double t = gap/grad;
        if( iprj == 0 ){
          min = t; 
          max = t;
        }
        if( t < min ){ min = t; }
        if( t > max ){ max = t; }
      }
      if( min*max < 0 ){ 
//        std::cout << "out " << min << " " << max << std::endl;
        continue;
      }
      double t;
      if( fabs(min) < fabs(max) ){ t = max; }
      else{ t = min; }
      bool is_out = false;
      for(unsigned int ifix=0;ifix<aFix.size();ifix++){
        double bnd = aBend[ifix] + (A[ifix*(nRBMode)+irbm]*r0 + A[ifix*(nRBMode)+jrbm]*r1)*t;
        if( fabs(bnd) > aBendCap[ifix] ){
//          std::cout << "   out " << ifix << " " << bnd << " " << aBendCap[ifix] << std::endl;
          is_out = true;
        }
      }
      if( is_out ){ continue; }
//      std::cout << "in  " << min << " " << max << " " << t << std::endl;       
      CRayCollide rc;
      rc.t = t;
      rc.r0 = r0;
      rc.r1 = r1;
      rc.imo0 = irbm;
      rc.imo1 = jrbm;
      rc.imo2 = -1;
      setRayCollide.insert(rc);            
      /*
      for(unsigned int iprj=0;iprj<aPrjIndGap.size();iprj++){
        unsigned int ind = aPrjIndGap[iprj].first;        
        double flg = (aBend[ind] > 0 ) ? -1 : 1;
        double c = aBend[ind] -flg*aBendCap[ind];
        double a0 = A[ind*(nRBMode)+irbm];
        double a1 = A[ind*(nRBMode)+jrbm];   
      } 
       */
    }
  }
  for(unsigned int  irbm=0;irbm<nRBMode;irbm++){    
    if( aFix.size() < 3 ) continue;
    for(unsigned int  jrbm=irbm+1;jrbm<nRBMode;jrbm++){
      {
        bool is_explored = false;
        for(std::set<CRayCollide>::iterator itr=setRayCollide.begin();itr!=setRayCollide.end();itr++){
          if( (itr->imo0 == irbm || itr->imo0 == jrbm) && itr->imo1 == -1 ){
            is_explored = true;
            break;
          }
        }
        if( is_explored ) continue;
      }      
      for(unsigned int  krbm=jrbm+1;krbm<nRBMode;krbm++){
      {
        bool is_explored = false;
        for(std::set<CRayCollide>::iterator itr=setRayCollide.begin();itr!=setRayCollide.end();itr++){
          if( itr->imo1 == -1 ){
            int imo0 = itr->imo0;
            if( imo0 == irbm || imo0 == jrbm || imo0 == krbm ){ is_explored = true; break; }
          }
          else if( itr->imo2 == -1 ){
            int imo0 = itr->imo0, imo1 = itr->imo1;            
            if( (imo0 == irbm && imo1 == jrbm ) || 
                (imo0 == jrbm && imo1 == krbm ) || 
                (imo0 == irbm && imo1 == krbm ) ){ is_explored = true; break; } 
          }
        }
        if( is_explored ) continue;
      }
      std::vector<double> B; B.resize(aFix.size()*3);
      for(unsigned int ifix=0;ifix<aFix.size();ifix++){
        B[ifix*3+0] = A[ifix*(nRBMode)+irbm];
        B[ifix*3+1] = A[ifix*(nRBMode)+jrbm];
        B[ifix*3+2] = A[ifix*(nRBMode)+krbm];        
      }
      double C[9] = {0,0,0, 0,0,0, 0,0,0};
      double b[3] = {0,0,0};
      for(unsigned int ifix=0;ifix<aFix.size();ifix++){
        C[0] += B[ifix*3+0]*B[ifix*3+0];
        C[1] += B[ifix*3+0]*B[ifix*3+1];                          
        C[2] += B[ifix*3+0]*B[ifix*3+2]; 
        C[3] += B[ifix*3+1]*B[ifix*3+0];                         
        C[4] += B[ifix*3+1]*B[ifix*3+1];                         
        C[5] += B[ifix*3+1]*B[ifix*3+2];  
        C[6] += B[ifix*3+2]*B[ifix*3+0];                         
        C[7] += B[ifix*3+2]*B[ifix*3+1];                         
        C[8] += B[ifix*3+2]*B[ifix*3+2];          
        b[0] += -aBend[ifix]*B[ifix*3+0];
        b[1] += -aBend[ifix]*B[ifix*3+1];
        b[2] += -aBend[ifix]*B[ifix*3+2];
      }
      int info;
      CalcInvMat(C, 3, info);
      double x[3] = { 
        C[0]*b[0]+C[1]*b[1]+C[2]*b[2], 
        C[3]*b[0]+C[4]*b[1]+C[5]*b[2], 
        C[6]*b[0]+C[7]*b[1]+C[8]*b[2] };
      std::vector<double> D; D.resize(nFix);
      double sqn_d = 0;
      for(unsigned int ifix=0;ifix<nFix;ifix++){
        D[ifix] = B[ifix*3+0]*x[0] + B[ifix*3+1]*x[1] + B[ifix*3+2]*x[2] + aBend[ifix];
        sqn_d = D[ifix]*D[ifix];
      }
      double l = sqrt(x[0]*x[0]+x[1]*x[1]);        
      if( l > 1 ){ continue; }        
      if( fabs(x[0]) < l*0.1 || fabs(x[1]) < l*0.1 || fabs(x[2]) < l*0.1 ){ continue; }
//      std::cout << "sqn " << irbm << " " << jrbm << " " << krbm << " " << sqrt(sqn_d) << " " << x[0] << " " << x[1] << std::endl;
      double r0 = x[0]/l;
      double r1 = x[1]/l;      
      double r2 = x[2]/l;      
      double min, max;
      for(unsigned int iprj=0;iprj<aPrjIndGap.size();iprj++){
        unsigned int ind = aPrjIndGap[iprj].first;
        double gap = aPrjIndGap[iprj].second;
        double grad = A[ind*(nRBMode)+irbm]*r0 + A[ind*(nRBMode)+jrbm]*r1 + A[ind*(nRBMode)+krbm]+r2;
        double t = gap/grad;
        if( iprj == 0 ){
          min = t; 
          max = t;
        }
        if( t < min ){ min = t; }
        if( t > max ){ max = t; }
      }
      if( min*max < 0 ){ 
        //        std::cout << "out " << min << " " << max << std::endl;
        continue;
      }
      double t;
      if( fabs(min) < fabs(max) ){ t = max; }
      else{ t = min; }
      bool is_out = false;
      for(unsigned int ifix=0;ifix<aFix.size();ifix++){
        double bnd = aBend[ifix] + (A[ifix*(nRBMode)+irbm]*r0 + A[ifix*(nRBMode)+jrbm]*r1 + A[ifix*(nRBMode)+krbm]*r2)*t;
        if( fabs(bnd) > aBendCap[ifix] ){
          is_out = true;
        }
      }
      if( is_out ){ continue; }
      //      std::cout << "in  " << min << " " << max << " " << t << std::endl;       
      CRayCollide rc;
      rc.t = t;
      rc.r0 = r0;
      rc.r1 = r1;
      rc.r2 = r2;
      rc.imo0 = irbm;
      rc.imo1 = jrbm;
      rc.imo2 = krbm;        
      setRayCollide.insert(rc);            
      /*
       for(unsigned int iprj=0;iprj<aPrjIndGap.size();iprj++){
       unsigned int ind = aPrjIndGap[iprj].first;        
       double flg = (aBend[ind] > 0 ) ? -1 : 1;
       double c = aBend[ind] -flg*aBendCap[ind];
       double a0 = A[ind*(nRBMode)+irbm];
       double a1 = A[ind*(nRBMode)+jrbm];   
       } 
       */
      }
    }
  }  
  std::cout << "  suggestion candidates" << setRayCollide.size() << std::endl;
  clock_middle = clock();            
  std::cout << "  candidate generaiton " << (double)(clock_middle-clock_start)/CLOCKS_PER_SEC << std::endl;
  
  aDesModeBound.clear();
  const unsigned int max_test = 20;  
  unsigned int itest = 0;
  for(std::set<CRayCollide>::iterator itr=setRayCollide.begin();itr!=setRayCollide.end();itr++){    
    itest++;
    if( itest > max_test ) break;
    CDesModeBound dmb;
    dmb.irb0 = -1; dmb.irb1 = -1; dmb.irb2 = -1;
    dmb.desMode.Initialize(aRI.size(),aFix.size());
    dmb.desMode.AddScaled(aDesSens[itr->imo0],itr->r0);
    {
      unsigned int imo0 = itr->imo0;
      dmb.irb0 = aRBMode[imo0].first;
      dmb.imode0 = aRBMode[imo0].second;
      dmb.r0 = itr->r0;
    }
    if( itr->imo1 >= 0 ){
      unsigned int imo1 = itr->imo1;      
      dmb.desMode.AddScaled(aDesSens[imo1],itr->r1);      
      dmb.irb1 = aRBMode[imo1].first;
      dmb.imode1 = aRBMode[imo1].second;
      dmb.r1 = itr->r1;      
    }
    else{
      dmb.irb1 = -1;
    }
    if( itr->imo2 >= 0 ){
      unsigned int imo2 = itr->imo2;      
      dmb.desMode.AddScaled(aDesSens[imo2],itr->r2);      
      dmb.irb2 = aRBMode[imo2].first;
      dmb.imode2 = aRBMode[imo2].second;
      dmb.r2 = itr->r2;            
    }
    else{
      dmb.irb2 = -1;
    }

    double n = 20;    
    double min0 = itr->t*0.9;
    double max0 = itr->t*2.0;
    if( itr->t > 0 ){ max0 += 0.5; min0 -= 0.3; }
    else { max0 -= 0.5; min0 += 0.3; }
    double min1 = min0, max1 = max0;
    bool stable_flg = false;
    bool is_stable_region = false;
    for(unsigned int i=0;i<=n;i++){    
      double t = min0+(max0-min0)/n*i;
      bool res = IsSoundStructure(t, dmb.desMode, aRI,aFix,aFixDir, 
                                  gravity,contProp,aFP,aExForce,aNailJoint);
      if( res ){ is_stable_region = true; }
      if( res && !stable_flg ){
        min1 = t;
        max1 = t+(max0-min0)*0.0001;
        stable_flg = true;
      }
      else if( res && stable_flg ){
        max1 = t;        
      }
      else if( !res && stable_flg ){
        break;
      }
    }  
    if( !is_stable_region ){
//      std::cout << "there is no stable region " << aDesModeBound.size() << " " << itr->imo0 << " " << itr->imo1 << " " << itr->imo2 << std::endl;
      continue;
    }
//    std::cout << "modes " << itr->t << " " << itr->imo0 << " " << itr->imo1 << " " << itr->imo2 << std::endl;
    if( max1 < min1 ){
      dmb.tmax = min1;
      dmb.tmin = max1;
    }
    else{
      dmb.tmin = min1;
      dmb.tmax = max1;      
    }
    aDesModeBound.push_back(dmb);    
    if( aDesModeBound.size() >= 8 ) break;
  }  
  for(unsigned int imo=0;imo<aDesModeBound.size();imo++){
    for(unsigned int jmo=0;jmo<aDesModeBound.size()-imo-1;jmo++){
      double range1 = aDesModeBound[jmo+0].tmax-aDesModeBound[jmo+0].tmin;
      double range2 = aDesModeBound[jmo+1].tmax-aDesModeBound[jmo+1].tmin; 
      if( range2 > range1 ){
        CDesModeBound dmb = aDesModeBound[jmo];
        aDesModeBound[jmo] = aDesModeBound[jmo+1];
        aDesModeBound[jmo+1] = dmb;
      }
    }
  }
  for(unsigned int imo=0;imo<aDesModeBound.size();imo++){
    double range1 = aDesModeBound[imo+0].tmax-aDesModeBound[imo+0].tmin;
//    double range2 = aDesModeBound[imo+1].tmax-aDesModeBound[imo+1].tmin; 
//    std::cout << imo << " " << range1 << std::endl;
  }
  clock_end = clock();            
  std::cout << "  line search " << (double)(clock_end-clock_middle)/CLOCKS_PER_SEC << std::endl;  
  std::cout << "  # of continuous " << aDesModeBound.size() << " " << aRI.size() << " " << aFix.size() << std::endl;
}



void GetSuggestionModes_TopplingStable
(std::vector<CDesModeBound>& aDesModeBound,
 const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,                                        
 const std::vector<CJoint_Fix>& aFix, 
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir,
 const std::vector<CNailJoint>& aNailJoint,
 const std::vector<CExternalForceOnRigidBody>& aExForce,  
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 std::vector<CContactPoints_RigidBody3D>& aFP)
{
  aDesModeBound.clear();
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };  
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);  
  
  std::vector< std::pair<unsigned int,unsigned int> > aRBMode;
  std::vector< CDesignSensitivity > aDesSens;
  aRBMode.reserve(aRI.size()*8);  
  aDesSens.reserve(aRI.size()*8);  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    int icont_flg = 0;
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x    
    if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x
    if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y        
    if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y
    int ijnt_flg = 0;
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      unsigned int irb0 = aFix[ifix].irb0;
      unsigned int irb1 = aFix[ifix].irb1;      
      unsigned int idir0 = aFixDir[ifix].first;
      unsigned int idir1 = aFixDir[ifix].second;
      if( irb0 != irb && irb1 != irb ) continue;
      unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
      if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
      if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
      if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
      if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y        
    }    
    for(unsigned int imode=0;imode<8;imode++){      
      bool is_alive = IsAliveMode(aRI,aFix,aFixDir,irb,imode,icont_flg,ijnt_flg);
      if( !is_alive ) continue;
      aRBMode.push_back( std::make_pair(irb,imode) );
      CDesignSensitivity desSens;
      GetDesignSensitivity
      (desSens,
       aRI,aFix,aFixDir,                
       irb,imode,icont_flg,contProp);      
      aDesSens.push_back(desSens);
    }
  }
  const unsigned int nRBMode = aDesSens.size();
  std::vector<double> aDepth;
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];        
    const CRigid3D_Def& rd = aRD[irb];        
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);  
      Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);
      const double depth1 = Com::Dot(contProp.cont_norm,contProp.cont_base-v1);
      aDepth.push_back(depth1);
    }
  }  
  std::vector<unsigned int> aPrjInd;
  bool is_prj_need = false;
  for(unsigned int icp=0;icp<aDepth.size();icp++){
//    std::cout << icp << " " << aDepth[icp] << std::endl;
    if( aDepth[icp]<0 ){
      is_prj_need = true;
      aPrjInd.push_back(icp);
    }
  }
  //  std::cout << std::endl;
  if( !is_prj_need ){ 
    std::cout << "no-need suggestion toppling stable" << std::endl;
    return; 
  }
  std::cout << "generate suggestion toppling stable" << std::endl;
  ////
  std::vector<double> A;
  A.resize( nRBMode*aDepth.size() );
//  std::cout << "depthsize " << aDepth.size() << std::endl;
  for(unsigned int irbm=0;irbm<nRBMode;irbm++){
    const CDesignSensitivity& desSens = aDesSens[irbm];
    CSimulationSensitivity simSens;
    GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
    (simSens,
     desSens,
     aRI, aRD, aFix, aJFixForce, aExForce,
     gravity, contProp, aFP);
    unsigned int icp = 0;
    for(unsigned int irb=0;irb<aRI.size();irb++){
      const CRigid3D_Ini& ri = aRI[irb];        
      const CRigid3D_Def& rd = aRD[irb];        
      const CContactPoints_RigidBody3D& aCP = aFP[irb];
      for(unsigned int i=0;i<aCP.Size();i++){
        if( !aCP.aFP[i].is_force_cont ) continue;
        Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);  
        Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);
        Com::CMatrix3  B0 = aCP.GetSensMat_ContPointParam(i);      
        Com::CVector3D sns = GetSensitivityCurrentPointOnRigidBody(desSens.aPlank[irb].p,desSens.aPlank[irb].r,desSens.aPlank[irb].l,
                                                                   simSens.aPlankPR[irb].first, simSens.aPlankPR[irb].second,
                                                                   ri,rd,
                                                                   v0,
                                                                   B0);        
        A[icp*(nRBMode)+irbm] = -Com::Dot(sns,contProp.cont_norm);        
        icp++;
      }
    }
    /*
    std::cout << irbm << " ";
    for(unsigned int icp=0;icp<aDepth.size();icp++){
      std::cout << B[icp*(nRBMode)+irbm] << " ";
    }
    std::cout << std::endl;
     */
  }
  std::set<CRayCollide> setRayCollide;
  for(unsigned int  irbm=0;irbm<nRBMode;irbm++){
    std::cout << "check mode " << aRBMode[irbm].first << " " << aRBMode[irbm].second << std::endl;    
    double min, max;
    for(unsigned int iprj=0;iprj<aPrjInd.size();iprj++){
      unsigned int ind = aPrjInd[iprj];
      double gap = aDepth[ind]; // have some minus value
      double grad = -A[ind*(nRBMode)+irbm];
      double t = gap/grad;
      if( iprj == 0 ){ min = t;  max = t; continue; } 
      if( t < min ){ min = t; }
      if( t > max ){ max = t; }
    }
//    std::cout << irbm << " " << "min max" << min << " " << max << std::endl;
    if( min*max < 0 ){ continue; }
    double t;
    if( fabs(min) < fabs(max) ){ t = max; }
    else{ t = min; }
    if( fabs(t) > 1 ){ continue; }
    bool is_out = false;
    unsigned int icp = 0;
    for(unsigned int irb=0;irb<aRI.size();irb++){
      const CContactPoints_RigidBody3D& aCP = aFP[irb];
      for(unsigned int i=0;i<aCP.Size();i++){
        if( !aCP.aFP[i].is_force_cont ) continue;        
        double depth = aDepth[icp] + A[icp*(nRBMode)+irbm]*t;
        icp++;
        if( depth < 0 ){ is_out = true; break; }
      }
      if( is_out ){ break; }
    }
    if( is_out ){ continue; }
//    std::cout << "in  " << aRBMode[irbm].first << " " << aRBMode[irbm].second << "    " << min << " " << max << " " << t << std::endl; 
    CRayCollide rc;
    rc.t = t;
    rc.r0 = 1;
    rc.imo0 = irbm;
    rc.imo1 = -1;
    setRayCollide.insert(rc);
  }
  unsigned int nCP = aDepth.size();
  for(unsigned int  irbm=0;irbm<nRBMode;irbm++){      
    for(unsigned int  jrbm=irbm+1;jrbm<nRBMode;jrbm++){
      {
        bool is_explored = false;
        for(std::set<CRayCollide>::iterator itr=setRayCollide.begin();itr!=setRayCollide.end();itr++){
          if( (itr->imo0 == irbm || itr->imo0 == jrbm) && itr->imo1 == -1 ){
            is_explored = true;
            break;
          }
        }
        if( is_explored ) continue;
      }
      std::vector<double> B; B.resize(aDepth.size()*2);
      for(unsigned int icp=0;icp<nCP;icp++){
        B[icp*2+0] = A[icp*(nRBMode)+irbm];
        B[icp*2+1] = A[icp*(nRBMode)+jrbm];
      }
      double C[4] = {0,0,0,0};
      double b[2] = {0,0};
      for(unsigned int icp=0;icp<nCP;icp++){
        C[0] += B[icp*2+0]*B[icp*2+0];                  
        C[1] += B[icp*2+0]*B[icp*2+1];                          
        C[2] += B[icp*2+1]*B[icp*2+0];                          
        C[3] += B[icp*2+1]*B[icp*2+1];   
        if( aDepth[icp] < 0 ){
          b[0] += -aDepth[icp]*B[icp*2+0];
          b[1] += -aDepth[icp]*B[icp*2+1];
        }
      }
      double detC = C[0]*C[3]-C[1]*C[2];
      double invdetC = 1.0/detC;
      double invC[4] = {invdetC*C[3],-invdetC*C[1],-invdetC*C[2],invdetC*C[0]};
      double x[2] = { invC[0]*b[0]+invC[1]*b[1], invC[2]*b[0]+invC[3]*b[1] };
      std::vector<double> D; D.resize(nCP);
      double sqn_d = 0;
      for(unsigned int icp=0;icp<nCP;icp++){
        D[icp] = B[icp*2+0]*x[0] + B[icp*2+1]*x[1] + aDepth[icp];
//        std::cout << icp << " " << aDepth[icp] << " " << D[icp] << " " << x[0] << " " << x[1] << std::endl;
      }
      if( fabs(x[0]) < fabs(x[1])*0.1 || fabs(x[1]) < fabs(x[0])*0.1 ){
        continue;
      }
      double l = sqrt(x[0]*x[0]+x[1]*x[1]);
      if( l > 1 ){ continue; }
//      std::cout << "sqn " << irbm << " " << jrbm << "   " << sqrt(sqn_d) << " " << x[0] << " " << x[1] << std::endl;
      double r0 = x[0]/l;
      double r1 = x[1]/l;      
      double min, max;
      for(unsigned int iprj=0;iprj<aPrjInd.size();iprj++){
        unsigned int ind = aPrjInd[iprj];
        double grad = A[ind*(nRBMode)+irbm]*r0 + A[ind*(nRBMode)+jrbm]*r1;
        double t = -aDepth[ind]/grad;
        if( iprj == 0 ){
          min = t; 
          max = t;
          continue;
        }
        if( t < min ){ min = t; }
        if( t > max ){ max = t; }
      }
      if( min*max < 0 ){ 
        //        std::cout << "out " << min << " " << max << std::endl;
        continue;
      }
      double t;
      if( fabs(min) < fabs(max) ){ t = max; }
      else{ t = min; }
      bool is_out = false;
      for(unsigned int ifix=0;ifix<aFix.size();ifix++){
        double depth = aDepth[ifix] + (A[ifix*(nRBMode)+irbm]*r0 + A[ifix*(nRBMode)+jrbm]*r1)*t;
        if( depth < 0 ){
          //          std::cout << "   out " << ifix << " " << bnd << " " << aBendCap[ifix] << std::endl;
          is_out = true;
        }
      }
      if( is_out ){ continue; }
      //      std::cout << "in  " << min << " " << max << " " << t << std::endl;       
      CRayCollide rc;
      rc.t = t;
      rc.r0 = r0;
      rc.r1 = r1;
      rc.imo0 = irbm;
      rc.imo1 = jrbm;
      setRayCollide.insert(rc);            
    }
  }
  aDesModeBound.clear();
  for(std::set<CRayCollide>::iterator itr=setRayCollide.begin();itr!=setRayCollide.end();itr++){
    CDesModeBound dmb;
    dmb.desMode.Initialize(aRI.size(),aFix.size());
    dmb.desMode.AddScaled(aDesSens[itr->imo0],itr->r0);
    {
      unsigned int imo0 = itr->imo0;
      dmb.irb0 = aRBMode[imo0].first;
      dmb.imode0 = aRBMode[imo0].second;
      dmb.r0 = itr->r0;
    }
    if( itr->imo1 >= 0 ){
      dmb.desMode.AddScaled(aDesSens[itr->imo1],itr->r1);      
      unsigned int imo1 = itr->imo1;
      dmb.irb1 = aRBMode[imo1].first;
      dmb.imode1 = aRBMode[imo1].second;
      dmb.r1 = itr->r1;      
    }
    else{
      dmb.irb1 = -1;
    }
    //    std::cout << itr->t << " " << itr->imo0 << " " << itr->imo1 << std::endl;
    double n = 20;
    double min0 = itr->t*0.9;  
    double max0 = itr->t*1.1;    
    if( itr->t > 0 ){ max0 += 1.0; }
    else{ max0 -= 1.0; }
    double min1 = min0, max1 = max0;
    bool stable_flg = false;
    bool is_stable_region = false;
    for(unsigned int i=1;i<=n;i++){    
      double t = min0+(max0-min0)/n*i;
      bool res = IsSoundStructure(t, dmb.desMode, aRI,aFix,aFixDir, 
                                  gravity,contProp,aFP,aExForce,aNailJoint,
                                  false);
//      std::cout << "   " << t << "  " << res << std::endl;
      if( res ){ is_stable_region = true; }
      if( res && !stable_flg ){
        min1 = t;
        max1 = min1 + (max0-min0)*0.00001;
        stable_flg = true;
      }
      else if( res && stable_flg ){
        max1 = t;
      }
      else if( !res && stable_flg ){
        break;
      }
    }  
    if( !is_stable_region ){
      std::cout << "there is no stable region " << aDesModeBound.size() << std::endl;
      continue;
    }
    if( max1 < min1 ){
      dmb.tmax = min1;
      dmb.tmin = max1;
    }
    else{
      dmb.tmin = min1;
      dmb.tmax = max1;      
    }
//    std::cout << "stable " << dmb.tmin << " " << dmb.tmax << " " << itr->t << std::endl;
    aDesModeBound.push_back(dmb);    
  }  
  for(unsigned int imo=0;imo<aDesModeBound.size();imo++){
    for(unsigned int jmo=0;jmo<aDesModeBound.size()-imo-1;jmo++){
      double range1 = aDesModeBound[jmo+0].tmax-aDesModeBound[jmo+0].tmin;
      double range2 = aDesModeBound[jmo+1].tmax-aDesModeBound[jmo+1].tmin; 
      if( range2 > range1 ){
        CDesModeBound dmb = aDesModeBound[jmo];
        aDesModeBound[jmo] = aDesModeBound[jmo+1];
        aDesModeBound[jmo+1] = dmb;
      }
    }
  }
  for(unsigned int imo=0;imo<aDesModeBound.size();imo++){
    double range1 = aDesModeBound[imo+0].tmax-aDesModeBound[imo+0].tmin;
//    double range2 = aDesModeBound[imo+1].tmax-aDesModeBound[imo+1].tmin; 
//    std::cout << "stable mode" << imo << " " << aDesModeBound[imo+0].tmax << " " << aDesModeBound[imo+0].tmin << std::endl;
  }
}

void PrintOutForceSpace
(std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,                                        
 std::vector<CJoint_Fix>& aFix, 
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir,
 std::vector<CNailJoint>& aNailJoint,
 const std::vector<CExternalForceOnRigidBody>& aExForce,  
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 std::vector<CContactPoints_RigidBody3D>& aFP)
{
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };  
  
  EnforceJointConnectivity_ChangeBoxLen(aRI,aFix,aFixDir,false);          
  EnforceContact_ChangeBoxLen(aRI,aFP,contProp.cont_norm,contProp.cont_base);
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);  
  
  std::vector< std::pair<unsigned int,unsigned int> > aRBMode;
  std::vector< CDesignSensitivity > aDesSens;
  aRBMode.reserve(aRI.size()*8);  
  aDesSens.reserve(aRI.size()*8);  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    int icont_flg = 0;
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    if( aCP.aFP[1].is_force_cont && aCP.aFP[2].is_force_cont ){ icont_flg += 1; } // +x    
    if( aCP.aFP[3].is_force_cont && aCP.aFP[0].is_force_cont ){ icont_flg += 2; } // -x
    if( aCP.aFP[2].is_force_cont && aCP.aFP[3].is_force_cont ){ icont_flg += 4; } // +y        
    if( aCP.aFP[0].is_force_cont && aCP.aFP[1].is_force_cont ){ icont_flg += 8; } // -y
    int ijnt_flg = 0;
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      unsigned int irb0 = aFix[ifix].irb0;
      unsigned int irb1 = aFix[ifix].irb1;      
      unsigned int idir0 = aFixDir[ifix].first;
      unsigned int idir1 = aFixDir[ifix].second;
      if( irb0 != irb && irb1 != irb ) continue;
      unsigned int idir2 = ( irb0 == irb ) ? idir0 : idir1;
      if( idir2 == 0 ){ ijnt_flg = ijnt_flg | 1; } // +x
      if( idir2 == 1 ){ ijnt_flg = ijnt_flg | 2; } // -x
      if( idir2 == 2 ){ ijnt_flg = ijnt_flg | 4; } // +y
      if( idir2 == 3 ){ ijnt_flg = ijnt_flg | 8; } // -y        
    }    
    for(unsigned int imode=0;imode<8;imode++){      
      if( imode == 1 ) continue;
      bool is_alive = IsAliveMode(aRI,aFix,aFixDir,irb,imode,icont_flg,ijnt_flg);
      if( !is_alive ) continue;
      aRBMode.push_back( std::make_pair(irb,imode) );
      CDesignSensitivity desSens;
      GetDesignSensitivity
      (desSens,
       aRI,aFix,aFixDir,                
       irb,imode,icont_flg,contProp);      
      aDesSens.push_back(desSens);
    }
  }
  unsigned int nRBMode = aRBMode.size();  
  const unsigned int nFix = aFix.size();  
  std::vector<Com::CVector3D> aQh;
  std::vector<double> aBend;
  std::vector<double> aBendCap;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
    const CRigid3D_Ini& ri0 = aRI[aFix[ifix].irb0];
    const CRigid3D_Ini& ri1 = aRI[aFix[ifix].irb1];
    const CNailJoint& nj = aNailJoint[ifix];
    assert( idir0 == 4 || idir0 == 5 );
    if( idir1!=4 && idir1!=5 ){    
      const Com::CMatrix3& Q1 = ri1.GetRotMatrix_Ini();
      const Com::CVector3D& Qn1 = Q1.MatVec( aFaceNormal[idir1] );      
      const Com::CVector3D& Qh1 = Q1.MatVec( Cross( Com::CVector3D(0,0,1), aFaceNormal[idir1] ) );
      aQh.push_back(Qh1);
      Com::CVector3D lam_r(aJFixForce[ifix].lambda[3],aJFixForce[ifix].lambda[4],aJFixForce[ifix].lambda[5]);
      double pull_max = (nj.lennail-ri0.zlen)*nj.pull_max_perlen;
      Com::CVector3D lam_t(aJFixForce[ifix].lambda[0],aJFixForce[ifix].lambda[1],aJFixForce[ifix].lambda[2]);
      double bend_cap = (pull_max*nj.aPosHead.size()-Dot(Qn1,lam_t))*0.5*ri1.zlen; 
      double bend = Com::Dot(lam_r,Qh1);
      aBend.push_back(bend);
      aBendCap.push_back(bend_cap);      
    }
  }
  
  for(unsigned int ifix=0;ifix<nFix;ifix++){
    std::cout << "bndcap : " << ifix << " " << aBendCap[ifix] << std::endl;
  }
      
  std::vector<double> aDepth0;
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];        
    const CRigid3D_Def& rd = aRD[irb];        
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);  
      Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);
      const double depth1 = Com::Dot(contProp.cont_norm,contProp.cont_base-v1);
      aDepth0.push_back(depth1);
    }
  }    
  std::vector<double> Af;
  Af.resize( nRBMode*aFix.size() );
  assert( aQh.size() == aFix.size() );  
  for(unsigned int irbm=0;irbm<nRBMode;irbm++){
    const CDesignSensitivity& desSens = aDesSens[irbm];
    CSimulationSensitivity simSens;
    GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
    (simSens,
     desSens,
     aRI, aRD, aFix, aJFixForce, aExForce,
     gravity, contProp, aFP);
    //    std::cout << irbm << " ";
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      double dtmp1 = Com::Dot(simSens.aJointLM[ifix].second,aQh[ifix]);
      unsigned int irb1 = aFix[ifix].irb1;
      Com::CVector3D vtmp1 = Cross(desSens.aPlank[irb1].r,aQh[ifix]);
      Com::CVector3D vtmp2(aJFixForce[ifix].lambda[3],aJFixForce[ifix].lambda[4],aJFixForce[ifix].lambda[5]);
      double dtmp2 = Com::Dot(vtmp1,vtmp2);
      Af[ifix*(nRBMode)+irbm] = +dtmp1 + dtmp2;      
      //      std::cout << A[ifix*(nRBMode)+irbm] << " ";
    }
    //    std::cout << std::endl;
  }  
  ///////
  std::vector<double> Ad;
  Ad.resize( nRBMode*aDepth0.size() );
  std::cout << "depthsize " << aDepth0.size() << std::endl;  
  for(unsigned int irbm=0;irbm<nRBMode;irbm++){
    const CDesignSensitivity& desSens = aDesSens[irbm];
    CSimulationSensitivity simSens;
    GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
    (simSens,
     desSens,
     aRI, aRD, aFix, aJFixForce, aExForce,
     gravity, contProp, aFP);
    unsigned int icp = 0;
    for(unsigned int irb=0;irb<aRI.size();irb++){
      const CRigid3D_Ini& ri = aRI[irb];        
      const CRigid3D_Def& rd = aRD[irb];        
      const CContactPoints_RigidBody3D& aCP = aFP[irb];
      for(unsigned int i=0;i<aCP.Size();i++){
        if( !aCP.aFP[i].is_force_cont ) continue;
        Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);  
        Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);
        Com::CMatrix3  B0 = aCP.GetSensMat_ContPointParam(i);      
        Com::CVector3D sns = GetSensitivityCurrentPointOnRigidBody(desSens.aPlank[irb].p,desSens.aPlank[irb].r,desSens.aPlank[irb].l,
                                                                   simSens.aPlankPR[irb].first, simSens.aPlankPR[irb].second,
                                                                   ri,rd,
                                                                   v0,
                                                                   B0);        
        Ad[icp*(nRBMode)+irbm] = -Com::Dot(sns,contProp.cont_norm);        
        icp++;
      }
    }
    /*
     std::cout << irbm << " ";
     for(unsigned int icp=0;icp<aDepth.size();icp++){
     std::cout << B[icp*(nRBMode)+irbm] << " ";
     }
     std::cout << std::endl;
     */
  }  
  unsigned int nCP = aDepth0.size();
  double range = 0.3;
  double ndiv = 20;
  double rdiv = range/ndiv;
  std::vector< std::vector<double> > atJForce; atJForce.resize(nRBMode);
  std::vector< std::vector<double> > atDepth;  atDepth.resize(nRBMode);
  for(unsigned int imo=0;imo<nRBMode;imo++){
    for(unsigned int idiv=0;idiv<=ndiv;idiv++){
      double tmode = rdiv*idiv;
      std::vector<double> aJForce,aDepth;
      bool res = GetJointForceDepth
      (aJForce,
       aDepth,
       tmode,
       aDesSens[imo],aRI,aFix,aFixDir,
       gravity,contProp,aFP,aExForce,aNailJoint);
      if( !res  ){ break; }        
      for(unsigned int i=0;i<aJForce.size();i++){
        atJForce[imo].push_back(aJForce[i]);
      }
      for(unsigned int i=0;i<aJForce.size();i++){
        atJForce[imo].push_back(aBend[i]+Af[i*(nRBMode)+imo]*tmode);
      }
      assert( aDepth.size() == nCP );
      for(unsigned int i=0;i<nCP;i++){
        atDepth[imo].push_back(aDepth[i]);
      }
      for(unsigned int i=0;i<nCP;i++){
        atDepth[imo].push_back(aDepth0[i]+Ad[i*(nRBMode)+imo]*tmode);
      }      
    }
    {
      std::vector<double> tmp = atJForce[imo];
      atJForce[imo].clear();
      unsigned int n = tmp.size()/(nFix*2);      
      for(unsigned int i=0;i<n;i++){
        for(unsigned int ifix=0;ifix<nFix*2;ifix++){
          atJForce[imo].push_back(tmp[(n-i-1)*(nFix*2)+ifix]);
        }
      }
    }
    {
      std::vector<double> tmp = atDepth[imo];
      atDepth[imo].clear();      
      unsigned int n = tmp.size()/(nCP*2); 
      for(unsigned int i=0;i<n;i++){
        for(unsigned int icp=0;icp<nCP*2;icp++){
          atDepth[imo].push_back(tmp[(n-i-1)*(nCP*2)+icp]);
        }
      }      
    }
    for(unsigned int idiv=0;idiv<=ndiv;idiv++){
      double tmode = -rdiv*idiv;
      std::vector<double> aJForce,aDepth;
      bool res = GetJointForceDepth
      (aJForce,
       aDepth,
       tmode,
       aDesSens[imo],aRI,aFix,aFixDir,
       gravity,contProp,aFP,aExForce,aNailJoint);
      if( !res  ){ break; }        
      for(unsigned int i=0;i<aJForce.size();i++){
        atJForce[imo].push_back(aJForce[i]);
      }
      for(unsigned int i=0;i<aJForce.size();i++){
        atJForce[imo].push_back(aBend[i]+Af[i*(nRBMode)+imo]*tmode);
      }      
      for(unsigned int i=0;i<nCP;i++){
        atDepth[imo].push_back(aDepth[i]);
      }
      for(unsigned int i=0;i<nCP;i++){
        atDepth[imo].push_back(aDepth0[i]+Ad[i*(nRBMode)+imo]*tmode);
      }            
    }    
  }
      
  std::cout << "print out stat" << std::endl;
  FILE* fp;
  fp = fopen("force.in","w");
  if( fp == NULL ){
    std::cout << "cannot opn file" << std::endl;
    return;
  }
  for(unsigned int imo=0;imo<nRBMode;imo++){
    fprintf(fp,"# %d %d %d\n",imo,aRBMode[imo].first,aRBMode[imo].second);
    unsigned int n = atJForce[imo].size()/(nFix*2);
    for(unsigned int i=0;i<n;i++){
      for(unsigned int ifix=0;ifix<nFix*2;ifix++){
        fprintf(fp,"%lf ",atJForce[imo][i*(nFix*2)+ifix]);
      }
      fprintf(fp,"\n");      
    }    
    fprintf(fp,"\n");      
    fprintf(fp,"\n");          
  }  
  fclose(fp);
  
  fp = fopen("dforce.in","w");
  if( fp == NULL ){
    std::cout << "cannot opn file" << std::endl;
    return;
  }
  for(unsigned int imo=0;imo<nRBMode;imo++){
    fprintf(fp,"# %d %d %d\n",imo,aRBMode[imo].first,aRBMode[imo].second);
    unsigned int n = atDepth[imo].size()/(nCP*2);
    for(unsigned int i=0;i<n;i++){
      for(unsigned int icp=0;icp<nCP*2;icp++){
        fprintf(fp,"%lf ",atDepth[imo][i*(nCP*2)+icp]*10000);
      }
      fprintf(fp,"\n");      
    }    
    fprintf(fp,"\n");      
    fprintf(fp,"\n");          
  }  
  fclose(fp);
  
}


