/*
 *  rigid_contact.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/6/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

//#include "rigidbody.h"
#include <math.h>

#include "rigid_contact.h"


double GetPenetrationDepth
(const Com::CVector3D& d, const Com::CMatrix3& Q,  
 const Com::CVector3D& u, const Com::CMatrix3& P,    
 //// 
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm)
{
  Com::CVector3D v1 = u + d + P.MatVec( Q.MatVec(v0) );
  const double h = -Com::Dot(v1-cont_base,cont_norm);
  return h;  
}

double GetCurContactEnergy
(const Com::CVector3D& d, const Com::CMatrix3& Q,  
 const Com::CVector3D& u, const Com::CMatrix3& P,    
// const CRigidBody3D& rb,
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff)
{
  const Com::CVector3D& v1 = u + d + P.MatVec( Q.MatVec(v0) );
  const double h = Dot(cont_base-v1,cont_norm);
  return 0.5*h*h*cont_stiff;
}

void GetDerCur_CurContactEnergy
(Com::CVector3D& dce_du, Com::CVector3D& dce_dr,
// const CRigidBody3D& rb,
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,   
 const Com::CVector3D& v0,
 const Com::CVector3D& b, const Com::CVector3D& n, double k)
{
  const Com::CVector3D& v1 = d + u + P.MatVec( Q.MatVec(v0) );
//  const Com::CMatrix3& P = rb.GetRotMatrix_Inc();  
  const double h = Dot(b-v1,n);
  Com::CVector3D Qa = Q.MatVec(v0);
  Com::CVector3D Ptn = P.MatVecTrans(n);
  Com::CVector3D wQaPtn = Com::CMatrix3(Qa).MatVec(Ptn);  
  dce_du = -h*k*n;
  dce_dr = -h*k*wQaPtn;
}

void GetDerIni_IniContactEnergy
(Com::CVector3D& dce_dd, Com::CVector3D& dce_ds, Com::CVector3D& dce_dl,
// const CRigidBody3D& rb,
 const Com::CVector3D& d,
 const Com::CMatrix3& Q, 
 const Com::CVector3D& a,
 const Com::CVector3D& b, const Com::CVector3D& n, double k)
{
//  const Com::CMatrix3& Q = rb.GetRotMatrix_Ini();
  const Com::CVector3D& Qa = Q.MatVec(a);
//  const Com::CVector3D& d = rb.GetIniPosCG();
  double h = Com::Dot(d+Q.MatVec(a)-b,n);
  dce_dd = (k*h)*n;
  dce_ds = (k*h)*Com::CMatrix3(Qa).MatVec(n);
  dce_dl = (k*h)*Q.MatVecTrans(n);
}

void GetDerIniIni_IniContactEnergy
(Com::CMatrix3& dce_dddd, Com::CMatrix3& dce_ddds, Com::CMatrix3& dce_ddda,
 Com::CMatrix3& dce_dsdd, Com::CMatrix3& dce_dsds, Com::CMatrix3& dce_dsda, 
 Com::CMatrix3& dce_dadd, Com::CMatrix3& dce_dads, Com::CMatrix3& dce_dada, 
// const CRigidBody3D& rb,
 const Com::CVector3D& d,
 const Com::CMatrix3& Q,
 const Com::CVector3D& a,
 const Com::CVector3D& b, const Com::CVector3D& n, double k)
{
//  const Com::CMatrix3& Q = rb.GetRotMatrix_Ini();
  const Com::CVector3D& Qa = Q.MatVec(a);
//  const Com::CVector3D& d = rb.GetIniPosCG();
  double h = Com::Dot(d+Q.MatVec(a)-b,n);
  const Com::CVector3D& dhdd = n;
  const Com::CVector3D& dhds = Com::CMatrix3(Qa).MatVec(n);
  const Com::CVector3D& dhda = Q.MatVecTrans(n);
  dce_dddd = k*Com::CMatrix3(dhdd,dhdd);
  dce_ddds = k*Com::CMatrix3(dhdd,dhds);
  dce_ddda = k*Com::CMatrix3(dhdd,dhda);
  dce_dsdd = k*Com::CMatrix3(dhds,dhdd);
  dce_dsds = k*Com::CMatrix3(dhds,dhds) + (k*h)*Com::CMatrix3(n).MatMat( Com::CMatrix3(Qa) );
  dce_dsda = k*Com::CMatrix3(dhds,dhda) - (k*h)*Com::CMatrix3(n).MatMat(Q);
  dce_dadd = k*Com::CMatrix3(dhda,dhdd);
  dce_dads = k*Com::CMatrix3(dhda,dhds) + (k*h)*Q.MatMatTrans( Com::CMatrix3(n) );
  dce_dada = k*Com::CMatrix3(dhda,dhda);
}



void GetDerCurCur_CurContactEnergy
(Com::CMatrix3& dce_dudu, Com::CMatrix3& dce_dudr, Com::CMatrix3& dce_drdr,
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,    
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& n, double cont_stiff)
{
  const Com::CVector3D& v1 = d + u + P.MatVec( Q.MatVec(v0) );
//  const Com::CMatrix3& P = rb.GetRotMatrix_Inc(); 
  const double h = Com::Dot(cont_base-v1,n);
  Com::CVector3D Qa = Q.MatVec(v0);
  Com::CVector3D Ptn = P.MatVecTrans(n);
  Com::CVector3D dhdr = Com::CMatrix3(Ptn).MatVec(Qa);  
  Com::CMatrix3 m3 = -cont_stiff*h*Com::CMatrix3(Qa).MatMat( Com::CMatrix3(Ptn) );
  Com::CMatrix3 m4; m4.SetOuterProduct(dhdr,dhdr); m4 *= cont_stiff;
  
  dce_dudu = Com::CMatrix3(n,n); dce_dudu*=cont_stiff;
  dce_drdr = m3+m4;  
  dce_dudr = Com::CMatrix3(n,dhdr); dce_dudr *= -cont_stiff;
}

void GetDerCurIni_CurContactEnergy
(Com::CMatrix3& dce_dudd, Com::CMatrix3& dce_drdd,
 Com::CMatrix3& dce_duds, Com::CMatrix3& dce_drds,
 Com::CMatrix3& dce_dudl, Com::CMatrix3& dce_drdl, 
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,     
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff)
{
  const Com::CMatrix3 PQ = P.MatMat(Q);
  const Com::CVector3D v1 = u + d + PQ.MatVec(v0);
  const double h = Dot(cont_norm, cont_base-v1);//GetPenetrationDepth(rb,v0,cont_base,cont_norm);  
  Com::CVector3D Qv = Q.MatVec(v0);
  Com::CVector3D Ptn = P.MatVecTrans(cont_norm);
  Com::CMatrix3  Qwa = Q.MatMat( Com::CMatrix3(v0) );
  Com::CVector3D dhdr = Com::CMatrix3(Ptn).MatVec(Qv);  
  Com::CVector3D Rtn = PQ.MatVecTrans(cont_norm);
  Com::CMatrix3 tmp1 = h*Q-Com::CMatrix3(Qv,Rtn);
  ////
  dce_dudd = cont_stiff*Com::CMatrix3(cont_norm,cont_norm);
  dce_drdd = cont_stiff*Com::CMatrix3(dhdr,-cont_norm);
  ////
  dce_duds = cont_stiff*Com::CMatrix3(-cont_norm,dhdr);
  dce_drds = cont_stiff*Com::CMatrix3(dhdr,dhdr)-cont_stiff*h*Com::CMatrix3(Ptn).MatMat( Com::CMatrix3(Qv) );
  ////
  dce_dudl = cont_stiff*Com::CMatrix3(cont_norm,Rtn);
  dce_drdl = cont_stiff*Com::CMatrix3(Ptn).MatMat(tmp1);  
}

/*
void GetSensitivityLength_1stDerContactEnergy
(Com::CMatrix3& dce_dudl, Com::CMatrix3& dce_drdl, 
 const CRigidBody3D& rb,
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff)
{
  const double h = GetPenetrationDepth(rb,v0,cont_base,cont_norm);  
  Com::CVector3D Qv = rb.GetRotMatrix_Ini().MatVec(v0);
  Com::CVector3D Ptn = rb.GetRotMatrix_Inc().MatVecTrans(cont_norm);
  Com::CVector3D dhdr = Com::CMatrix3(Ptn).MatVec(Qv);  
}*/


// Contact Test for CRigidBody3D

/////////////////////////


void GetKineticFrictionForce
(Com::CVector3D& fv, Com::CVector3D& fr,
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,     
 const Com::CVector3D& v, const Com::CVector3D& O,
// const CRigidBody3D& rb,
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, double myu_k)
{
//  const Com::CVector3D& uv = rb.GetCurrentVelocity(v0);
//  {
//    const Com::CVector3D& Qa = this->GetRotMatrix_Ini().MatVec(vec);
//  }
  const Com::CVector3D& uv = v + P.MatVec( Cross(O,Q.MatVec(v0)) );      
  
  Com::CVector3D n0 = cont_norm;
  const double pd = GetPenetrationDepth(d,Q, u,P, v0,cont_base,cont_norm);
  
  Com::CVector3D velo_t = uv - Com::Dot(uv,n0)*n0;
  const double force_n = pd*stiff_n;
  
  Com::CVector3D n_velo_t = velo_t; n_velo_t.SetNormalizedVector(); // normal tangent velocity
  fv = -(force_n*myu_k)*n_velo_t;
  Com::CVector3D Qv = Q.MatVec(v0);
  Com::CMatrix3  wQvPt = Com::CMatrix3(Qv).MatMat( P.Trans() );
  fr = wQvPt.MatVec(fv); 
}

void GetJacobian_KineticFrictionForce
(Com::CMatrix3& dfv_dv, Com::CMatrix3& dfv_do, Com::CMatrix3& dfr_dv, Com::CMatrix3& dfr_do,
 Com::CMatrix3& dfv_du, Com::CMatrix3& dfv_dr, Com::CMatrix3& dfr_du, Com::CMatrix3& dfr_dr,
// const CRigidBody3D& rb,
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,     
 const Com::CVector3D& v, const Com::CVector3D& O, 
 ////
 const Com::CVector3D& a,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, double myu_k)
{
//  const Com::CVector3D& uv = rb.GetCurrentVelocity(a);
  const Com::CVector3D& uv = v + P.MatVec( Cross(O,Q.MatVec(a)) );        
  
  Com::CVector3D n0 = cont_norm;
  const double pd = GetPenetrationDepth(d,Q, u,P, a,cont_base,cont_norm);
  
  const Com::CVector3D velo_t = uv - Com::Dot(uv,n0)*n0;
  const double force_n = pd*stiff_n;
  
  const double len_velo_t = velo_t.Length();
  const Com::CVector3D n_velo_t = velo_t / len_velo_t; // normal tangent velocity
  const Com::CVector3D Qa = Q.MatVec(a);  
  const Com::CMatrix3  wQaPt = Com::CMatrix3(Qa).MatMat( P.Trans() );
  const Com::CVector3D wQaPtn = wQaPt.MatVec(n0);
  const Com::CVector3D Ptvnt = P.MatVecTrans(n_velo_t);
  const Com::CMatrix3 dVdr = -P.MatMat( Com::CMatrix3( Com::Cross(O,Qa) ) );
  const Com::CMatrix3 dVdo = -P.MatMat( Com::CMatrix3(Qa) );  
  
  dfv_dv = (force_n*myu_k/len_velo_t)*(Com::CMatrix3(1)-Com::CMatrix3(n_velo_t,n_velo_t));
  dfv_dv = -dfv_dv.MatMat( Com::CMatrix3(1)-Com::CMatrix3(n0,n0) ); 
  
  dfr_dv = wQaPt.MatMat(dfv_dv);
  
  dfv_do = dfv_dv.MatMat( dVdo );
  dfr_do = wQaPt.MatMat(dfv_do); 
  
  ////
  
  dfv_du = stiff_n*myu_k*Com::CMatrix3(n_velo_t,n0);
  dfr_du = wQaPt.MatMat(dfv_du);
  
  dfv_dr = dfv_du.MatMat(wQaPt.Trans()) + dfv_dv.MatMat(dVdr); 
  dfr_dr = wQaPt.MatMat(dfv_dr) - stiff_n*myu_k*pd*Com::CMatrix3(Qa).MatMat( Com::CMatrix3(Ptvnt) );
}




/////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////

double GetStaticFrictionEnergy
(const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,    
 ////
// const CRigidBody3D& rb,
 const Com::CVector3D& v0,
 const Com::CVector3D& aloc,
 const Com::CVector3D& n0, 
 double stiff_f)
{
  const Com::CVector3D& v1 = u + d + P.MatVec( Q.MatVec(v0) );
  Com::CVector3D ap_t = v1-aloc;  // relative position from anchor
  ap_t -= n0*Com::Dot(n0,ap_t);  // tangent vector from anchor to point
  return 0.5*stiff_f*SquareLength(ap_t);
}

void GetStaticFrictionForce
(Com::CVector3D& fv, Com::CVector3D& fr,
 //// 
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,     
 //// 
// const CRigidBody3D& rb,
 const Com::CVector3D& v0,
 const Com::CVector3D& aloc,
 const Com::CVector3D& n0, 
 double stiff_f)
{
  const Com::CVector3D& v1 = u + d + P.MatVec( Q.MatVec(v0) );
  Com::CVector3D ap_t = v1-aloc;  // relative position from anchor
  ap_t -= n0*Com::Dot(n0,ap_t);  // tangent vector from anchor to point
  const Com::CVector3D Qa = Q.MatVec(v0);  
//  const Com::CMatrix3& P = rb.GetRotMatrix_Inc();
  const Com::CMatrix3  wQaPt = Com::CMatrix3(Qa).MatMat( P.Trans() );
  fv = -stiff_f*ap_t;// calc a spring force between anchor
  fr = wQaPt.MatVec(fv);
}

void GetJacobian_StaticFrictionForce
(Com::CMatrix3& dfv_du, Com::CMatrix3& dfv_dr, Com::CMatrix3& dfr_du, Com::CMatrix3& dfr_dr,
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,                                      
 ////
 const Com::CVector3D& v0,
 const Com::CVector3D& aloc,
 const Com::CVector3D& n0, 
 double stiff_f)
{
  const Com::CVector3D v1 = u + d + P.MatVec( Q.MatVec(v0) );
//  const Com::CMatrix3 P = rb.GetRotMatrix_Inc();
  const Com::CVector3D Qa = Q.MatVec(v0);    
  const Com::CMatrix3 PwQa = P.MatMat( Com::CMatrix3(Qa) );
  const Com::CMatrix3  wQaPt = Com::CMatrix3(Qa).MatMat( P.Trans() );
  Com::CVector3D ap_t = v1-aloc;  // relative position from anchor  
  ap_t -= n0*Com::Dot(n0,ap_t);  // tangent vector from anchor to pointw
  Com::CVector3D Ptapt = P.MatVecTrans(ap_t);
  Com::CMatrix3 m3 = -stiff_f*Com::CMatrix3(Qa).MatMat(Ptapt);
  
  dfv_du = -stiff_f*(Com::CMatrix3(1)-Com::CMatrix3(n0,n0));
  //  dfv_du = -stiff_f*Com::CMatrix3(1);  
  dfv_dr = -dfv_du.MatMat(PwQa);
  dfr_du = wQaPt.MatMat(dfv_du);
  dfr_dr = wQaPt.MatMat(dfv_dr) + m3;
}








void GetSensitivity_StaticFrictionForce_FixAnchor
(Com::CMatrix3& dfv_dd, Com::CMatrix3& dfr_dd, 
 Com::CMatrix3& dfv_ds, Com::CMatrix3& dfr_ds,
 Com::CMatrix3& dfv_dl, Com::CMatrix3& dfr_dl,
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,     
 //// 
 const Com::CVector3D& a,
 const Com::CVector3D& aloc,
 const Com::CVector3D& n0, 
 double stiff_f)
{
  const Com::CVector3D Qa = Q.MatVec(a);    
  const Com::CMatrix3 PwQa = P.MatMat( Com::CMatrix3(Qa) );
  const Com::CMatrix3  wQaPt = Com::CMatrix3(Qa).MatMat( P.Trans() );
  const Com::CVector3D& co = u + d + P.MatVec( Q.MatVec(a) );
  Com::CVector3D ap_t = co-aloc;  // relative position from anchor
  ap_t -= n0*Com::Dot(n0,ap_t);  // tangent vector from anchor to pointw
  Com::CVector3D Ptapt = P.MatVecTrans(ap_t);
//  Com::CVector3D Rtuda = rb.GetRotMatrix().MatVecTrans( rb.GetDispCG() + rb.GetIniPosCG() - aloc );
  Com::CMatrix3 R = P.MatMat(Q);
  Com::CVector3D Rtapt = R.MatVecTrans(ap_t);
  Com::CMatrix3 RtNR = R.MatMatTrans( (Com::CMatrix3(1)-Com::CMatrix3(n0,n0)).MatMat(R) );
  
  dfv_dd = -stiff_f*(Com::CMatrix3(1)-Com::CMatrix3(n0,n0));
//  dfv_dd = -stiff_f*Com::CMatrix3(1);  
  dfr_dd = wQaPt.MatMat(dfv_dd);
  ////
  dfv_ds = -dfv_dd.MatMat(PwQa);  
  dfr_ds = wQaPt.MatMat(dfv_ds) - stiff_f*Com::CMatrix3(Ptapt).MatMat(Com::CMatrix3(Qa));
  ////
  dfv_dl = dfv_dd.MatMat( R );
  dfr_dl = stiff_f*Q.MatMat( Com::CMatrix3(Rtapt) - Com::CMatrix3(a).MatMat(RtNR) );
}

// predict how the frictional force changes when initial box (pos,rot,relpos) config was changed
// the anchor moves in the initial config
void GetSensitivity_StaticFrictionForce_InitialAnchor
(Com::CMatrix3& dfv_dd, Com::CMatrix3& dfr_dd, 
 Com::CMatrix3& dfv_ds, Com::CMatrix3& dfr_ds,
 Com::CMatrix3& dfv_da, Com::CMatrix3& dfr_da,
 //// 
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,     
 ////
 const Com::CVector3D& a,
 const Com::CVector3D& n0, 
 double stiff_f)
{
  const Com::CVector3D Qa = Q.MatVec(a);    
  const Com::CMatrix3 PI = P-Com::CMatrix3(1);
  const Com::CMatrix3 PIwQa = PI.MatMat( Com::CMatrix3(Qa) );
  Com::CMatrix3 N = Com::CMatrix3(1)-Com::CMatrix3(n0,n0);  
  Com::CVector3D s = N.MatVec( PI.MatVec(Qa) + u );
  Com::CVector3D Pts = P.MatVecTrans(s);
  const Com::CMatrix3  wQaPt = Com::CMatrix3(Qa).MatMat( P.Trans() );  
  /*
  const Com::CVector3D& co = rb.GetCurrentPosition(a);    
  const Com::CVector3D& aloc = rb.GetInitialPosition(a);
  Com::CVector3D ap_t = co-aloc;  // relative position from anchor
  ap_t -= n0*Com::Dot(n0,ap_t);  // tangent vector from anchor to pointw
  Com::CVector3D Ptapt = rb.GetRotMatrix_Inc().MatVecTrans(ap_t);
  //  Com::CVector3D Rtuda = rb.GetRotMatrix().MatVecTrans( rb.GetDispCG() + rb.GetIniPosCG() - aloc );
  Com::CMatrix3 R = rb.GetRotMatrix();
  Com::CVector3D Rtapt = R.MatVecTrans(ap_t);
  Com::CMatrix3 RtNR = R.MatMatTrans( N.MatMat(R) );
   */    
  dfv_dd = Com::CMatrix3(0.0);
  dfr_dd = Com::CMatrix3(0.0);
  ////
  dfv_ds = stiff_f*N.MatMat(PIwQa);  
  dfr_ds = wQaPt.MatMat(dfv_ds) - stiff_f*Com::CMatrix3(Pts).MatMat(Com::CMatrix3(Qa));
  ////
  dfv_da = -stiff_f*N.MatMat( PI.MatMat(Q) );
  dfr_da = wQaPt.MatMat(dfv_da) + stiff_f*Com::CMatrix3(Pts).MatMat(Q);
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////

