/*
 *  rigid_contact.h
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/6/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if !defined(RIGID_CONTACT_H)
#define RIGID_CONTACT_H

// don't include linear_system class here
// don't include corner point of box here

//#include "rigidbody.h"
#include "delfem/matrix3d.h"
#include "delfem/vector3d.h"

double GetPenetrationDepth
(//const CRigidBody3D& rb,
 const Com::CVector3D& d, const Com::CMatrix3& Q,  
 const Com::CVector3D& u, const Com::CMatrix3& P,    
 ///// 
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm);

////////////////////////////////////////////////////////////////////////////////////////////////////

double GetCurContactEnergy
(const Com::CVector3D& d, const Com::CMatrix3& Q,  
 const Com::CVector3D& u, const Com::CMatrix3& P,    
 //// 
// const CRigidBody3D& rb,
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff);

void GetDerCur_CurContactEnergy
(Com::CVector3D& dce_du, Com::CVector3D& dce_dr,
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q,  
 const Com::CVector3D& u, const Com::CMatrix3& P,    
 ////
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff);

void GetDerIni_IniContactEnergy
(Com::CVector3D& dce_dd, Com::CVector3D& dce_ds, Com::CVector3D& dce_dl,
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 ////
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff);

void GetDerIniIni_IniContactEnergy
(Com::CMatrix3& dce_dddd, Com::CMatrix3& dce_ddds, Com::CMatrix3& dce_dddl,
 Com::CMatrix3& dce_dsdd, Com::CMatrix3& dce_dsds, Com::CMatrix3& dce_dsdl, 
 Com::CMatrix3& dce_dldd, Com::CMatrix3& dce_dlds, Com::CMatrix3& dce_dldl, 
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 ////
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff);

void GetDerCurCur_CurContactEnergy
(Com::CMatrix3& dce_dudu, Com::CMatrix3& dce_dudr, Com::CMatrix3& dce_drdr,
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,    
 ////
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff);

void GetDerCurIni_CurContactEnergy
(Com::CMatrix3& dce_dudd, Com::CMatrix3& dce_drdd,
 Com::CMatrix3& dce_duds, Com::CMatrix3& dce_drds, 
 Com::CMatrix3& dce_dudl, Com::CMatrix3& dce_drdl,  
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,    
 ////
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff);

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

// predict how the frictional force changes when initial box (pos,rot,relpos) config was changed
void GetSensitivity_StaticFrictionForce_FixAnchor
(Com::CMatrix3& dfv_dd, Com::CMatrix3& dfr_dd, 
 Com::CMatrix3& dfv_ds, Com::CMatrix3& dfr_ds,
 Com::CMatrix3& dfv_da, Com::CMatrix3& dfr_da,
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,     
 ////
// const CRigidBody3D& rb,
 const Com::CVector3D& a,
 const Com::CVector3D& aloc,
 const Com::CVector3D& n0, 
 double stiff_f);

// predict how the frictional force changes when initial box (pos,rot,relpos) config was changed
// the anchor moves in the initial config
void GetSensitivity_StaticFrictionForce_InitialAnchor
(Com::CMatrix3& dfv_dd, Com::CMatrix3& dfr_dd, 
 Com::CMatrix3& dfv_ds, Com::CMatrix3& dfr_ds,
 Com::CMatrix3& dfv_da, Com::CMatrix3& dfr_da,
 ///
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,     
 ///
 const Com::CVector3D& a,
 const Com::CVector3D& n0, 
 double stiff_f);



//////

////////////////////////

void GetJacobian_KineticFrictionForce
(Com::CMatrix3& dfv_dv, Com::CMatrix3& dfv_do, Com::CMatrix3& dfr_dv, Com::CMatrix3& dfr_do,
 Com::CMatrix3& dfv_du, Com::CMatrix3& dfv_dr, Com::CMatrix3& dfr_du, Com::CMatrix3& dfr_dr,
// const CRigidBody3D& rb,
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,
 const Com::CVector3D& v, const Com::CVector3D& O,  
 ////
 const Com::CVector3D& a,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, double myu_k);

void GetKineticFrictionForce
(Com::CVector3D& fv, Com::CVector3D& fr,
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,
 const Com::CVector3D& v, const Com::CVector3D& O,  
// const CRigidBody3D& rb,
 const Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, double myu_k);

////////////////////////

double GetStaticFrictionEnergy
(
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,
 ////
 const Com::CVector3D& a,
 const Com::CVector3D& aloc,
 const Com::CVector3D& n0,
 double stiff_f);

void GetStaticFrictionForce
(Com::CVector3D& fv, Com::CVector3D& fr,
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,
 /////
 const Com::CVector3D& a,
 const Com::CVector3D& aloc,
 const Com::CVector3D& cont_norm, 
 double stiff_f);

// used for solving static dynamics
void GetJacobian_StaticFrictionForce
(Com::CMatrix3& dfv_du, Com::CMatrix3& dfv_dr, Com::CMatrix3& dfr_du, Com::CMatrix3& dfr_dr,
 ////
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,                                     
 ////
 const Com::CVector3D& a, 
 const Com::CVector3D& aloc,
 const Com::CVector3D& cont_norm, 
 double stiff_f);

////////////////////////

////////////////////////

#endif