/*
 *  rigid_contact_ls.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/24/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include "rigid_contact_ls.h"
#include "rigid_contact.h"
#include "furniture_utility.h"


void AddLinSys_FrictionalContact_Penalty_NonStatic_BackwardEular
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 /////
 unsigned int irb, //const CRigidBody3D& rb, 
 const CRigid3D_Ini& ri, 
 const CRigid3D_Def& rd,    
 Com::CVector3D& a, double dt,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, double myu_s, double myu_k,
 CFrictionPoint& fp )
{  
  const Com::CVector3D& uv = GetCurrentVelocity(a,ri,rd);
  const Com::CVector3D& co = GetCurrentPosition(a,ri,rd);  
  
  Com::CVector3D n0 = cont_norm;
//  const CRigid3D_Def& rd = rb.rd;      
//  const CRigid3D_Ini& ri = rb.ri;        
  const double pd = GetPenetrationDepth(ri.GetIniPosCG(),ri.GetRotMatrix_Ini(),
                                        rd.GetDispCG(),  rd.GetRotMatrix_Inc(),
                                        a,cont_base,cont_norm);
  fp.pd = pd;
  
  Com::CMatrix3 dxdr = -rd.GetRotMatrix_Inc().MatMat( Com::CMatrix3( ri.GetRotMatrix_Ini().MatVec(a) ) );
  
  if( pd < 0 ){
    fp.itype_contact = 0;
    return;
  }
  
  {
    Com::CVector3D dce_du, dce_dr;
    GetDerCur_CurContactEnergy(dce_du,dce_dr,
                               ri.GetIniPosCG(), ri.GetRotMatrix_Ini(), 
                               rd.GetDispCG(), rd.GetRotMatrix_Inc(),
                               a,cont_base,cont_norm,stiff_n);
    Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
    GetDerCurCur_CurContactEnergy(dce_dudu, dce_dudr, dce_drdr, 
                                  ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                  rd.GetDispCG(),   rd.GetRotMatrix_Inc(), 
                                  a, cont_base, cont_norm, stiff_n);
    {
      AddMatrix(mat, irb,0,  irb,0,     dce_dudu, dt*dt, true  );
      AddMatrix(mat, irb,3,  irb,0,     dce_dudr, dt*dt, false );
      AddMatrix(mat, irb,0,  irb,3,     dce_dudr, dt*dt, true  );
      AddMatrix(mat, irb,3,  irb,3,     dce_drdr, dt*dt, true  );          
    }
    {
      Com::CVector3D ru = dce_du;
      ru += dt*dce_dudu.MatVec(rd.GetVeloCG());
      ru += dt*dce_dudr.MatVec(rd.GetOmega());
      AddVector(r, irb,0,ru,-dt);
    }
    {
      Com::CVector3D rr = dce_dr;
      rr += dt*dce_dudr.MatVecTrans(rd.GetVeloCG());
      rr += dt*dce_drdr.MatVec(rd.GetOmega());
      AddVector(r, irb,3,rr,-dt);              
    }
  }
  
  // friction handling
  Com::CVector3D ap_t = co-fp.aloc;  // relative position from anchor
  ap_t -= n0*Com::Dot(n0,ap_t);  // tangent vector from anchor to point
  
  Com::CVector3D velo_t = uv - Com::Dot(uv,n0)*n0;
  const double len_ap_t = Com::Length(ap_t);
  const double len_velo_t = Com::Length(velo_t);
  const double force_f = len_ap_t*stiff_f;
  const double force_n = pd*stiff_n;
  if( force_f < force_n*myu_s && len_velo_t < 1.0e+0 ){
    fp.itype_contact = 1; // static contact
    Com::CVector3D fv, fr;
    GetStaticFrictionForce(fv,fr,
                           ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                           rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                           a,fp.aloc,cont_norm,stiff_f);
    Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
    GetJacobian_StaticFrictionForce(dfv_du,dfv_dr,dfr_du,dfr_dr,
                                    ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                    rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                                    a,fp.aloc,cont_norm,stiff_f);  
    {
      Com::CVector3D ru = fv;
      ru += dt*dfv_du.MatVec(rd.GetVeloCG());
      ru += dt*dfv_dr.MatVec(rd.GetOmega());        
      AddVector(r, irb,0, ru,+dt);
    }
    {
      Com::CVector3D rr = fr;        
      rr += dt*dfr_du.MatVec(rd.GetVeloCG());
      rr += dt*dfr_dr.MatVec(rd.GetOmega());        
      AddVector(r, irb,3, rr,+dt);
    }    
    AddMatrix(mat, irb,0,  irb,0,     dfv_du, -dt*dt, true );
    AddMatrix(mat, irb,0,  irb,3,     dfv_dr, -dt*dt, true );    
    AddMatrix(mat, irb,3,  irb,0,     dfr_du, -dt*dt, true );
    AddMatrix(mat, irb,3,  irb,3,     dfr_dr, -dt*dt, true );                      
  }
  else{
    fp.itype_contact = 2; // dynamic contact
    if( len_velo_t > 1.0e-10 ){
      Com::CVector3D fv, fr;
      GetKineticFrictionForce(fv,fr, 
                              ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                              rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                              rd.GetVeloCG(),   rd.GetOmega(),  
                              a,cont_base,cont_norm,stiff_n,stiff_f,myu_k);
      Com::CMatrix3 dfv_dv, dfv_do, dfr_dv, dfr_do;
      Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
      GetJacobian_KineticFrictionForce(dfv_dv,dfv_do,dfr_dv,dfr_do, 
                                       dfv_du,dfv_dr,dfr_du,dfr_dr,
                                       ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                       rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                                       rd.GetVeloCG(),   rd.GetOmega(),                                                                   
                                       a,cont_base,cont_norm,stiff_n,stiff_f,myu_k);  
      {
        Com::CVector3D ru = fv;
        ru += dt*dfv_du.MatVec(rd.GetVeloCG());
        ru += dt*dfv_dr.MatVec(rd.GetOmega());        
        AddVector(r, irb,0, ru,+dt);
      }
      {
        Com::CVector3D rr = fr;        
        rr += dt*dfr_du.MatVec(rd.GetVeloCG());
        rr += dt*dfr_dr.MatVec(rd.GetOmega());        
        AddVector(r, irb,3, rr,+dt);
      }
      AddMatrix(mat, irb,0,  irb,0,     dfv_dv, -dt, true );
      AddMatrix(mat, irb,3,  irb,0,     dfr_dv, -dt, true );
      AddMatrix(mat, irb,0,  irb,3,     dfv_do, -dt, true );
      AddMatrix(mat, irb,3,  irb,3,     dfr_do, -dt, true );
      AddMatrix(mat, irb,0,  irb,0,     dfv_du, -dt*dt, true );
      AddMatrix(mat, irb,3,  irb,0,     dfr_du, -dt*dt, true );
      AddMatrix(mat, irb,0,  irb,3,     dfv_dr, -dt*dt, true );
      AddMatrix(mat, irb,3,  irb,3,     dfr_dr, -dt*dt, true );                  
    }
  }
}

void AddLinSys_FrictionalContact_Penalty_Static_Contact
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 unsigned int irb, const CRigid3D_Ini& ri, const CRigid3D_Def& rd,    
 Com::CVector3D& a, 
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, 
 const CFrictionPoint& fp)
{
  Com::CVector3D n0 = cont_norm;
  
  Com::CMatrix3 dxdr = -rd.GetRotMatrix_Inc().MatMat( Com::CMatrix3( ri.GetRotMatrix_Ini().MatVec(a) ) );
  
  { // contact part
    Com::CVector3D dce_du, dce_dr;
    GetDerCur_CurContactEnergy(dce_du,dce_dr,
                               ri.GetIniPosCG(), ri.GetRotMatrix_Ini(), 
                               rd.GetDispCG(), rd.GetRotMatrix_Inc(),                               
                               a,cont_base,cont_norm,stiff_n);
    Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
    GetDerCurCur_CurContactEnergy(dce_dudu, dce_dudr, dce_drdr, 
                                  ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                  rd.GetDispCG(),   rd.GetRotMatrix_Inc(),                                   
                                  a, cont_base, cont_norm, stiff_n);
    {
      AddMatrix(mat, irb,0,  irb,0,     dce_dudu, 1, true  );
      AddMatrix(mat, irb,3,  irb,0,     dce_dudr, 1, false );
      AddMatrix(mat, irb,0,  irb,3,     dce_dudr, 1, true  );
      AddMatrix(mat, irb,3,  irb,3,     dce_drdr, 1, true  );          
    }
    {
      Com::CVector3D ru = dce_du;
//      ru += dt*dce_dudu.MatVec(rd.GetVeloCG());
//      ru += dt*dce_dudr.MatVec(rd.GetOmega());
      AddVector(r, irb,0,ru,-1);
    }
    {
      Com::CVector3D rr = dce_dr;
//      rr += dt*dce_dudr.MatVecTrans(rd.GetVeloCG());
//      rr += dt*dce_drdr.MatVec(rd.GetOmega());
      AddVector(r, irb,3,rr,-1);              
    }
  }
  
  { // static frcition
    Com::CVector3D fv, fr;
    GetStaticFrictionForce(fv,fr, 
                           ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                           rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                           a,fp.aloc,cont_norm,stiff_f);
    Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
    GetJacobian_StaticFrictionForce(dfv_du,dfv_dr,dfr_du,dfr_dr,
                                    ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                    rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                                    a,fp.aloc,cont_norm,stiff_f);  
    {
      Com::CVector3D ru = fv;
//      ru += dt*dfv_du.MatVec(rd.GetVeloCG());
//      ru += dt*dfv_dr.MatVec(rd.GetOmega());        
      AddVector(r, irb,0, ru,+1);
    }
    {
      Com::CVector3D rr = fr;
//      rr += dt*dfr_du.MatVec(rd.GetVeloCG());
//      rr += dt*dfr_dr.MatVec(rd.GetOmega());        
      AddVector(r, irb,3, rr,+1);
    }    
    AddMatrix(mat, irb,0,  irb,0,     dfv_du, -1, true );
    AddMatrix(mat, irb,0,  irb,3,     dfv_dr, -1, true );    
    AddMatrix(mat, irb,3,  irb,0,     dfr_du, -1, true );
    AddMatrix(mat, irb,3,  irb,3,     dfr_dr, -1, true );                      
  }  
}

void AddLinSys_FrictionalContact_Penalty_NonStatic_BackwardEular_Contact
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 unsigned int irb,//const CRigidBody3D& rb, 
 const CRigid3D_Ini& ri, 
 const CRigid3D_Def& rd,    
 Com::CVector3D& a, double dt,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, 
 const CFrictionPoint& fp )
{   
  Com::CVector3D n0 = cont_norm;
  
  Com::CMatrix3 dxdr = -rd.GetRotMatrix_Inc().MatMat( Com::CMatrix3( ri.GetRotMatrix_Ini().MatVec(a) ) );
  
  { // contact part
    Com::CVector3D dce_du, dce_dr;
    GetDerCur_CurContactEnergy(dce_du,dce_dr,
                               ri.GetIniPosCG(), ri.GetRotMatrix_Ini(), 
                               rd.GetDispCG(), rd.GetRotMatrix_Inc(),                               
                               a,cont_base,cont_norm,stiff_n);
    Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
    GetDerCurCur_CurContactEnergy(dce_dudu, dce_dudr, dce_drdr, 
                                  ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                  rd.GetDispCG(),   rd.GetRotMatrix_Inc(),                                   
                                  a, cont_base, cont_norm, stiff_n);
    {
      AddMatrix(mat, irb,0,  irb,0,     dce_dudu, dt*dt, true  );
      AddMatrix(mat, irb,3,  irb,0,     dce_dudr, dt*dt, false );
      AddMatrix(mat, irb,0,  irb,3,     dce_dudr, dt*dt, true  );
      AddMatrix(mat, irb,3,  irb,3,     dce_drdr, dt*dt, true  );          
    }
    {
      Com::CVector3D ru = dce_du;
      ru += dt*dce_dudu.MatVec(rd.GetVeloCG());
      ru += dt*dce_dudr.MatVec(rd.GetOmega());
      AddVector(r, irb,0,ru,-dt);
    }
    {
      Com::CVector3D rr = dce_dr;
      rr += dt*dce_dudr.MatVecTrans(rd.GetVeloCG());
      rr += dt*dce_drdr.MatVec(rd.GetOmega());
      AddVector(r, irb,3,rr,-dt);              
    }
  }
  
  { // static frcition
    Com::CVector3D fv, fr;
    GetStaticFrictionForce(fv,fr, 
                           ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                           rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                           a,fp.aloc,cont_norm,stiff_f);
    Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
    GetJacobian_StaticFrictionForce(dfv_du,dfv_dr,dfr_du,dfr_dr,
                                    ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                    rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                                    a,fp.aloc,cont_norm,stiff_f);  
    {
      Com::CVector3D ru = fv;
      ru += dt*dfv_du.MatVec(rd.GetVeloCG());
      ru += dt*dfv_dr.MatVec(rd.GetOmega());        
      AddVector(r, irb,0, ru,+dt);
    }
    {
      Com::CVector3D rr = fr;
      rr += dt*dfr_du.MatVec(rd.GetVeloCG());
      rr += dt*dfr_dr.MatVec(rd.GetOmega());        
      AddVector(r, irb,3, rr,+dt);
    }    
    AddMatrix(mat, irb,0,  irb,0,     dfv_du, -dt*dt, true );
    AddMatrix(mat, irb,0,  irb,3,     dfv_dr, -dt*dt, true );    
    AddMatrix(mat, irb,3,  irb,0,     dfr_du, -dt*dt, true );
    AddMatrix(mat, irb,3,  irb,3,     dfr_dr, -dt*dt, true );                      
  }
}


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////

/*
void AddContact_NewmarkBetaAPrime
(const CRigidBody3D& rb,
 Ls::CLinearSystem_RigidBody& ls, unsigned int irb,
 const double dt, const double newmark_gamma, const double newmark_beta,
 bool is_initial,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double cont_stiff)
{
  if( rb.imode != 1 ){ return; }  
  const double corner[8][3] = { 
    {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
    {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };    
  for(unsigned int i=0;i<8;i++){
    Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
    const double h = GetPenetrationDepth(rb,v0,cont_base,cont_norm);
    if( h < 0 ) continue;      
    Com::CVector3D dce_du, dce_dr;
    Get1stDer_ContactEnergy(dce_du,dce_dr,rb,v0,cont_base,cont_norm,cont_stiff);
    Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
    Get2ndDer_ContactEnergy(dce_dudu, dce_dudr, dce_drdr, rb, v0, cont_base, cont_norm, cont_stiff);
    
    {   // 慣性力から来る回転の残差
      ls.AddResidual(irb,true,0,dce_du,-1);
      ls.AddResidual(irb,true,3,dce_dr,-1);              
    }
    {   // 剛性行列を作る
      const double dtmp1 = dt*dt*newmark_beta;
      ls.AddMatrix(irb,true,0,  irb,true,0,  dce_dudu,dtmp1, true);
      ls.AddMatrix(irb,true,3,  irb,true,0,  dce_dudr,dtmp1, false);        
      ls.AddMatrix(irb,true,0,  irb,true,3,  dce_dudr,dtmp1, true);                
      ls.AddMatrix(irb,true,3,  irb,true,3,  dce_drdr,dtmp1, true);        
    }           
  }  
}
 */
/*
void AddContact_BackwardEular
(const CRigidBody3D& rb,
 Ls::CLinearSystem_RigidBody& ls, unsigned int irb,
 const double dt, 
 bool is_initial,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n)
{
  if( rb.imode != 1 ){ return; }  
  const double corner[8][3] = { 
    {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
    {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };    
  for(unsigned int i=0;i<8;i++){
    Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
    const double h = GetPenetrationDepth(rb,v0,cont_base,cont_norm);
    if( h < 0 ) continue;      
    Com::CVector3D dce_du, dce_dr;
    Get1stDer_ContactEnergy(dce_du,dce_dr,rb,v0,cont_base,cont_norm,stiff_n);
    Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
    Get2ndDer_ContactEnergy(dce_dudu, dce_dudr, dce_drdr, rb, v0, cont_base, cont_norm, stiff_n);
    
    {   // 慣性力から来る回転の残差
      {
        Com::CVector3D ru = dce_du;
        ru += dt*dce_dudu.MatVec(rb.rd.GetVeloCG());
        ru += dt*dce_dudr.MatVec(rb.rd.GetOmega());
        ls.AddResidual(irb,true,0,ru,-dt);
      }
      {
        Com::CVector3D rr = dce_dr;
        rr += dt*dce_dudr.MatVecTrans(rb.rd.GetVeloCG());
        rr += dt*dce_drdr.MatVec(rb.rd.GetOmega());
        ls.AddResidual(irb,true,3,rr,-dt);              
      }
    }
    {   // 剛性行列を作る
      const double dtmp1 = dt*dt;
      ls.AddMatrix(irb,true,0,  irb,true,0,  dce_dudu,dtmp1, true);
      ls.AddMatrix(irb,true,3,  irb,true,0,  dce_dudr,dtmp1, false);        
      ls.AddMatrix(irb,true,0,  irb,true,3,  dce_dudr,dtmp1, true);                
      ls.AddMatrix(irb,true,3,  irb,true,3,  dce_drdr,dtmp1, true);        
    }           
  }  
}
*/


void AddLinearSystem_RigidBody_BackwardEular
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 unsigned int irb, const CRigid3D_Ini& ri, const CRigid3D_Def& rd,     
 const double dt, 
 const Com::CVector3D& gravity, 
 bool is_initial)
{
//  const CRigid3D_Ini& ri = rb.ri;  
  const Com::CVector3D& Omega = rd.GetOmega();
//  const Com::CVector3D& dOmega = rb.GetdOmega();
  const double mass = ri.GetMass();
  const double* mineatia = ri.mineatia;
  AddVector(r, irb,0,  dt*gravity, ri.GetMass() );
	{
    double res[6] = {0,0,0, 0,0,0};
//		res[3] = mineatia[0]*dOmega.x;
//    res[4] = mineatia[1]*dOmega.y;
//    res[5] = mineatia[2]*dOmega.z;
		const double wOmg[9] = { 0,-Omega.z,+Omega.y, +Omega.z,0,-Omega.x,  -Omega.y,+Omega.x, 0 };
    const double jo[3] = { mineatia[0]*Omega.x, mineatia[1]*Omega.y, mineatia[2]*Omega.z };
		for(unsigned int i=0;i<3;i++){
      for(unsigned int j=0;j<3;j++){
        res[i+3] += wOmg[i*3+j]*jo[j];
      }
    }
    AddVector(r, irb,0,6, res,-dt);    
  }
	{ 
    Com::CMatrix3 mtmp;
    mtmp.SetIdentity(mass);
    AddMatrix(mat, irb,0,  irb,0,  mtmp,1, true);
    mtmp.mat[0] = mineatia[0]; mtmp.mat[1] = 0;           mtmp.mat[2] = 0;
    mtmp.mat[3] = 0;           mtmp.mat[4] = mineatia[1]; mtmp.mat[5] = 0;
    mtmp.mat[6] = 0;           mtmp.mat[7] = 0;           mtmp.mat[8] = mineatia[2];
    AddMatrix(mat, irb,3,  irb,3,  mtmp,1, true);
	}
}

void AddLinearSystem_ExternalForce_Static
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 unsigned int irb, const CRigid3D_Ini& ri, const CRigid3D_Def& rd,  
 const Com::CVector3D& pos, const Com::CVector3D& force)
{
  const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
  const Com::CMatrix3& P = rd.GetRotMatrix_Inc();  
  const Com::CVector3D& Qa = Q.MatVec(pos);
  const Com::CMatrix3& wQa = Com::CMatrix3(Qa);
  const Com::CVector3D& Ptf = P.Trans().MatVec(force);
  const Com::CVector3D& fv = force;
  const Com::CVector3D& fr = wQa.MatVec(Ptf);
  AddVector(r, irb,0, fv,1); 
  AddVector(r, irb,3, fr,1);   
  //////
  Com::CMatrix3 wQawPtf = wQa.MatMat( Com::CMatrix3(Ptf) );
  AddMatrix(mat, irb,3,  irb,3, wQawPtf,1, true);      
}

void AddLinearSystem_ExternalForce_BackwardEular
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 unsigned int irb, 
 const CRigid3D_Ini& ri, 
 const CRigid3D_Def& rd,     
 const double dt, 
 bool is_initial,
 const Com::CVector3D& pos, const Com::CVector3D& force)
{
  const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
  const Com::CMatrix3& P = rd.GetRotMatrix_Inc();  
  const Com::CVector3D& Qa = Q.MatVec(pos);
  const Com::CMatrix3& wQa = Com::CMatrix3(Qa);
  const Com::CVector3D& Ptf = P.Trans().MatVec(force);
  const Com::CVector3D& fv = force;
  const Com::CVector3D& fr = wQa.MatVec(Ptf);
  AddVector(r, irb,0, fv,dt); 
  AddVector(r, irb,3, fr,dt);   
  //////
  Com::CMatrix3 wQawPtf = wQa.MatMat( Com::CMatrix3(Ptf) );
  AddMatrix(mat, irb,3,  irb,3, wQawPtf,dt*dt, true);  
}

void AddResidual_ExternalForce_Static
(Com::CVector3D& res_v, Com::CVector3D& res_r,
 const CRigid3D_Ini& ri, 
 const CRigid3D_Def& rd,
 const Com::CVector3D& pos, const Com::CVector3D& force)
{
  const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
  const Com::CMatrix3& P = rd.GetRotMatrix_Inc();  
  const Com::CVector3D& Qa = Q.MatVec(pos);
  const Com::CMatrix3& wQa = Com::CMatrix3(Qa);
  const Com::CVector3D& Ptf = P.Trans().MatVec(force);
  const Com::CVector3D& fv = force;
  const Com::CVector3D& fr = wQa.MatVec(Ptf);
  res_v += fv;
  res_r += fr;
}

void AddMatrix_ExternalForce_Static
(MatVec::CMatDia_BlkCrs& mat,
 unsigned int irb, 
 const CRigid3D_Ini& ri, 
 const CRigid3D_Def& rd, 
 const Com::CVector3D& pos, const Com::CVector3D& force)
{
//  const CRigid3D_Def& rd = rb.rd;        
//  const CRigid3D_Ini& ri = rb.ri;          
  const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
  const Com::CMatrix3& P = rd.GetRotMatrix_Inc();  
  const Com::CVector3D& Qa = Q.MatVec(pos);
  const Com::CMatrix3& wQa = Com::CMatrix3(Qa);
  const Com::CVector3D& Ptf = P.Trans().MatVec(force);  
  Com::CMatrix3 wQawPtf = wQa.MatMat( Com::CMatrix3(Ptf) );
  AddMatrix(mat, irb,3,  irb,3, wQawPtf,1, true);  
}


/*
void AddContact_BackwardEular
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 unsigned int irb, const CRigidBody3D& rb,
 const double dt, 
 bool is_initial,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, double myu_s, double myu_k, 
 CContactPoints_RigidBody3D& aCP )
{
//  if( rb.imode != 1 ){ return; }  
  const double corner[8][3] = { 
    {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
    {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };    
  for(unsigned int i=0;i<aCP.Size();i++){
    Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
    CFrictionPoint& fp = aCP.aFP[i];
    AddLinSys_FrictionalContact_Penalty_NonStatic_BackwardEular
    (mat,r,
     irb, rb,
     v0, dt,
     cont_base,cont_norm,stiff_n,stiff_f,myu_s,myu_k,fp);
  }
}
 */
/*
void AddContact_BackwardEular_IntactContactState
(MatVec::CMatDia_BlkCrs& mat, MatVec::CVector_Blk& r,
 unsigned int irb, const CRigidBody3D& rb,
 const double dt, 
 bool is_initial,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, 
 const CContactPoints_RigidBody3D& aCP )
{
//  if( rb.imode != 1 ){ return; }  
  const double corner[8][3] = { 
    {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
    {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };    
  for(unsigned int i=0;i<aCP.Size();i++){
    Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
    const CFrictionPoint& fp = aCP.aFP[i];
    AddLinSys_FrictionalContact_Penalty_NonStatic_BackwardEular_UntactContact
    (mat,r,
     irb, rb,
     v0, dt,
     cont_base,cont_norm,stiff_n,stiff_f,fp);
  }
}*/


void UpdateSolution_RigidBody_BackwardEular
(CRigidBody3D& rb, 
 const double* upd, 
 double dt,
 bool is_first_iter )
{
  assert( is_first_iter );
  
  const Com::CVector3D vel_update( upd[0], upd[1], upd[2] );
  const Com::CVector3D omg_update( upd[3], upd[4], upd[5] );
  
  CRigid3D_Def& rd = rb.rd;  
  Com::CVector3D& velo_cg = rd.velo_cg;
  Com::CVector3D& disp_cg = rd.disp_cg;  
  Com::CVector3D& Omega = rd.Omega;    
  
  velo_cg += vel_update; // update velocity
  disp_cg += dt*velo_cg; // update displacemnet
  
  ////////////////////////////////////////////////
  Omega += omg_update;  
	{	// 回転のアップデート
    Com::CMatrix3 m; m.SetRotMatrix_Cartesian(dt*Omega);
    rd.AddRotation_IncRight( m );
	}
}
/*
void AddLinearSystem_StaticContact
(unsigned int irb, const CRigidBody3D& rb,
 Ls::CLinearSystem_RigidBody& ls,
 PARAM_TYPE itype_s, int irb_s, unsigned int idim_s,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n, double stiff_f,
 const std::vector<CFrictionPoint>& aFP)
{
  if( rb.imode != 1 ){ return; }  
  const double corner[8][3] = { 
    {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
    {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };    
  for(unsigned int i=0;i<8;i++){
    Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
    const double h = GetPenetrationDepth(rb,v0,cont_base,cont_norm);
    if( h < 0 ) continue;      
    Com::CVector3D dce_du, dce_dr;
    Get1stDer_ContactEnergy(dce_du,dce_dr,rb,v0,cont_base,cont_norm,stiff_n);
    Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
    Get2ndDer_ContactEnergy(dce_dudu, dce_dudr, dce_drdr, rb, v0, cont_base, cont_norm, stiff_n);    
    if( (irb==irb_s) && (itype_s == BOX_POS || itype_s == BOX_ROT || itype_s == BOX_LEN) ){
      Com::CVector3D e_s(0,0,0); e_s[idim_s] = 1.0;
      Com::CMatrix3 dce_dudd, dce_drdd, dce_duds, dce_drds;
      GetSensitivity_1stDerContactEnergy(dce_dudd, dce_drdd, dce_duds, dce_drds, rb, v0, cont_base,cont_norm, stiff_n);          
      if( itype_s == BOX_POS ){
        ls.AddResidual(irb,true,0,dce_dudd.MatVec(e_s),-1);
        ls.AddResidual(irb,true,3,dce_drdd.MatVec(e_s),-1);                                    
      }
      else if( itype_s == BOX_ROT ){
        ls.AddResidual(irb,true,0,dce_duds.MatVec(e_s),-1);
        ls.AddResidual(irb,true,3,dce_drds.MatVec(e_s),-1);                      
      }
      else if( itype_s == BOX_LEN ){
        Com::CVector3D v1(0,0,0);
        v1[idim_s] = corner[i][idim_s]*0.5;
        Com::CMatrix3 dce_dudl, dce_drdl;
        GetSensitivityLength_1stDerContactEnergy(dce_dudl,dce_drdl, rb,v0,cont_base,cont_norm,stiff_n);    
        ls.AddResidual(irb,true,0,dce_dudl.MatVec(v1),-1);
        ls.AddResidual(irb,true,3,dce_drdl.MatVec(v1),-1);
      }      
    }
    {   // 剛性行列を作る
      ls.AddMatrix(irb,true,0,  irb,true,0,  dce_dudu,1, true);
      ls.AddMatrix(irb,true,3,  irb,true,0,  dce_dudr,1, false);        
      ls.AddMatrix(irb,true,0,  irb,true,3,  dce_dudr,1, true);                
      ls.AddMatrix(irb,true,3,  irb,true,3,  dce_drdr,1, true);        
    }    
    if( aFP[irb*8+i].itype_contact == 1 )
    {
      Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
      GetJacobian_StaticFrictionForce(dfv_du,dfv_dr,dfr_du,dfr_dr,
                                      rb,v0,aFP[irb*8+i].aloc,cont_norm,stiff_f);  
      ls.AddMatrix(irb,true,0,  irb,true,0,     dfv_du, -1, true );
      ls.AddMatrix(irb,true,3,  irb,true,0,     dfr_du, -1, true );
      ls.AddMatrix(irb,true,0,  irb,true,3,     dfv_dr, -1, true );
      ls.AddMatrix(irb,true,3,  irb,true,3,     dfr_dr, -1, true );                            
    }
  }      
  {   // 剛性行列を作る
    Com::CMatrix3 m0; m0.SetIdentity();
    ls.AddMatrix(irb,true,0,  irb,true,0,  m0,1.0e-8, true);
    ls.AddMatrix(irb,true,3,  irb,true,3,  m0,1.0e-8, true);
  }
}
*/



void UpdateContact_RigidBody
(const CRigid3D_Ini& ri, const CRigid3D_Def& rd,     
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 double stiff_n, double stiff_f, double myu_s, double myu_k, 
 CContactPoints_RigidBody3D& aCP )
{
  for(unsigned int i=0;i<aCP.Size();i++){
    Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);          
    CFrictionPoint& fp = aCP.aFP[i];
    Update_FrictionalContact(ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                             rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                             rd.GetVeloCG(),   rd.GetOmega(), 
                             v0,cont_base,cont_norm,stiff_n,stiff_f,myu_s,myu_k,fp);
  }
}


//////////////////////////////////////////////////////

void AddMatrix_StaticContact_Sensitivity
(MatVec::CMatDia_BlkCrs& mat, 
 ////
 unsigned int irb, const CRigid3D_Ini& ri, const CRigid3D_Def& rd,   
 PARAM_TYPE itype_s, int irb_s, unsigned int idim_s,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n, double stiff_f,
 const CContactPoints_RigidBody3D& aCP,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc)
{
  for(unsigned int i=0;i<aCP.Size();i++){
    if( !aCP.aFP[i].is_force_cont ) continue;
    Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);          
    Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
    GetDerCurCur_CurContactEnergy(dce_dudu, dce_dudr, dce_drdr, 
                                  ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                  rd.GetDispCG(),   rd.GetRotMatrix_Inc(),                                   
                                  v0, cont_base, cont_norm, stiff_n);
    {   // build stiffness matrix
      AddMatrix(mat, irb,0,  irb,0,  dce_dudu, 1, true );
      AddMatrix(mat, irb,0,  irb,3,  dce_dudr, 1, true );
      AddMatrix(mat, irb,3,  irb,0,  dce_dudr, 1, false);              
      AddMatrix(mat, irb,3,  irb,3,  dce_drdr, 1, true );        
    }    
    if(      itype_anchor_reloc == DONT_MOVE || itype_anchor_reloc == MOVE_INITIAL ){
      Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
      GetJacobian_StaticFrictionForce(dfv_du,dfv_dr,dfr_du,dfr_dr,
                                      ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                      rd.GetDispCG(),   rd.GetRotMatrix_Inc(),                                      
                                      v0,aCP.aFP[i].aloc,cont_norm,stiff_f);  
      AddMatrix(mat, irb,0,  irb,0,     dfv_du, -1, true);
      AddMatrix(mat, irb,0,  irb,3,     dfv_dr, -1, true);
      AddMatrix(mat, irb,3,  irb,0,     dfr_du, -1, true);
      AddMatrix(mat, irb,3,  irb,3,     dfr_dr, -1, true);                                 
    }
    else if( itype_anchor_reloc == MOVE_TOGETHER){
      if( irb != irb_s ){
        Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
        GetJacobian_StaticFrictionForce(dfv_du,dfv_dr,dfr_du,dfr_dr,
                                        ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                        rd.GetDispCG(),   rd.GetRotMatrix_Inc(),                                        
                                        v0,aCP.aFP[i].aloc,cont_norm,stiff_f);  
        AddMatrix(mat, irb,0,  irb,0,     dfv_du, -1, true);
        AddMatrix(mat, irb,0,  irb,3,     dfv_dr, -1, true);
        AddMatrix(mat, irb,3,  irb,0,     dfr_du, -1, true);
        AddMatrix(mat, irb,3,  irb,3,     dfr_dr, -1, true);
      }
    }
  }        
}

/*
void AddMatrix_StaticContact_Sensitivity_AnchorInitial
(MatVec::CMatDia_BlkCrs& mat, 
 unsigned int irb, const CRigidBody3D& rb,  
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n, double stiff_f,
 const CContactPoints_RigidBody3D& aCP,
 ANCHOR_RELOCATION_TYPE itype_anchor_relocation)
{
  //  if( rb.imode != 1 ){ return; }  
  for(unsigned int i=0;i<aCP.Size();i++){
    if( aCP.aFP[i].itype_contact != 1 ) continue;
    Com::CVector3D v0 = aCP.GetPosContactPoint(i,rb.xlen,rb.ylen,rb.zlen);          
    Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
    GetDerCurCur_CurContactEnergy(dce_dudu, dce_dudr, dce_drdr, rb, v0, cont_base, cont_norm, stiff_n);
    {   // 剛性行列を作る
      AddMatrix(mat, irb,0,  irb,0,  dce_dudu, 1, true );
      AddMatrix(mat, irb,0,  irb,3,  dce_dudr, 1, true );
      AddMatrix(mat, irb,3,  irb,0,  dce_dudr, 1, false);              
      AddMatrix(mat, irb,3,  irb,3,  dce_drdr, 1, true );        
    }    
    {
      Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
      GetJacobian_StaticFrictionForce(dfv_du,dfv_dr,dfr_du,dfr_dr,
                                      rb,v0,aCP.aFP[i].aloc,cont_norm,stiff_f);  
      AddMatrix(mat, irb,0,  irb,0,     dfv_du, -1, true);
      AddMatrix(mat, irb,0,  irb,3,     dfv_dr, -1, true);
      AddMatrix(mat, irb,3,  irb,0,     dfr_du, -1, true);
      AddMatrix(mat, irb,3,  irb,3,     dfr_dr, -1, true);                                 
    }
  }          
}
*/

void AddMatrix_StaticContact
(MatVec::CMatDia_BlkCrs& mat, 
 unsigned int irb, //const CRigidBody3D& rb,
 const CRigid3D_Ini& ri, 
 const CRigid3D_Def& rd, 
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n, double stiff_f,
 const CContactPoints_RigidBody3D& aCP)
{
  for(unsigned int i=0;i<aCP.Size();i++){
    if( !aCP.aFP[i].is_force_cont ) continue;
    Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);                  
    Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
    GetDerCurCur_CurContactEnergy(dce_dudu, dce_dudr, dce_drdr, 
                                  ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                  rd.GetDispCG(),   rd.GetRotMatrix_Inc(),                                   
                                  v0, cont_base, cont_norm, stiff_n);
    {   // build stiffness matrix
      AddMatrix(mat, irb,0,  irb,0,  dce_dudu, 1, true );
      AddMatrix(mat, irb,0,  irb,3,  dce_dudr, 1, true );
      AddMatrix(mat, irb,3,  irb,0,  dce_dudr, 1, false);              
      AddMatrix(mat, irb,3,  irb,3,  dce_drdr, 1, true );        
    }    
    {
      Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
      GetJacobian_StaticFrictionForce(dfv_du,dfv_dr,dfr_du,dfr_dr,
                                      ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                      rd.GetDispCG(),   rd.GetRotMatrix_Inc(),                                                                              
                                      v0,aCP.aFP[i].aloc,cont_norm,stiff_f);  
      AddMatrix(mat, irb,0,  irb,0,     dfv_du, -1, true);
      AddMatrix(mat, irb,0,  irb,3,     dfv_dr, -1, true);
      AddMatrix(mat, irb,3,  irb,0,     dfr_du, -1, true);
      AddMatrix(mat, irb,3,  irb,3,     dfr_dr, -1, true);                                       
    }
  }        
}


/*
void AddSensitivityResidual_StaticContact
(unsigned int irb, const CRigidBody3D& rb, 
 std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResPR, 
 PARAM_TYPE itype_s, int iobj_s, unsigned int idim_s, double alpha,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n, double stiff_f,
 const CContactPoints_RigidBody3D& aCP,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc)
{
//  if( rb.imode != 1 ){ return; }  
  if( irb!=iobj_s ){ assert(0); return; }  // residual changes when rigid body was moved
  if( itype_s != BOX_POS && itype_s != BOX_ROT && itype_s != BOX_LEN){ assert(0); return; }
  for(unsigned int i=0;i<aCP.Size();i++){
    if( aCP.aFP[i].itype_contact != 1 ){ continue; }        
//    Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
    Com::CVector3D v0 = aCP.GetPosContactPoint(i,rb.xlen,rb.ylen,rb.zlen);              
    Com::CMatrix3 B0 = aCP.GetSensMat_ContPointParam(i);
    Com::CVector3D e_s(0,0,0); e_s[idim_s] = alpha;
//    if( itype_s == BOX_POS || itype_s == BOX_ROT ){
    {
      Com::CMatrix3 dce_dudd, dce_drdd, dce_duds, dce_drds, dce_dudl, dce_drdl;
      GetDerCurIni_CurContactEnergy(dce_dudd, dce_drdd, dce_duds, dce_drds, dce_dudl, dce_drdl, rb, v0, cont_base,cont_norm, stiff_n);
      if(      itype_s == BOX_POS ){
        aResPR[irb].first  -= dce_dudd.MatVec(e_s);
        aResPR[irb].second -= dce_drdd.MatVec(e_s);
      }
      else if( itype_s == BOX_ROT ){
        aResPR[irb].first  -= dce_duds.MatVec(e_s);
        aResPR[irb].second -= dce_drds.MatVec(e_s);
      }
      else if( itype_s == BOX_LEN ){
        Com::CVector3D v1 = B0.MatVec(e_s);                
//        Com::CVector3D v1(0,0,0);
//        v1[idim_s] = corner[i][idim_s]*0.5;
        aResPR[irb].first  -= dce_dudl.MatVec(v1);
        aResPR[irb].second -= dce_drdl.MatVec(v1);
      }
    }          
    /////
    // static friction spring
    if( itype_anchor_reloc == MOVE_TOGETHER ){ continue; }
    if( itype_anchor_reloc == DONT_MOVE ){
      Com::CMatrix3 dfv_dd, dfr_dd, dfv_ds, dfr_ds, dfv_dl,dfr_dl;
      GetSensitivity_StaticFrictionForce_FixAnchor(dfv_dd,dfr_dd, dfv_ds,dfr_ds, dfv_dl,dfr_dl,
                                                   rb, v0, aCP.aFP[i].aloc,cont_norm, stiff_f);
      if(      itype_s == BOX_POS ){
        aResPR[irb].first  += dfv_dd.MatVec(e_s);
        aResPR[irb].second += dfr_dd.MatVec(e_s);
      }
      else if( itype_s == BOX_ROT ){
        aResPR[irb].first  += dfv_ds.MatVec(e_s);
        aResPR[irb].second += dfr_ds.MatVec(e_s);
      }
      else if( itype_s == BOX_LEN ){
//        Com::CVector3D v1(0,0,0); v1[idim_s] = corner[i][idim_s]*0.5;
        Com::CVector3D v1 = B0.MatVec(e_s);        
        aResPR[irb].first  += dfv_dl.MatVec(v1);
        aResPR[irb].second += dfr_dl.MatVec(v1);        
      }          
    }
    if( itype_anchor_reloc == MOVE_INITIAL ){
      Com::CMatrix3 dfv_dd, dfr_dd, dfv_ds, dfr_ds, dfv_dl,dfr_dl;
      GetSensitivity_StaticFrictionForce_InitialAnchor(dfv_dd,dfr_dd, dfv_ds,dfr_ds, dfv_dl,dfr_dl,
                                                       rb, v0,cont_norm, stiff_f);
      if(      itype_s == BOX_POS ){
        aResPR[irb].first  += dfv_dd.MatVec(e_s);
        aResPR[irb].second += dfr_dd.MatVec(e_s);
      }
      else if( itype_s == BOX_ROT ){
        aResPR[irb].first  += dfv_ds.MatVec(e_s);
        aResPR[irb].second += dfr_ds.MatVec(e_s);
      }
      else if( itype_s == BOX_LEN ){        
//        Com::CVector3D v1(0,0,0); v1[idim_s] = corner[i][idim_s]*0.5;
        Com::CVector3D v1 = B0.MatVec(e_s);
        aResPR[irb].first  += dfv_dl.MatVec(v1);
        aResPR[irb].second += dfr_dl.MatVec(v1);        
      }                      
    }
  }      
}
*/


void AddSensitivityResidual_StaticContact
(unsigned int irb, 
// const CRigidBody3D& rb, 
 const CRigid3D_Ini& ri, 
 const CRigid3D_Def& rd,
 Com::CMatrix3& ResP,Com::CMatrix3& ResR, 
 PARAM_TYPE itype_s, int iobj_s, 
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n, double stiff_f,
 const CContactPoints_RigidBody3D& aCP,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc)
{
//  const CRigid3D_Ini& ri = rb.ri;  
//  const CRigid3D_Def& rd = rb.rd;    
  if( irb!=iobj_s ){ assert(0); return; }  // residual changes when rigid body was moved
  if( itype_s != BOX_POS && itype_s != BOX_ROT && itype_s != BOX_LEN){ assert(0); return; }
  for(unsigned int i=0;i<aCP.Size();i++){
    if( !aCP.aFP[i].is_force_cont ){ continue; }        
    Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);              
    Com::CMatrix3 B0 = aCP.GetSensMat_ContPointParam(i);
    {
      Com::CMatrix3 dce_dudd, dce_drdd, dce_duds, dce_drds, dce_dudl, dce_drdl;
      GetDerCurIni_CurContactEnergy(dce_dudd, dce_drdd, dce_duds, dce_drds, dce_dudl, dce_drdl, 
                                    ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                    rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                                    v0, cont_base,cont_norm, stiff_n);
      if(      itype_s == BOX_POS ){
        ResP -= dce_dudd;
        ResR -= dce_drdd;
      }
      else if( itype_s == BOX_ROT ){
        ResP -= dce_duds;
        ResR -= dce_drds;
      }
      else if( itype_s == BOX_LEN ){
        ResP -= dce_dudl.MatMat(B0);
        ResR -= dce_drdl.MatMat(B0);
      }
    }          
    /////
    // static friction spring
    if( itype_anchor_reloc == MOVE_TOGETHER ){ continue; }
    if( itype_anchor_reloc == DONT_MOVE ){
      Com::CMatrix3 dfv_dd, dfr_dd, dfv_ds, dfr_ds, dfv_dl,dfr_dl;
      GetSensitivity_StaticFrictionForce_FixAnchor(dfv_dd,dfr_dd, dfv_ds,dfr_ds, dfv_dl,dfr_dl,
                                                   ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                                   rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                                                   v0, aCP.aFP[i].aloc,cont_norm, stiff_f);
      if(      itype_s == BOX_POS ){
        ResP += dfv_dd;
        ResR += dfr_dd;
      }
      else if( itype_s == BOX_ROT ){
        ResP += dfv_ds;
        ResR += dfr_ds;
      }
      else if( itype_s == BOX_LEN ){
        ResP += dfv_dl.MatMat(B0);
        ResR += dfr_dl.MatMat(B0);        
      }          
    }
    if( itype_anchor_reloc == MOVE_INITIAL ){
      Com::CMatrix3 dfv_dd, dfr_dd, dfv_ds, dfr_ds, dfv_dl,dfr_dl;
      GetSensitivity_StaticFrictionForce_InitialAnchor(dfv_dd,dfr_dd, dfv_ds,dfr_ds, dfv_dl,dfr_dl,
                                                       ri.GetIniPosCG(),  ri.GetRotMatrix_Ini(),
                                                       rd.GetDispCG(),    rd.GetRotMatrix_Inc(),
                                                       v0,cont_norm, stiff_f);
      if(      itype_s == BOX_POS ){
        ResP += dfv_dd;
        ResR += dfr_dd;
      }
      else if( itype_s == BOX_ROT ){
        ResP += dfv_ds;
        ResR += dfr_ds;
      }
      else if( itype_s == BOX_LEN ){        
        ResP += dfv_dl.MatMat(B0);
        ResR += dfr_dl.MatMat(B0);        
      }                      
    }
  }      
}
 

/*
void UpdateAnchorRigidBody
(std::vector<CFrictionPoint>& aFP,
 const CRigidBody3D& rb0,
 PARAM_TYPE itype_s,unsigned int irb_s, unsigned int idim_s,
 double eps,
 const Com::CVector3D& n0)
{
    if( itype_s == JNT_POS ) return;
    if( rb0.imode != 1 ){ return; }  
    CRigidBody3D rb1 = rb0;
    if( itype_s == BOX_POS ){
      Com::CVector3D e_s(0,0,0); e_s[idim_s] = eps;
      rb1.SetIniPosCG( rb1.GetIniPosCG()+e_s );
    }
    else if( itype_s == BOX_ROT ){
      Com::CVector3D e_s(0,0,0); e_s[idim_s] = eps;
      Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e_s);
      rb1.AddRotation_IniLeft(rot);  
    }      
    else if( itype_s == BOX_LEN ){
      double xlen=rb1.xlen; 
      double ylen=rb1.ylen;
      double zlen=rb1.zlen;    
      double rho = rb1.GetMass()/(xlen*ylen*zlen);
      if( idim_s == 0 ){ xlen += eps; }
      if( idim_s == 1 ){ ylen += eps; }
      if( idim_s == 2 ){ zlen += eps; }            
      rb1.SetModeBox(xlen, ylen, zlen, rho);
    }  
    const double corner[8][3] = { 
      {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
      {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };    
    for(unsigned int i=0;i<8;i++){
      CFrictionPoint& fp = aFP[irb_s*8+i];
      if( fp.itype_contact == 0 ) continue;
      Com::CVector3D a = Com::CVector3D(corner[i][0]*rb0.xlen, corner[i][1]*rb0.ylen, corner[i][2]*rb0.zlen )*0.5;
      Com::CVector3D aloc0 = fp.aloc;
      Com::CVector3D x0 = rb0.GetCurrentPosition(a);    
      Com::CVector3D x1 = rb1.GetCurrentPosition(a); 
      Com::CVector3D d = x1-x0;
      d -= n0*Com::Dot(n0,d);
      fp.aloc += d;
    }    
}*/

Com::CVector3D GetSensitivityCurrentPointOnRigidBody
(PARAM_TYPE itype_s, unsigned int irb_s, unsigned int idim_s,
 unsigned int irb, 
// const CRigidBody3D& rb, 
 const CRigid3D_Ini& ri, 
 const CRigid3D_Def& rd,   
 unsigned int i, 
 const CContactPoints_RigidBody3D& aCP,
 const std::pair< Com::CVector3D, Com::CVector3D>& sns_prlm )
{
  Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);       
  Com::CMatrix3 B0 = aCP.GetSensMat_ContPointParam(i);
  Com::CVector3D dv1c(0,0,0);
  {
    const Com::CVector3D& Qa = ri.GetRotMatrix_Ini().MatVec(v0);
    const Com::CVector3D& wQasr = Cross(Qa,sns_prlm.second);
    dv1c = sns_prlm.first - rd.GetRotMatrix_Inc().MatVec(wQasr);            
  }
  Com::CVector3D dv1i(0,0,0);
  if( irb == irb_s ){ // initial position moves
    if( itype_s == BOX_POS ){
      dv1i[idim_s] = 1;
    }
    else if( itype_s == BOX_ROT ){
      Com::CVector3D e_s(0,0,0); e_s[idim_s] = 1;
      const Com::CVector3D& Qa = ri.GetRotMatrix_Ini().MatVec(v0);
      const Com::CVector3D& wQaes = Cross(Qa,e_s);
      dv1i = -rd.GetRotMatrix_Inc().MatVec(wQaes);          
    }
    else if( itype_s == BOX_LEN ){
//      Com::CVector3D e_s(0,0,0); e_s[idim_s] = corner[i][idim_s]*0.5;                    
      Com::CVector3D e_s(0,0,0); e_s[idim_s] = 1;      
      Com::CVector3D v1 = B0.MatVec(e_s);        
      Com::CMatrix3 R = rd.GetRotMatrix_Inc().MatMat( ri.GetRotMatrix_Ini() );
      dv1i = R.MatVec(v1);
    }
  }
  return dv1c + dv1i;  
}

Com::CVector3D GetSensitivityCurrentPointOnRigidBody
(const Com::CVector3D& sD,
 const Com::CVector3D& sQ,
 const Com::CVector3D& sL,
 const Com::CVector3D& sU, 
 const Com::CVector3D& sP,
// const CRigidBody3D& rb, 
const CRigid3D_Ini& ri, 
const CRigid3D_Def& rd,  
 const Com::CVector3D& v0,
 const Com::CMatrix3& B0 )
{
//  const CRigid3D_Ini& ri = rb.ri;    
//  const CRigid3D_Def& rd = rb.rd;      
  Com::CVector3D dv1c(0,0,0);
  {
    const Com::CVector3D& Qa = ri.GetRotMatrix_Ini().MatVec(v0);
    const Com::CVector3D& wQasr = Cross(Qa,sP);
    dv1c = sU - rd.GetRotMatrix_Inc().MatVec(wQasr);            
  }
  Com::CVector3D dv1i(0,0,0);
  {
    dv1i = sD;
    const Com::CVector3D& Qa = ri.GetRotMatrix_Ini().MatVec(v0);
    const Com::CVector3D& wQaes = Cross(Qa,sQ);
    dv1i += -rd.GetRotMatrix_Inc().MatVec(wQaes);          
    Com::CVector3D v1 = B0.MatVec(sL);   
    Com::CMatrix3 R = rd.GetRotMatrix_Inc().MatMat( ri.GetRotMatrix_Ini() );
    dv1i += R.MatVec(v1);
  }
  return dv1c + dv1i;       
}

void GetSensitivityInitialPointOnRigidBody
(Com::CMatrix3& sns_p, 
 Com::CMatrix3& sns_r,
 Com::CMatrix3& sns_l,
 const CRigid3D_Ini& ri,
 const Com::CVector3D& v0, 
 const Com::CMatrix3& B0)
{
  const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
  const Com::CVector3D& Qv0 = Q.MatVec(v0);
  sns_p = Com::CMatrix3(1);
  sns_r = -Com::CMatrix3(Qv0);
  sns_l = Q.MatMat(B0);
}



void GetSensitivityInitialPointOnRigidBody
(const CRigid3D_Ini& ri,
// const CRigidBody3D& rb, 
 unsigned int i, 
 const CContactPoints_RigidBody3D& aCP, 
 Com::CMatrix3& sns_p, 
 Com::CMatrix3& sns_r,
 Com::CMatrix3& sns_l)
{

  const Com::CVector3D& v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
  Com::CMatrix3 B0 = aCP.GetSensMat_ContPointParam(i);
  GetSensitivityInitialPointOnRigidBody(sns_p, sns_r, sns_l, ri,v0,B0);
}





void Update_FrictionalContact
(//const CRigidBody3D& rb, 
 const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P,      
 const Com::CVector3D& v, const Com::CVector3D& O, 
 Com::CVector3D& v0,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm,
 double stiff_n, double stiff_f, double myu_s, double myu_k,
 CFrictionPoint& fp )
{	
  //  const Com::CVector3D& uv = rb.GetCurrentVelocity(v0);
  const Com::CVector3D& uv = v + P.MatVec( Cross(O,Q.MatVec(v0)) );          
  const Com::CVector3D& co = u + d + P.MatVec( Q.MatVec(v0) );
  const Com::CVector3D& n1 = cont_norm;
  const double pd1 = Dot(cont_norm, cont_base-co);//GetPenetrationDepth(rb,v0,cont_base,cont_norm);
  const double pd0 = fp.pd;
  
  if( pd1 < 0 ){ // not contact
  }
  else if( pd0 < 0 ){	// h1>0 && h0<0	: contact in next step
    fp.aloc = co + pd1*n1;  // put anchor at projected point
  }
  else{	// h1>0 && h0>0
    // update the anchor that spring give force equal to dynamic friction direction to the velocity
    if( fp.itype_contact == 2 ){ // sliding contact
      Com::CVector3D n_v_t = uv-Com::Dot(uv,n1)*n1; n_v_t.SetNormalizedVector();  // normalize tangent velocity
      const double dist = pd1*stiff_n*myu_k/stiff_f;
      fp.aloc = co-n_v_t*dist; // update anchor      
    }
    else if( fp.itype_contact == 1 ){ // no update of anchor in case of static friction
    }
  }
}



void AddResidual_StaticContact
(const Com::CVector3D& d, const Com::CMatrix3& Q, 
 const Com::CVector3D& u, const Com::CMatrix3& P, 
 const Com::CVector3D& a,
 unsigned int irb,
 std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aResPR,
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_n, double stiff_f,
 const CFrictionPoint& fp)
{  
  Com::CVector3D n0 = cont_norm;
  const double pd = GetPenetrationDepth(d,Q,u,P,a,cont_base,cont_norm);
  
  if( pd < 0 ){ return; }
  
  {
    Com::CVector3D dce_du, dce_dr;
    GetDerCur_CurContactEnergy(dce_du,dce_dr,
                               d,Q, u,P,
                               a,cont_base,cont_norm,stiff_n);
    aResPR[irb].first  -= dce_du;
    aResPR[irb].second -= dce_dr;
  }  
  
  if( fp.itype_contact == 1 ){
    Com::CVector3D fv, fr;
    GetStaticFrictionForce(fv,fr, 
                           d,Q, u,P,
                           a,fp.aloc,cont_norm,stiff_f);
    aResPR[irb].first  += fv;
    aResPR[irb].second += fr;
  }
}







