/*
 *  rigid_contact_test.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/21/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include <math.h>
#include <stdlib.h>
#include "delfem/matrix3d.h"
#include "delfem/vector3d.h"

#include "rigidbody.h"
#include "rigid_contact.h"
#include "rigid_contact_test.h"


static double rand2(double min, double max){
  return rand()/(RAND_MAX+1.0)*(max-min)+min;
}

void TestContact()
{
  const Com::CVector3D v0( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  CRigidBody3D rb;
  rb.ri.SetIniPosCG( Com::CVector3D(0,0,0) );
  {
    const unsigned int PI = 3.1415;
    Com::CMatrix3 m;
    m.SetRotMatrix_BryantAngle( rand2(-PI,PI), rand2(-PI,PI), rand2(-PI,PI) );
    rb.ri.SetIniRotation(m);
  }
  {
    const unsigned int PI = 3.1415;
    Com::CMatrix3 m;
    m.SetRotMatrix_BryantAngle( rand2(-PI,PI), rand2(-PI,PI), rand2(-PI,PI) );
    rb.rd.SetIncRotation(m);
  }    
  Com::CVector3D cont_base( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  Com::CVector3D cont_norm( rand2(-1,1), rand2(-1,1), rand2(-1,1) );  
  const double cont_stiff = rand2(0.01,100);
  double ce = GetCurContactEnergy(rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(), 
                                  rb.rd.GetDispCG(), rb.rd.GetRotMatrix_Inc(), 
                                  v0,cont_base,cont_norm,cont_stiff);
  Com::CVector3D dce_du, dce_dr;
  GetDerCur_CurContactEnergy(dce_du,dce_dr,
                             rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(), 
                             rb.rd.GetDispCG(), rb.rd.GetRotMatrix_Inc(), 
                             v0,cont_base,cont_norm,cont_stiff);
  Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
  GetDerCurCur_CurContactEnergy(dce_dudu, dce_dudr, dce_drdr, 
                                rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(),
                                rb.rd.GetDispCG(),   rb.rd.GetRotMatrix_Inc(),                                 
                                v0, cont_base, cont_norm, cont_stiff);  
  /////
  /*
   double Ce = GetContactEnergy(rb,v0,cont_base,cont_norm,cont_stiff,false);
   Com::CVector3D dCe_du, dCe_dr;
   Get1stDer_ContactEnergy(dCe_du,dCe_dr,rb,v0,cont_base,cont_norm,cont_stiff,false);
   Com::CMatrix3 dCe_dudu, dCe_dudr, dCe_drdr;
   Get2ndDer_ContactEnergy(dCe_dudu, dCe_dudr, dCe_drdr, rb, v0, cont_base, cont_norm, cont_stiff,false);    
   */
  /////
  Com::CMatrix3 dce_dudd, dce_drdd, dce_duds, dce_drds,  dce_dudl, dce_drdl;
  GetDerCurIni_CurContactEnergy(dce_dudd, dce_drdd, dce_duds, dce_drds, dce_dudl, dce_drdl, 
                                rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(),
                                rb.rd.GetDispCG(),   rb.rd.GetRotMatrix_Inc(),
                                v0, cont_base,cont_norm, cont_stiff);    
  
  double eps = 1.0e-4;
  
  Com::CVector3D e( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  e *= eps;
  {
    std::cout << "dce_du" << std::endl;
    CRigidBody3D rb1 = rb;    
    rb1.rd.AddDispCG( e );
    double ce1 = GetCurContactEnergy(rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(), 
                                     rb1.rd.GetDispCG(),   rb1.rd.GetRotMatrix_Inc(), 
                                     v0,cont_base,cont_norm,cont_stiff);
    std::cout << (ce1-ce)/eps << " " << Dot(dce_du,e)/eps << std::endl; 
  }
  {
    std::cout << "dce_dr" << std::endl;    
    CRigidBody3D rb1 = rb;    
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
    rb1.rd.AddRotation_IncRight(rot);
    double ce1 = GetCurContactEnergy(rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(), 
                                     rb1.rd.GetDispCG(),   rb1.rd.GetRotMatrix_Inc(),
                                     v0,cont_base,cont_norm,cont_stiff);    
    std::cout << (ce1-ce)/eps << " " << Dot(dce_dr,e)/eps << std::endl; 
  }
  /*
   {
   std::cout << "dCe_du" << std::endl;
   CRigidBody3D rb1 = rb;    
   rb1.SetIniPosCG( e + rb1.ri.GetIniPosCG() );
   double Ce1 = GetContactEnergy(rb1,v0,cont_base,cont_norm,cont_stiff,false);
   std::cout << (Ce1-Ce)/eps << " " << Dot(dCe_du,e)/eps << std::endl; 
   }  
   {
   std::cout << "dCe_dr" << std::endl;
   CRigidBody3D rb1 = rb;    
   Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
   rb1.AddRotation_IniLeft(rot);
   double Ce1 = GetContactEnergy(rb1,v0,cont_base,cont_norm,cont_stiff,false);    
   std::cout << (Ce1-Ce)/eps << " " << Dot(dCe_dr,e)/eps << std::endl; 
   }  
   */
  ///
  {
    CRigidBody3D rb1 = rb;
    rb1.rd.AddDispCG( e );
    Com::CVector3D dce_du1, dce_dr1;
    GetDerCur_CurContactEnergy(dce_du1,dce_dr1,
                               rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(), 
                               rb1.rd.GetDispCG(),   rb1.rd.GetRotMatrix_Inc(),                                
                               v0,cont_base,cont_norm,cont_stiff);    
    Com::CVector3D d0 = (dce_du1-dce_du)*=1.0/eps;
    Com::CVector3D d1 = dce_dudu.MatVec(e); d1*=1.0/eps;
    std::cout << "dce_dudu" << std::endl;
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    std::cout << "dce_drdu" << std::endl;
    Com::CVector3D e0 = (dce_dr1-dce_dr)*=1.0/eps;
    Com::CVector3D e1 = dce_dudr.MatVecTrans(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;        
  }
  {
    CRigidBody3D rb1 = rb;    
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
    rb1.rd.AddRotation_IncRight(rot);
    Com::CVector3D dce_du1, dce_dr1;
    GetDerCur_CurContactEnergy(dce_du1,dce_dr1,
                               rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(), 
                               rb1.rd.GetDispCG(), rb1.rd.GetRotMatrix_Inc(), 
                               v0,cont_base,cont_norm,cont_stiff);    
    Com::CVector3D d0 = (dce_du1-dce_du)*=1.0/eps;
    Com::CVector3D d1 = dce_dudr.MatVec(e); d1*=1.0/eps;    
    std::cout << "dce_dudr" << std::endl;    
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    std::cout << "dce_drdr" << std::endl;        
    Com::CVector3D e0 = (dce_dr1-dce_dr)*=1.0/eps;
    Com::CVector3D e1 = dce_drdr.MatVec(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;    
  }
  ////  
  /*
   {
   CRigidBody3D rb1 = rb;
   rb1.SetIniPosCG( e + rb1.ri.GetIniPosCG() );
   Com::CVector3D dCe_du1, dCe_dr1;
   Get1stDer_ContactEnergy(dCe_du1,dCe_dr1,rb1,v0,cont_base,cont_norm,cont_stiff,false);    
   Com::CVector3D d0 = (dCe_du1-dCe_du)*=1.0/eps;
   Com::CVector3D d1 = dCe_dudu.MatVec(e); d1*=1.0/eps;
   std::cout << "dCe_dudu" << std::endl;
   std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
   std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
   std::cout << "dCe_drdu" << std::endl;
   Com::CVector3D e0 = (dCe_dr1-dCe_dr)*=1.0/eps;
   Com::CVector3D e1 = dCe_dudr.MatVecTrans(e); e1*=1.0/eps;
   std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
   std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;        
   }
   {
   CRigidBody3D rb1 = rb;    
   Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
   rb1.AddRotation_IniLeft(rot);
   Com::CVector3D dCe_du1, dCe_dr1;
   Get1stDer_ContactEnergy(dCe_du1,dCe_dr1,rb1,v0,cont_base,cont_norm,cont_stiff,false);    
   Com::CVector3D d0 = (dCe_du1-dCe_du)*=1.0/eps;
   Com::CVector3D d1 = dCe_dudr.MatVec(e); d1*=1.0/eps;    
   std::cout << "dCe_dudr" << std::endl;    
   std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
   std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
   std::cout << "dCe_drdr" << std::endl;        
   Com::CVector3D e0 = (dCe_dr1-dCe_dr)*=1.0/eps;
   Com::CVector3D e1 = dCe_drdr.MatVec(e); e1*=1.0/eps;
   std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
   std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;    
   }  
   */
  ////
  std::cout << "sensitive" << std::endl;
  {
    CRigidBody3D rb1 = rb;    
    rb1.ri.SetIniPosCG( rb1.ri.GetIniPosCG()+e ); 
    Com::CVector3D dce_du1, dce_dr1;
    GetDerCur_CurContactEnergy(dce_du1,dce_dr1,
                               rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(), 
                               rb1.rd.GetDispCG(),   rb1.rd.GetRotMatrix_Inc(), 
                               v0,cont_base,cont_norm,cont_stiff);    
    Com::CVector3D d0 = (dce_du1-dce_du)*=1.0/eps;
    Com::CVector3D d1 = dce_dudd.MatVec(e); d1*=1.0/eps;    
    std::cout << "dce_dudd" << std::endl;    
    std::cout << (d0-d1).Length() / d0.Length() << std::endl;    
    std::cout << "   " << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << "   " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    
    Com::CVector3D e0 = (dce_dr1-dce_dr)*=1.0/eps;
    Com::CVector3D e1 = dce_drdd.MatVec(e); e1*=1.0/eps;
    std::cout << "dce_drdd" << std::endl;
    std::cout << (e0-e1).Length() / e0.Length() << std::endl;        
    std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;    
  }    
  {
    CRigidBody3D rb1 = rb;    
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
    rb1.ri.AddRotation_IniLeft(rot);
    Com::CVector3D dce_du1, dce_dr1;
    GetDerCur_CurContactEnergy(dce_du1,dce_dr1,
                               rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(), 
                               rb1.rd.GetDispCG(),   rb1.rd.GetRotMatrix_Inc(),   
                               v0,cont_base,cont_norm,cont_stiff);    
    Com::CVector3D d0 = (dce_du1-dce_du)*=1.0/eps;    
    Com::CVector3D d1 = dce_duds.MatVec(e); d1*=1.0/eps;    
    std::cout << "dce_duds" << std::endl;
    std::cout << (d0-d1).Length() / d0.Length() << std::endl;        
    std::cout << "    " << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << "    " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    
    Com::CVector3D e0 = (dce_dr1-dce_dr)*=1.0/eps;
    Com::CVector3D e1 = dce_drds.MatVec(e); e1*=1.0/eps;
    std::cout << "dce_drds" << std::endl;    
    std::cout << (e0-e1).Length() / e0.Length() << std::endl;            
    std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;    
  }
  {
    Com::CVector3D v1 = v0 + e;
    Com::CVector3D dce_du1, dce_dr1;    
    GetDerCur_CurContactEnergy(dce_du1,dce_dr1,
                               rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(), 
                               rb.rd.GetDispCG(), rb.rd.GetRotMatrix_Inc(),  
                               v1,cont_base,cont_norm,cont_stiff);    
    Com::CVector3D d0 = (dce_du1-dce_du)*=1.0/eps;
    Com::CVector3D d1 = dce_dudl.MatVec(e); d1*=1.0/eps;    
    std::cout << "dce_dudl" << std::endl;    
    std::cout << (d0-d1).Length() / d0.Length() << std::endl;            
    std::cout << "    " << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << "    " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    
    Com::CVector3D e0 = (dce_dr1-dce_dr)*=1.0/eps;
    Com::CVector3D e1 = dce_drdl.MatVec(e); e1*=1.0/eps;
    std::cout << "dce_drdl" << std::endl;        
    std::cout << (e0-e1).Length() / e0.Length() << std::endl;            
    std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;    
  }
}








void TestKineticFriction()
{
  const Com::CVector3D v0( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  CRigidBody3D rb;
  rb.ri.SetIniPosCG( Com::CVector3D(rand2(-1,1), rand2(-1,1), rand2(-1,1)) );
  rb.rd.SetOmega( Com::CVector3D(rand2(-1,1), rand2(-1,1), rand2(-1,1)) );
  rb.rd.SetVeloCG( Com::CVector3D(rand2(-1,1), rand2(-1,1), rand2(-1,1)) );
  {
    const unsigned int PI = 3.1415;
    Com::CMatrix3 m;
    m.SetRotMatrix_BryantAngle( rand2(-PI,PI), rand2(-PI,PI), rand2(-PI,PI) );
    rb.ri.SetIniRotation(m);
  }
  {
    const unsigned int PI = 3.1415;
    Com::CMatrix3 m;
    m.SetRotMatrix_BryantAngle( rand2(-PI,PI), rand2(-PI,PI), rand2(-PI,PI) );
    rb.rd.SetIncRotation(m);
  }    
  Com::CVector3D cont_base( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  Com::CVector3D cont_norm( rand2(-1,1), rand2(-1,1), rand2(-1,1) );  
  const double stiff_n = rand2(0.01,100);
  const double stiff_f = rand2(0.01,100);
  const double myu_k = rand2(0.001,1);
  Com::CVector3D fv, fr;
  GetKineticFrictionForce(fv,fr, 
                          rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(),
                          rb.rd.GetDispCG(), rb.rd.GetRotMatrix_Inc(),
                          rb.rd.GetVeloCG(), rb.rd.GetOmega(),
                          v0,cont_base,cont_norm,stiff_n,stiff_f,myu_k);
  Com::CMatrix3 dfv_dv, dfv_do, dfr_dv, dfr_do;
  Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
  GetJacobian_KineticFrictionForce(dfv_dv,dfv_do,dfr_dv,dfr_do, 
                                   dfv_du,dfv_dr,dfr_du,dfr_dr,
                                   //                                   rb,
                                   rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(),
                                   rb.rd.GetDispCG(), rb.rd.GetRotMatrix_Inc(),
                                   rb.rd.GetVeloCG(), rb.rd.GetOmega(),                                   
                                   v0,cont_base,cont_norm,stiff_n,stiff_f,myu_k);  
  
  double eps = 1.0e-8;  
  Com::CVector3D e( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  e *= eps;
  
  std::cout << "Test Kinetic Friction" << std::endl;
  std::cout << fv << " *****   " << fr << std::endl;
  
  { // derivative of fv about v
    CRigidBody3D rb1 = rb;
    rb1.rd.SetVeloCG( rb1.rd.GetVeloCG()+e );
    Com::CVector3D fv1, fr1;
    GetKineticFrictionForce(fv1,fr1,
                            rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(),
                            rb1.rd.GetDispCG(), rb1.rd.GetRotMatrix_Inc(),
                            rb1.rd.GetVeloCG(), rb1.rd.GetOmega(),                            
                            v0,cont_base,cont_norm,stiff_n,stiff_f,myu_k);
    Com::CVector3D d0 = (fv1-fv)*=1.0/eps;
    Com::CVector3D d1 = dfv_dv.MatVec(e); d1*=1.0/eps;
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    
    Com::CVector3D e0 = (fr1-fr)*=1.0/eps;
    Com::CVector3D e1 = dfr_dv.MatVec(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;        
  }
  {
    CRigidBody3D rb1 = rb; 
    rb1.rd.SetOmega( rb1.rd.GetOmega()+e );
    Com::CVector3D fv1, fr1;
    GetKineticFrictionForce(fv1,fr1,
                            rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(),
                            rb1.rd.GetDispCG(), rb1.rd.GetRotMatrix_Inc(),
                            rb1.rd.GetVeloCG(), rb1.rd.GetOmega(),                            
                            v0,cont_base,cont_norm,stiff_n,stiff_f,myu_k);
    Com::CVector3D d0 = (fv1-fv)*=1.0/eps;
    Com::CVector3D d1 = dfv_do.MatVec(e); d1*=1.0/eps;    
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    
    Com::CVector3D e0 = (fr1-fr)*=1.0/eps;
    Com::CVector3D e1 = dfr_do.MatVec(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;    
  }
  {
    CRigidBody3D rb1 = rb;
    rb1.rd.AddDispCG( e );
    Com::CVector3D fv1, fr1;
    GetKineticFrictionForce(fv1,fr1,
                            rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(),
                            rb1.rd.GetDispCG(), rb1.rd.GetRotMatrix_Inc(),
                            rb1.rd.GetVeloCG(), rb1.rd.GetOmega(),                                                        
                            v0,cont_base,cont_norm,stiff_n,stiff_f,myu_k);
    Com::CVector3D d0 = (fv1-fv)*=1.0/eps;
    Com::CVector3D d1 = dfv_du.MatVec(e); d1*=1.0/eps;
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    
    Com::CVector3D e0 = (fr1-fr)*=1.0/eps;
    Com::CVector3D e1 = dfr_du.MatVec(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;        
  }
  {
    CRigidBody3D rb1 = rb;    
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
    rb1.rd.AddRotation_IncRight(rot);
    Com::CVector3D fv1, fr1;
    GetKineticFrictionForce(fv1,fr1,
                            rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(),
                            rb1.rd.GetDispCG(), rb1.rd.GetRotMatrix_Inc(),
                            rb1.rd.GetVeloCG(), rb1.rd.GetOmega(),                                                        
                            v0,cont_base,cont_norm,stiff_n,stiff_f,myu_k);    
    Com::CVector3D d0 = (fv1-fv)*=1.0/eps;
    Com::CVector3D d1 = dfv_dr.MatVec(e); d1*=1.0/eps;   
    std::cout << "dfv_dr" << std::endl;
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    std::cout << "dfr_dr" << std::endl;
    Com::CVector3D e0 = (fr1-fr)*=1.0/eps;
    Com::CVector3D e1 = dfr_dr.MatVec(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;    
  }  
}



void TestStaticFriction()
{
  const Com::CVector3D v0( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  const Com::CVector3D aloc( rand2(-1,1), rand2(-1,1), rand2(-1,1) );  
  CRigidBody3D rb;
  rb.ri.SetIniPosCG( Com::CVector3D(rand2(-1,1), rand2(-1,1), rand2(-1,1)) );
  {
    const unsigned int PI = 3.1415;
    Com::CMatrix3 m;
    m.SetRotMatrix_BryantAngle( rand2(-PI,PI), rand2(-PI,PI), rand2(-PI,PI) );
    rb.ri.SetIniRotation(m);
  }
  {
    const unsigned int PI = 3.1415;
    Com::CMatrix3 m;
    m.SetRotMatrix_BryantAngle( rand2(-PI,PI), rand2(-PI,PI), rand2(-PI,PI) );
    rb.rd.SetIncRotation(m);
  }  
  Com::CVector3D cont_norm( rand2(-1,1), rand2(-1,1), rand2(-1,1) );  cont_norm.SetNormalizedVector();
  const double stiff_f = rand2(0.01,100);
  Com::CVector3D fv, fr;
  GetStaticFrictionForce(fv,fr,
                         rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(),
                         rb.rd.GetDispCG(),   rb.rd.GetRotMatrix_Inc(), 
                         v0,aloc,cont_norm,stiff_f);
  Com::CMatrix3 dfv_du, dfv_dr, dfr_du, dfr_dr;
  GetJacobian_StaticFrictionForce(dfv_du,dfv_dr,dfr_du,dfr_dr,                                  
                                  rb.ri.GetIniPosCG(),  rb.ri.GetRotMatrix_Ini(),
                                  rb.rd.GetDispCG(),    rb.rd.GetRotMatrix_Inc(),
                                  v0,aloc,cont_norm,stiff_f);  
  ////
  /*
   Com::CVector3D Fv, Fr;
   GetStaticFrictionForce(Fv,Fr, rb,v0,aloc,cont_norm,stiff_f,false);
   Com::CMatrix3 dFv_du, dFv_dr, dFr_du, dFr_dr;
   GetJacobian_StaticFrictionForce(dFv_du,dFv_dr,dFr_du,dFr_dr,   
   rb,v0,aloc,cont_norm,stiff_f,false);    
   */
  ////
  Com::CMatrix3 dfv_dd,dfr_dd, dfv_ds,dfr_ds, dfv_dl,dfr_dl; 
  GetSensitivity_StaticFrictionForce_FixAnchor(dfv_dd,dfr_dd,  dfv_ds,dfr_ds,  dfv_dl,dfr_dl,
                                               rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(),
                                               rb.rd.GetDispCG(),   rb.rd.GetRotMatrix_Inc(),
                                               v0,aloc,cont_norm,stiff_f);    
  
  double eps = 1.0e-8;  
  Com::CVector3D e( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  e *= eps;
  
  std::cout << "Test Static Friction" << std::endl;
  std::cout << fv << " *****   " << fr << std::endl;
  
  {
    CRigidBody3D rb1 = rb;
    rb1.rd.AddDispCG( e );
    Com::CVector3D fv1, fr1;
    GetStaticFrictionForce(fv1,fr1,
                           rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(),
                           rb1.rd.GetDispCG(),   rb1.rd.GetRotMatrix_Inc(),                            
                           v0,aloc,cont_norm,stiff_f);
    Com::CVector3D d0 = (fv1-fv)*=1.0/eps;
    Com::CVector3D d1 = dfv_du.MatVec(e); d1*=1.0/eps;
    std::cout << "dfv_du" << std::endl;    
    std::cout << "    " << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << "    " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    
    Com::CVector3D e0 = (fr1-fr)*=1.0/eps;
    Com::CVector3D e1 = dfr_du.MatVec(e); e1*=1.0/eps;
    std::cout << "dfr_du" << std::endl;        
    std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;        
  }
  {
    CRigidBody3D rb1 = rb;    
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
    rb1.rd.AddRotation_IncRight(rot);
    Com::CVector3D fv1, fr1;
    GetStaticFrictionForce(fv1,fr1,
                           rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(),
                           rb1.rd.GetDispCG(),   rb1.rd.GetRotMatrix_Inc(), 
                           v0,aloc,cont_norm,stiff_f);    
    Com::CVector3D d0 = (fv1-fv)*=1.0/eps;
    Com::CVector3D d1 = dfv_dr.MatVec(e); d1*=1.0/eps;   
    std::cout << "dfv_dr" << std::endl;
    std::cout << "    " << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << "    " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    std::cout << "dfr_dr" << std::endl;
    Com::CVector3D e0 = (fr1-fr)*=1.0/eps;
    Com::CVector3D e1 = dfr_dr.MatVec(e); e1*=1.0/eps;
    std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;    
  }  
  /////
  /*
   {
   CRigidBody3D rb1 = rb;
   rb1.SetIniPosCG( e  + rb1.ri.GetIniPosCG() );
   Com::CVector3D Fv1, Fr1;
   GetStaticFrictionForce(Fv1,Fr1,rb1,v0,aloc,cont_norm,stiff_f,false);
   Com::CVector3D d0 = (Fv1-Fv)*=1.0/eps;
   Com::CVector3D d1 = dFv_du.MatVec(e); d1*=1.0/eps;
   std::cout << "dFv_du" << std::endl;    
   std::cout << "    " << d0.x << " " << d0.y << " " << d0.z << std::endl;
   std::cout << "    " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
   
   Com::CVector3D e0 = (Fr1-Fr)*=1.0/eps;
   Com::CVector3D e1 = dFr_du.MatVec(e); e1*=1.0/eps;
   std::cout << "dFr_du" << std::endl;        
   std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
   std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;        
   }
   {
   CRigidBody3D rb1 = rb;    
   Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
   rb1.AddRotation_IniLeft(rot);
   Com::CVector3D Fv1, Fr1;
   GetStaticFrictionForce(Fv1,Fr1,rb1,v0,aloc,cont_norm,stiff_f,false);    
   Com::CVector3D d0 = (Fv1-Fv)*=1.0/eps;
   Com::CVector3D d1 = dFv_dr.MatVec(e); d1*=1.0/eps;   
   std::cout << "dFv_dr" << std::endl;
   std::cout << "    " << d0.x << " " << d0.y << " " << d0.z << std::endl;
   std::cout << "    " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
   std::cout << "dFr_dr" << std::endl;
   Com::CVector3D e0 = (Fr1-Fr)*=1.0/eps;
   Com::CVector3D e1 = dFr_dr.MatVec(e); e1*=1.0/eps;
   std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
   std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;    
   }    
   */
  /////
  std::cout << "sensitive" << std::endl;
  {
    CRigidBody3D rb1 = rb;    
    rb1.ri.SetIniPosCG( rb1.ri.GetIniPosCG()+e ); 
    Com::CVector3D fv1, fr1;
    GetStaticFrictionForce(fv1,fr1,
                           rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(),
                           rb1.rd.GetDispCG(),   rb1.rd.GetRotMatrix_Inc(), 
                           v0,aloc,cont_norm,stiff_f);
    Com::CVector3D d0 = (fv1-fv)*=1.0/eps;
    Com::CVector3D d1 = dfv_dd.MatVec(e); d1*=1.0/eps;
    std::cout << "dfv_dd" << std::endl;    
    std::cout << "    " << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << "    " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    
    Com::CVector3D e0 = (fr1-fr)*=1.0/eps;
    Com::CVector3D e1 = dfr_dd.MatVec(e); e1*=1.0/eps;
    std::cout << "dfr_dd" << std::endl;        
    std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;            
  }    
  {
    CRigidBody3D rb1 = rb;    
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
    rb1.ri.AddRotation_IniLeft(rot);
    Com::CVector3D fv1, fr1;
    GetStaticFrictionForce(fv1,fr1,
                           rb1.ri.GetIniPosCG(), rb1.ri.GetRotMatrix_Ini(),
                           rb1.rd.GetDispCG(),   rb1.rd.GetRotMatrix_Inc(),                           
                           v0,aloc,cont_norm,stiff_f);    
    Com::CVector3D d0 = (fv1-fv)*=1.0/eps;
    Com::CVector3D d1 = dfv_ds.MatVec(e); d1*=1.0/eps;   
    std::cout << "dfv_ds" << std::endl;
    std::cout << "    " << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << "    " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    std::cout << "dfr_ds" << std::endl;
    Com::CVector3D e0 = (fr1-fr)*=1.0/eps;
    Com::CVector3D e1 = dfr_ds.MatVec(e); e1*=1.0/eps;
    std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;        
  }
  {
    Com::CVector3D v1 = v0 + e;
    Com::CVector3D fv1, fr1;
    GetStaticFrictionForce(fv1,fr1,
                           rb.ri.GetIniPosCG(), rb.ri.GetRotMatrix_Ini(),
                           rb.rd.GetDispCG(),   rb.rd.GetRotMatrix_Inc(),
                           v1,aloc,cont_norm,stiff_f);    
    Com::CVector3D d0 = (fv1-fv)*=1.0/eps;
    Com::CVector3D d1 = dfv_dl.MatVec(e); d1*=1.0/eps;    
    std::cout << "dfv_dl" << std::endl;    
    std::cout << (d0-d1).Length() / d0.Length() << std::endl;            
    std::cout << "    " << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << "    " << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    
    Com::CVector3D e0 = (fr1-fr)*=1.0/eps;
    Com::CVector3D e1 = dfr_dl.MatVec(e); e1*=1.0/eps;
    std::cout << "dfr_dl" << std::endl;        
    std::cout << (e0-e1).Length() / e0.Length() << std::endl;            
    std::cout << "    " << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << "    " << e1.x << " " << e1.y << " " << e1.z << std::endl;    
  }
}



