/*
 *  untitled.h
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/24/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if !defined(RIGID_POINTSET_H)
#define RIGID_POINTSET_H

#include "delfem/vector3d.h"
#include "delfem/matrix3d.h"

double GetContactEnergy
(const Com::CVector3D& c,
 const Com::CVector3D& v,
 const Com::CVector3D& b, 
 const Com::CVector3D& n, 
 double cont_stiff);

void Get1stDer_ContactEnergy
(Com::CVector3D& dce_du, Com::CVector3D& dce_dr,
 const Com::CVector3D& c, 
 const Com::CVector3D& v,
 const Com::CVector3D& b, const Com::CVector3D& n, double cont_stiff);

void Get2ndDer_ContactEnergy
(Com::CMatrix3& dce_dudu, Com::CMatrix3& dce_dudr, Com::CMatrix3& dce_drdr,
 const Com::CVector3D& c, 
 const Com::CVector3D& v, 
 const Com::CVector3D& cont_base, const Com::CVector3D& n, double cont_stiff);

void TestContact_RigidPointSet();

///////////////////////////////////////////////////////////////////////////

double GetStaticFrictionEnergy
(const Com::CVector3D& c,
 const Com::CVector3D& v,
 const Com::CVector3D& a, 
 const Com::CVector3D& n, 
 const double stiff_f);

void Get1stDer_StaticFrictionEnergy
(Com::CVector3D& fv, Com::CVector3D& fr,
 const Com::CVector3D& c,
 const Com::CVector3D& v,
 const Com::CVector3D& a,
 const Com::CVector3D& n, 
 const double stiff_f);

void Get2ndDer_StaticFrictionEnergy
(Com::CMatrix3& dfv_du, Com::CMatrix3& dfv_dr, Com::CMatrix3& dfr_du, Com::CMatrix3& dfr_dr,
 const Com::CVector3D& c, 
 const Com::CVector3D& v,
 const Com::CVector3D& a,
 const Com::CVector3D& n, 
 double stiff_f);

void TestStaticFriction_RigidPointSet();

///////////////////////////////////////////////////////////////////////////

double GetSpringEnergy
(const Com::CVector3D& v, // dist from center
 const Com::CVector3D& a); // anchor

void Get1stDer_SpringEnergy
(Com::CVector3D& fv, Com::CVector3D& fr,
 const Com::CVector3D& v,
 const Com::CVector3D& a);

void Get2ndDer_SpringEnergy
(Com::CMatrix3& dfv_du, Com::CMatrix3& dfv_dr, Com::CMatrix3& dfr_du, Com::CMatrix3& dfr_dr,
 const Com::CVector3D& v,
 const Com::CVector3D& a);

#endif

