/*
 *  rigidbody.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/6/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include "rigidbody.h"
#include "furniture_utility.h"

CRigidBody3D::CRigidBody3D(const CRigidBody3D& rb)
{
  ri = rb.ri;
  rd = rb.rd;
}





Com::CMatrix3 CRigidBody3D::GetRotMatrix() const
{
  Com::CMatrix3 m0, m1;
  m0.SetRotMatrix_CRV(ri.crv_ini);
  m1.SetRotMatrix_CRV(rd.crv_inc);
  Com::CMatrix3 m = m1.MatMat(m0);
  return m;
}


void UpdateSolution_BackwardEular
(CRigid3D_Def& rd, 
 const double* upd, 
 double dt,
 bool is_first_iter )
{
  assert( is_first_iter );
  
  const Com::CVector3D vel_update( upd[0], upd[1], upd[2] );
  const Com::CVector3D omg_update( upd[3], upd[4], upd[5] );
  
  rd.velo_cg += vel_update; // update velocity
  rd.disp_cg += dt*rd.velo_cg; // update displacemnet
  
  ////////////////////////////////////////////////
  rd.Omega += omg_update;  
	{	// 回転のアップデート
    Com::CMatrix3 m; m.SetRotMatrix_Cartesian(dt*rd.Omega);
    rd.AddRotation_IncRight( m );
	}
}

bool IsRayCollide_Ini
(const Com::CVector3D& src, 
 const Com::CVector3D& dir,
 Com::CVector3D& hit_pos,
 unsigned int& iface,
 const CRigid3D_Ini& ri)
{
  const Com::CVector3D aFaceNormal[6][3] = {
    { Com::CVector3D(+1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0,+1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-0,-1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0, 0,+1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) },
    { Com::CVector3D(-0, 0,-1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) } };  
  double mindepth; bool is_some_hit = false;  
  for(unsigned int i=0;i<6;i++){
    Com::CMatrix3 B(ri.xlen,ri.ylen,ri.zlen);
    Com::CMatrix3 R = ri.GetRotMatrix_Ini();
    Com::CMatrix3 RB = R.MatMat(B);
    const Com::CVector3D& n = R.MatVec( aFaceNormal[i][0] );    
    const Com::CVector3D& org   = R.MatVec( B.MatVec( aFaceNormal[i][0]*0.5) ) + ri.GetIniPosCG();
    const Com::CVector3D& vec_h = R.MatVec( B.MatVec( aFaceNormal[i][1]*0.5) );
    const Com::CVector3D& vec_v = R.MatVec( B.MatVec( aFaceNormal[i][2]*0.5) );
    double rh, rv;
    bool res = IsRayCollideQuad(src,dir, org,n,vec_h,vec_v, rh,rv);
    if( !res ) continue;
    Com::CVector3D pos = org + rh*vec_h + rv*vec_v;
    const double depth = Dot(pos-src,dir);    
    if( !is_some_hit ){
      hit_pos = pos;
      is_some_hit = true;      
      mindepth = depth;
      iface = i;
    }
    else{
      if( depth < mindepth ){
        mindepth = depth;
        hit_pos = pos;
        iface = i;        
      }
    }
  }
  return is_some_hit;
}



