/*
 *  rigidbody.h
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/6/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#if !defined(RIGID_BODY_3D_H)
#define RIGID_BODY_3D_H

#include "delfem/vector3d.h"
#include "delfem/matrix3d.h"



class CRigid3D_Ini
{
public:
  CRigid3D_Ini(){
    this->ini_pos_cg.SetVector(0,0,0);
    crv_ini[0]=0;  crv_ini[1]=0;  crv_ini[2]=0;  
    this->mineatia[0] = 1.0; 
    this->mineatia[1] = 1.0;
    this->mineatia[2] = 1.0;
    this->mass = 1.0;          
  }
  CRigid3D_Ini(const CRigid3D_Ini& rb){
    this->crv_ini[0] = rb.crv_ini[0];
    this->crv_ini[1] = rb.crv_ini[1];
    this->crv_ini[2] = rb.crv_ini[2];      
    this->ini_pos_cg = rb.ini_pos_cg;  
    this->mass = rb.mass;
    this->mineatia[0] = rb.mineatia[0];
    this->mineatia[1] = rb.mineatia[1];
    this->mineatia[2] = rb.mineatia[2];  
    this->xlen = rb.xlen;
    this->ylen = rb.ylen;
    this->zlen = rb.zlen;    
  }
  double GetMass() const { return mass; }    
  void SetModeBox(double xlen, double ylen, double zlen, double rho ){
    //    imode = 1;
    this->xlen = xlen;
    this->ylen = ylen;
    this->zlen = zlen;
    mass = xlen*ylen*zlen*rho;
    mineatia[0] = 1.0/12.0*(ylen*ylen+zlen*zlen)*mass;
    mineatia[1] = 1.0/12.0*(zlen*zlen+xlen*xlen)*mass;
    mineatia[2] = 1.0/12.0*(xlen*xlen+ylen*ylen)*mass;
  }    
  void SetMass(double mass, const double mineatia[3]){
    this->mass = mass;
    this->mineatia[0] = mineatia[0];
    this->mineatia[1] = mineatia[1];
    this->mineatia[2] = mineatia[2];
  }  
  void SetIniRotation(const Com::CMatrix3& m ){
    m.GetCRV_RotMatrix(crv_ini);  
  }  
  void AddRotation_IniLeft( const Com::CMatrix3& rot )
  {
    Com::CMatrix3 m; m.SetRotMatrix_CRV(crv_ini);
    m = rot.MatMat(m);
    m.GetCRV_RotMatrix(crv_ini);
  }
//  void AddRotation_IniLeft(  const Com::CMatrix3& rot ){ ri.AddRotation_IniLeft( rot); }  
  Com::CMatrix3 GetRotMatrix_Ini() const
  {
    Com::CMatrix3 m;
    m.SetRotMatrix_CRV(crv_ini);
    return m;
  }
  void SetIniPosCG( const Com::CVector3D& pos){ ini_pos_cg = pos; }  
  const Com::CVector3D& GetIniPosCG() const { return ini_pos_cg; }      
  Com::CVector3D GetInitialPosition(const Com::CVector3D& vec) const  // input is a relative position
  {
    return GetRotMatrix_Ini().MatVec(vec)+ini_pos_cg;
  }          
public:
  double crv_ini[3]; // initial_rotation    
  Com::CVector3D ini_pos_cg;  // center of gravity
  double mass;
  double mineatia[3];      
  ////
  double xlen, ylen, zlen; // separate this line
};

class CRigid3D_Def
{
public:
  CRigid3D_Def(){
    this->disp_cg.SetVector(0,0,0);
    this->velo_cg.SetVector(0,0,0);  
    crv_inc[0]=0;  crv_inc[1]=0;  crv_inc[2]=0;      
    Omega.SetVector(0,0,0);    
  }
  CRigid3D_Def(const CRigid3D_Def& rb){    
    this->crv_inc[0] = rb.crv_inc[0];
    this->crv_inc[1] = rb.crv_inc[1];
    this->crv_inc[2] = rb.crv_inc[2];
    this->disp_cg = rb.disp_cg;  
    this->velo_cg = rb.velo_cg;  
    this->Omega = rb.Omega;      
  }
  void Clear(){
    crv_inc[0] = 0; crv_inc[1] = 0; crv_inc[2] = 0;
    Omega.SetVector(0,0,0);
    velo_cg.SetVector(0,0,0);
    disp_cg.SetVector(0,0,0);
  }  
  void SetIncRotation( const Com::CMatrix3& m ){
    m.GetCRV_RotMatrix(crv_inc);
  }    
  void AddRotation_IncRight( const Com::CMatrix3& rot )
  {
    Com::CMatrix3 m; m.SetRotMatrix_CRV(crv_inc);
    m = m.MatMat(rot);
    m.GetCRV_RotMatrix(crv_inc);
  }  
  Com::CMatrix3 GetRotMatrix_Inc() const
  {
    Com::CMatrix3 m;
    m.SetRotMatrix_CRV(crv_inc);
    return m;
  }  
  const Com::CVector3D& GetDispCG() const { return disp_cg; }
  void AddDispCG(const Com::CVector3D& d){ disp_cg += d; }
  const Com::CVector3D& GetVeloCG() const { return velo_cg; }
  void SetVeloCG( const Com::CVector3D& vcg ){ this->velo_cg = vcg; }      
  const Com::CVector3D& GetOmega() const { return Omega; }
  void SetOmega( const Com::CVector3D& omg ){ this->Omega = omg; }  
public:
  Com::CVector3D disp_cg;    
  double crv_inc[3]; // incremental_rotation  
  Com::CVector3D velo_cg;  
  Com::CVector3D Omega;
};

class CRigidBody3D
{
public:
  CRigidBody3D(){}
  CRigidBody3D(const CRigidBody3D& rb);
  Com::CMatrix3 GetRotMatrix() const;  
public:  
  CRigid3D_Ini ri;
  CRigid3D_Def rd;
};

void UpdateSolution_BackwardEular
(CRigid3D_Def& rd, 
 const double* upd, double dt, bool is_first_iter );

static Com::CVector3D GetCurrentPositionFromInitial(const Com::CVector3D& vec,                                             
                                             const CRigid3D_Ini& ri,
                                             const CRigid3D_Def& rd )
{
  const Com::CMatrix3& rot = rd.GetRotMatrix_Inc();
  return rot.MatVec(vec-ri.ini_pos_cg)+ri.ini_pos_cg+rd.disp_cg;
}

bool IsRayCollide_Ini(const Com::CVector3D& src, const Com::CVector3D& dir, 
                      Com::CVector3D& hitpos,
                      unsigned int& iface,
                      const CRigid3D_Ini& ri);


static double GetKineticEnergy(const CRigid3D_Ini& ri,
                        const CRigid3D_Def& rd)
{
  double e = 0;
  e += 0.5*( rd.Omega.x*rd.Omega.x*ri.mineatia[0]
            +rd.Omega.y*rd.Omega.y*ri.mineatia[1]
            +rd.Omega.z*rd.Omega.z*ri.mineatia[2]);
  e += 0.5*Com::Dot(rd.velo_cg,rd.velo_cg)*ri.mass;      
  return e;
}

static Com::CVector3D GetCurrentPosition(const Com::CVector3D& vec, 
                                  const CRigid3D_Ini& ri,
                                  const CRigid3D_Def& rd)  // input is a relative position
{
  const Com::CMatrix3& R = rd.GetRotMatrix_Inc().MatMat( ri.GetRotMatrix_Ini() );
  return R.MatVec(vec)+ri.ini_pos_cg+rd.disp_cg;
}

static Com::CVector3D GetCurrentVelocity(const Com::CVector3D& vec,                                  
                                  const CRigid3D_Ini& ri,
                                  const CRigid3D_Def& rd )
{
  const Com::CVector3D& Qa = ri.GetRotMatrix_Ini().MatVec(vec);
  return rd.velo_cg + rd.GetRotMatrix_Inc().MatVec( Cross(rd.Omega,Qa) );
}

class CFrictionPoint
{
public:
  CFrictionPoint(){
    pd = -1;
    itype_contact = 0;
    is_force_cont = false;
  }
public:
	double pd;  // penetration depth
  Com::CVector3D aloc; // anchor location
	int itype_contact;	// 0:not_contact 1:static contact 2:dynamic contact  
  bool is_force_cont;
};


class CContactPoints_RigidBody3D
{
public:
  CContactPoints_RigidBody3D(){
    unsigned int nct = this->Size();
    aFP.resize(nct);
  }
  /*
  unsigned int Size() const { return 8; }  
  Com::CVector3D GetPosContactPoint(unsigned int i, double xlen, double ylen, double zlen) const {
    const double corner[8][3] = { 
      {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
      {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };            
    return Com::CVector3D(corner[i][0]*xlen, corner[i][1]*ylen, corner[i][2]*zlen )*0.5;    
  }
  Com::CMatrix3 GetSensMat_ContPointParam(unsigned int i) const {
    const double corner[8][3] = { 
      {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
      {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };        
    return Com::CMatrix3(corner[i][0]*0.5,corner[i][1]*0.5,corner[i][2]*0.5);
  }
  */
  unsigned int Size() const { return 4; }  
  Com::CVector3D GetPosContactPoint(unsigned int i, double xlen, double ylen, double zlen) const {
    const double corner[4][3] = { 
      {-1,-1,-0}, {+1,-1,+0}, {+1,+1,-0}, {-1,+1,+0} };
    return Com::CVector3D(corner[i][0]*xlen, corner[i][1]*ylen, corner[i][2]*zlen )*0.5;    
  }
  Com::CMatrix3 GetSensMat_ContPointParam(unsigned int i) const {
    const double corner[4][3] = { 
      {-1,-1,-0}, {+1,-1,+0}, {+1,+1,-0}, {-1,+1,+0} };
    return Com::CMatrix3(corner[i][0]*0.5,corner[i][1]*0.5,corner[i][2]*0.5);
  }  
public:
  std::vector<CFrictionPoint> aFP;
};

static Com::CVector3D GetGravityCenter(bool is_current,
                                       const std::vector<CRigid3D_Ini>& aRI,
                                       const std::vector<CRigid3D_Def>& aRD)
{
  Com::CVector3D res(0,0,0);
  double tot_mass = 0;
  for(unsigned int irb=0;irb<aRI.size();irb++){        
    const CRigid3D_Ini& ri = aRI[irb];
    const CRigid3D_Def& rd = aRD[irb];    
    tot_mass += ri.GetMass();
    if( !is_current ){
      res += ri.GetMass()*ri.GetIniPosCG();                
    }
    else{
      res += ri.GetMass()*(ri.GetIniPosCG()+rd.GetDispCG());
    }      
  }
  res /= tot_mass;
  return res;
}



#endif